#include <stdlib.h>
#include <stdio.h>
#include <stdint.h>
#include <string.h>
#include <time.h>
#include <stdarg.h>

#include "stubs_load.h"
#include "firmware_load.h"

//------------------------------------------------------------------------------------------------------------
// #define LIST_IMPORTANT_FUNCTIONS 1   // always list functions with 'LIST_ALWAYS' flag, even when not found
// #define LIST_PHYSW_TABLE 1           // print all physw events to file
// #define PRINT_LEVENT_TABLE 1         // print all levents to file
//------------------------------------------------------------------------------------------------------------

// Buffer output into header and body sections

FILE    *out_fp;
char    out_buf[32*1024] = "";
int     out_len = 0;
char    hdr_buf[32*1024] = "";
int     hdr_len = 0;
int     out_hdr = 1;

void bprintf(char *fmt, ...)
{
    va_list argp;
    va_start(argp, fmt);

    if (out_hdr)
        hdr_len += vsprintf(hdr_buf+hdr_len,fmt,argp);
    else
        out_len += vsprintf(out_buf+out_len,fmt,argp);

    va_end(argp);
}

void add_blankline()
{
    if (strcmp(hdr_buf+hdr_len-2,"\n\n") != 0)
    {
        hdr_buf[hdr_len++] = '\n';
        hdr_buf[hdr_len] = 0;
    }
}

void write_output()
{
    add_blankline();
    if (out_fp)
    {
        fprintf(out_fp,"%s",hdr_buf);
        fprintf(out_fp,"%s",out_buf);
    }
}

//------------------------------------------------------------------------------------------------------------

void usage(char *err)
{
    bprintf("finsig <primary> <base> <outputfilename> [alt base] - Error = %s\n",err);
    write_output();
    fprintf(stderr,"finsig <primary> <base> <outputfilename> [alt base] - Error = %s\n",err);
    exit(1);
}

void error(char *fmt, int n)
{
    bprintf(fmt, n);
    write_output();
    exit(1);
}

//------------------------------------------------------------------------------------------------------------

// Signature match handling

// Structure to store match info
typedef struct {
    uint32_t ptr;
    uint32_t fail;
    uint32_t success;
    int sig;
} Match;

// qsort compare function the Match struture
int match_compare(const Match *p1, const Match *p2)
{
    /* NOTE: If a function has *more* matches, it will be prefered, even if it has a lower percent matches */
    if (p1->success > p2->success)
    {
        if ((p2->fail == 0) && (p1->fail > 0))
        {
            return 1;
        }
        else
        {
            return -1;
        }
    }
    else if (p1->success < p2->success)
    {
        if ((p1->fail == 0) && (p2->fail > 0))
        {
            return -1;
        }
        else
        {
            return 1;
        }
    }
    else
    {
        if (p1->fail < p2->fail)
        {
            return -1;
        }
        else if (p1->fail > p2->fail)
        {
            return 1;
        }
    }

    if (p1->sig < p2->sig)
    {
        return -1;
    }
    else if (p1->sig > p2->sig)
    {
        return 1;
    }

    /* scores are equal. prefer lower address */

    if (p1->ptr < p2->ptr)
    {
        return -1;
    }
    else if (p1->ptr > p2->ptr)
    {
        return 1;
    }

    return 0;
}

#define MAX_MATCHES (8192)

Match matches[MAX_MATCHES];
int count;

// Add a new Match
void addMatch(uint32_t fadr, int s, int f, int sig)
{
    matches[count].ptr = fadr;
    matches[count].success = s;
    matches[count].fail = f;
    matches[count].sig = sig;
    count++;
}

// Add a Match, adjust address for stuff in the block copied to RAM
void fwAddMatch(firmware *fw, uint32_t fadr, int s, int f, int sig)
{
    if ((fadr >= fw->base_copied) && (fadr < (fw->base_copied + fw->size2*4)))
    {
        addMatch(fadr - fw->base_copied + fw->base2,s,f,sig);
    }
    else
    {
        addMatch(fadr,s,f,sig);
    }
}

//------------------------------------------------------------------------------------------------------------

// Signature structures generated by gensig2.exe

typedef struct {
    uint32_t offs;
    uint32_t value;
    uint32_t mask;
} FuncSig;

typedef struct {
    const char *name;
    FuncSig *sig;
    int ver;
} FuncsList;

//------------------------------------------------------------------------------------------------------------

// Master list of functions / addresses to find

#define DONT_EXPORT     1
#define OPTIONAL        2
#define UNUSED          4
#define BAD_MATCH       8
#define EV_MATCH        16
#define LIST_ALWAYS     32

typedef struct {
    char        *name;
    int         flags;
    uint32_t    val;
} func_entry;

int next_func_entry = 0;

#define MAX_FUNC_ENTRY  5000

func_entry  func_names[MAX_FUNC_ENTRY] =
{
    // Do these first as they are needed to find others
    { "ExportToEventProcedure_FW", UNUSED|DONT_EXPORT },
    { "CreateJumptable", UNUSED },
    { "_uartr_req", UNUSED },
    { "StartRecModeMenu", UNUSED },
    { "LogCameraEvent", UNUSED|DONT_EXPORT },
    { "getImageDirName", UNUSED|DONT_EXPORT },

    { "AllocateMemory", UNUSED|LIST_ALWAYS },
    { "AllocateUncacheableMemory" },
    { "Close" },
    { "CreateBinarySemaphore", UNUSED|LIST_ALWAYS },
    { "CreateCountingSemaphore", UNUSED|LIST_ALWAYS },
    { "CreateTask" },
    { "DebugAssert", OPTIONAL|LIST_ALWAYS },
    { "DeleteDirectory_Fut" },
    { "DeleteFile_Fut" },
    { "DeleteSemaphore", UNUSED|LIST_ALWAYS },
    { "DoAELock" },
    { "DoAFLock" },
    { "EnterToCompensationEVF" },
    { "ExecuteEventProcedure" },
    { "ExitFromCompensationEVF" },
    { "ExitTask" },
    { "ExpCtrlTool_StartContiAE" },
    { "ExpCtrlTool_StopContiAE" },
    { "Fclose_Fut" },
    { "Feof_Fut" },
    { "Fflush_Fut" },
    { "Fgets_Fut" },
    { "Fopen_Fut" },
    { "Fread_Fut" },
    { "FreeMemory", UNUSED|LIST_ALWAYS },
    { "FreeUncacheableMemory" },
    { "Fseek_Fut" },
    { "Fwrite_Fut" },
    { "GetBatteryTemperature" },
    { "GetCCDTemperature" },
    { "GetCurrentAvValue" },
    { "GetDrive_ClusterSize" },
    { "GetDrive_FreeClusters" },
    { "GetDrive_TotalClusters" },
    { "GetFocusLensSubjectDistance" },
    { "GetFocusLensSubjectDistanceFromLens" },
    { "GetImageFolder", OPTIONAL },
    { "GetKbdState" },
    { "GetMemInfo" },
    { "GetOpticalTemperature" },
    { "GetParameterData" },
    { "GetPropertyCase" },
    { "GetSystemTime" },
    { "GetVRAMHPixelsSize" },
    { "GetVRAMVPixelsSize" },
    { "GetZoomLensCurrentPoint" },
    { "GetZoomLensCurrentPosition" },
    { "GiveSemaphore", OPTIONAL|LIST_ALWAYS },
    { "IsStrobeChargeCompleted" },
    { "LEDDrive", OPTIONAL },
    { "LocalTime" },
    { "LockMainPower" },
    { "Lseek", UNUSED|LIST_ALWAYS },
    { "MakeDirectory_Fut" },
    { "MakeSDCardBootable", OPTIONAL },
    { "MoveFocusLensToDistance" },
    { "MoveIrisWithAv", OPTIONAL },
    { "MoveZoomLensWithPoint" },
    { "NewTaskShell", UNUSED },
    { "Open" },
    { "PB2Rec" },
    { "PT_MoveDigitalZoomToWide", OPTIONAL },
    { "PT_MoveOpticalZoomAt", OPTIONAL },
    { "PT_PlaySound" },
    { "PostLogicalEventForNotPowerType" },
    { "PostLogicalEventToUI" },
    { "PutInNdFilter", OPTIONAL },
    { "PutOutNdFilter", OPTIONAL },
    { "Read" },
    { "ReadFastDir" },
    { "Rec2PB" },
    { "RefreshPhysicalScreen" },
    { "Remove", OPTIONAL|UNUSED },
    { "RenameFile_Fut" },
    { "Restart" },
    { "ScreenLock" },
    { "ScreenUnlock" },
    { "SetAE_ShutterSpeed" },
    { "SetAutoShutdownTime" },
    { "SetCurrentCaptureModeType" },
    { "SetFileAttributes" },
    { "SetFileTimeStamp" },
    { "SetLogicalEventActive" },
    { "SetParameterData" },
    { "SetPropertyCase" },
    { "SetScriptMode" },
    { "SleepTask" },
    { "TakeSemaphore" },
    { "TurnOffBackLight" },
    { "TurnOnBackLight" },
    { "TurnOnDisplay" },
    { "TurnOffDisplay" },
    { "UIFS_WriteFirmInfoToFile" },
    { "UnlockAE" },
    { "UnlockAF" },
    { "UnlockMainPower" },
    { "UnsetZoomForMovie", OPTIONAL },
    { "UpdateMBROnFlash" },
    { "VbattGet" },
    { "Write" },
    { "WriteSDCard" },

    { "_log" },
    { "_log10" },
    { "_pow" },
    { "_sqrt" },
    { "add_ptp_handler" },
    { "apex2us" },
    { "close" },
    { "err_init_task", OPTIONAL },
    { "exmem_alloc" },
    { "exmem_free", OPTIONAL|LIST_ALWAYS },
    { "free" },

    { "kbd_p1_f" },
    { "kbd_p1_f_cont" },
    { "kbd_p2_f" },
    { "kbd_read_keys" },
    { "kbd_read_keys_r2" },

    { "kbd_pwr_off", OPTIONAL },
    { "kbd_pwr_on", OPTIONAL },
    { "lseek" },
    { "malloc" },
    { "memcmp" },
    { "memcpy" },
    { "memset" },
    { "mkdir" },
    { "mktime_ext" },
    { "open" },
    { "OpenFastDir" },
    { "closedir" },
    { "get_fstype", OPTIONAL|LIST_ALWAYS },
    { "qsort" },
    { "rand" },
    { "read", UNUSED|OPTIONAL },
    { "realloc", OPTIONAL|LIST_ALWAYS },
    { "reboot_fw_update" },
    { "set_control_event" },
    { "srand" },
    { "stat" },
    { "strcat" },
    { "strchr" },
    { "strcmp" },
    { "strcpy" },
    { "strftime" },
    { "strlen" },
    { "strncmp" },
    { "strncpy" },
    { "strrchr" },
    { "strtol" },
    { "strtolx" },

    { "task_CaptSeq" },
    { "task_DvlpSeqTask", OPTIONAL },
    { "task_ExpDrv" },
    { "task_FileWrite", OPTIONAL },
    { "task_InitFileModules" },
    { "task_MovieRecord" },
    { "task_PhySw", OPTIONAL },
    { "task_RotaryEncoder", OPTIONAL },
    { "task_TouchPanel", OPTIONAL },

    { "hook_CreateTask" },

    { "time" },
    { "vsprintf" },
    { "write", UNUSED|OPTIONAL },

    { "EngDrvIn", OPTIONAL|UNUSED|LIST_ALWAYS },
    { "EngDrvOut", OPTIONAL|UNUSED|LIST_ALWAYS },
    { "EngDrvRead" },
    { "EngDrvBits", OPTIONAL|UNUSED|LIST_ALWAYS },

    { "PTM_GetCurrentItem" },
    { "PTM_SetCurrentItem", UNUSED|LIST_ALWAYS },
    { "PTM_NextItem", OPTIONAL|UNUSED|LIST_ALWAYS },
    { "PTM_PrevItem", OPTIONAL|UNUSED|LIST_ALWAYS },
    { "PTM_SetPropertyEnable", OPTIONAL|UNUSED|LIST_ALWAYS },

    { "DisableISDriveError", OPTIONAL },

    // OS functions, mostly to aid firmware analysis. Order is important!
    { "_GetSystemTime", OPTIONAL|UNUSED|LIST_ALWAYS }, // only for locating timer functions
    { "SetTimerAfter", OPTIONAL|UNUSED|LIST_ALWAYS },
    { "SetTimerWhen", OPTIONAL|UNUSED|LIST_ALWAYS },
    { "CancelTimer", OPTIONAL|UNUSED|LIST_ALWAYS },
    { "CancelHPTimer" },
    { "SetHPTimerAfterTimeout", OPTIONAL|UNUSED|LIST_ALWAYS },
    { "SetHPTimerAfterNow" },
    { "CreateTaskStrictly", OPTIONAL|UNUSED|LIST_ALWAYS },
    { "CreateMessageQueue", OPTIONAL|UNUSED|LIST_ALWAYS },
    { "CreateRecursiveLock", OPTIONAL|UNUSED|LIST_ALWAYS },
    { "GetSemaphoreValue", OPTIONAL|UNUSED|LIST_ALWAYS },
    { "TryTakeSemaphore", OPTIONAL|UNUSED|LIST_ALWAYS },
    { "CreateMessageQueueStrictly", OPTIONAL|UNUSED|LIST_ALWAYS },
    { "CreateEventFlagStrictly", OPTIONAL|UNUSED|LIST_ALWAYS },
    { "CreateBinarySemaphoreStrictly", OPTIONAL|UNUSED|LIST_ALWAYS },
    { "CreateCountingSemaphoreStrictly", OPTIONAL|UNUSED|LIST_ALWAYS },
    { "CreateRecursiveLockStrictly", OPTIONAL|UNUSED|LIST_ALWAYS },
    { "TakeSemaphoreStrictly", OPTIONAL|UNUSED|LIST_ALWAYS }, // r23+
    { "ReceiveMessageQueueStrictly", OPTIONAL|UNUSED|LIST_ALWAYS }, // r23+
    { "PostMessageQueueStrictly", OPTIONAL|UNUSED|LIST_ALWAYS },    // r23+
    { "WaitForAnyEventFlagStrictly", OPTIONAL|UNUSED|LIST_ALWAYS }, // r23+
    { "WaitForAllEventFlagStrictly", OPTIONAL|UNUSED|LIST_ALWAYS }, // r23+
    { "AcquireRecursiveLockStrictly", OPTIONAL|UNUSED|LIST_ALWAYS }, // r23+
    { "DeleteMessageQueue", OPTIONAL|UNUSED|LIST_ALWAYS },
    { "PostMessageQueue", OPTIONAL|UNUSED|LIST_ALWAYS },
    { "ReceiveMessageQueue", OPTIONAL|UNUSED|LIST_ALWAYS },
    { "TryReceiveMessageQueue", OPTIONAL|UNUSED|LIST_ALWAYS },
    { "TryPostMessageQueue", OPTIONAL|UNUSED|LIST_ALWAYS },
    { "GetNumberOfPostedMessages", OPTIONAL|UNUSED|LIST_ALWAYS },
    { "DeleteRecursiveLock", OPTIONAL|UNUSED|LIST_ALWAYS },
    { "AcquireRecursiveLock", OPTIONAL|UNUSED|LIST_ALWAYS },
    { "ReleaseRecursiveLock", OPTIONAL|UNUSED|LIST_ALWAYS },
    { "WaitForAnyEventFlag", OPTIONAL|UNUSED|LIST_ALWAYS },
    { "WaitForAllEventFlag", OPTIONAL|UNUSED|LIST_ALWAYS },
    { "ClearEventFlag", OPTIONAL|UNUSED|LIST_ALWAYS },
    { "SetEventFlag", OPTIONAL|UNUSED|LIST_ALWAYS },
    { "GetEventFlagValue", OPTIONAL|UNUSED|LIST_ALWAYS },
    { "CreateEventFlag", OPTIONAL|UNUSED|LIST_ALWAYS },
    { "DeleteEventFlag", OPTIONAL|UNUSED|LIST_ALWAYS },
    { "CheckAnyEventFlag", OPTIONAL|UNUSED|LIST_ALWAYS },
    { "CheckAllEventFlag", OPTIONAL|UNUSED|LIST_ALWAYS },
    { "RegisterInterruptHandler", OPTIONAL|UNUSED|LIST_ALWAYS },
    { "UnregisterInterruptHandler", OPTIONAL|UNUSED|LIST_ALWAYS },
    { "GetSRAndDisableInterrupt", OPTIONAL|UNUSED|LIST_ALWAYS }, // disables IRQ, returns a value
    { "SetSR", OPTIONAL|UNUSED|LIST_ALWAYS }, // enables IRQ, puts back value returned by GetSR
    { "EnableInterrupt", OPTIONAL|UNUSED|LIST_ALWAYS }, // enables IRQ
    { "_divmod_signed_int", OPTIONAL|UNUSED|LIST_ALWAYS}, // division for signed integers, remainder is returned in r1
    { "_divmod_unsigned_int", OPTIONAL|UNUSED|LIST_ALWAYS}, // division for unsigned integers, remainder is returned in r1
    { "_dflt", OPTIONAL|UNUSED|LIST_ALWAYS}, // int -> double
    { "_dfltu", OPTIONAL|UNUSED|LIST_ALWAYS}, // uint -> double
    { "_dfix", OPTIONAL|UNUSED|LIST_ALWAYS}, // double -> int
    { "_dfixu", OPTIONAL|UNUSED|LIST_ALWAYS}, // double -> uint
    { "_dmul", OPTIONAL|UNUSED|LIST_ALWAYS}, // double precision float multiplication
    { "_ddiv", OPTIONAL|UNUSED|LIST_ALWAYS}, // double precision float division
    { "_dadd", OPTIONAL|UNUSED|LIST_ALWAYS}, // addition for doubles
    { "_dsub", OPTIONAL|UNUSED|LIST_ALWAYS}, // subtraction for doubles
    { "_drsb", OPTIONAL|UNUSED|LIST_ALWAYS}, // reverse subtraction for doubles (?)
    { "_dcmp", OPTIONAL|UNUSED|LIST_ALWAYS}, // comparison of 2 doubles, only updates condition flags
    { "_dcmp_reverse", OPTIONAL|UNUSED|LIST_ALWAYS}, // like _dcmp, but operands in reverse order, only updates condition flags
    { "_safe_sqrt", OPTIONAL|UNUSED|LIST_ALWAYS}, // only calls _sqrt for numbers >= 0
    { "_scalbn", OPTIONAL|UNUSED|LIST_ALWAYS}, // double scalbn (double x, long exp), returns x * FLT_RADIX ** exp
    { "_fflt", OPTIONAL|UNUSED|LIST_ALWAYS}, // int -> float
    { "_ffltu", OPTIONAL|UNUSED|LIST_ALWAYS}, // uint -> float
    { "_ffix", OPTIONAL|UNUSED|LIST_ALWAYS}, // float -> int
    { "_ffixu", OPTIONAL|UNUSED|LIST_ALWAYS}, // float -> uint
    { "_fmul", OPTIONAL|UNUSED|LIST_ALWAYS}, // single precision float multiplication
    { "_fdiv", OPTIONAL|UNUSED|LIST_ALWAYS}, // single precision float division
    { "_f2d", OPTIONAL|UNUSED|LIST_ALWAYS}, // float -> double
    { "DisplayBusyOnScreen", OPTIONAL|UNUSED|LIST_ALWAYS}, // displays full screen "busy" message
    { "UndisplayBusyOnScreen", OPTIONAL|UNUSED|LIST_ALWAYS},
    { "CreateDialogBox", OPTIONAL|UNUSED|LIST_ALWAYS},
    { "DisplayDialogBox", OPTIONAL|UNUSED|LIST_ALWAYS},
    { "add_ui_to_dialog", OPTIONAL|UNUSED|LIST_ALWAYS}, // name made up, assigns resources to a dialog
    { "get_string_by_id", OPTIONAL|UNUSED|LIST_ALWAYS}, // name made up, retrieves a localised or unlocalised string by its ID
    { "malloc_strictly", OPTIONAL|UNUSED|LIST_ALWAYS }, // name made up
    { "GetCurrentMachineTime", OPTIONAL|UNUSED|LIST_ALWAYS }, // reads usec counter, name from ixus30
    { "HwOcReadICAPCounter", OPTIONAL|UNUSED|LIST_ALWAYS }, // reads usec counter, name from ixus30
    { "get_self_task_id", OPTIONAL|UNUSED|LIST_ALWAYS }, // gets ID of own task
    { "get_task_properties", OPTIONAL|UNUSED|LIST_ALWAYS }, // gets copy of task's data (different struct, not the real TCB)
    { "get_self_task_errno_pointer", OPTIONAL|UNUSED|LIST_ALWAYS }, // gets pointer to own task's errno

    // Other stuff needed for finding misc variables - don't export to stubs_entry.S
    { "GetSDProtect", UNUSED },
    { "DispCon_ShowBitmapColorBar", UNUSED },
    { "ResetZoomLens", OPTIONAL|UNUSED },
    { "ResetFocusLens", OPTIONAL|UNUSED },
    { "NR_GetDarkSubType", OPTIONAL|UNUSED },
    { "NR_SetDarkSubType", OPTIONAL|UNUSED },
    { "SavePaletteData", OPTIONAL|UNUSED },
    { "GUISrv_StartGUISystem", OPTIONAL|UNUSED|LIST_ALWAYS },
    { "get_resource_pointer", OPTIONAL|UNUSED|LIST_ALWAYS }, // name made up, gets a pointer to a certain resource (font, dialog, icon)
    { "CalcLog10", OPTIONAL|UNUSED|LIST_ALWAYS }, // helper

    { "MFOn", OPTIONAL },
    { "MFOff", OPTIONAL },
    { "PT_MFOn", OPTIONAL },
    { "PT_MFOff", OPTIONAL },
    { "SS_MFOn", OPTIONAL },
    { "SS_MFOff", OPTIONAL },

    { "GetAdChValue", OPTIONAL },

    { 0, 0, 0 }
};

// Return the array index of a named function in the array above
int find_saved_sig(const char *name)
{
    int i;
    for (i=0; func_names[i].name != 0; i++)
    {
        if (strcmp(name,func_names[i].name) == 0)
        {
            return i;
        }
    }
    return -1;
}

// Save the address value found for a function in the above array
void save_sig(const char *name, uint32_t val)
{
    int i = find_saved_sig(name);
    if (i >= 0)
    {
        func_names[i].val = val;
    }
}

// Get the saved address value for a named function
// If the address value is 0 then assume the function search has not occurred yet, so go search for it.
int get_saved_sig(firmware *fw, const char *name)
{
    int i = find_saved_sig(name);
    if (i >= 0)
    {
        if (func_names[i].val == 0)
        {
            // See if the function is in the 'func_list' array below
            int find_func(const char* name);
            int k1 = find_func(name);
            if (k1 >= 0)
            {
                // if found do full search
                void find_matches(firmware*,const char*);
                find_matches(fw, name);
                count = 0;
            }
            else
            {
                // not found, only do string matching search
                void find_str_sig_matches(firmware*,const char*);
                find_str_sig_matches(fw, name);
                count = 0;
            }
        }
        if (func_names[i].val == 0)
        {
            // If not found return invalid index
            i = -1;
        }
    }
    return i;
}

// Search for something relative to the location stored for a previously matched function
// Matching is done via the 'func' function, searching continues until 'func' returns non-zero
// Starts searching at 'ofst' from the saved address for a max of 'len' instructions
// Returns the 'func' value or 0 if no match found
int search_saved_sig(firmware *fw, char *sig, int (*func)(firmware*, int, int), int v, int ofst, int len)
{
    int k = get_saved_sig(fw, sig);
    if (k >= 0)
    {
        int idx = adr2idx(fw, func_names[k].val);
        for (k=idx+ofst; k<idx+ofst+len; k++)
        {
            int rv = func(fw, k, v);
            if (rv)
                return rv;
        }
    }
    return 0;
}

//------------------------------------------------------------------------------------------------------------

// Signature min / max DryOS versions

typedef struct {
    char        *name;
    uint32_t    val;
} sig_stuff;

sig_stuff min_ver[] = {
    { "ScreenLock", 39 },
    { "ScreenUnlock", 39 },
    { "MakeSDCardBootable", 47 },
    { "hook_CreateTask", 51 },

    { 0, 0 }
};

sig_stuff max_ver[] = {
    { "UpdateMBROnFlash", 45 },
    { "kbd_pwr_on", 43 },
    { "kbd_pwr_off", 43 },

    { 0, 0 }
};

int find_min_ver(const char *name)
{
    int i;
    for (i=0; min_ver[i].name != 0; i++)
    {
        if (strcmp(name,min_ver[i].name) == 0)
        {
            return min_ver[i].val;
        }
    }
    return 0;
}

int find_max_ver(const char *name)
{
    int i;
    for (i=0; max_ver[i].name != 0; i++)
    {
        if (strcmp(name,max_ver[i].name) == 0)
        {
            return max_ver[i].val;
        }
    }
    return 99999;
}

//------------------------------------------------------------------------------------------------------------

// New string / signature matching structure

// Structure for matching

typedef struct {
    int     type;               // 1 = func*, string, 2 = string, ... string*, func*, 3 = ADR Rx, func, ADR Ry, string, BL, ... string, etc
    char    *name;              // function name
    char    *ev_name;           // event / other name to match in the firmware
    int     offset;             // offset for following branches, or other tests
    // DryOS version specific offsets
    int     dryos20_offset;     // ***** UPDATE for new DryOS version *****
    int     dryos23_offset;
    int     dryos31_offset;
    int     dryos39_offset;
    int     dryos43_offset;
    int     dryos45_offset;
    int     dryos47_offset;
    int     dryos49_offset;
    int     dryos50_offset;
    int     dryos51_offset;
    int     dryos52_offset;
    int     dryos54_offset;
    int     dryos55_offset;
    int     dryos57_offset;
    int     dryos58_offset;
} string_sig;

// Load old signature matching data generated by gensig_dryos
#include "signatures_dryos.h"

//------------------------------------------------------------------------------------------------------------

// Data for matching 'apex2us' function
uint32_t apex2us_test[] = { 0x3D09000, 0x3BBA304, 0x3A728D2, 0x3931EF4, 0x37F8303, 0x36C52A2, 0x3598B85, 0x3472B6A, 0 };
uint32_t apex2us_test2[] = { 0x3d090000, 0x3bba3040, 0x3a728d1f, 0x3931ef45, 0x37f8302c, 0x36c52a26, 0x3598b852, 0x3472b699, 0 }; // r52+?

// Special case for apex2us
int match_apex2us(firmware *fw, int k, uint32_t v1, uint32_t v2)
{
    if (isLDR_PC(fw,k) && (LDR2val(fw,k) == v1) && ((fwRd(fw,k) == 1) || (fwRd(fw,k) == 2)))
    {
        k = find_inst_rev(fw, isSTMFD_LR, k, 200);
        if (k != 0)
        {
            if (fwval(fw,k-2) == 0xE3700D09)    // CMN R0, #0x240
                k -= 2;
            uint32_t fadr = idx2adr(fw,k);
            fwAddMatch(fw,fadr,32,0,121);
            return 1;
        }
    }
    return 0;
}
int match_apex2us2(firmware *fw, int k, uint32_t v1, uint32_t v2) // r52+?
{
    if (isLDR_PC(fw,k) && (LDR2val(fw,k) == v1) && ((fwRd(fw,k) == 1) || (fwRd(fw,k) == 2)))
    {
        k = find_inst_rev(fw, isSTMFD_LR, k, 200);
        if (k != 0)
        {
            if (fwval(fw,k+1) != 0xe3700d0f)    // CMN R0, #0x3c0
                return 0;
            uint32_t fadr = idx2adr(fw,k);
            fwAddMatch(fw,fadr,32,0,121);
            return 1;
        }
    }
    return 0;
}
int find_apex2us(firmware *fw, string_sig *sig, int j)
{
    int i;
    int fnd = 1;
    for (i=0; apex2us_test[i] != 0; i++)
    {
        if (fwval(fw,j+i) != apex2us_test[i])
        {
            fnd = 0;
            break;
        }
    }
    if (fnd)
    {
        return search_fw(fw, match_apex2us, idx2adr(fw,j), 0, 1);
    }

    for (i=0; apex2us_test2[i] != 0; i++)
    {
        if (fwval(fw,j+i) != apex2us_test2[i])
        {
            return 0;
        }
    }
    return search_fw(fw, match_apex2us2, idx2adr(fw,j), 0, 1);
}

// Special case for mkdir
int find_mkdir(firmware *fw, string_sig *sig, int k)
{
    if (fwval(fw,k) == 0x12CEA600)
    {
        k = find_inst_rev(fw, isSTMFD_LR, k-20, 200);
        if (k != 0)
        {
            if ((((fwval(fw,k+12) & 0xFFF0FFFF) == 0xE350002F) && ((fwval(fw,k+15) & 0xFFF0FFFF) == 0xE3500021) && ((fwval(fw,k+19) & 0xFFF0FFFF) == 0xE3500020)) ||
                (((fwval(fw,k+11) & 0xFFF0FFFF) == 0xE350002F) && ((fwval(fw,k+14) & 0xFFF0FFFF) == 0xE3500021) && ((fwval(fw,k+18) & 0xFFF0FFFF) == 0xE3500020)))
            {
                uint32_t fadr = 0;
                if (isBL(fw,k+47))
                {
                    fadr = followBranch(fw, idx2adr(fw,k+47), 0x01000001);
                }
                else if (isBL(fw,k+48))
                {
                    fadr = followBranch(fw, idx2adr(fw,k+48), 0x01000001);
                }
                if (fadr != 0)
                {
                    fwAddMatch(fw,fadr,32,0,121);
                    return 1;
                }
            }
        }
    }
    return 0;
}

// Special case for _pow
int find_pow(firmware *fw, string_sig *sig, int j)
{
    // Find values passed to _pow
    if ((fwval(fw,j) == 0x00000000) && (fwval(fw,j+1) == 0x40000000) && (fwval(fw,j+2) == 0x00000000) && (fwval(fw,j+3) == 0x408F4000))
    {
        uint32_t adr1 = idx2adr(fw,j);        // address of 1st value
        uint32_t adr2 = idx2adr(fw,j+2);    // address of 2nd value
        int j1;

        for (j1 = j-5; j1>0; j1--)
        {
            if (isADR_PC_cond(fw,j1) &&                 // ADR ?
                (fwval(fw,j1+1) == 0xE8900003) &&       // LDMIA R0,{R0,R1}
                isBL(fw,j1+2) &&                        // BL
                isADR_PC_cond(fw,j1+4))                 // ADR ?
            {
                if ((ADR2adr(fw,j1) == adr1) && (ADR2adr(fw,j1+4) == adr2))
                {
                    uint32_t fadr = followBranch(fw,idx2adr(fw,j1+2),0x01000001);
                    fwAddMatch(fw,fadr,32,0,121);
                    return 1;
                }
            }
            else
            if (isADR_PC_cond(fw,j1) &&                 // ADR ?
                (fwval(fw,j1+2) == 0xE8900003) &&       // LDMIA R0,{R0,R1}
                isBL(fw,j1+3) &&                        // BL
                isADR_PC_cond(fw,j1+4))                 // ADR ?
            {
                if ((ADR2adr(fw,j1) == adr1) && (ADR2adr(fw,j1+4) == adr2))
                {
                    uint32_t fadr = followBranch(fw,idx2adr(fw,j1+3),0x01000001);
                    fwAddMatch(fw,fadr,32,0,121);
                    return 1;
                }
            }
        }
    }

    return 0;
}

// Special case for rand
int find_rand(firmware *fw, string_sig *sig, int j)
{
    if (fwval(fw,j) == 0x41C64E6D)
    {
        int j1;

        for (j1 = j-1; j1>j-30; j1--)
        {
            if (isLDR_PC_cond(fw,j1) &&         // LDR Rx, =0x41C64E6D
                (LDR2val(fw,j1) == 0x41C64E6D)) // LDMIA R0,{R0,R1}
            {
                int k = find_inst_rev(fw, isBX_LR,j1-1,15);
                if (k >= 0)
                {
                    uint32_t fadr = idx2adr(fw, k+1);
                    fwAddMatch(fw,fadr,32,0,121);
                    return 1;
                }
            }
        }
    }

    return 0;
}

// Special case for 'closedir'
int find_closedir(firmware *fw)
{
    int j = get_saved_sig(fw,"OpenFastDir");
    if (j >= 0)
    {
        int k = find_inst(fw, isSTMFD_LR, adr2idx(fw,func_names[j].val)+1, 100);
        if (isB(fw,k-1) && isBL(fw,k-2))
        {
            uint32_t fadr = followBranch(fw, idx2adr(fw, k-2), 0x01000001);
            fwAddMatch(fw,fadr,32,0,121);
            return 1;
        }
    }

    return 0;
}

// Special case for 'get_fstype'
int find_get_fstype(firmware *fw)
{
    int j = get_saved_sig(fw,"OpenFastDir");
    if (j >= 0)
    {
        int k = find_Nth_inst(fw, isBL, adr2idx(fw,func_names[j].val)+1, 6, 2);
        if (k > 0)
        {
            // sanity check 1
            if ( (fwval(fw, k+1) & 0xffff0fff) != 0xe1b00000 ) // movs rx, r0
                return 0;
            // sanity check 2
            uint32_t cmpinst = ((fwval(fw, k+1) & 0x0000f000)<<4) + 0xe3500004; // cmp rx, #4
            int l;
            int m = 0;
            for (l=0; l<32; l++)
            {
                if ( fwval(fw, k+1+l) == cmpinst )
                    m++;
            }
            if (m != 2)
                return 0;
            // confirmed
            uint32_t fadr = followBranch(fw, idx2adr(fw, k), 0x01000001);
            fwAddMatch(fw,fadr,32,0,122);
            return 1;
        }
    }

    return 0;
}

// Special case for 'Restart'
int find_Restart(firmware *fw)
{
    int j = get_saved_sig(fw,"reboot_fw_update");
    if (j >= 0)
    {
        int k = get_saved_sig(fw,"StopWDT_FW");
        if (k >= 0)
        {
            j = adr2idx(fw, func_names[j].val);
            int i;
            for (i=j+1; i<j+100; i++)
            {
                if (isBL(fw,i) && isBL(fw,i+2))
                {
                    // find call to StopWDT_FW
                    uint32_t fadr = followBranch(fw, idx2adr(fw, i), 0x01000001);
                    if (func_names[k].val == fadr)
                    {
                        fadr = followBranch(fw, idx2adr(fw, i+2), 0x01000001);
                        fwAddMatch(fw,fadr,32,0,122);
                        return 1;
                    }
                }
            }
        }
    }

    return 0;
}

// Special case for 'add_ptp_handler'
int find_add_ptp_handler(firmware *fw, string_sig *sig, int k)
{
    uint32_t vals[] = { 0x9801, 0x9802, 0x9803, 0x9804, 0x9805, 0 };
    uint32_t fadr = 0;

    int i = 0;
    while ((vals[i] != 0) && isLDR_PC(fw,k) && (fwRd(fw,k) == 0) && (LDR2val(fw,k) == vals[i]))
    {
        k = find_inst(fw, isBL, k+1, 5);
        if (k == 0) return 0;
        if (fadr == 0)
            fadr = followBranch(fw, idx2adr(fw,k), 0x01000001);
        k = find_inst(fw, isLDR_PC, k+1, 5);
        if (k == 0) return 0;
        i++;
    }

    if (fadr != 0)
    {
        fwAddMatch(fw,fadr,32,0,121);
        return 1;
    }

    return 0;
}

// Special case for 'PT_PlaySound'
int find_PT_PlaySound(firmware *fw)
{
    int j, k;
    int k1 = get_saved_sig(fw,"LogCameraEvent");

    if (k1 >= 0)
    {
        j = find_str_ref(fw,"BufAccBeep");
        if (j >= 0)
        {
            k = find_inst(fw, isBL, j+1, 4);
            if (k >= 0)
            {
                uint32_t fadr = followBranch(fw, idx2adr(fw,k), 0x01000001);
                if (func_names[k1].val == fadr)
                {
                    k = find_inst(fw, isB, k+1, 10);
                    fadr = followBranch(fw, idx2adr(fw, k), 1);
                    fwAddMatch(fw,fadr,32,0,122);
                    return 1;
                }
            }
        }
    }

    return 0;
}

// Special case for 'getImageDirName'
int find_getImageDirName(firmware *fw)
{
    int k = find_str_ref(fw,"%3d_%02d%02d");
    if (k >= 0)
    {
        k = find_inst_rev(fw, isLDMFD_PC, k-1, 16);
        if (k >= 0)
        {
            uint32_t fadr = idx2adr(fw,k+1);
            fwAddMatch(fw,fadr,32,0,122);
            return 1;
        }
    }
    else
    {
        k = find_str_ref(fw,"___%02d");
        if (k >= 0)
        {
            k = find_inst_rev(fw, isLDMFD_PC, k-1, 18);
            if (k >= 0)
            {
                if (isMOV(fw,k+1) && isMOV(fw,k+2)) // sanity check
                {
                    uint32_t fadr = idx2adr(fw,k+1);
                    fwAddMatch(fw,fadr,32,0,122);
                    return 1;
                }
            }
        }
    }

    return 0;
}

// Special case for 'GetImageFolder'
uint32_t strGIF = 0;
int match_GetImageFolder(firmware *fw, int k, uint32_t a_getImageDirName, uint32_t a_TakeSemaphore)
{
    int k1, fnd;

    if (isBL(fw,k))
    {
        uint32_t fadr = followBranch2(fw,idx2adr(fw,k),0x01000001);
        if (fadr == a_getImageDirName)
        {
            int s = find_inst_rev(fw, isSTMFD_LR, k-1, 80);
            int e = find_inst(fw, isLDMFD_PC, k+1, 80);
            if ((s >= 0) && (e >= 0))
            {
                fnd = 0;
                for (k1=s+1; k1<k-1; k1++)
                {
                    if (isBL(fw,k1))
                    {
                        fadr = followBranch2(fw,idx2adr(fw,k1),0x01000001);
                        if (fadr == a_TakeSemaphore)
                        {
                            fnd++;
                            break;
                        }
                    }
                }
                if (fnd != 0)
                {
                    for (k1=k+1; k1<e-1; k1++)
                    {
                        if ((isLDR_PC(fw,k1) || isADR_PC(fw,k1)) && (idx2adr(fw,k1) == strGIF))
                        {
                            fnd--;
                            break;
                        }
                    }
                }
                if (fnd != 0)
                {
                    fwAddMatch(fw,idx2adr(fw,s),32,0,122);
                    return 1;
                }
            }
        }
    }
    return 0;
}
int find_GetImageFolder(firmware *fw)
{
    int j = find_str_ref(fw,"GetCameraObjectTmpPath ERROR[ID:%lx] [TRY:%lx]\n");
    if (j < 0)
        j = find_str_ref(fw,"_GetCameraObjectTmpPath ERROR[ID:%lx] [TRY:%lx]\n");
    if (j >= 0)
    {
        strGIF = idx2adr(fw,j);
        int j = get_saved_sig(fw,"TakeSemaphore");
        int k = get_saved_sig(fw,"getImageDirName");
        if ((k >= 0) && (j >= 0))
        {
            return search_fw(fw, match_GetImageFolder, func_names[k].val, func_names[j].val, 1);
        }
    }

    return 0;
}

// Special case for 'GetDrive_ClusterSize'
int match_GetDrive_ClusterSize(firmware *fw, int k, uint32_t v1, uint32_t v2)
{
    if (isBL_cond(fw,k))
    {
        uint32_t fadr = followBranch2(fw,idx2adr(fw,k),0xF1000001);
        if (fadr == v1)
        {
            int fnd = 0;
            if (isLDR_cond(fw,k-1) && idx_valid(fw,adr2idx(fw,LDR2val(fw,k-1))) && (strcmp(adr2ptr(fw,LDR2val(fw,k-1)),"Mounter.c") == 0))
            {
                fnd = 1;
            }
            else if (isLDR_cond(fw,k-2) && idx_valid(fw,adr2idx(fw,LDR2val(fw,k-2))) && (strcmp(adr2ptr(fw,LDR2val(fw,k-2)),"Mounter.c") == 0))
            {
                fnd = 1;
            }
            else if (isLDR_cond(fw,k-3) && idx_valid(fw,adr2idx(fw,LDR2val(fw,k-3))) && (strcmp(adr2ptr(fw,LDR2val(fw,k-3)),"Mounter.c") == 0))
            {
                fnd = 1;
            }
            else if (isADR_PC_cond(fw,k-1) && (strcmp(adr2ptr(fw,ADR2adr(fw,k-1)),"Mounter.c") == 0))
            {
                fnd = 1;
            }
            else if (isADR_PC_cond(fw,k-2) && (strcmp(adr2ptr(fw,ADR2adr(fw,k-2)),"Mounter.c") == 0))
            {
                fnd = 1;
            }
            if ((fnd == 1) &&
                isLDR_PC(fw,k+1) &&
                ((fwval(fw,k+2) & 0xFFF00FF0) == 0xE0800200) && ((fwval(fw,k+3) & 0xFFF00FF0) == 0xE0800100) &&
                (fwval(fw,k+4) == 0xE5901004) && (fwval(fw,k+5) == 0xE5900008) && (fwval(fw,k+6) == 0xE0000091) &&
                isLDMFD_PC(fw,k+7))
            {
                k = find_inst_rev(fw,isSTMFD_LR,k-1,8);
                if (k >= 0)
                {
                    if (fwval(fw,k-1) == 0xE3500001)    // CMP R0, #1
                        k--;
                    fwAddMatch(fw,idx2adr(fw,k),32,0,122);
                    return 1;
                }
            }
        }
    }

    return 0;
}
int find_GetDrive_ClusterSize(firmware *fw)
{
    int k = get_saved_sig(fw,"DebugAssert");
    if (k >= 0)
    {
        return search_fw(fw, match_GetDrive_ClusterSize, func_names[k].val, 0, 16);
    }

    return 0;
}

int find_GetDrive_TotalClusters(firmware *fw)
{
    extern uint32_t find_str_bytes(firmware *fw, char *str);
    // restrict usage (we rely on a function preceding a string)
    if (fw->dryos_ver < 52)
        return 0;
    uint32_t j = find_str_bytes(fw,"DriveLetterManager.c");
    if (j > 0)
    {
        int k = adr2idx(fw,j);
        k = find_inst_rev(fw,isLDMFD_PC,k-1,2);
        if ((k > 0) && ( (fwval(fw,k-1)&0xfffff0f0)==0xe0810090 )) // umull r0, r1, rx, ry
        {
            if (isBL(fw,k-2))
            {
                k = idxFollowBranch(fw,k-2,0x01000001);
                fwAddMatch(fw,idx2adr(fw,k),32,0,122);
                return 1;
            }
        }
    }

    return 0;
}

// Special case for srand
int find_srand(firmware *fw)
{
    int k = get_saved_sig(fw,"rand");
    if (k >= 0)
    {
        k = adr2idx(fw, func_names[k].val) - 3;
        if (isLDR_PC(fw,k) && isSTR(fw,k+1) && isBX_LR(fw,k+2))
            fwAddMatch(fw,idx2adr(fw,k),32,0,122);
    }

    return 0;
}

int find_malloc_strictly(firmware *fw)
{

    int s1 = find_str(fw,"Size: %ld");
    int s2 = find_str(fw,"Memory.c");
    int f1 = get_saved_sig(fw,"malloc");
    if ((s1 < 0)||(s2 < 0)||(f1 < 0))
        return 0;
    if (s1 < s2-16) // the two strings should be close
        s1 = find_Nth_str(fw,"Size: %ld",2); // this string has multiple instances, try the next one
    f1 = adr2idx(fw, func_names[f1].val);
    
    int r1 = find_nxt_str_ref(fw, s1, 0);
    int r2 = find_nxt_str_ref(fw, s2, 0);
    int l1 = 0;
    while((r1>0) && (r2>0) && (l1<2))
    {
        if (r2 == r1 + 3)
        {
            int m1 = find_inst_rev(fw,isBL,r1,6);
            if (m1 > 0)
            {
                int m2 = idxFollowBranch(fw,m1,0x01000001);
                if (m2 == f1)
                {
                    m1 = find_inst_rev(fw,isSTMFD_LR,m1,3);
                    if (m1 > 0)
                    {
                        fwAddMatch(fw,idx2adr(fw,m1),32,0,122);
                        return 1;
                    }
                }
            }
        }
        r1 = find_nxt_str_ref(fw, s1, r1+1);
        r2 = find_nxt_str_ref(fw, s2, r2+1);
        l1++;
    }
    return 0;
}

static int idx_createdialogbox=-1, idx_displaydialogbox=-1, idx_adduitodialog=-1, idx_getstring=-1;

int find_DisplayBusyOnScreen(firmware *fw)
{

    int s1 = find_str(fw,"ErrorMessageController.c");
    int s2 = find_str(fw,"StrMan.c");
    if (s1 < 0)
        s1 = find_str(fw,"MessageController.c");
    int j = find_str_ref(fw,"_PBBusyScrn");
    if (j < 0)
        j = find_str_ref(fw,"_PlayBusyScreen");

    if ((j>=0)&&(s1>=0)&&(s2>=0))
    {
        int m1 = find_Nth_inst(fw,isBL,j+1,12,fw->dryos_ver<54?4:3);
        int m2, k;

        if (fw->dryos_ver > 57)
        {
            // 1st B after str ref branches to DisplayBusyOnScreen
            m1 = find_inst(fw,isB,j+1,12);
            if (m1 > 0)
            {
                m2 = idxFollowBranch(fw,m1,0x00000001);
                k = find_nxt_str_ref(fw, s1, m2);
            }
            else
            {
                return 0;
            }
        }
        else if (fw->dryos_ver > 56)
        {
            // these functions are called indirectly in this part of fw on r57+
            int found = 0;
            for (k=-1; k>-3; k--)
            {
                if ((fwval(fw,m1+k) & 0xFE1FF000) == 0xE41F0000) // ldr r0, =func
                {
                    uint32_t u1 = LDR2val(fw, m1+k);
                    if ( u1 > fw->base )
                    {
                        if (isSTMFD_LR(fw, adr2idx(fw, u1)))
                        {
                            found = 1;
                            m2 = adr2idx(fw, u1);
                            k = find_nxt_str_ref(fw, s1, m2);
                            break;
                        }
                    }
                }
            }
            if (!found)
            {
                return 0;
            }
        }
        else
        {
            m2 = idxFollowBranch(fw,m1,0x01000001);
            k = find_nxt_str_ref(fw, s1, m2);
            if ((k <= 0)||(k-m2 >= 22))
            {
                // not found, try the next BL
                m1 = find_inst(fw,isBL,m1+1,4);
                m2 = idxFollowBranch(fw,m1,0x01000001);
                k = find_nxt_str_ref(fw, s1, m2);
            }
        }


        if ((k > 0)&&(k-m2 < 22))
        {
            // the string is referenced near enough
            idx_createdialogbox = find_inst_rev(fw, isBL, k-1, 4);
            // DisplayBusyOnScreen is found
            fwAddMatch(fw,idx2adr(fw,m2),32,0,122);
            // find additional stuff
            idx_adduitodialog = find_inst(fw, isBL, k+1, 7);
            int m3;
            for (m3=k; m3<k+128; m3++)
            {
                // mov r0, #imm or ldr r0, [pc,... between two BLs
                if (isBL(fw,m3)&&isBL(fw,m3+2)&&
                    (((fwval(fw,m3+1)&0xfffff000)==0xe3a00000)||((fwval(fw,m3+1)&0xff7ff000)==0xe51f0000)))
                {
                    // further check
                    int m30 = m3+2;
                    int m4 = idxFollowBranch(fw,m30,0x01000001);
                    if (m4 > 0)
                    {
                        // search for ref. to assert string
                        int m5 = find_inst(fw, isLDMFD_PC, m4+1, 64);
                        int m6 = find_nxt_str_ref(fw, s2, m4);
                        if ((m6 > 0)&&(m6 < m5))
                        {
                            idx_getstring = m30;
                            break;
                        }
                        // no string, search for reference to 0x00000020 (string consisting of a single space char)
                        m5 = find_inst(fw, isADR_PC, m4+1, 10);
                        if (m5>0)
                        {
                            uint32_t u1 = ADR2adr(fw, m5);
                            if (fwval(fw, adr2idx(fw, u1)) == 0x00000020)
                            {
                                idx_getstring = m30;
                                break;
                            }
                        }
                    }
                }
            }
            if (fw->dryos_ver < 54)
            {
                m3 = find_inst(fw, isLDMFD_PC, k+30, 64);
                if (m3>0)
                {
                    m3 = find_Nth_inst_rev(fw, isBL, m3-1, 8, 2);
                    if (m3>0)
                    {
                        idx_displaydialogbox = m3;
                    }
                }
            }
            else
            {
                m3 = find_inst(fw, isLDMFD, k+30, 20);
                if (m3>0)
                {
                    m3 = find_inst_rev(fw, isBL, m3-1, 4);
                    if (m3>0)
                    {
                        idx_displaydialogbox = m3;
                    }
                }
            }
            return 1;
        }
    }

    return 0;
}

int find_UndisplayBusyOnScreen(firmware *fw)
{
    if (get_saved_sig(fw,"DisplayBusyOnScreen") < 0) return 0;
    int j;
    if (fw->dryos_ver > 57)
    {
        j = find_str_ref(fw,"_PBBusyScrnToCtrlSrvTask");
    }
    else
    {
        j = find_str_ref(fw,"_PBBusyScrn");
    }
    if (j < 0)
        j = find_str_ref(fw,"_PlayBusyScreen");

    if (fw->dryos_ver < 57)
    {
        if (idx_createdialogbox > 0)
        {
            int m;
            for (m=0; m<2; m++)
            {
                int n = find_Nth_inst(fw, isSTMFD_LR, idx_createdialogbox + 30, 140, m+1);
                if (n>0)
                {
                    uint32_t a1 = idx2adr(fw,n);
                    if (j > 0)
                    {
                        int k;
                        for (k=j; k<j+24; k++)
                        {
                            if (isBL_cond(fw,k)&&(idx2adr(fw,idxFollowBranch(fw,k,0xe1000001))==a1)) // BLEQ
                            {
                                fwAddMatch(fw,a1,32,0,122);
                                return 1;
                            }
                        }
                    }
                }
            }
        }
    }
    else
    {
        int m1 = find_Nth_inst(fw,isBLEQ,j+1,20,1);
        if (m1 > 0)
        {
            // these functions are called indirectly in this part of fw on r57+
            int k;
            for (k=-1; k>-3; k--)
            {
                if ((fwval(fw,m1+k) & 0xFE1FF000) == 0x041F0000) // ldreq r0, =func
                {
                    uint32_t u1 = LDR2val(fw, m1+k);
                    if ( u1 > fw->base )
                    {
                        if (isSTMFD_LR(fw, adr2idx(fw, u1)))
                        {
                            fwAddMatch(fw,u1,32,0,122);
                            return 1;
                        }
                    }
                }
            }
        }
    }
    return 0;
}
// TODO: redundant functions follow
int find_CreateDialogBox(firmware *fw)
{
    if (get_saved_sig(fw,"DisplayBusyOnScreen") < 0) return 0;
    if (idx_createdialogbox > 0)
    {
        int n = idxFollowBranch(fw,idx_createdialogbox,0x01000001);
        if (n>0)
        {
            fwAddMatch(fw,idx2adr(fw,n),32,0,122);
            return 1;
        }
    }
    return 0;
}
int find_DisplayDialogBox(firmware *fw)
{
    if (get_saved_sig(fw,"DisplayBusyOnScreen") < 0) return 0;
    if (idx_displaydialogbox > 0)
    {
        int n = idxFollowBranch(fw,idx_displaydialogbox,0x01000001);
        if (n>0)
        {
            fwAddMatch(fw,idx2adr(fw,n),32,0,122);
            return 1;
        }
    }
    return 0;
}
int find_add_ui_to_dialog(firmware *fw)
{
    if (get_saved_sig(fw,"DisplayBusyOnScreen") < 0) return 0;
    if (idx_adduitodialog > 0)
    {
        int n = idxFollowBranch(fw,idx_adduitodialog,0x01000001);
        if (n>0)
        {
            fwAddMatch(fw,idx2adr(fw,n),32,0,122);
            return 1;
        }
    }
    return 0;
}
int find_get_string_by_id(firmware *fw)
{
    if (get_saved_sig(fw,"DisplayBusyOnScreen") < 0) return 0;
    if (idx_getstring > 0)
    {
        int n = idxFollowBranch(fw,idx_getstring,0x01000001);
        if (n>0)
        {
            fwAddMatch(fw,idx2adr(fw,n),32,0,122);
            return 1;
        }
    }
    return 0;
}

int find_get_self_task_errno_pointer(firmware *fw)
{
    int f1 = get_saved_sig(fw,"malloc");
    int f2 = get_saved_sig(fw,"close");
    if ((f1<0) && (f2<0))
        return 0;
    f1 = adr2idx(fw, func_names[f1].val);
    f1 = find_inst(fw, isLDMFD_PC, f1, 24);
    if (f1>0)
    {
        f1 = find_inst_rev(fw, isBL, f1, 6);
        if (f1>0)
        {
            if (fwval(fw,f1+2) == 0xe5801000) // str r1, [r0]
            {
                f1 = idxFollowBranch(fw,f1,0x01000001);
                fwAddMatch(fw,idx2adr(fw,f1),32,0,122);
                return 1;
            }
        }
    }
    // older cams don't set errno on malloc failure
    f1 = adr2idx(fw, func_names[f2].val);
    f1 = find_Nth_inst(fw, isBL, f1, 8, 2); // second BL
    if (f1>0)
    {
        if (fwval(fw,f1+2) == 0xe5801000) // str r1, [r0]
        {
            f1 = idxFollowBranch(fw,f1,0x01000001);
            fwAddMatch(fw,idx2adr(fw,f1),32,0,122);
            return 1;
        }
    }
    return 0;
}

int find_getcurrentmachinetime(firmware *fw)
{
    int f1 = get_saved_sig(fw,"SetHPTimerAfterTimeout");
    if (f1 < 0)
        return 0;
    f1 = adr2idx(fw, func_names[f1].val);
    f1 = find_inst(fw, isBL, f1, 16);
    if (f1>0)
    {
        f1 = idxFollowBranch(fw,f1,0x01000001);
        fwAddMatch(fw,idx2adr(fw,f1),32,0,122);
        return 1;
    }
    return 0;
}

// for cases where method 15 fails (sx170)
int find_sethptimeraftertimeout(firmware *fw)
{
    int sadr = find_str(fw, "FrameRateGenerator.c");
    int j = find_nxt_str_ref(fw, sadr, -1);
    if (j < 0)
        return 0;
    int f1, f2, n;
    for (n=0; n<2; n++)
    {
        f1 = find_inst_rev(fw, isBL, j-1, 7);
        f2 = find_Nth_inst_rev(fw, isBL, j-1, 128, 2);
        // check whether previous BL is too close
        if ((f1 < 1) || (f1-f2<8))
        {
            j = find_nxt_str_ref(fw, sadr, j+1);
            if (j < 0)
                return 0;
        }
        else
        {
            f1 = idxFollowBranch(fw,f1,0x01000001);
            fwAddMatch(fw,idx2adr(fw,f1),32,0,122);
            return 1;
        }
    }
    return 0;
}

//------------------------------------------------------------------------------------------------------------

// Data for matching the '_log' function
uint32_t log_test[] = {
    0x1526E50E, 0x3FDBCB7B, 0
};

// Data for matching 'Fut' functions
uint32_t DeleteDirectory_Fut_test[] = { 0x09400017 };
uint32_t MakeDirectory_Fut_test[]   = { 0x09400015 };
uint32_t RenameFile_Fut_test[]      = { 0x09400013 };

//------------------------------------------------------------------------------------------------------------

// Signature matching data
string_sig string_sigs[] =
{
    // Same as previously found eventproc - do these first
    {20, "AllocateMemory", "AllocateMemory_FW", 1 },
    {20, "Close", "Close_FW", 1 },
    {20, "CreateCountingSemaphore", "CreateCountingSemaphore_FW", 1 },
    {20, "CreateTask", "CreateTask_FW", 1 },
    {20, "DeleteSemaphore", "DeleteSemaphore_FW", 1 },
    {20, "DispCon_ShowBitmapColorBar", "DispCon_ShowBitmapColorBar_FW", 1 },
    {20, "ExitTask", "ExitTask_FW", 1 },
    {20, "Fclose_Fut", "Fclose_Fut_FW", 1 },
    {20, "Fopen_Fut", "Fopen_Fut_FW", 1 },
    {20, "Fread_Fut", "Fread_Fut_FW", 1 },
    {20, "FreeMemory", "FreeMemory_FW", 1 },
    {20, "Fseek_Fut", "Fseek_Fut_FW", 1 },
    {20, "Fwrite_Fut", "Fwrite_Fut_FW", 1 },
    {20, "GetSDProtect", "GetSDProtect_FW", 1 },
    {20, "GetSystemTime", "GetSystemTime_FW", 1 },
    {20, "GetCurrentAvValue", "GetCurrentAvValue_FW", 1 },
    {20, "GetOpticalTemperature", "GetOpticalTemperature_FW", 1 },
    {20, "GetVRAMHPixelsSize", "GetVRAMHPixelsSize_FW", 1 },
    {20, "GetVRAMVPixelsSize", "GetVRAMVPixelsSize_FW", 1 },
    {20, "GetZoomLensCurrentPoint", "GetZoomLensCurrentPoint_FW", 1 },
    {20, "GiveSemaphore", "GiveSemaphore_FW", 1 },
    {20, "GUISrv_StartGUISystem", "GUISrv_StartGUISystem_FW", 1 },
    {20, "LEDDrive", "LEDDrive_FW", 1 },
    {20, "LockMainPower", "LockMainPower_FW", 1 },
    {20, "lseek", "Lseek_FW", 1 },
    {20, "Lseek", "Lseek_FW", 1 },
    {20, "MoveIrisWithAv", "MoveIrisWithAv_FW", 1 },
    {20, "MoveZoomLensWithPoint", "MoveZoomLensWithPoint_FW", 1 },
    {20, "memcmp", "memcmp_FW", 1 },
    {20, "memcpy", "memcpy_FW", 1 },
    {20, "memset", "memset_FW", 1 },
    {20, "NewTaskShell", "NewTaskShell_FW", 1 },
    {20, "NR_GetDarkSubType", "NR_GetDarkSubType_FW", 1 },
    {20, "NR_SetDarkSubType", "NR_SetDarkSubType_FW", 1 },
    {20, "Open", "Open_FW", 1 },
    {20, "PostLogicalEventForNotPowerType", "PostLogicalEventForNotPowerType_FW", 1 },
    {20, "PostLogicalEventToUI", "PostLogicalEventToUI_FW", 1 },
    {20, "PT_MoveDigitalZoomToWide", "PT_MoveDigitalZoomToWide_FW", 1 },
    {20, "PT_MoveOpticalZoomAt", "PT_MoveOpticalZoomAt_FW", 1 },
    {20, "Read", "Read_FW", 1 },
    {20, "RefreshPhysicalScreen", "RefreshPhysicalScreen_FW", 1 },
    {20, "ResetFocusLens", "ResetFocusLens_FW", 1 },
    {20, "ResetZoomLens", "ResetZoomLens_FW", 1 },
    {20, "SavePaletteData", "SavePaletteData_FW", 1 },
    {20, "SetAutoShutdownTime", "SetAutoShutdownTime_FW", 1 },
    {20, "SetCurrentCaptureModeType", "SetCurrentCaptureModeType_FW", 1 },
    {20, "SetScriptMode", "SetScriptMode_FW", 1 },
    {20, "SleepTask", "SleepTask_FW", 1 },
    {20, "strcmp", "j_strcmp_FW", 0 },
    {20, "strcmp", "strcmp_FW", 0 },
    {20, "strcpy", "strcpy_FW", 1 },
    {20, "strlen", "strlen_FW", 1 },
    {20, "StartRecModeMenu", "StartRecModeMenu_FW", 1 },
    {20, "TakeSemaphore", "TakeSemaphore_FW", 1 },
    {20, "UIFS_WriteFirmInfoToFile", "UIFS_WriteFirmInfoToFile_FW", 1 },
    {20, "UnlockMainPower", "UnlockMainPower_FW", 1 },
    {20, "VbattGet", "VbattGet_FW", 1 },
    //{20, "write", "Write_FW", 1 },
    {20, "Write", "Write_FW", 1 },
    {20, "task_CaptSeq", "task_CaptSeqTask", 1 },
    {20, "task_ExpDrv", "task_ExpDrvTask", 1 },
    {20, "task_FileWrite", "task_FileWriteTask", 1 },
    {20, "task_RotaryEncoder", "task_JogDial", 1 },
    {20, "task_RotaryEncoder", "task_RotarySw", 1 },
    {20, "TurnOnDisplay", "DispCon_TurnOnDisplay_FW", 0 },
    {20, "TurnOffDisplay", "DispCon_TurnOffDisplay_FW", 0 },
    {20, "DoAELock", "PT_DoAELock_FW", 0x01000002 },
    {20, "DoAELock", "SS.DoAELock_FW", 0x01000002 },
    {20, "DoAFLock", "PT_DoAFLock_FW", 0x01000002 },
    {20, "DoAFLock", "SS.DoAFLock_FW", 0x01000002 },
    {20, "UnlockAE", "PT_UnlockAE_FW", 0x01000002 },
    {20, "UnlockAE", "SS.UnlockAE_FW", 0x01000002 },
    {20, "UnlockAF", "PT_UnlockAF_FW", 0x01000002 },
    {20, "UnlockAF", "SS.UnlockAF_FW", 0x01000002 },
    {20, "MFOn", "MFOn_FW", 1 },
    {20, "MFOff", "MFOff_FW", 1 },
    {20, "PT_MFOn", "PT_MFOn_FW", 1 },
    {20, "PT_MFOff", "PT_MFOff_FW", 1 },
    {20, "SS_MFOn", "SS.MFOn_FW", 1 },
    {20, "SS_MFOff", "SS.MFOff_FW", 1 },
    {20, "SetLogicalEventActive", "UiEvnt_SetLogicalEventActive_FW", 1 },
    {20, "GetAdChValue", "GetAdChValue_FW", 0 },
    {20, "CalcLog10", "CalcLog10_FW", 4 },
    {20, "HwOcReadICAPCounter", "GetCurrentMachineTime", 1 },
    {20, "DisableISDriveError", "DisableISDriveError_FW", 1},

    { 1, "ExportToEventProcedure_FW", "ExportToEventProcedure", 1 },
    { 1, "AllocateMemory", "AllocateMemory", 1 },
    { 1, "Close", "Close", 1 },
    { 1, "CreateTask", "CreateTask", 1 },
    //{ 1, "DoAFLock", "PT_DoAFLock", 0x01000002 },
    { 1, "ExitTask", "ExitTask", 1 },
    { 1, "exmem_alloc", "ExMem.AllocCacheable", 4 },
    { 1, "exmem_free", "ExMem.FreeCacheable", 0x01000003 },
    { 1, "Fclose_Fut", "Fclose_Fut", 1 },
    { 1, "Feof_Fut", "Feof_Fut", 1 },
    { 1, "Fflush_Fut", "Fflush_Fut", 1 },
    { 1, "Fgets_Fut", "Fgets_Fut", 1 },
    { 1, "Fopen_Fut", "Fopen_Fut", 1 },
    { 1, "Fread_Fut", "Fread_Fut", 1 },
    { 1, "FreeMemory", "FreeMemory", 1 },
    { 1, "Fseek_Fut", "Fseek_Fut", 1 },
    { 1, "Fwrite_Fut", "Fwrite_Fut", 1 },
    { 1, "GetParameterData", "PTM_RestoreUIProperty", 0xF0000004 },
    { 1, "GetPropertyCase", "PT_GetPropertyCaseString", 1 },
    { 1, "GetPropertyCase", "PT_GetPropertyCaseInt", 0x0100000F },
    { 1, "GetPropertyCase", "GetPropertyCase", 0x0100000F },
    { 1, "GetSDProtect", "GetSDProtect", 1 },
    { 1, "GetSystemTime", "GetSystemTime", 1 },
    { 1, "LEDDrive", "LEDDrive", 1 },
    { 1, "LockMainPower", "LockMainPower", 1 },
    { 1, "Lseek", "Lseek", 1 },
    { 1, "lseek", "Lseek", 1 },
    { 1, "memcpy", "memcpy", 1 },
    { 1, "memcmp", "memcmp", 1 },
    { 1, "memset", "memset", 1 },
    { 1, "NewTaskShell", "NewTaskShell", 1 },
    { 1, "Open", "Open", 1 },
    { 1, "PostLogicalEventToUI", "PostLogicalEventToUI", 1 },
    { 1, "PostLogicalEventForNotPowerType", "PostLogicalEventForNotPowerType", 1 },
    //{ 1, "Read", "Read", 1 },
    //{ 1, "read", "Read", 1 },
    { 1, "RefreshPhysicalScreen", "RefreshPhysicalScreen", 1 },
    { 1, "SetAutoShutdownTime", "SetAutoShutdownTime", 1 },
    { 1, "SetCurrentCaptureModeType", "SetCurrentCaptureModeType", 1 },
    { 1, "SetLogicalEventActive", "UiEvnt_SetLogicalEventActive", 1 },
    { 1, "SetParameterData", "PTM_BackupUIProperty", 1 },
    { 1, "SetPropertyCase", "PT_SetPropertyCaseInt", 0x01000003 },
    { 1, "SetPropertyCase", "SetPropertyCase", 0x01000004 },
    { 1, "SetScriptMode", "SetScriptMode", 1 },
    { 1, "SleepTask", "SleepTask", 1 },
    { 1, "strcmp", "strcmp", 0 },
    { 1, "strcpy", "strcpy", 1 },
    { 1, "strlen", "strlen", 1 },
    { 1, "strtol", "atol", 3 },
    { 1, "TakeSemaphore", "TakeSemaphore", 1 },
    { 1, "UIFS_WriteFirmInfoToFile", "UIFS_WriteFirmInfoToFile", 1 },
    //{ 1, "UnlockAF", "PT_UnlockAF", 0x01000002 },
    { 1, "UnlockMainPower", "UnlockMainPower", 1 },
    { 1, "VbattGet", "VbattGet", 1 },
    //{ 1, "Write", "Write", 1 },
    //{ 1, "write", "Write", 1 },
    { 1, "GUISrv_StartGUISystem", "GUISrv_StartGUISystem", 1 },

    { 2, "GetBatteryTemperature", "GetBatteryTemperature", 1 },
    { 2, "GetCCDTemperature", "GetCCDTemperature", 1 },
    { 2, "GetOpticalTemperature", "GetOpticalTemperature", 1 },
    { 2, "GetFocusLensSubjectDistance", "GetCurrentTargetDistance", 1 },
    { 2, "GetZoomLensCurrentPoint", "GetZoomLensCurrentPoint", 1 },
    { 2, "GetZoomLensCurrentPosition", "GetZoomLensCurrentPosition", 1 },
    { 2, "MoveFocusLensToDistance", "MoveFocusLensToDistance", 1 },
    { 2, "MoveZoomLensWithPoint", "MoveZoomLensWithPoint", 1 },
    { 2, "GetCurrentAvValue", "GetCurrentAvValue", 1 },
    { 2, "PT_MoveOpticalZoomAt", "PT_MoveOpticalZoomAt", 1 },
    { 2, "PT_MoveOpticalZoomAt", "SS.MoveOpticalZoomAt", 1 },
    { 2, "PT_MoveDigitalZoomToWide", "PT_MoveDigitalZoomToWide", 1 },
    { 2, "PT_MoveDigitalZoomToWide", "SS.MoveDigitalZoomToWide", 1 },
    { 2, "MoveIrisWithAv", "MoveIrisWithAv", 1},
    { 2, "PutInNdFilter", "TurnOnNdFilter", 1 },
    { 2, "PutOutNdFilter", "TurnOffNdFilter", 1 },
    { 2, "PutInNdFilter", "PutInNdFilter", 1 },
    { 2, "PutOutNdFilter", "PutOutNdFilter", 1 },
    { 2, "IsStrobeChargeCompleted", "EF.IsChargeFull", 1 },
    { 2, "GetPropertyCase", "PT_GetPropertyCaseInt", 0x01000012 },
    { 2, "SetPropertyCase", "PT_SetPropertyCaseInt", 0x01000008 },
    { 2, "SetPropertyCase", "PT_SetPropertyCaseInt", 0x01000009 },
    //{ 2, "UnlockAF", "PT_UnlockAF", 0x01000002 },
    //{ 2, "UnlockAF", "SS.UnlockAF", 0x01000002 },
    //{ 2, "DoAFLock", "PT_DoAFLock", 0x01000002 },
    //{ 2, "DoAFLock", "SS.DoAFLock", 0x01000002 },
    { 2, "GetSystemTime", "PT_GetSystemTime", 0x01000003 },
    { 2, "PT_PlaySound", "PT_PlaySound", 0x01000005 },
    { 2, "StartRecModeMenu", "StartRecModeMenu", 1 },
    { 2, "GetSDProtect", "GetSDProtect", 1 },
    { 2, "DispCon_ShowBitmapColorBar", "DispCon_ShowBitmapColorBar", 1 },
    { 2, "SetAE_ShutterSpeed", "SetAE_ShutterSpeed", 1 },
    { 2, "ResetZoomLens", "ResetZoomLens", 1 },
    { 2, "ResetFocusLens", "ResetFocusLens", 1 },
    { 2, "NR_GetDarkSubType", "NR_GetDarkSubType", 1 },
    { 2, "NR_GetDarkSubType", "NRTBL.GetDarkSubType", 1 },
    { 2, "NR_SetDarkSubType", "NR_SetDarkSubType", 1 },
    { 2, "NR_SetDarkSubType", "NRTBL.SetDarkSubType", 1 },
    { 2, "SavePaletteData", "SavePaletteData", 1 },
    { 2, "GetVRAMHPixelsSize", "GetVRAMHPixelsSize", 1 },
    { 2, "GetVRAMVPixelsSize", "GetVRAMVPixelsSize", 1 },
    { 2, "EngDrvIn", "EngDrvIn", 2 },
    { 2, "EngDrvOut", "EngDrvOut", 0x01000005 },
    { 2, "EngDrvRead", "EngDrvRead", 2 },
    { 2, "EngDrvBits", "EngDrvBits", 0x01000007 },
    { 2, "EngDrvBits", "EngDrvBits", 0x01000005 },
    { 2, "exmem_alloc", "ExMem.AllocCacheable", 4 },
    { 2, "exmem_free", "ExMem.FreeCacheable", 0x01000003 },

    { 2, "PTM_GetCurrentItem", "PTM_GetCurrentItem", 2 }, // s5is
    { 2, "PTM_SetCurrentItem", "PTM_SetCurrentItem", 4 }, // s5is
    { 2, "PTM_NextItem", "PTM_NextItem", 2 }, // s5is
    { 2, "PTM_PrevItem", "PTM_PrevItem", 2 }, // s5is
    { 2, "PTM_SetPropertyEnable", "PTM_SetProprietyEnable", 4 }, // s5is

    { 3, "AllocateMemory", "AllocateMemory", 1 },
    { 3, "FreeMemory", "FreeMemory", 1 },
    { 3, "PostLogicalEventToUI", "PostLogicalEventToUI", 1 },
    { 3, "PostLogicalEventForNotPowerType", "PostLogicalEventForNotPowerType", 1 },
    { 3, "LockMainPower", "LockMainPower", 1 },
    { 3, "UnlockMainPower", "UnlockMainPower", 1 },
    { 3, "SetAutoShutdownTime", "SetAutoShutdownTime", 1 },
    { 3, "NewTaskShell", "NewTaskShell", 1 },
    { 3, "VbattGet", "VbattGet", 1 },
    { 3, "LEDDrive", "LEDDrive", 1 },
    { 3, "SetPropertyCase", "PT_SetPropertyCaseInt", 0x01000003 },
    //{ 3, "UnlockAF", "PT_UnlockAF", 0x01000002 },
    //{ 3, "DoAFLock", "PT_DoAFLock", 0x01000002 },
    { 3, "UIFS_WriteFirmInfoToFile", "UIFS_WriteFirmInfoToFile", 1 },
    { 3, "PT_MoveOpticalZoomAt", "PT_MoveOpticalZoomAt", 1 },
    { 3, "PT_MoveDigitalZoomToWide", "PT_MoveDigitalZoomToWide", 1 },
    { 3, "PT_PlaySound", "PT_PlaySound", 1 },
    { 3, "exmem_alloc", "ExMem.AllocCacheable", 4 },
    { 3, "exmem_free", "ExMem.FreeCacheable", 0x01000003 },
    { 3, "GetSDProtect", "GetSDProtect", 1 },

    { 4, "TurnOnBackLight", "TurnOnBackLight", 1 },
    { 4, "TurnOffBackLight", "TurnOffBackLight", 1 },
    { 4, "EnterToCompensationEVF", "SSAPI::EnterToCompensationEVF", 1 },
    { 4, "EnterToCompensationEVF", "ExpComp On", 1 },
    { 4, "EnterToCompensationEVF", "ExpOn", 1 },
    { 4, "ExitFromCompensationEVF", "SSAPI::ExitFromCompensationEVF", 1 },
    { 4, "ExitFromCompensationEVF", "ExpComp Off", 1 },
    { 4, "ExitFromCompensationEVF", "ExpOff", 1 },
    { 4, "PB2Rec", "AC:PB2Rec", 1 },
    { 4, "PB2Rec", "AC:PB2Rec", 6 },
    { 4, "PB2Rec", "AC:PB2Rec", 9 },
    { 4, "PB2Rec", "AC:PB2Rec", 11 },
    { 4, "Rec2PB", "AC:Rec2PB", 1 },
    { 4, "Rec2PB", "AC:Rec2PB", 2 },
    { 4, "RefreshPhysicalScreen", "ScreenUnLock", 1 },
    { 4, "RefreshPhysicalScreen", "ScreenUnLock", 7 },
    { 4, "RefreshPhysicalScreen", "ScreenUnLock", 15 },
    { 4, "RefreshPhysicalScreen", "Reduce ScreenUnLock", 5 },
    { 4, "RefreshPhysicalScreen", "Window:IneffectiveLockPhysicalScreen", 8 },
    { 4, "UnsetZoomForMovie", "ZoomCon_UnsetZoomForMovie", 1 },
    { 4, "ExpCtrlTool_StopContiAE", "StopContiAE", 9 },
    { 4, "ExpCtrlTool_StopContiAE", "StopContiAE", 10 },
    { 4, "ExpCtrlTool_StopContiAE", "StopContiAE", 11 },
    { 4, "ExpCtrlTool_StartContiAE", "StartContiAE", 9 },
    { 4, "ExpCtrlTool_StartContiAE", "StartContiAE", 10 },
    { 4, "ExpCtrlTool_StartContiAE", "StartContiAE", 11 },
    { 4, "ExecuteEventProcedure", "Can not Execute ", 14 },

    { 5, "UIFS_WriteFirmInfoToFile", "UIFS_WriteFirmInfoToFile", 1 },
    { 5, "CreateTask", "CreateTask", 1 },
    { 5, "hook_CreateTask", "CreateTask", 1 },
    { 5, "ExitTask", "ExitTask", 1 },
    { 5, "SleepTask", "SleepTask", 1 },
    { 5, "DeleteSemaphore", "DeleteSemaphore", 1 },
    { 5, "CreateCountingSemaphore", "CreateCountingSemaphore", 1 },
    //                                                                   R20   R23   R31   R39   R43   R45   R47   R49   R50   R51   R52   R54   R55   R57   R58
    { 5, "UpdateMBROnFlash", "MakeBootDisk", 0x01000003,                  11,   11,   11,   11,   11,   11,    1,    1,    1,    1,    1,    1,    1,    1,    1 },
    { 5, "MakeSDCardBootable", "MakeBootDisk", 0x01000003,                 1,    1,    1,    1,    1,    1,    8,    8,    8,    8,    8,    9,    9,    9,    9 },

    { 5, "PTM_GetCurrentItem", "PTM_GetCurrentItem", 1 },
    { 5, "PTM_SetCurrentItem", "PTM_SetCurrentItem", 1 },
    { 5, "PTM_NextItem", "PTM_NextItem", 1 },
    { 5, "PTM_PrevItem", "PTM_PrevItem", 1 },
    { 5, "PTM_SetPropertyEnable", "PTM_SetProprietyEnable", 1 },

    //{ 6, "Restart", "Bye", 0 },
    { 6, "reboot_fw_update", "FirmUpgrade.c", 0 },

    { 7, "CreateTaskStrictly", "PhySw", 0x01000001 },
    { 7, "RegisterInterruptHandler", "WdtInt", 0x01000001 },
    { 7, "LogCameraEvent", "BufAccBeep", 0x01000001 },
    { 7, "LogCameraEvent", "MyCamFunc_PlaySound_MYCAM_COVER_OPEN", 0x01000001 },
    { 7, "DebugAssert", "Console.c", 0x01000001 },

    { 8, "WriteSDCard", "Mounter.c", 0 },

    // Ensure ordering in func_names is correct for dependencies here
    //                                                                   R20   R23   R31   R39   R43   R45   R47   R49   R50   R51   R52   R54   R55   R57   R58
    { 9, "kbd_p1_f", "task_PhySw", 0,                                      5,    5,    5,    5,    5,    5,    5,    5,    5,    5,    5,    5,    5,    5,    5 },
    { 9, "kbd_p2_f", "task_PhySw", 0,                                      7,    7,    7,    7,    7,    7,    7,    7,    7,    7,    7,    7,    7,    7,    7 },
    { 9, "kbd_read_keys", "kbd_p1_f", 0,                                   2,    2,    2,    2,    2,    2,    2,    2,    2,    2,    2,    2,    2,    2,    2 },
    { 9, "kbd_p1_f_cont", "kbd_p1_f", -1,                                  3,    3,    3,    3,    3,    3,    3,    3,    3,    3,    3,    3,    3,    3,    3 },
    { 9, "kbd_read_keys_r2", "kbd_read_keys", 0,                          11,   11,   11,   11,   11,   11,   11,   11,   11,   11,   11,   11,   11,   11,   11 },
    { 9, "GetKbdState", "kbd_read_keys", 0,                                8,    8,    8,    8,    8,    8,    8,    8,    8,    8,    8,    8,    8,    8,    8 },
    { 9, "GetKbdState", "kbd_read_keys", 0,                                9,    9,    9,    9,    9,    9,    9,    9,    9,    9,    9,    9,    9,    9,    9 },
    { 9, "strtolx", "strtol", 0,                                           1,    1,    1,    1,    1,    1,    1,    1,    1,    1,    1,    1,    1,    1,    1 },
    { 9, "mkdir", "MakeDirectory_Fut", 0x01000001,                        17,   17,   17,   17,   17,   17,   17,   17,   17,   17,   17,   17,   17,   17,   17 },
    { 9, "mkdir", "MakeDirectory_Fut", 0x01000002,                        17,   17,   17,   17,   17,   17,   17,   17,   17,   17,   17,   17,   17,   17,   17 },
    { 9, "time", "MakeDirectory_Fut", 0,                                  12,   12,   12,   12,   12,   12,   12,   12,   12,   12,   12,   12,   12,   12,   12 },
    { 9, "stat", "_uartr_req", 0,                                          0,    0,    0,    4,    4,    4,    4,    4,    4,    4,    4,    4,    4,    4,    4 },
    { 9, "PostMessageQueue", "PostMessageQueueStrictly", 0,                3,    3,    3,    3,    3,    3,    3,    3,    3,    3,    3,    3,    3,    3,    3 },
    { 9, "WaitForAnyEventFlag", "WaitForAnyEventFlagStrictly", 0,          3,    3,    3,    3,    3,    3,    3,    3,    3,    3,    3,    3,    3,    3,    3 },
    { 9, "WaitForAllEventFlag", "WaitForAllEventFlagStrictly", 0,          3,    3,    3,    3,    3,    3,    3,    3,    3,    3,    3,    3,    3,    3,    3 },
    { 9, "CreateMessageQueue", "CreateMessageQueueStrictly", 0,            1,    1,    1,    1,    1,    1,    1,    1,    1,    1,    1,    1,    1,    1,    1 },
    { 9, "CreateRecursiveLock", "CreateRecursiveLockStrictly", 0,          1,    1,    1,    1,    1,    1,    1,    1,    1,    1,    1,    1,    1,    1,    1 },
    { 9, "CreateEventFlag", "CreateEventFlagStrictly", 0,                  1,    1,    1,    1,    1,    1,    1,    1,    1,    1,    1,    1,    1,    1,    1 },
    { 9, "_GetSystemTime", "GetSystemTime", 0,                             2,    2,    2,    2,    2,    2,    2,    2,    2,    2,    2,    2,    2,    2,    2 },
    { 9, "close", "Close", 0,                                              2,    2,    2,    2,    2,    2,    2,    2,    2,    2,    2,    2,    2,    2,    2 },
    { 9, "open", "Open", 0,                                                3,    3,    3,    3,   16,   16,   35,   35,   35,   35,   35,   35,   35,   35,   35 },
    { 9, "open", "Open", 0,                                                3,    3,    3,   13,   16,   16,   35,   35,   35,   35,   35,   35,   35,   35,   35 },
    { 9, "_divmod_signed_int", "PT_mod_FW", 0,                             4,    7,    7,    7,    7,    7,    6,    6,    6,    6,    6,    6,    6,    6,    6 },
    { 9, "_divmod_signed_int", "PT_mod_FW", 0,                             0,    4,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0 },
    { 9, "_divmod_signed_int", "mod_FW", 0,                                0,    0,    0,    0,    0,    0,    0,    0,    0,    4,    4,    4,    4,    4,    4 },
    { 9, "_divmod_unsigned_int", "SetTimerAfter", 0,                      23,   23,   23,   23,   23,   23,   23,   23,   23,   23,   23,    0,    0,    0,    0 },
    { 9, "_divmod_unsigned_int", "DispCon_ShowWhiteChart_FW", 0,           0,    0,    0,    0,    0,    0,    0,    0,    0,    0,   15,   15,   15,   15,   13 },
    { 9, "_divmod_unsigned_int", "DispCon_ShowWhiteChart_FW", 0,           0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    8,    8,    8,    0,    0 },
    { 9, "_dflt", "CalcLog10", 0,                                          9,    9,    9,    9,    9,    9,    9,    9,    9,    9,    9,    9,    9,    9,    9 },
    { 9, "_dfltu", "CalcLog10", 0,                                         4,    4,    4,    4,    4,    4,    4,    4,    4,    4,    4,    4,    4,    4,    4 },
    { 9, "_dmul", "CalcLog10", 0,                                         12,   12,   12,   12,   12,   12,   12,   12,   12,   12,   12,   12,   12,   12,   12 },
    { 9, "_dfix", "CalcLog10", 0,                                         14,   14,   14,   14,   14,   14,   14,   14,   14,   14,   14,   14,   14,   14,   14 },
    { 9, "_dadd", "_pow", 0,                                              26,   26,   26,   26,   26,   26,   29,   29,   29,   29,   29,   29,   29,   29,   29 },
    { 9, "_dadd", "_pow", 0,                                               1,    1,    1,    1,    1,    1,   24,   24,    1,    1,    1,    1,    1,    1,    1 }, // s100, sx230 (100c only...)
    { 9, "_scalbn", "_log", 0,                                            19,   19,   19,   19,   19,   19,   18,   18,   18,   18,   18,   18,   18,   18,   18 },
    { 9, "_scalbn", "_log", 0,                                             1,    1,    1,    1,    1,    1,   14,   14,    1,    1,    1,    1,    1,    1,    1 }, // s100, sx230 (100c only...)
    { 9, "_safe_sqrt", "CalcSqrt_FW", 0,                                   0,    0,   -3,   -6,   -6,   -6,   -6,   -6,   -6,   -6,   -6,   -6,   -6,   -6,   -6 },
    { 9, "_ddiv", "ConvertApexStdToApex_FW", 0,                            0,    0,    0,   21,   21,   21,    0,    0,    0,    0,    0,    0,    0,    0,    0 },
    { 9, "_fflt", "ConvertApexStdToApex_FW", 0,                           -7,   -7,   -7,   -7,   -7,   -7,   -7,   -7,   -7,   -7,   -7,   -7,   -7,   -7,   -7 },
    { 9, "_ffix", "ConvertApexStdToApex_FW", 0,                           -4,   -4,   -4,   -4,   -4,   -4,   -4,   -4,   -4,   -4,   -4,   -4,   -4,   -4,   -4 },
    { 9, "_fmul", "ConvertApexStdToApex_FW", 0,                           -5,   -5,   -5,   -5,   -5,   -5,   -5,   -5,   -5,   -5,   -5,   -5,   -5,   -5,   -5 },
    { 9, "_fdiv", "ConvertApexToApexStd_FW", 0,                           -3,   -3,   -3,   -3,   -3,   -3,   -3,   -3,   -3,   -3,   -3,   -3,   -3,   -3,   -3 },
    { 9, "GetSRAndDisableInterrupt", "_GetSystemTime", 0,                  9,    9,    9,    9,    9,    9,    9,    9,    9,    9,    9,    9,    9,    9,    9 },
    { 9, "SetSR", "_GetSystemTime", 0,                                    12,   12,   12,   12,   12,   12,   12,   12,   12,   12,   12,   12,   12,   12,   12 },

    //                                                                   R20   R23   R31   R39   R43   R45   R47   R49   R50   R51   R52   R54   R55   R57   R58
//    { 11, "DebugAssert", "\nAssert: File %s Line %d\n", 0,                 5,    5,    5,    5,    5,    5,    5,    5,    5,    5,    1,    6 },
    { 11, "err_init_task", "\n-- %s() error in init_task() --", 0,         2,    2,    2,    2,    2,    2,    2,    2,    2,    2,    2,    2,    2,    2,    2 },
    { 11, "set_control_event", "Button:0x%08X:%s", 0x01000001,            14,   14,   14,   14,   14,   14,   14,   14,   14,   14,   14,   14,   14,   14,   14 },
    { 11, "set_control_event", "Button:0x%08X:%s", 0xf1000001,            15,   15,   15,   15,   15,   15,   15,   15,   15,   15,   15,   15,   15,   15,   15 },
    { 11, "set_control_event", "Button:0x%08X:%s", 0x01000001,            19,   19,   19,   19,   19,   19,   19,   19,   19,   19,   19,   19,   19,   19,   19 },
    { 11, "set_control_event", "Button:0x%08X:%s", 0x01000001,            20,   20,   20,   20,   20,   20,   20,   20,   20,   20,   20,   20,   20,   20,   20 },
    { 11, "_log", (char*)log_test, 0x01000001,                             1,    1,    1,    1,    1,    1,    1,    1,    1,    1,    1,    1,    1,    1,    1 },
    { 11, "_uartr_req", "A/uartr.req", 0,                                  3,    3,    3,    3,    3,    3,    3,    3,    3,    3,    3,    3,    3,    3,    3 },

    //                                                                   R20   R23   R31   R39   R43   R45   R47   R49   R50   R51   R52   R54   R55   R57   R58
    { 12, "DeleteFile_Fut", "DeleteFile_Fut", 1,                        0x38, 0x38, 0x4C, 0x4C, 0x4C, 0x54, 0x54, 0x54, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00 },
    { 12, "AllocateUncacheableMemory", "AllocateUncacheableMemory", 1,  0x2C, 0x2C, 0x2C, 0x2C, 0x2C, 0x34, 0x34, 0x34, 0x4C, 0x4C, 0x4C, 0x4C, 0x54, 0x54, 0x54 },
    { 12, "FreeUncacheableMemory", "FreeUncacheableMemory", 1,          0x30, 0x30, 0x30, 0x30, 0x30, 0x38, 0x38, 0x38, 0x50, 0x50, 0x50, 0x50, 0x58, 0x58, 0x58 },
    { 12, "free", "free", 1,                                            0x28, 0x28, 0x28, 0x28, 0x28, 0x30, 0x30, 0x30, 0x48, 0x48, 0x48, 0x48, 0x50, 0x50, 0x50 },
    { 12, "malloc", "malloc", 0x01000003,                               0x24, 0x24, 0x24, 0x24, 0x24, 0x2C, 0x2C, 0x2C, 0x44, 0x44, 0x44, 0x44, 0x4c, 0x4c, 0x4c }, // uses 'malloc_strictly'
    { 12, "TakeSemaphore", "TakeSemaphore", 1,                          0x14, 0x14, 0x14, 0x14, 0x14, 0x1C, 0x1C, 0x1C, 0x1C, 0x1C, 0x1C, 0x1C, 0x1C, 0x1C, 0x1C },
    { 12, "GiveSemaphore", "GiveSemaphore", 1,                          0x18, 0x18, 0x18, 0x18, 0x18, 0x20, 0x20, 0x20, 0x20, 0x20, 0x20, 0x20, 0x20, 0x20, 0x20 },
    { 12, "_log10", "_log10", 0x01000006,                              0x278,0x280,0x280,0x284,0x294,0x2FC,0x2FC,0x31C,0x354,0x35C,0x35C,0x35C,0x388,0x38c,0x390 }, // uses 'CalcLog10'
    { 12, "_log10", "_log10", 0x01000006,                              0x000,0x278,0x27C,0x000,0x000,0x000,0x000,0x000,0x000,0x000,0x000,0x000,0x38c,0x000,0x000 },
    { 12, "_log10", "_log10", 0x01000006,                              0x000,0x000,0x2C4,0x000,0x000,0x000,0x000,0x000,0x000,0x000,0x000,0x000,0x000,0x000,0x000 },
    { 12, "ClearEventFlag", "ClearEventFlag", 1,                        0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04 },
    { 12, "SetEventFlag", "SetEventFlag", 1,                            0x08, 0x08, 0x08, 0x08, 0x08, 0x08, 0x08, 0x08, 0x08, 0x08, 0x08, 0x08, 0x08, 0x08, 0x08 },
    { 12, "WaitForAnyEventFlag", "WaitForAnyEventFlag", 1,              0x0c, 0x0c, 0x0c, 0x0c, 0x0c, 0x0c, 0x0c, 0x0c, 0x0c, 0x0c, 0x0c, 0x0c, 0x0c, 0x0c, 0x0c },
    { 12, "WaitForAllEventFlag", "WaitForAllEventFlag", 1,              0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10 },

    { 13, "strftime", "Sunday", 1 },

    { 15, "LocalTime", "%04d:%02d:%02d %02d:%02d:%02d", 0x01000001 },
    { 15, "GetMemInfo", "Malloc Information\n", 0x01000001 },
    { 15, "GetMemInfo", "Malloc Information (%s type)\n", 0x01000001 },
    { 15, "vsprintf", "\nCPrintf Size Over!!", 0x01000001 },
    { 15, "ReadFastDir", "ReadFast_ERROR\n", 0x01000001 },
    { 15, "OpenFastDir", "OpenFastDir_ERROR\n", 0x01000001 },
    { 15, "realloc", "fatal error - scanner input buffer overflow", 0x01000001 },
    { 15, "CreateBinarySemaphore", "SdPower.c", 0x01000001 },
    { 15, "get_resource_pointer", "Not found icon resource.\r\n", 0x01000001 },
    { 15, "get_self_task_id", "ASSERT!! %s Line %d\n", 0x01000001 },
    //                                                                           R20     R23     R31     R39     R43     R45     R47     R49     R50     R51     R52     R54     R55     R57     R58
    { 15, "SetHPTimerAfterTimeout", "FrameRateGenerator.c", 0x01000001,          0x0007, 0x0007, 0x0007, 0x0007, 0x0007, 0x0007, 0x0007, 0x0007, 0x0007, 0x0007, 0x0001, 0x0007, 0x0007, 0x0007, 0x0007 },
    { 15, "get_task_properties", "Task ID: %d\n", 0x01000001,                    0x0003, 0x0003, 0x0003, 0x0003, 0x0003, 0x0003, 0x0003, 0x0003, 0x0003, 0x0003, 0x0003, 0x0003, 0x0003, 0x0003, 0x0003 }, // string not unique!

    { 16, "DeleteDirectory_Fut", (char*)DeleteDirectory_Fut_test, 0x01000001 },
    { 16, "MakeDirectory_Fut", (char*)MakeDirectory_Fut_test, 0x01000001 },
    { 16, "RenameFile_Fut", (char*)RenameFile_Fut_test, 0x01000001 },

    { 17, "ScreenLock", "StartRecModeMenu", 0 },
    { 17, "ScreenUnlock", "StartRecModeMenu", 0 },

    // Ensure ordering in func_names is correct for dependencies here
    //                                                                           R20     R23     R31     R39     R43     R45     R47     R49     R50     R51     R52     R54     R55     R57     R58
    { 19, "GetSemaphoreValue", "GiveSemaphore", 0,                               0x000e, 0x000e, 0x000e, 0x000e, 0x000e, 0x000e, 0x000e, 0x000e, 0x000e, 0x000e, 0x000e, 0x0014, 0x0014, 0x0014, 0x0014 },
    { 19, "GetSemaphoreValue", "GiveSemaphore", 0,                               0xf000, 0xf000, 0xf000, 0xf000, 0xf000, 0xf000, 0xf000, 0xf000, 0xf000, 0xf000, 0x000f, 0xf000, 0xf000, 0xf000, 0xf000 },
    { 19, "CreateMessageQueueStrictly", "CreateTaskStrictly", 0,                 0x000d, 0x000d, 0x000d, 0x000d, 0x000d, 0x000d, 0x000d, 0x000d, 0x000d, 0x000d, 0x000d, 0x000d, 0x000d, 0x000d, 0x000d },
    { 19, "CreateMessageQueueStrictly", "CreateTaskStrictly", 0,                 0x000d, 0x000d, 0x000d, 0x000d, 0x000d, 0x000d, 0x000d, 0x000d, 0x000d, 0x000d, 0x000e, 0x000e, 0x000e, 0x000e, 0x000e },
    { 19, "CreateEventFlagStrictly", "CreateMessageQueueStrictly", 0,            0x0009, 0x0009, 0x0009, 0x0009, 0x0009, 0x0009, 0x0009, 0x0009, 0x0009, 0x0009, 0x0009, 0x0009, 0x0009, 0x0009, 0x0009 },
    { 19, "CreateEventFlagStrictly", "CreateMessageQueueStrictly", 0,            0x0009, 0x0009, 0x0009, 0x0009, 0x0009, 0x0009, 0x0009, 0x0009, 0x0009, 0x0009, 0x000a, 0x000a, 0x000a, 0x000a, 0x000a },
    { 19, "CreateBinarySemaphoreStrictly", "CreateEventFlagStrictly", 0,         0x0009, 0x0009, 0x0009, 0x0009, 0x0009, 0x0009, 0x0009, 0x0009, 0x0009, 0x0009, 0x0009, 0x0009, 0x0009, 0x0009, 0x0009 },
    { 19, "CreateBinarySemaphoreStrictly", "CreateEventFlagStrictly", 0,         0x0009, 0x0009, 0x0009, 0x0009, 0x0009, 0x0009, 0x0009, 0x0009, 0x0009, 0x0009, 0x000a, 0x000a, 0x000a, 0x000a, 0x000a },
    { 19, "CreateCountingSemaphoreStrictly", "CreateBinarySemaphoreStrictly", 0, 0x0009, 0x0009, 0x0009, 0x0009, 0x0009, 0x0009, 0x0009, 0x0009, 0x0009, 0x0009, 0x0009, 0x0009, 0x0009, 0x0009, 0x0009 },
    { 19, "CreateCountingSemaphoreStrictly", "CreateBinarySemaphoreStrictly", 0, 0x0009, 0x0009, 0x0009, 0x0009, 0x0009, 0x0009, 0x0009, 0x0009, 0x0009, 0x0009, 0x000a, 0x000a, 0x000a, 0x000a, 0x000a },
    { 19, "CreateRecursiveLockStrictly", "CreateCountingSemaphoreStrictly", 0,   0x0009, 0x0009, 0x0009, 0x0009, 0x0009, 0x0009, 0x0009, 0x0009, 0x0009, 0x0009, 0x0009, 0x0009, 0x0009, 0x0009, 0x0009 },
    { 19, "CreateRecursiveLockStrictly", "CreateCountingSemaphoreStrictly", 0,   0x0009, 0x0009, 0x0009, 0x0009, 0x0009, 0x0009, 0x0009, 0x0009, 0x0009, 0x0009, 0x000a, 0x000a, 0x000a, 0x000a, 0x000a },
    { 19, "TakeSemaphoreStrictly", "CreateRecursiveLockStrictly", 0,             0x0001, 0x0009, 0x0009, 0x0009, 0x0009, 0x0009, 0x0009, 0x0009, 0x0009, 0x0009, 0x0009, 0x0009, 0x0009, 0x0009, 0xf000 }, // name made up
    { 19, "TakeSemaphoreStrictly", "CreateRecursiveLockStrictly", 0,             0x0001, 0x0009, 0x0009, 0x0009, 0x0009, 0x0009, 0x0009, 0x0009, 0x0009, 0x0009, 0x000a, 0x000a, 0x000a, 0x000a, 0x0014 }, // name made up
    { 19, "ReceiveMessageQueueStrictly", "TakeSemaphoreStrictly", 0,             0x000b, 0x000b, 0x000b, 0x000b, 0x000b, 0x000b, 0x000b, 0x000b, 0x000b, 0x000b, 0x000b, 0x000b, 0x000b, 0x000b, 0x000b }, // name made up
    { 19, "PostMessageQueueStrictly", "ReceiveMessageQueueStrictly", 0,          0x000b, 0x000b, 0x000b, 0x000b, 0x000b, 0x000b, 0x000b, 0x000b, 0x000b, 0x000b, 0x000b, 0x000b, 0x000b, 0x000b, 0x000b }, // name made up
    { 19, "WaitForAnyEventFlagStrictly", "PostMessageQueueStrictly", 0,          0x000b, 0x000b, 0x000b, 0x000b, 0x000b, 0x000b, 0x000b, 0x000b, 0x000b, 0x000b, 0x000b, 0x000b, 0x000b, 0x000b, 0x000b }, // name made up
    { 19, "WaitForAllEventFlagStrictly", "WaitForAnyEventFlagStrictly", 0,       0x000b, 0x000b, 0x000b, 0x000b, 0x000b, 0x000b, 0x000b, 0x000b, 0x000b, 0x000b, 0x000b, 0x000b, 0x000b, 0x000b, 0x000b }, // name made up
    { 19, "AcquireRecursiveLockStrictly", "WaitForAllEventFlagStrictly", 0,      0x000b, 0x000b, 0x000b, 0x000b, 0x000b, 0x000b, 0x000b, 0x000b, 0x000b, 0x000b, 0x000b, 0x000b, 0x000b, 0x000b, 0x000b }, // name made up
                                                                                                                                                                                               
    { 19, "PostMessageQueue", "TryReceiveMessageQueue", 0,                       0x091f, 0x0001, 0x0001, 0x0001, 0x0001, 0x0001, 0x0001, 0x0001, 0x0001, 0x0001, 0x0001, 0x0001, 0x0001, 0x0001, 0x0001 },
    { 19, "DeleteMessageQueue", "CreateMessageQueue", 0,                         0x1021, 0x1021, 0x1021, 0x1021, 0x1021, 0x1021, 0x1021, 0x1021, 0x1021, 0x1021, 0x1021, 0x002c, 0x002c, 0x002c, 0x002c },
    { 19, "DeleteMessageQueue", "CreateMessageQueue", 0,                         0xf000, 0xf000, 0xf000, 0xf000, 0xf000, 0xf000, 0xf000, 0xf000, 0xf000, 0xf000, 0x0027, 0xf000, 0xf000, 0xf000, 0xf000 },
    { 19, "ReceiveMessageQueue", "DeleteMessageQueue", 0,                        0x1024, 0x1024, 0x1024, 0x1024, 0x1024, 0x1024, 0x1024, 0x1024, 0x1024, 0x1024, 0x1024, 0x1029, 0x1029, 0x1029, 0x1028 },
    { 19, "ReceiveMessageQueue", "DeleteMessageQueue", 0,                        0xf000, 0xf000, 0xf000, 0xf000, 0xf000, 0xf000, 0xf000, 0xf000, 0xf000, 0xf000, 0x1025, 0xf000, 0xf000, 0xf000, 0xf000 },
    { 19, "TryReceiveMessageQueue", "ReceiveMessageQueue", 0,                    0x002b, 0x002b, 0x1032, 0x1032, 0x1032, 0x1032, 0x1032, 0x1032, 0x1032, 0x1032, 0x1032, 0x073d, 0x073d, 0x073d, 0x073d },
    { 19, "TryReceiveMessageQueue", "ReceiveMessageQueue", 0,                    0xf000, 0xf000, 0xf000, 0xf000, 0xf000, 0xf000, 0xf000, 0xf000, 0xf000, 0xf000, 0x183e, 0xf000, 0xf000, 0xf000, 0xf000 },
    { 19, "TryPostMessageQueue", "PostMessageQueue", 0,                          0x0027, 0x0027, 0x102e, 0x102e, 0x102e, 0x102e, 0x102e, 0x102e, 0x102e, 0x102e, 0x102e, 0x0031, 0x0031, 0x0031, 0x0031 },
    { 19, "TryPostMessageQueue", "PostMessageQueue", 0,                          0xf000, 0xf000, 0xf000, 0xf000, 0xf000, 0xf000, 0xf000, 0xf000, 0xf000, 0xf000, 0x1031, 0xf000, 0xf000, 0xf000, 0xf000 },
    { 19, "GetNumberOfPostedMessages", "TryPostMessageQueue", 0,                 0x0010, 0x0010, 0x0010, 0x0010, 0x0010, 0x0010, 0x0010, 0x0010, 0x0010, 0x0010, 0x0010, 0x0015, 0x0015, 0x0015, 0x0015 },
    { 19, "DeleteRecursiveLock", "CreateRecursiveLock", 0,                       0x0014, 0x0014, 0x0014, 0x0014, 0x0014, 0x0014, 0x0014, 0x0014, 0x0014, 0x0014, 0x0014, 0x0016, 0x0016, 0x0016, 0x0016 },
    { 19, "DeleteRecursiveLock", "CreateRecursiveLock", 0,                       0xf000, 0xf000, 0xf000, 0xf000, 0xf000, 0xf000, 0xf000, 0xf000, 0xf000, 0xf000, 0x0016, 0xf000, 0xf000, 0xf000, 0xf000 },
    { 19, "AcquireRecursiveLock", "DeleteRecursiveLock", 0,                      0x0014, 0x0014, 0x0014, 0x0014, 0x0014, 0x0014, 0x0014, 0x0014, 0x0014, 0x0014, 0x0014, 0x001a, 0x001a, 0x001a, 0x001a },
    { 19, "AcquireRecursiveLock", "DeleteRecursiveLock", 0,                      0xf000, 0xf000, 0xf000, 0xf000, 0xf000, 0xf000, 0xf000, 0xf000, 0xf000, 0xf000, 0x0015, 0xf000, 0xf000, 0xf000, 0xf000 },
    { 19, "ReleaseRecursiveLock", "AcquireRecursiveLock", 0,                     0x0041, 0x0041, 0x0048, 0x0048, 0x0048, 0x0048, 0x0048, 0x0048, 0x0048, 0x0048, 0x0048, 0x085a, 0x085a, 0x085a, 0x085a },
    { 19, "ReleaseRecursiveLock", "AcquireRecursiveLock", 0,                     0xf000, 0xf000, 0xf000, 0xf000, 0xf000, 0xf000, 0xf000, 0xf000, 0xf000, 0xf000, 0x004d, 0xf000, 0xf000, 0xf000, 0xf000 },
    { 19, "GetEventFlagValue", "ClearEventFlag", 0,                              0x000e, 0x000e, 0x000e, 0x000e, 0x000e, 0x000e, 0x000e, 0x000e, 0x000e, 0x000e, 0x000e, 0x0013, 0x0013, 0x0013, 0x0013 },
    { 19, "DeleteEventFlag", "CreateEventFlag", 0,                               0x0016, 0x0016, 0x0016, 0x0016, 0x0016, 0x0016, 0x0016, 0x0016, 0x0016, 0x0016, 0x0016, 0x0018, 0x0018, 0x0018, 0x0018 },
    { 19, "DeleteEventFlag", "CreateEventFlag", 0,                               0xf000, 0xf000, 0xf000, 0xf000, 0xf000, 0xf000, 0xf000, 0xf000, 0xf000, 0xf000, 0x0018, 0xf000, 0xf000, 0xf000, 0xf000 },
    { 19, "CheckAnyEventFlag", "DeleteEventFlag", 0,                             0x0014, 0x0014, 0x0014, 0x0014, 0x0014, 0x0014, 0x0014, 0x0014, 0x0014, 0x0014, 0x0014, 0x001a, 0x001a, 0x001a, 0x001a },
    { 19, "CheckAnyEventFlag", "DeleteEventFlag", 0,                             0xf000, 0xf000, 0xf000, 0xf000, 0xf000, 0xf000, 0xf000, 0xf000, 0xf000, 0xf000, 0x0015, 0xf000, 0xf000, 0xf000, 0xf000 },
    { 19, "CheckAllEventFlag", "CheckAnyEventFlag", 0,                           0x0012, 0x0012, 0x0012, 0x0012, 0x0012, 0x0012, 0x0012, 0x0012, 0x0012, 0x0012, 0x0012, 0x0017, 0x0017, 0x0017, 0x0017 },
    { 19, "TryTakeSemaphore", "DeleteSemaphore", 0,                              0x0016, 0x0016, 0x0016, 0x0016, 0x0016, 0x0016, 0x0016, 0x0016, 0x0016, 0x0016, 0x0016, 0x001d, 0x001d, 0x001d, 0x001d },
    { 19, "TryTakeSemaphore", "DeleteSemaphore", 0,                              0xf000, 0xf000, 0xf000, 0xf000, 0xf000, 0xf000, 0xf000, 0xf000, 0xf000, 0xf000, 0x0018, 0xf000, 0xf000, 0xf000, 0xf000 },
    { 19, "SetTimerAfter", "_GetSystemTime", 0,                                  0x004e, 0x004e, 0x004e, 0x004e, 0x004e, 0x004e, 0x004e, 0x004e, 0x004e, 0x004e, 0x004e, 0x004e, 0x004e, 0x0054, 0x0054 },
    { 19, "SetTimerWhen", "SetTimerAfter", 0,                                    0x0020, 0x0020, 0x0020, 0x0020, 0x0020, 0x0020, 0x0020, 0x0020, 0x0020, 0x0020, 0x0020, 0x001b, 0x001b, 0x001b, 0x001b },
    { 19, "SetTimerWhen", "SetTimerAfter", 0,                                    0xf000, 0xf000, 0xf000, 0xf000, 0xf000, 0xf000, 0xf000, 0xf000, 0xf000, 0xf000, 0x001b, 0xf000, 0xf000, 0xf000, 0xf000 },
    { 19, "CancelTimer", "SetTimerWhen", 0,                                      0x0019, 0x0019, 0x0019, 0x0019, 0x0019, 0x0019, 0x0019, 0x0019, 0x0019, 0x0019, 0x0019, 0x0019, 0x0019, 0x0019, 0x0019 },
    { 19, "CancelHPTimer", "SetHPTimerAfterTimeout", 0,                          0x0022, 0x0022, 0x0022, 0x0022, 0x0022, 0x0022, 0x0022, 0x0022, 0x0022, 0x0022, 0x0022, 0x0022, 0x0022, 0x0022, 0x0022 },
    { 19, "SetHPTimerAfterNow", "SetHPTimerAfterTimeout", 0,                    -0x0020,-0x0020,-0x0020,-0x0020,-0x0020,-0x0020,-0x0020,-0x0020,-0x0020,-0x0020,-0x0020,-0x0020,-0x0020,-0x0020,-0x0020 },
    //{ 19, "UnregisterInterruptHandler", "RegisterInterruptHandler", 0,         0x0009, 0x0009, 0x0009, 0x0009, 0x0009, 0x0009, 0x0009, 0x0009, 0x0009, 0x0009, 0x0009, 0x0013, 0x0013, 0x0013, 0x0013 }, // unreliable on r52+
    //{ 19, "GetSRAndDisableInterrupt", "UnregisterInterruptHandler", 0,         0x0006, 0x0006, 0x0006, 0x0006, 0x0006, 0x0006, 0x0006, 0x0006, 0x0006, 0x0006, 0x0006, 0x004c, 0x004c, 0x004c, 0x004c }, // unreliable on r52+
    //{ 19, "SetSR", "UnregisterInterruptHandler", 0,                            0x1007, 0x1007, 0x1007, 0x1007, 0x1007, 0x1007, 0x1007, 0x1007, 0x1007, 0x1007, 0x1007, 0x104d, 0x104d, 0x104d, 0x104d }, // unreliable on r52+
    { 19, "EnableInterrupt", "UnregisterInterruptHandler", 0,                    0x170f, 0x170f, 0x170f, 0x170f, 0x170f, 0x170f, 0x170f, 0x170f, 0x170f, 0x170f, 0x170f, 0x1755, 0x1755, 0x1755, 0x1755 },
    { 19, "EnableInterrupt", "SetSR", 0,                                         0xf000, 0xf000, 0xf000, 0xf000, 0xf000, 0xf000, 0xf000, 0xf000, 0xf000, 0xf000, 0x1708, 0x1708, 0x1708, 0x1708, 0x1708 },
    //                                                                           R20     R23     R31     R39     R43     R45     R47     R49     R50     R51     R52     R54     R55     R57     R58
    { 19, "GetDrive_TotalClusters", "GetDrive_ClusterSize", 0,                   0x0001, 0x0001, 0x0001, 0x0001, 0x0001, 0x000d, 0x000c, 0x000c, 0x000c, 0x000c, 0x0001, 0x0001, 0x0001, 0x0001, 0x0001 },
    { 19, "GetDrive_FreeClusters", "GetDrive_TotalClusters", 0,                  0x0001, 0x0001, 0x0001, 0x0001, 0x0001, 0x000b, 0x000a, 0x000a, 0x000a, 0x000a, 0x000a, 0x000b, 0x000b, 0x000b, 0x000b },
    { 19, "GetDrive_FreeClusters", "GetDrive_TotalClusters", 0,                  0x0001, 0x0001, 0x0001, 0x0001, 0x0001, 0x0001, 0x0001, 0x0001, 0x0001, 0x0001, 0x000b, 0x0001, 0x0001, 0x0001, 0x0001 }, // alt r52 (sx510)
                                                                                                                                                                                               
    { 19, "time", "GetTimeOfSystem_FW", 0,                                       0x0001, 0x0001, 0x0001, 0x0001, 0x0001, 0x0001, 0x0001, 0x0001, 0x0001, 0x0001,-0x002c,-0x002d,-0x0c2d,-0x0c2d,-0x0c2d },
    { 19, "time", "GetTimeOfSystem_FW", 0,                                       0x0001, 0x0001, 0x0001, 0x0001, 0x0001, 0x0001, 0x0001, 0x0001, 0x0001, 0x0001,-0x002d, 0x0001, 0x0001, 0x0001, 0x0001 }, // alt r52 (sx510)
                                                                                                                                                                                               
    { 19, "CalcLog10", "CalcLog10_FW", 0,                                       -0x100f,-0x100f,-0x100f,-0x100f,-0x100f,-0x100f,-0x100f,-0x100f,-0x100f,-0x100f,-0x100f,-0x100f,-0x100f,-0x100f,-0x100f },
    { 19, "_dfixu", "_dfix", 0,                                                  0x1f2c, 0x1f2c, 0x1f2c, 0x1f2c, 0x1f2c, 0x1f2c, 0x1f2c, 0x1f2c, 0x1f2c, 0x1f2c, 0x1f2c, 0x1f2c, 0x1f2c, 0x1f2c, 0x1f2c },
    { 19, "_dsub", "_sqrt", 0,                                                   0x165d, 0x165d, 0x165d, 0x165d, 0x165d, 0x165d, 0x165d, 0x165d, 0x165d, 0x165d, 0x165d, 0x165d, 0x165d, 0x165d, 0x165d },
    { 19, "_drsb", "_sqrt", 0,                                                  -0x1114,-0x1114,-0x1114,-0x1114,-0x1114,-0x1114,-0x1106,-0x1106,-0x1106,-0x1106,-0x1106,-0x1106,-0x1106,-0x1106,-0x1106 },
    { 19, "_dcmp_reverse", "_dmul", 0,                                           0xf000, 0xf000, 0x33fc, 0x33fc, 0x33fc, 0x33fc, 0x43e2, 0x43e2, 0x43e2, 0x43e2, 0xf000, 0xf000, 0xf000, 0xf000, 0xf000 }, // has to come first
    { 19, "_dcmp_reverse", "_dmul", 0,                                           0x2e80, 0x2e80, 0x3580, 0x3580, 0x3580, 0x3580, 0x3474, 0x3474, 0x3474, 0x3474, 0x3474, 0x3474, 0x3474, 0x3474, 0x3474 },
    { 19, "_dcmp", "_dfltu", 0,                                                  0x1003, 0x1003, 0x1003, 0x1003, 0x1003, 0x1003, 0x1003, 0x1003, 0x1003, 0x1003, 0x1003, 0x1003, 0x1003, 0x1003, 0x1003 },
    { 19, "_safe_sqrt", "_dadd", 0,                                              0xf000, 0xf000, 0xf000, 0xf000, 0xf000, 0xf000,-0x1d13,-0x1d13,-0x1d13,-0x1d13,-0x1d13,-0x1d13,-0x1d13,-0x1d13,-0x1d13 },
    { 19, "_safe_sqrt", "_log", 0,                                              -0x132f,-0x132f,-0x1695,-0x132f,-0x132f,-0x132f, 0xf000, 0xf000, 0xf000, 0xf000, 0xf000, 0xf000, 0xf000, 0xf000, 0xf000 },
    { 19, "_safe_sqrt", "_log", 0,                                               0xf000, 0xf000,-0x132f,-0x1695, 0xf000, 0xf000, 0xf000, 0xf000, 0xf000, 0xf000, 0xf000, 0xf000, 0xf000, 0xf000, 0xf000 },
    { 19, "_ddiv", "_dadd", 0,                                                   0x10aa, 0x10aa, 0x10aa, 0x10aa, 0x10aa, 0x10c3, 0x10b6, 0x10b6, 0x10b6, 0x10b6, 0x10b6, 0x10b6, 0x10b6, 0x10b6, 0x10b6 },
    { 19, "_ffixu", "_ffix", 0,                                                  0x311d, 0x311d, 0x311d, 0x311d, 0x311d, 0x311d, 0x311d, 0x311d, 0x311d, 0x311d, 0x311d, 0x311d, 0x311d, 0x311d, 0x311d },
    { 19, "_ffltu", "_fflt", 0,                                                  0x400d, 0x400d, 0x400d, 0x400d, 0x400d, 0x400d, 0x400d, 0x400d, 0x400d, 0x400d, 0x400d, 0x400d, 0x400d, 0x400d, 0x400d },
    { 19, "_f2d", "_fdiv", 0,                                                    0xf000,-0x301f,-0x301f,-0x301f,-0x301f,-0x301f,-0x3061,-0x3061,-0x3061,-0x3061,-0x3061,-0x3061,-0x3061,-0x3061,-0x3061 },
    { 19, "_f2d", "_fdiv", 0,                                                   -0x306b,-0x306b,-0x306b,-0x306b,-0x306b,-0x306b, 0xf000, 0xf000, 0xf000, 0xf000, 0xf000, 0xf000, 0xf000, 0xf000, 0xf000 },

    { 21, "add_ptp_handler", (char*)find_add_ptp_handler, 0 },
    { 21, "apex2us", (char*)find_apex2us, 0 },
    { 21, "mkdir", (char*)find_mkdir, 0 },
    { 21, "_pow", (char*)find_pow, 0 },
    { 21, "rand", (char*)find_rand, 0 },

    { 22, "closedir", (char*)find_closedir, 0 },
    { 22, "PT_PlaySound", (char*)find_PT_PlaySound, 0 },
    { 22, "getImageDirName", (char*)find_getImageDirName, 0 },
    { 22, "GetImageFolder", (char*)find_GetImageFolder, 0 },
    { 22, "GetDrive_ClusterSize", (char*)find_GetDrive_ClusterSize, 0 },
    { 22, "GetDrive_TotalClusters", (char*)find_GetDrive_TotalClusters, 0 },
    { 22, "srand", (char*)find_srand, 0 },
    { 22, "Restart", (char*)find_Restart, 0 },
    { 22, "DisplayBusyOnScreen", (char*)find_DisplayBusyOnScreen, 0 },
    { 22, "UndisplayBusyOnScreen", (char*)find_UndisplayBusyOnScreen, 0 },
    { 22, "CreateDialogBox", (char*)find_CreateDialogBox, 0 },
    { 22, "DisplayDialogBox", (char*)find_DisplayDialogBox, 0 },
    { 22, "add_ui_to_dialog", (char*)find_add_ui_to_dialog, 0 },
    { 22, "get_string_by_id", (char*)find_get_string_by_id, 0 },
    { 22, "get_fstype", (char*)find_get_fstype, 0 },
    { 22, "malloc_strictly", (char*)find_malloc_strictly, 0 },
    { 22, "SetHPTimerAfterTimeout", (char*)find_sethptimeraftertimeout, 0},
    { 22, "GetCurrentMachineTime", (char*)find_getcurrentmachinetime, 0},
    { 22, "get_self_task_errno_pointer", (char*)find_get_self_task_errno_pointer, 0},

    //                                                                           R20     R23     R31     R39     R43     R45     R47     R49     R50     R51     R52     R54     R55     R57     R58
    { 23, "UnregisterInterruptHandler", "HeadInterrupt1", 76,                    1,      1,      1,      1,      1,      1,      1,      1,      1,      1,      1,      1,      1,      1,      1 },

    //                                                                           R20     R23     R31     R39     R43     R45     R47     R49     R50     R51     R52     R54     R55     R57     R58
    { 24, "get_string_by_id", "StringID[%d] is not installed!!\n", 64,           0xf000, 0xf000, 0xf000, 0xf000, 0xf000, 0xf000, 0xf000, 0xf000, 0xf000, 0xf000, 0xf000, 0xf000, 0x0000, 0x0000, 0x0000 },

    { 0, 0, 0, 0 }
};

// Find the named function in the 'func_list' array above, return index of entry
int find_func(const char* name)
{
    int i;
    for (i=0; func_list[i].name != 0; i++)
    {
        if (strcmp(name, func_list[i].name) == 0)
        {
            return i;
        }
    }
    return -1;  // Not found
}

// Get DryOS version specific offset
int dryos_offset(firmware *fw, string_sig *sig)
{
    switch (fw->dryos_ver)
    {
    case 20:    return sig->dryos20_offset;
    case 23:    return sig->dryos23_offset;
    case 31:    return sig->dryos31_offset;
    case 39:    return sig->dryos39_offset;
    case 43:    return sig->dryos43_offset;
    case 45:    return sig->dryos45_offset;
    case 47:    return sig->dryos47_offset;
    case 49:    return sig->dryos49_offset;
    case 50:    return sig->dryos50_offset;
    case 51:    return sig->dryos51_offset;
    case 52:    return sig->dryos52_offset;
    case 54:    return sig->dryos54_offset;
    case 55:    return sig->dryos55_offset;
    case 57:    return sig->dryos57_offset;
    case 58:    return sig->dryos58_offset;
    }
    return 0;
}

//------------------------------------------------------------------------------------------------------------

// Loop through firmware looking for instances of a sig string
// For each one found call the check_match function to see if it matches the sig
// Return 1 if match found, else return 0
int fw_string_process(firmware *fw, string_sig *sig, int (*check_match)(firmware *fw, string_sig *sig, int j), int inc_eos)
{
    int nlen = strlen(sig->ev_name);
    uint32_t nm0 = *((uint32_t*)sig->ev_name);
    uint32_t *p;
    int j;
    BufRange *br;

    for (br = fw->br; br != 0; br = br->next)
    {
        for (p = br->p, j = br->off; j < br->off+br->len-nlen/4; p++, j++)
        {
            if ((nm0 == *p) && (memcmp(p+1,sig->ev_name+4,nlen-4+inc_eos) == 0))
            {
                if (check_match(fw,sig,j))
                    return 1;
            }
        }
    }

    return 0;
}

// As above; but scan the firmware byte-by-byte rather than by words
// Slower; but required when strings are not stores on 32 bit boundaries
int fw_string_process_unaligned(firmware *fw, string_sig *sig, int (*check_match)(firmware *fw, string_sig *sig, int j))
{
    int nlen = strlen(sig->ev_name);
    char *p;
    int j;
    BufRange *br;

    for (br = fw->br; br != 0; br = br->next)
    {
        for (p = (char*)br->p, j = 0; j < br->len*4-nlen; p++, j++)
        {
            if (strcmp(p,sig->ev_name) == 0)
            {
                if (check_match(fw,sig,j+br->off*4))
                    return 1;
            }
        }
    }

    return 0;
}

// Loop through the firmware from the start, calling 'check_match' for each location
// If 'check_match' returns non-zero, exit returning 1.
// Otherwise returns 0 - no match found.
int fw_process(firmware *fw, string_sig *sig, int (*check_match)(firmware *fw, string_sig *sig, int j))
{
    uint32_t *p;
    int j;
    BufRange *br;

    for (br = fw->br; br != 0; br = br->next)
    {
        for (p = br->p, j = br->off; j < br->off+br->len; p++, j++)
        {
            if (check_match(fw,sig,j))
                return 1;
        }
    }

    return 0;
}

//------------------------------------------------------------------------------------------------------------

// New string / signature matching functions

// Sig pattern:
//      Function pointer    -   DCD func
//      String              -   DCB "func"
// Note: 'func' may not be the target address, the sig->offset value allows selection of another function
//       called at a fixed number of instructions from the 'func' address found
int match_strsig1(firmware *fw, string_sig *sig, int j)
{
    uint32_t fadr = fwval(fw,j-1);      // function address
    if (idx_valid(fw,adr2idx(fw,fadr))) // is function address valid
    {
        // If function address is a B, and we are following branches, then follow the first B
        if (sig->offset > 1) fadr = followBranch(fw, fadr, 1);
        // Follow any subsequent branch at the given offset
        fadr = followBranch2(fw, fadr, sig->offset);
        fwAddMatch(fw,fadr,32,0,101);
        return 1;
    }
    return 0;
}

// Sig pattern:
//      String pointer      -       DCD str
//      Function pointer    -       DCD func
//                ...
//      String              -   str DCB "func"
// Note: 'func' may not be the target address, the offset value allows selection of another function
//       called at a fixed number of instructions from the 'func' address found
int match_strsig2a(firmware *fw, int k, uint32_t sadr, uint32_t offset)
{
    if (fwval(fw,k) == sadr)                // pointer to string?
    {
        uint32_t fadr = fwval(fw,k+1);      // function address
        if (idx_valid(fw,adr2idx(fw,fadr))) // is function address valid
        {
            uint32_t bfadr = followBranch2(fw, fadr, offset);
            if ((offset <= 1) || (bfadr != fadr))
            {
                fwAddMatch(fw,bfadr,32,0,102);
                return 1;
            }
        }
    }
    return 0;
}
int match_strsig2(firmware *fw, string_sig *sig, int j)
{
    // Note - 'j' is byte offset in firmware not instruction index (called from fw_string_process_unaligned)
    return search_fw(fw, match_strsig2a, fw->base + j, sig->offset, 2);
}

// Sig pattern:
//      Load Func Address   -   ADR Rx, func
//      Load String Address -   ADR Rx, "func"
//      Branch              -   BL
//              ...
//      String              -   DCB "func"
// or
//      Load Func Address   -   ADR Rx, func
//                              B   loc
//              ...
//                          loc:
//      Load String Address -   ADR    Rx, "func"
//      Branch              -   BL
//              ...
//      String              -   DCB "func"
// 'sadr' = address of "func" string
// Note: 'func' may not be the target address, the offset value allows selection of another function
//       called at a fixed number of instructions from the 'func' address found
int match_strsig3a(firmware *fw, int k, uint32_t sadr, uint32_t offset)
{
    if (isADR_PC(fw,k+1) &&    // ADR ?
        isBorBL(fw,k+2))       // B or BL ?
    {
        uint32_t padr = ADR2adr(fw,k+1);    // get address pointed to by 2nd ADR instructioin
        if (padr == sadr)                   // does it match target string
        {
            int j2 = k;
            int found = 0;
            if (isADR_PC(fw,k))             // ADR ?
                found = 1;
            else
            {
                // May be DCD block between 1st and 2nd ADR
                for (j2 = k-2; j2 >= 0 && j2 >= k-4096; j2--)
                {
                    if (isADR_PC(fw,j2) &&  // ADR ?
                        isB(fw,j2+1))       // B
                    {
                        uint32_t fa = idx2adr(fw,j2+1);
                        fa = followBranch(fw,fa,1);
                        if (adr2idx(fw,fa) == k+1)
                        {
                            found = 1;
                            break;
                        }
                    }
                }
            }
            if (found)
            {
                uint32_t fadr = ADR2adr(fw,j2);
                if (offset > 1) fadr = followBranch(fw, fadr, 1);
                fadr = followBranch2(fw, fadr, offset);
                fwAddMatch(fw,fadr,32,0,103);
                return 1;
            }
        }
    }
    return 0;
}
int match_strsig3(firmware *fw, string_sig *sig, int j)
{
    return search_fw(fw, match_strsig3a, idx2adr(fw,j), sig->offset, 3);
}

// Sig pattern:
//      Save Regs           -   STMFD
//                ... (offset)
//      Load String Address -   ADR Rx, "func"
//                ...
//      String              -   DCB "func"
int match_strsig4a(firmware *fw, int k, uint32_t sadr, uint32_t offset)
{
    if (isSTMFD(fw,k) &&        // STMFD
        isADR_PC(fw,k+offset))  // ADR ?
    {
        uint32_t padr = ADR2adr(fw,k+offset);
        if (padr == sadr)
        {
            uint32_t fadr = idx2adr(fw,k);
            fwAddMatch(fw,fadr,32,0,104);
            return 1;
        }
    }
    return 0;
}
int match_strsig4(firmware *fw, string_sig *sig, int j)
{
    return search_fw(fw, match_strsig4a, idx2adr(fw,j), sig->offset, sig->offset+1);
}

// Sig pattern:
//      Load Func Address   -   LDR Rx, =func
//      Load String Address -   xDR Rx, "func"  (LDR or ADR)
//      Branch              -   BL
//              ...
//      String              -   DCB "func"
// or
//      Load Func Address   -   LDR Rx, =func
//                              B   loc
//              ...
//                          loc:
//      Load String Address -   xDR Rx, "func"  (LDR or ADR)
//      Branch              -   BL
//              ...
//      String              -   DCB "func"
static int dryos_ofst;
int match_strsig5a(firmware *fw, int k, uint32_t sadr, uint32_t offset)
{
    if ((isADR_PC(fw,k+1) || isLDR_PC(fw,k+1)) &&   // LDR or ADR ?
        isBorBL(fw,k+2))                            // B or BL ?
    {
        uint32_t padr;
        if (isLDR_PC(fw,k+1))                       // LDR ?
            padr = LDR2val(fw,k+1);
        else
            padr = ADR2adr(fw,k+1);
        if (padr == sadr)
        {
            int j2 = k;
            int found = 0;
            if (isLDR_PC(fw,k))                     // LDR ?
                found = 1;
            else
            {
                for (j2 = k-2; j2 >= 0 && j2 >= k-4096; j2--)
                {
                    if (isLDR_PC(fw,j2) &&  // LDR ?
                        isB(fw,j2+1))       // B
                    {
                        if (idxFollowBranch(fw,j2+1,1) == k+1)
                        {
                            found = 1;
                            break;
                        }
                    }
                }
            }
            if (found)
            {
                uint32_t fadr = LDR2val(fw,j2);
                if (offset > 1) fadr = followBranch(fw, fadr, 1);
                fadr = followBranch2(fw, fadr, offset);
                if (dryos_ofst != 0)
                {
                    uint32_t fadr2 = followBranch(fw, fadr, dryos_ofst);
                    if (fadr == fadr2) return 0;
                    fadr = fadr2;
                }
                fwAddMatch(fw,fadr,32,0,105);
                return 1;
            }
        }
    }
    return 0;
}
int match_strsig5(firmware *fw, string_sig *sig, int j)
{
    dryos_ofst = dryos_offset(fw,sig);
    return search_fw(fw, match_strsig5a, idx2adr(fw,j), sig->offset, 3);
}

// Sig pattern:
//    Function immediately preceeding string
int match_strsig6(firmware *fw, string_sig *sig, int j)
{
    int j1 = find_inst_rev(fw, isSTMFD_LR, j-1, j-1);
    if (j1 > 0)
    {
        uint32_t fadr = idx2adr(fw,j1);
        fwAddMatch(fw,fadr,32,0,106);
        return 1;
    }

    return 0;
}

// Sig pattern:
//      Str ref -   xDRnn Rx, =str_ptr
//            ...
//                  BL  func
//            ...
//      String      DCB "str"
int match_strsig7a(firmware *fw, int k, uint32_t sadr, uint32_t offset)
{
    if (isADR_PC_cond(fw,k) || isLDR_PC_cond(fw,k)) // LDR or ADR ?
    {
        uint32_t padr;
        if (isLDR_PC_cond(fw,k)) // LDR ?
            padr = LDR2val(fw,k);
        else
            padr = ADR2adr(fw,k);
        if (padr == sadr)
        {
            int j2 = find_inst(fw, isBL, k+1, 10);
            if (j2 > 0)
            {
                uint32_t fa = idx2adr(fw,j2);
                fa = followBranch2(fw,fa,offset);
                fwAddMatch(fw,fa,32,0,107);
                return 1;
            }
        }
    }
    return 0;
}
int match_strsig7(firmware *fw, string_sig *sig, int j)
{
    return search_fw(fw, match_strsig7a, idx2adr(fw,j), sig->offset, 2);
}

// Sig pattern:
//      Special case for WriteSDCard
int ofst;
int match_strsig8(firmware *fw, string_sig *sig, int j)
{
    int j1;
    for (j1=j-2; j1<j+8; j1++)
    {
        uint32_t fadr = idx2adr(fw,j1);
        if (fwval(fw,j1) >= fw->base)   // pointer ??
        {
            int j2;
            for (j2=j1-1; j2>=j1-1000 && j2>=0; j2--)
            {
                if (isLDR_PC(fw,j2) && (LDR2adr(fw,j2) == fadr))    // LDR ?
                {
                    if ((isSTR(fw,j2+1) && (fwOp2(fw,j2+1) == ofst)) ||    // STR ?
                        (isSTR(fw,j2+2) && (fwOp2(fw,j2+2) == ofst)))      // STR ?
                    {
                        fadr = fwval(fw,j1);
                        if (idx_valid(fw,adr2idx(fw,fadr)))
                        {
                            fwAddMatch(fw,fadr,32,0,108);
                            return 1;
                        }
                    }
                }
            }
        }
    }

    return 0;
}
int find_strsig8(firmware *fw, string_sig *sig)
{
    uint32_t fadr = 0;
    int srch = 20;

    // Find "UpdateMBROnFlash" code
    int j = get_saved_sig(fw,"UpdateMBROnFlash");
    if (j >= 0)
    {
        fadr = func_names[j].val;
    }
    else
    {
        j = get_saved_sig(fw,"MakeSDCardBootable");
        if (j >= 0)
        {
            fadr = func_names[j].val;
            srch = 32;
        }
    }

    if (fadr == 0) return 0;

    int idx = adr2idx(fw, fadr);
    ofst = -1;

    for (j=idx+srch; j<idx+srch+12; j++)
    {
        if (isLDR(fw,j) && isLDR(fw,j+1) && isLDR(fw,j+2))
        {
            ofst = fwOp2(fw,j) + fwOp2(fw,j+1) + fwOp2(fw,j+2);
            break;
        }
    }

    if (ofst == -1) return 0;

    return fw_string_process(fw, sig, match_strsig8, 1);
}

// Sig pattern:
//      Func is B/BL @ offset from previous found func
//          prev_func
//              ... (offset)
//              BL    func
int find_strsig9(firmware *fw, string_sig *sig)
{
    int j = get_saved_sig(fw,sig->ev_name);
    if (j >= 0)
    {
        if (func_names[j].val != 0)
        {
            int ofst = dryos_offset(fw, sig);
            uint32_t fadr = followBranch(fw, func_names[j].val+ofst*4, 0xF1000001);
            if ((sig->offset == -1) || (fadr != func_names[j].val+ofst*4))
            {
                uint32_t fadr2 = fadr;
                if (sig->offset != -1) fadr2 = followBranch2(fw, fadr2, sig->offset);
                if ((sig->offset <= 0) || (fadr2 != fadr))
                {
                    fwAddMatch(fw,fadr2,32,0,109);
                    return 1;
                }
            }
        }
    }

    return 0;
}

// Sig pattern:
//      Func            -   func
//            .... (offset)
//      Ref to string   -       ADR Rx, str
//            ....
//      String          -       DCB "str"
int match_strsig11(firmware *fw, string_sig *sig, int j)
{
    int ofst = dryos_offset(fw, sig);

    uint32_t sadr = idx2adr(fw,j);        // string address
    int j1;
    for (j1 = j-1; j1 >= 0; j1--)
    {
        if (isADR_PC_cond(fw,j1))   // ADR ?
        {
            uint32_t padr = ADR2adr(fw,j1);
            if (padr == sadr)
            {
                uint32_t fadr = idx2adr(fw,j1-ofst);
                uint32_t bfadr = followBranch(fw,fadr,sig->offset);
                // special case for 'set_control_event'
                int found = 0;
                if (strcmp(sig->name,"set_control_event") == 0)
                {
                    int j2 = j1 - ofst;
                    if (isBL_cond(fw,j2) &&                                             // BLxx
                        isLDR_SP(fw,j2+1) && (fwRd(fw,j2+1) == 0) &&                    // LDR R0,[SP,x]
                        isBL(fw,j2+2) &&                                                // BL
                        isMOV(fw,j2+3) && (fwRd(fw,j2+3) == 4) && (fwRn(fw,j2+3) == 0)) // LDR R4, R0
                    {
                        found = 1;
                    }
                }
                else
                    found = 1;
                if (found && ((sig->offset == 0) || (bfadr != fadr)))
                {
                    fwAddMatch(fw,bfadr,32,0,111);
                    return 1;
                }
            }
        }
    }

    return 0;
}

// Sig pattern:
//      Func is referenced in 'CreateJumptable'
//          LDR R1, =func
//          STR R1, [R0,nnn]
//      nnn - dryos version dependant offset
int find_strsig12(firmware *fw, string_sig *sig)
{
    int j = get_saved_sig(fw,"CreateJumptable");

    int ofst = dryos_offset(fw, sig);

    if (ofst == 0) return 0;

    if (j >= 0)
    {
        if (func_names[j].val != 0)
        {
            int idx = adr2idx(fw, func_names[j].val);
            for(; !isBX_LR(fw,idx); idx++)  // BX LR
            {
                if (((fwval(fw,idx+1) & 0xFFFFF000) == 0xE5801000) && // STR R1,[R0,nnn]
                    (fwOp2(fw,idx+1) == ofst))
                {
                    uint32_t fadr = LDR2val(fw,idx);
                    uint32_t bfadr = followBranch2(fw,fadr,sig->offset);
                    if ((sig->offset <= 1) || ((bfadr != fadr) && ((fw->buf[adr2idx(fw,fadr)] & 0xFFFF0000) == 0xE92D0000)))
                    {
                        fwAddMatch(fw,bfadr,32,0,112);
                        return 1;
                    }
                }
                else if (isB(fw,idx))    // B
                {
                    idx = adr2idx(fw,followBranch(fw,idx2adr(fw,idx),1)) - 1;
                }
            }
        }
    }

    return 0;
}

// Sig pattern:
//      Func    -   func
//          ... (offset)
//      Str ref -       LDR Rx, =str_ptr
//          ...
//      Str ptr -       DCD str_ptr
//          ...
//      Str ptr -   str_ptr
//                      DCD str
//          ...
//      String          DCB "str"
int match_strsig13a(firmware *fw, int k, uint32_t sadr, uint32_t offset)
{
    if (fwval(fw,k) == sadr)    // string ptr
    {
        uint32_t padr = idx2adr(fw,k);      // string ptr address
        int j2;
        for (j2 = k-1; j2 >= 0; j2--)
        {
            if (fwval(fw,j2) == padr)       // string ptr address
            {
                uint32_t ppadr = idx2adr(fw,j2);        // string ptr ptr address
                int j3;
                for (j3 = j2-1; j3 >= 0; j3--)
                {
                    if (isLDR_PC(fw,j3) && (LDR2adr(fw,j3) == ppadr))
                    {
                        uint32_t fadr = idx2adr(fw,j3-offset);
                        fwAddMatch(fw,fadr,32,0,113);
                        return 1;
                    }
                }
            }
        }
    }
    return 0;
}
int match_strsig13(firmware *fw, string_sig *sig, int j)
{
    // Note - 'j' is offset in firmware not instruction index (called from fw_string_process_unaligned)
    return search_fw(fw, match_strsig13a, fw->base + j, sig->offset, 1);
}

// Sig pattern:
//                  BL  func
//            ...
//      Str ref -   xDR Rx, =str_ptr
//            ...
//      String      DCB "str"
// dryos_ofst: search range limit for the previous bl instruction
int match_strsig15a(firmware *fw, int k, uint32_t sadr, uint32_t offset)
{
    if (isADR_PC_cond(fw,k) || isLDR_PC_cond(fw,k))   // LDR or ADR ?
    {
        uint32_t padr;
        if (isLDR_PC_cond(fw,k)) // LDR ?
            padr = LDR2val(fw,k);
        else
            padr = ADR2adr(fw,k);
        if (padr == sadr)
        {
            int j2 = find_inst_rev(fw, isBL, k-1, dryos_ofst);
            if (j2 > 0)
            {
                uint32_t fa = idx2adr(fw,j2);
                fa = followBranch2(fw,fa,offset);
                fwAddMatch(fw,fa,32,0,115);
                return 1;
            }
        }
    }
    return 0;
}
int match_strsig15(firmware *fw, string_sig *sig, int j)
{
    dryos_ofst = dryos_offset(fw,sig);
    if (dryos_ofst == 0) dryos_ofst = 50;
    return search_fw(fw, match_strsig15a, idx2adr(fw,j), sig->offset, 1);
}

// Sig pattern:
//      Function immediately preceeding usage of hex value
int match_strsig16(firmware *fw, string_sig *sig, int j)
{
    uint32_t nm0 = *((uint32_t*)sig->ev_name);

    if (isADR_PC_cond(fw,j) || isLDR_PC_cond(fw,j))   // LDR or ADR ?
    {
        uint32_t padr;
        if (isLDR_PC_cond(fw,j)) // LDR ?
            padr = LDR2val(fw,j);
        else
            padr = ADR2adr(fw,j);
        if (padr == nm0)
        {
            int j2 = find_inst_rev(fw, isSTMFD_LR, j-1, 50);
            if (j2 > 0)
            {
                uint32_t fa = idx2adr(fw,j2);
                fwAddMatch(fw,fa,32,0,116);
                return 1;
            }
        }
    }

    return 0;
}

// Sig pattern:
//      Special case for ScreenLock & ScreenUnlock
int find_strsig17(firmware *fw, string_sig *sig)
{
    int j = get_saved_sig(fw,"StartRecModeMenu");

    if (j >= 0)
    {
        if (func_names[j].val != 0)
        {
            int idx = adr2idx(fw, func_names[j].val);
            int k = 0;
            if (fw->dryos_ver < 58)
            {
                if (isLDR_PC(fw,idx-3) && isMOV_immed(fw,idx-2) && isB(fw,idx-1))
                {
                    k = adr2idx(fw,LDR2val(fw,idx-3));
                }
                else if (isMOV_immed(fw,idx-3) && isADR_PC(fw,idx-2) && isB(fw,idx-1))
                {
                    k = adr2idx(fw,ADR2adr(fw,idx-2));
                }
            }
            else
            {
                int l = find_inst_rev(fw, isBL, idx-1, 4);
                if (l > 0)
                {
                    if (isLDR_PC(fw,l-2) && isMOV_immed(fw,l-1))
                    {
                        k = adr2idx(fw,LDR2val(fw,l-2));
                    }
                }
            }
            if (k != 0)
            {
                uint32_t fadr = 0;
                if (strcmp(sig->name,"ScreenLock") == 0)
                    fadr = followBranch(fw,idx2adr(fw,k+1),0x01000001);
                else
                {
                    k = find_inst(fw, isLDMFD, k+1, 60);
                    if (fw->dryos_ver < 58)
                    {
                        fadr = followBranch(fw,idx2adr(fw,k-1),0x01000001);
                    }
                    else
                    {
                        k = find_inst_rev(fw, isBL, k-1, 4);
                        if (k > 0)
                        {
                            fadr = followBranch(fw,idx2adr(fw,k),0x01000001);
                        }
                    }
                }
                if (fadr > 0)
                {
                    fwAddMatch(fw,fadr,32,0,117);
                    return 1;
                }
            }
        }
    }

    return 0;
}

// Sig pattern:
//      Func is offset from previously found func
//          prev_func
//              ... (offset, less than 0xff)
//          func
// dryos_offset to be encoded as: 0xQRPP
// PP) offset, 0x00...0xff, in words
// Q) 'previous instruction flag': 0 for LDMFD, 1 for B, 2 for conditional B, 3 for mov pc,lr, 4 for bx lr
// R) 'additional gap size': 0x0 ... 0xf, to skip an additional gap, in words
// offset is negative if dryos_offset is negative
int find_strsig19(firmware *fw, string_sig *sig)
{
    int j = get_saved_sig(fw,sig->ev_name);
    if (j >= 0)
    {
        if (func_names[j].val != 0) // address, not idx
        {
            int ofst = dryos_offset(fw, sig);
            int neg = 1;
            if (ofst<0)
            {
                neg=-1;
                ofst=-ofst;
            }
            int addoffs = (ofst & 0xf00)>>8;
            uint32_t fadr = func_names[j].val+neg*(ofst&0xff)*4;
            int k = 0;
            switch ((ofst&0xf000)>>12)
            {
                case 0: k = isLDMFD_PC(fw, adr2idx(fw, fadr)-1-addoffs); break;
                case 1: k = isB(fw, adr2idx(fw, fadr)-1-addoffs); break;
                case 2: k = ((fwval(fw, adr2idx(fw, fadr)-1-addoffs) & 0x0f000000) == 0x0a000000); break; // B cond.
                case 3: k = (fwval(fw, adr2idx(fw, fadr)-1-addoffs) == 0xE1A0F00E); break; // mov pc, lr
                case 4: k = (fwval(fw, adr2idx(fw, fadr)-1-addoffs) == 0xE12FFF1E); break; // bx lr
                default: return 0;
            }
            if (k)
            {
                fwAddMatch(fw,fadr,32,0,119);
                return 1;
            }
        }
    }

    return 0;
}

// Sig pattern:
//      Str ref -   xDRnn Rx, =str_ptr
//            ...
//                  BL  func
//            ...
//      String      DCB "str"
// offset: interpreted as max search distance
// dryos_ofst: 0 for 1st B/BL after str ref, 1 for 2nd, etc.
// based on method 7
int match_strsig23a(firmware *fw, int k, uint32_t sadr, uint32_t maxdist)
{
    if (isADR_PC_cond(fw,k) || isLDR_PC_cond(fw,k)) // LDR or ADR ?
    {
        uint32_t padr;
        if (isLDR_PC_cond(fw,k)) // LDR ?
            padr = LDR2val(fw,k);
        else
            padr = ADR2adr(fw,k);
        if (padr == sadr)
        {
            int j2 = find_Nth_inst(fw, isBorBL, k+1, maxdist, dryos_ofst+1);
            if (j2 > 0)
            {
                uint32_t fa = idx2adr(fw,j2);
                fa = followBranch2(fw,fa,0x01000001);
                fwAddMatch(fw,fa,32,0,123);
                return 1;
            }
        }
    }
    return 0;
}
int match_strsig23(firmware *fw, string_sig *sig, int j)
{
    dryos_ofst = dryos_offset(fw,sig);
    return search_fw(fw, match_strsig23a, idx2adr(fw,j), sig->offset, 2);
}

// Sig pattern:
//      Func            -   func
//            .... [offset]
//      Nth instruction backwards from string ref
//      Ref to string   -       LDR Rx, pstr
//            ....
//      Ptr to string   -       DCD pstr
// dryos_offset to be encoded as: 0xQNPP
// P) additional offset
// Q) instruction to locate: 0 for STMFD_LR, 0xf to disable
// N) search the instruction this many times - 1 (i.e. 0 = first instruction backwards)
// prev instruction search range is limited by sig->offset
// the reference to the string has to occur within 256 words (fixed range)
// based on method 104 of finsig_vxworks
int match_strsig24(firmware *fw, string_sig *sig, int j)
{

    int ofst = dryos_offset(fw, sig);
    int prinst = (ofst&0xf000)>>12;
    int ninst = ((ofst&0xf00)>>8)+1;
    ofst &= 0xff;
    void *instid;
    switch (prinst) {
        case 0:
            instid = (void*)isSTMFD_LR;
            break;
        default:
            return 0;
    }

    uint32_t sadr = idx2adr(fw,j);        // string address
    int j1;
    for (j1 = j; j1 >= MAX(j-256,0); j1--)
    {
        if (isLDR(fw,j1))   // LDR ?
        {
            uint32_t pval = LDR2val(fw,j1);
            if (pval == sadr)
            {
                int j2 = find_Nth_inst_rev(fw,instid,j1-1,sig->offset,ninst);
                if (j2>0)
                {
                    fwAddMatch(fw,idx2adr(fw,j2-ofst),32,0,124);
                    return 1;
                }
            }
        }
        else if (isADR_PC(fw,j1))   // ADR ?
        {
            uint32_t pval = ADR2adr(fw,j1);
            if (pval == sadr)
            {
                int j2 = find_Nth_inst_rev(fw,instid,j1-1,sig->offset,ninst);
                if (j2>0)
                {
                    fwAddMatch(fw,idx2adr(fw,j2-ofst),32,0,124);
                    return 1;
                }
            }
        }
    }

    return 0;
}

// Call processing function based on type
int find_strsig(firmware *fw, string_sig *sig)
{
    switch (sig->type)
    {
    case 1:     return fw_string_process(fw, sig, match_strsig1, 1);
    case 2:     return fw_string_process_unaligned(fw, sig, match_strsig2);
    case 3:     return fw_string_process(fw, sig, match_strsig3, 1);
    case 4:     return fw_string_process(fw, sig, match_strsig4, 1);
    case 5:     return fw_string_process(fw, sig, match_strsig5, 1);
    case 6:     return fw_string_process(fw, sig, match_strsig6, 1);
    case 7:     return fw_string_process(fw, sig, match_strsig7, 1);
    case 8:     return find_strsig8(fw, sig);
    case 9:     return find_strsig9(fw, sig);
    case 11:    return fw_string_process(fw, sig, match_strsig11, 0);
    case 12:    return find_strsig12(fw, sig);
    case 13:    return fw_string_process_unaligned(fw, sig, match_strsig13);
    case 15:    return fw_string_process(fw, sig, match_strsig15, 1);
    case 16:    return fw_process(fw, sig, match_strsig16);
    case 17:    return find_strsig17(fw, sig);
    case 19:    return find_strsig19(fw, sig);
    case 20:
        {
            int j = find_saved_sig(sig->ev_name);
            if (j >= 0)
            {
                uint32_t fadr = followBranch2(fw,func_names[j].val,sig->offset);
                fwAddMatch(fw,fadr,32,0,120);
                return 1;
            }
            return 0;
        }
    case 21:    return fw_process(fw, sig, (int (*)(firmware*, string_sig*, int))(sig->ev_name));
    case 22:    return ((int (*)(firmware*))(sig->ev_name))(fw);
    case 23:    return fw_string_process(fw, sig, match_strsig23, 1);
    case 24:    return fw_string_process(fw, sig, match_strsig24, 0);
    }

    return 0;
}

//------------------------------------------------------------------------------------------------------------

// Matching functions

// Search using new style matching only
void find_str_sig_matches(firmware *fw, const char *curr_name)
{
    int i;

    int found_ev = 0;

    count = 0;

    for (i = 0; string_sigs[i].ev_name != 0 && !found_ev; i++)
    {
        if (strcmp(curr_name, string_sigs[i].name) == 0)
        {
            if (find_strsig(fw, &string_sigs[i]))
            {
                found_ev = 1;
                break;
            }
        }
    }

    if (count > 1)
    {
        qsort(matches, count, sizeof(Match), (void*)match_compare);
    }

    if (count > 0)
    {
        save_sig(curr_name, matches->ptr);
    }
}

// Search using new style matching first
// If not found used old style matching
void find_matches(firmware *fw, const char *curr_name)
{
    FuncSig *sig, *s;
    BufRange *n;
    uint32_t *p;
    int i, j;
    int fail, success;

    int found_ev = 0;

    count = 0;

    // Already found (eventproc)?
    i = find_saved_sig(curr_name);
    if (i >= 0)
    {
        if ((func_names[i].val != 0) && (func_names[i].flags & EV_MATCH) != 0)
        {
            fwAddMatch(fw,func_names[i].val,32,0,120);
            found_ev = 1;
        }
    }


    // Try and match using 'string' based signature matching first
    for (i = 0; string_sigs[i].ev_name != 0 && !found_ev; i++)
    {
        if (strcmp(curr_name, string_sigs[i].name) == 0)
        {
            if (find_strsig(fw, &string_sigs[i]))
            {
                found_ev = 1;
                break;
            }
        }
    }

    // If not found see if the name is in the old style instruction compare match table
    // Set start value for j in next section if found
    if (!found_ev)
    {
        found_ev = 1;
        for (j=0; func_list[j].name; j++)
        {
            if (strcmp(curr_name,func_list[j].name) == 0)
            {
                found_ev = 0;
                break;
            }
        }
    }

    // Not found so far, try instruction comparison matching
    while (!found_ev)
    {
        sig = func_list[j].sig;

        for (n = fw->br; n != 0; n = n->next)
        {
            for (p = n->p, i = 0; i < n->len; p++, i++)
            {
                fail = 0;
                success = 0;
                for (s = sig; s->offs != -1; s++)
                {
                    if ((p[s->offs] & s->mask) != s->value)
                        fail++;
                    else
                        success++;
                }
                // If sig starts with STMFD and first instruction does not match ignore it
                if (((p[sig->offs] & sig->mask) != sig->value) && (sig->offs == 0) && (sig->value == 0xe92d0000)) success = 0;
                if (success > fail)
                {
                    if (s->mask == -2)
                    {
                        int end_branch = 0;
                        int idx = 0;
                        uint32_t *p1 = 0;
                        if ((fw->buf[n->off+i+s->value] & 0x0F000000) == 0x0A000000)   // B
                        {
                            idx = adr2idx(fw, followBranch2(fw, idx2adr(fw,n->off+i+s->value), 0xF0000001));
                            if ((idx >= 0) && (idx < fw->size))
                            {
                                end_branch = 1;
                                p1 = &fw->buf[idx];
                            }
                        }
                        int fail2 = 0;
                        int success2 = 0;
                        //fprintf(stderr,"\t%s %d %08x %08x %d %d\n",curr_name,idx,idx2adr(fw,idx),idx2adr(fw,i+n->off),success,fail);
                        s++;
                        for (; s->offs != -1; s++)
                        {
                            if (!end_branch || (p1[s->offs] & s->mask) != s->value){
                                fail2++;
                            } else {
                                success2++;
                            }
                        }
                        if (fail2 == 0)
                        {
                            success = success + fail + success2;
                            fail = 0;
                        }
                        else
                        {
                            success = success + success2;
                            fail = fail + fail2;
                        }
                        //fprintf(stderr,"\t%s %d %08x %08x %d %d\n",curr_name,idx,idx2adr(fw,idx),idx2adr(fw,i+n->off),success,fail);
                    }
                }
                if (success > fail)
                {
                    // Special case for drive space functions, see if there is a refernce to "Mounter.c" in the function
                    // Increase match % if so, increase fail count if not
                    if ((strcmp(curr_name, "GetDrive_ClusterSize") == 0) ||
                        (strcmp(curr_name, "GetDrive_FreeClusters") == 0) ||
                        (strcmp(curr_name, "GetDrive_TotalClusters") == 0))
                    {
                        int fnd = 0;
                        for (s = sig; s->offs != -1; s++)
                        {
                            if (isLDR_PC_cond(fw,n->off+i+s->offs))
                            {
                                int m = adr2idx(fw,LDR2val(fw,n->off+i+s->offs));
                                if ((m >= 0) && (m < fw->size) && (strcmp((char*)(&fw->buf[m]),"Mounter.c") == 0))
                                {
                                    fnd = 1;
                                }
                            }
                            else if (isADR_PC_cond(fw,n->off+i+s->offs))
                            {
                                int m = adr2idx(fw,ADR2adr(fw,n->off+i+s->offs));
                                if ((m >= 0) && (m < fw->size) && (strcmp((char*)(&fw->buf[m]),"Mounter.c") == 0))
                                {
                                    fnd = 1;
                                }
                            }
                        }
                        if (fnd)
                            success++;
                        else
                            fail++;
                    }
                    fwAddMatch(fw,idx2adr(fw,i+n->off),success,fail,func_list[j].ver);
                    if (count >= MAX_MATCHES)
                    {
                        bprintf("// WARNING: too many matches for %s!\n", func_list[j].name);
                        break;
                    }
                }
            }
        }

        // same name, so we have another version of the same function
        if ((func_list[j+1].name == NULL) || (strcmp(curr_name, func_list[j+1].name) != 0))
        {
            found_ev = 1;
            break;
        }
        j++;
    }

    if (count > 1)
    {
        qsort(matches, count, sizeof(Match), (void*)match_compare);
    }

    if (count > 0)
    {
        save_sig(curr_name, matches->ptr);
    }
}

// Output match results for function
void print_results(firmware *fw, const char *curr_name, int k)
{
    int i;
    int err = 0;
    char line[500] = "";

    if (func_names[k].flags & DONT_EXPORT) return;

    // find best match and report results
    osig* ostub2 = find_sig(fw->sv->stubs,curr_name);

    if ((count == 0)
        || (matches->fail > 0)
        || (ostub2 && (matches->ptr != ostub2->val))
       )
    {
        if (!ostub2 || (ostub2->type != TYPE_IGNORE))
            err = 1;
        func_names[k].flags |= BAD_MATCH;
    }
    else
    {
        if (func_names[k].flags & UNUSED) return;
    }

    // write to header (if error) or body buffer (no error)
    out_hdr = err;

    char *macro = "NHSTUB";
    if (strncmp(curr_name,"task_",5) == 0 ||
        strncmp(curr_name,"hook_",5) == 0) macro = "   DEF";

    if (count == 0)
    {
        if (func_names[k].flags & OPTIONAL) return;
        char fmt[50] = "";
        sprintf(fmt, "// ERROR: %%s is not found. %%%ds//--- --- ", (int)(34-strlen(curr_name)));
        sprintf(line+strlen(line), fmt, curr_name, "");
    }
    else
    {
        if (ostub2 || (func_names[k].flags & UNUSED))
            sprintf(line+strlen(line),"//%s(%-37s,0x%08x) //%3d ", macro, curr_name, matches->ptr, matches->sig);
        else
            sprintf(line+strlen(line),"%s(%-39s,0x%08x) //%3d ", macro, curr_name, matches->ptr, matches->sig);

        if (matches->fail > 0)
            sprintf(line+strlen(line),"%2d%% ", matches->success*100/(matches->success+matches->fail));
        else
            sprintf(line+strlen(line),"    ");
    }

    if (ostub2)
    {
        if (ostub2->type == TYPE_IGNORE)
            sprintf(line+strlen(line),"       Overridden");
        else if ((count > 0) && (matches->ptr == ostub2->val))
            sprintf(line+strlen(line),"       == 0x%08x    ",ostub2->val);
        else
            sprintf(line+strlen(line),"   *** != 0x%08x    ",ostub2->val);
    }
    else
        sprintf(line+strlen(line),"                        ");

    for (i=strlen(line)-1; i>=0 && line[i]==' '; i--) line[i] = 0;
    bprintf("%s\n",line);

    for (i=1;i<count && matches[i].fail==matches[0].fail;i++)
    {
        if (matches[i].ptr != matches->ptr)
        {
            bprintf("// ALT: %s(%s, 0x%x) // %d %d/%d\n", macro, curr_name, matches[i].ptr, matches[i].sig, matches[i].success, matches[i].fail);
        }
    }
}

//------------------------------------------------------------------------------------------------------------

typedef struct
{
    uint16_t mode;
    char *nm;
} ModeMapName;

ModeMapName mmnames[] = {
    { 32768,"MODE_AUTO" },
    { 32769,"MODE_M" },
    { 32770,"MODE_AV" },
    { 32771,"MODE_TV" },
    { 32772,"MODE_P" },

    { 65535,"" }
};

char* mode_name(uint16_t v)
{
    int i;
    for (i=0; mmnames[i].mode != 65535; i++)
    {
        if (mmnames[i].mode == v)
            return mmnames[i].nm;
    }

    return "";
}

void output_modemap(firmware *fw, int k)
{
    int cnt = 0;

    if (isLDR_PC(fw,k))
    {
        k = adr2idx(fw,LDR2val(fw,k));
        bprintf("%08x\n",idx2adr(fw,k));
        uint16_t *p = (uint16_t*)(&fw->buf[k]);
        k = 0;
        while ((*p != 0xFFFF) && (k < 50))
        {
            if (((fw->dryos_ver < 47) && ((*p < 8000) || (*p > 8999))) || ((fw->dryos_ver >= 47) && ((*p < 4000) || (*p > 4999))))
            {
                osig *m = find_sig_val(fw->sv->modemap, *p);
                if (!m)
                {
                    char *s = mode_name(*p);
                    bprintf("// Mode %5d in firmware but not in current modemap",*p);
                    if (strcmp(s,"") != 0)
                        bprintf(" (%s)",s);
                    bprintf("\n");
                    cnt++;
                }
                else
                {
                    m->pct = 100;
                }
            }
            p++;
            k++;
        }
    }
    osig *m = fw->sv->modemap;
    while (m)
    {
        if (m->pct != 100)    // not matched above?
        {
            bprintf("// Current modemap entry not found in firmware - %-24s %5d\n",m->nm,m->val);
            cnt++;
        }
        m = m->nxt;
    }
    if (cnt == 0)
    {
        bprintf("// No problems found with modemap table.\n");
    }
}

int match_modelist(firmware *fw, int k, uint32_t fadr, uint32_t v2)
{
    if (isBX_LR(fw,k) && (fw->buf[k+4] == fadr))
    {
        fadr = fwval(fw,k+1);
        int k1;
        for (k1=k-1; k1>k-20; k1--)
        {
            if (isLDR_PC(fw,k1) && (LDR2val(fw,k1) == fadr))
            {
                bprintf("// Firmware modemap table found @%08x -> ",idx2adr(fw,k1));
                output_modemap(fw,k1);
                return 1;
            }
        }
    }

    return 0;
}

static uint32_t FlashParamsTable_address = 0;

int match_FlashParamsTable2(firmware *fw, int k, uint32_t v1, uint32_t v2)
{
    if (fw->buf[k] == v1)
    {
        FlashParamsTable_address = idx2adr(fw,k);
        return 1;
    }
    return 0;
}

int match_FlashParamsTable(firmware *fw, int k, uint32_t v1, uint32_t v2)
{
    if ((fw->buf[k] > fw->base) && (fw->buf[k+1] == 0x00010000) && (fw->buf[k+2] == 0xFFFF0002))
    {
        if (search_fw(fw, match_FlashParamsTable2, idx2adr(fw,k), 0, 1))
            return 1;
    }
    return 0;
}

void find_modemap(firmware *fw)
{
    out_hdr = 1;

    // Find 'FlashParamsTable'
    search_fw(fw, match_FlashParamsTable, 0, 0, 1);
    // Find 'modelist'
    if (FlashParamsTable_address != 0)
    {
        add_blankline();
        bprintf("// Check of modemap from 'platform/CAMERA/shooting.c':\n");
        search_fw(fw, match_modelist, FlashParamsTable_address, 0, 5);
    }
}

//------------------------------------------------------------------------------------------------------------

int match_CAM_UNCACHED_BIT(firmware *fw, int k, int v)
{
    if ((fw->buf[k] & 0x0FFFF000) == 0x03C00000)    // BIC
    {
        fw->uncached_adr = ALUop2(fw,k);
        fw->uncached_adr_idx = k;
    }

    return 0;
}

// Search for things that go in 'platform_camera.h'
void find_platform_vals(firmware *fw)
{
    int k,k1;

    out_hdr = 1;
    add_blankline();

    bprintf("// Values below go in 'platform_camera.h':\n");
    bprintf("//#define CAM_DRYOS         1\n");
    if (fw->dryos_ver >= 39)
        bprintf("//#define CAM_DRYOS_2_3_R39 1 // Defined for cameras with DryOS version R39 or higher\n");
    if (fw->dryos_ver >= 47)
        bprintf("//#define CAM_DRYOS_2_3_R47 1 // Defined for cameras with DryOS version R47 or higher\n");

    // Find 'RAW' image size
    uint32_t raw_width = 0;
    uint32_t raw_height = 0;
    uint32_t kw=0, kh=0;

    k = find_str_ref(fw, "\r[%ld] AdjDrvType[%02ld] -> DrvType[%02");
    if (k >= 0)
    {
        // Width
        for (k1 = k-1; k1 >= k-20; k1--)
        {
            if ((fw->buf[k1] & 0x0FFF0FFF) == 0x058D0034)           // STRxx Rn, [SP,#0x34]
            {
                if ((fw->buf[k1-1] & 0x0FFF0000) == 0x03A00000)     // MOVxx Rn, #YYY
                {
                    raw_width = ALUop2(fw, k1-1);
                    kw = k1-1;
                }
                else if ((fw->buf[k1-2] & 0x0FFF0000) == 0x03A00000)// MOVxx Rn, #YYY
                {
                    raw_width = ALUop2(fw, k1-2);
                    kw = k1-2;
                }
                else if (isLDR_PC_cond(fw,k1-1))
                {
                    raw_width = LDR2val(fw,k1-1);
                    kw = k1-1;
                }
                else if (isLDR_PC_cond(fw,k1-2))
                {
                    raw_width = LDR2val(fw,k1-2);
                    kw = k1-2;
                }
            }
        }
        // Height
        for (k1 = k-1; k1 >= k-20; k1--)
        {
            if ((fw->buf[k1] & 0x0FFF0FFF) == 0x058D0030)           // STRxx Rn, [SP,#0x30]
            {
                if ((fw->buf[k1-1] & 0x0FFF0000) == 0x03A00000)     // MOVxx Rn, #YYY
                {
                    raw_height = ALUop2(fw, k1-1);
                    kh = k1-1;
                }
                else if ((fw->buf[k1-2] & 0x0FFF0000) == 0x03A00000)// MOVxx Rn, #YYY
                {
                    raw_height = ALUop2(fw, k1-2);
                    kh = k1-2;
                }
                else if (isLDR_PC_cond(fw,k1-1))
                {
                    raw_height = LDR2val(fw,k1-1);
                    kh = k1-1;
                }
                else if (isLDR_PC_cond(fw,k1-2))
                {
                    raw_height = LDR2val(fw,k1-2);
                    kh = k1-2;
                }
                if ((fw->buf[k1-1] & 0x0FFF0000) == 0x02400000)     // SUBxx Rn, #YYY
                {
                    raw_height = raw_width - ALUop2(fw, k1-1);
                    kh = k1-1;
                }
            }
        }
    }

    if ((raw_width == 0) && (raw_height == 0))
    {
        k = find_str_ref(fw, " CrwAddress %lx, CrwSize H %ld V %ld\r");
        if (k >= 0)
        {
            // Width
            for (k1=k-1; k1>=k-5; k1--)
            {
                if ((fw->buf[k1] & 0xFFFFF000) == 0xE3A02000)       // MOV R2, #nnn
                {
                    raw_width = ALUop2(fw,k1);
                    kw = k1;
                }
                else
                if (isLDR_PC(fw,k1) && ((fw->buf[k1]& 0x0000F000) == 0x00002000))   // LDR R2, =nnn
                {
                    raw_width = LDR2val(fw,k1);
                    kw = k1;
                }
            }
            // Height
            for (k1=k-1; k1>=k-5; k1--)
            {
                if ((fw->buf[k1] & 0xFFFFF000) == 0xE3A03000)       // MOV R3, #nnn
                {
                    raw_height = ALUop2(fw,k1);
                    kh = k1;
                }
                else
                if (isLDR_PC(fw,k1) && ((fw->buf[k1]& 0x0000F000) == 0x00003000))   // LDR R3, =nnn
                {
                    raw_height = LDR2val(fw,k1);
                    kh = k1;
                }
                else
                if ((fw->buf[k1] & 0xFFFFF000) == 0xE2423000)       // SUB R3, R2, #nnn
                {
                    raw_height = raw_width - ALUop2(fw,k1);
                    kh = k1;
                }
            }
        }
    }

    if (raw_width != 0)
    {
        bprintf("//#define CAM_RAW_ROWPIX    %d // Found @0x%08x\n",raw_width,idx2adr(fw,kw));
    }
    else
    {
        bprintf("//#define CAM_RAW_ROWPIX    *** Not Found ***\n");
    }
    if (raw_height != 0)
    {
        bprintf("//#define CAM_RAW_ROWS      %d // Found @0x%08x\n",raw_height,idx2adr(fw,kh));
    }
    else
    {
        bprintf("//#define CAM_RAW_ROWS      *** Not Found ***\n");
    }

    // Find 'CAM_UNCACHED_BIT'
    if (fw->uncached_adr_idx != 0)
    {
        bprintf("//#undef  CAM_UNCACHED_BIT\n");
        bprintf("//#define CAM_UNCACHED_BIT  0x%08x // Found @0x%08x\n",fw->uncached_adr,idx2adr(fw,fw->uncached_adr_idx));
    }

    // Look for CAM_DATE_FOLDER_NAMING value
    k = get_saved_sig(fw,"GetImageFolder");
    if (k >= 0)
    {
        uint32_t fadr = func_names[k].val;
        int s = adr2idx(fw,fadr);
        int e = find_inst(fw, isLDMFD_PC, s+1, 160);
        for (k1=s+1; k1<s+16; k1++)
        {
            if (isMOV(fw,k1) && (fwRnMOV(fw,k1) == 2))
            {
                int r1 = fwRd(fw,k1);
                int k2;
                for (k2=e-32; k2<e; k2++)
                {
                    int b = 0;
                    if (isMOV(fw,k2) && isBL(fw,k2+1) && (fwRnMOV(fw,k2) == r1))
                        b = 1;
                    else if (isMOV(fw,k2) && isBL(fw,k2+7) && (fwRnMOV(fw,k2) == r1))
                        b = 7;
                    if (b != 0)
                    {
                        int r2 = fwRd(fw,k2);
                        fadr = followBranch2(fw,idx2adr(fw,k2+b),0x01000001);
                        k = adr2idx(fw,fadr);
                        int k3;
                        for (k3=k; k3<k+8; k3++)
                        {
                            if (isCMP(fw,k3) && (fwRn(fw,k3) == r2))
                            {
                                int val = ALUop2(fw,k3);
                                bprintf("//#define CAM_DATE_FOLDER_NAMING 0x%03x // Found @0x%08x (pass as 3rd param to GetImageFolder)\n",val,idx2adr(fw,k3));
                            }
                        }
                    }
                }
            }
        }
    }

    // Find 'PARAM_CAMERA_NAME'
    if (FlashParamsTable_address != 0)
    {
        k1 = adr2idx(fw,FlashParamsTable_address);
        for (k=k1; k<k1+20; k++)
        {
            uint32_t fadr = fwval(fw,k);
            int k2 = adr2idx(fw,fadr);
            if (idx_valid(fw,k2))
            {
                uint32_t sadr = fwval(fw,k2);
                k2 = adr2idx(fw,sadr);
                if (idx_valid(fw,k2))
                {
                    char *s = adr2ptr(fw,sadr);
                    if (((fw->cam != 0) && (strcmp(s,fw->cam) == 0)) || (strcmp(s,"Unknown") == 0))
                    {
                        bprintf("//#define PARAM_CAMERA_NAME %d // Found @0x%08x\n",k-k1,fadr);
                        break;
                    }
                }
            }
        }
    }
}

//------------------------------------------------------------------------------------------------------------

uint32_t find_viewport_address(firmware *fw, int *kout)
{
    int k, k1;

    // find viewwport address for 'vid_get_viewport_fb'
    k = find_str_ref(fw, "VRAM Address  : %p\r");
    if (k >= 0)
    {
        for (k1=k-1; k1>k-8; k1--)
        {
            if (isLDR(fw,k1) && isLDR(fw,k1+1))
            {
                uint32_t v1 = LDR2val(fw,k1);
                uint32_t v2 = LDR2val(fw,k1+1);
                if (v2 > v1) v1 = v2;
                *kout = k1;
                return v1;
            }
        }
    }

    *kout = -1;
    return 0;
}

int match_vid_get_bitmap_fb(firmware *fw, int k, int v)
{
    if (isBL(fw,k-1) && // BL
        isLDR_PC(fw,k))
    {
        uint32_t v1 = LDR2val(fw,k);
        bprintf("//void *vid_get_bitmap_fb()        { return (void*)0x%08x; }             // Found @0x%08x\n",v1,idx2adr(fw,k));
        return 1;
    }
    else
    if (isBL(fw,k-1) && // BL
        (isLDR_PC(fw,k+1)))
    {
        uint32_t v1 = LDR2val(fw,k+1);
        bprintf("//void *vid_get_bitmap_fb()        { return (void*)0x%08x; }             // Found @0x%08x\n",v1,idx2adr(fw,k));
        return 1;
    }

    return 0;
}

int match_get_flash_params_count(firmware *fw, int k, int v)
{
    if ((fw->buf[k] & 0xFFF00FFF) == 0xE3C00901)    // BIC Rn, Rn, #0x4000
    {
        uint32_t r = fw->buf[k] & 0x000F0000;       // Register
        if (((fw->buf[k+1] & 0xFFF00000) == 0xE3500000) && ((fw->buf[k+1] & 0x000F0000) == r))  // CMP, Rn #val
        {
            bprintf("//int get_flash_params_count(void) { return 0x%02x; }                          // Found @0x%08x\n",fw->buf[k+1]&0xFFF,idx2adr(fw,k+1));
            return 1;
        }
    }

    return 0;
}

// based on match_get_flash_params_count
int match_uiprop_count(firmware *fw, int k, int v)
{
    if ((fw->buf[k] & 0xFFF00FFF) == 0xe3c00902)    // BIC Rn, Rn, #0x8000
    {
        uint32_t r = fw->buf[k] & 0x000F0000;       // Register
        if (((fw->buf[k+1] & 0xFFF00000) == 0xE3500000) && ((fw->buf[k+1] & 0x000F0000) == r))  // CMP, Rn #val
        {
            char *name = "uiprop_count";
            uint32_t fadr = fw->buf[k+1]&0xFFF;
            osig *o = find_sig(fw->sv->stubs_min,name);
            if (o)
            {
                bprintf("//DEF_CONST(%-34s,0x%08x) // Found @0x%08x",name,fadr,idx2adr(fw,k+1));
                if (fadr != o->val)
                {
                    bprintf(", ** != ** stubs_min = 0x%08x (%s)",o->val,o->sval);
                }
                else
                {
                    bprintf(",          stubs_min = 0x%08x (%s)",o->val,o->sval);
                }
            }
            else
            {
                bprintf("DEF_CONST(%-34s,0x%08x) // Found @0x%08x",name,fadr,idx2adr(fw,k+1));
            }
            bprintf("\n");
            return 1;
        }
    }

    return 0;
}

// Search for things that go in 'lib.c'
void find_lib_vals(firmware *fw)
{
    int k,k1;

    out_hdr = 1;
    add_blankline();

    bprintf("// Values below go in 'lib.c':\n");

    // Find 'vid_get_bitmap_fb'
    search_saved_sig(fw, "DispCon_ShowBitmapColorBar", match_vid_get_bitmap_fb, 0, 1, 30);

    // find 'vid_get_viewport_fb'
    uint32_t v = find_viewport_address(fw,&k);
    if (k >= 0)
    {
        bprintf("//void *vid_get_viewport_fb()      { return (void*)0x%08x; }             // Found @0x%08x\n",v,idx2adr(fw,k));
    }

    // find 'vid_get_viewport_fb_d'
    static int fbd[3][3] =
    {
        { -2, -3,  1 },
        {  1,  3,  4 },
        { -1, -2,  1 },
    };
    int sadr = find_str(fw, "ImagePlayer.c");
    k = find_nxt_str_ref(fw, sadr, -1);
    int found = 0;
    while ((k >= 0) && !found)
    {
        int f;
        for (f=0; f<3 && !found; f++)
        {
            if (isLDR(fw,k+fbd[f][0]) && isLDR(fw,k+fbd[f][1]) && isLDR(fw,k+fbd[f][2]))
            {
                int reg = fw->buf[k+fbd[f][2]] & 0x000F0000;    // Index register used
                int ka = 0;
                if (((fw->buf[k+fbd[f][0]] & 0x0000F000) << 4) == reg)      { ka = k+fbd[f][0]; }
                else if (((fw->buf[k+fbd[f][1]] & 0x0000F000) << 4) == reg) { ka = k+fbd[f][1]; }
                if (ka > 0)
                {
                    uint32_t adr = LDR2val(fw,ka);
                    for (k1=k+2; k1<k+20; k1++)
                    {
                        if (isSTR(fw,k1) && ((fw->buf[k1] & 0x000F0000) == reg))
                        {
                            uint32_t ofst = fw->buf[k1] & 0x00000FFF;
                            bprintf("DEF(%-40s,0x%08x) // Found 0x%04x (@0x%08x) + 0x%02x (@0x%08x)\n","viewport_fb_d",adr+ofst,adr,idx2adr(fw,ka),ofst,idx2adr(fw,k1));
                            bprintf("//void *vid_get_viewport_fb_d()    { return (void*)(*(int*)(0x%04x+0x%02x)); } // Found @0x%08x & 0x%08x\n",adr,ofst,idx2adr(fw,ka),idx2adr(fw,k1));
                            found = 1;
                            break;
                        }
                    }
                }
            }
        }
        k = find_nxt_str_ref(fw, sadr, k);
    }

    // find 'camera_jpeg_count_str'
    k = find_str_ref(fw, "9999");
    if (k >= 0)
    {
        if (isLDR(fw,k-1) && isBL(fw,k+1))
        {
            uint32_t v1 = LDR2val(fw,k-1);
            bprintf("DEF(%-40s,0x%08x) // Found @0x%08x\n","jpeg_count_str",v1,idx2adr(fw,k-1));
            bprintf("//char *camera_jpeg_count_str()    { return (char*)0x%08x; }             // Found @0x%08x\n",v1,idx2adr(fw,k-1));
        }
    }

//    // find 'hook_raw_size'
//    k = find_str_ref(fw, "CRAW BUFF SIZE  %p");
//    if (k >= 0)
//    {
//        if (isLDR(fw,k-1))
//        {
//            uint32_t craw_bufsize = LDR2val(fw,k-1);
//            bprintf("//long hook_raw_size()             { return 0x%08x; }                    // Found @0x%08x\n",craw_bufsize,idx2adr(fw,k-1));
//        }
//    }

    // Find value for 'get_flash_params_count'
    search_saved_sig(fw, "GetParameterData", match_get_flash_params_count, 0, 0, 30);
}

//------------------------------------------------------------------------------------------------------------

void print_stubs_min(firmware *fw, const char *name, uint32_t fadr, uint32_t atadr)
{
    osig *o = find_sig(fw->sv->stubs_min,name);
    if (o)
    {
        bprintf("//DEF(%-40s,0x%08x) // Found @0x%08x",name,fadr,atadr);
        if (fadr != o->val)
        {
            bprintf(", ** != ** stubs_min = 0x%08x (%s)",o->val,o->sval);
        }
        else
        {
            bprintf(",          stubs_min = 0x%08x (%s)",o->val,o->sval);
        }
    }
    else
    {
        bprintf("DEF(%-40s,0x%08x) // Found @0x%08x",name,fadr,atadr);
    }
    bprintf("\n");
}

int match_levent_table(firmware *fw, int k, uint32_t v1, uint32_t v2)
{
    if ((fw->buf[k] > fw->base) && (fw->buf[k+1] == 0x00000800) && (fw->buf[k+2] == 0x00000002))
    {
        print_stubs_min(fw,"levent_table",idx2adr(fw,k),idx2adr(fw,k));
#ifdef PRINT_LEVENT_TABLE
        uint32_t levent_tbl = idx2adr(fw,k);
        void write_levent_table_dump(firmware*, uint32_t);
        write_levent_table_dump(fw, levent_tbl);
#endif
    }
    return 0;
}

int match_movie_status(firmware *fw, int k, uint32_t v1, uint32_t v2)
{
    if (isLDR_PC(fw, k) &&                              // LDR R0, =base
        ((fw->buf[k+1] & 0xFE0F0000) == 0xE20F0000) &&  // ADR R1, =sub
        isSTR(fw, k+2) &&                               // STR R1, [R0,N]
        (fw->buf[k+3] == 0xE3A01003) &&                 // MOV R1, 3
        isSTR(fw, k+4) &&                               // STR R1, [R0,ofst]
        (LDR2val(fw,k) < fw->base))
    {
        uint32_t base = LDR2val(fw,k);
        uint32_t ofst = fw->buf[k+4] & 0x00000FFF;
        print_stubs_min(fw,"movie_status",base+ofst,idx2adr(fw,k));
        return 1;
    }
    else
    if (isLDR_PC(fw, k) &&                              // LDR R1, =sub
        isLDR_PC(fw, k+1) &&                            // LDR R0, =base
        isSTR(fw, k+2) &&                               // STR R1, [R0,N]
        (fw->buf[k+3] == 0xE3A01003) &&                 // MOV R1, 3
        isSTR(fw, k+4) &&                               // STR R1, [R0,ofst]
        (LDR2val(fw,k+1) < fw->base))
    {
        uint32_t base = LDR2val(fw,k+1);
        uint32_t ofst = fw->buf[k+4] & 0x00000FFF;
        print_stubs_min(fw,"movie_status",base+ofst,idx2adr(fw,k));
        return 1;
    }
    else
    if (isLDR_PC(fw, k) &&                                      // LDR Rx, =base
        isLDR(fw, k+1) && (fwRd(fw,k) == fwRn(fw,k+1)) &&       // LDR R0, [Rx, ...]
        isCMP(fw, k+2) && (fwRd(fw,k+2) == fwRd(fw,k+1)) &&     // CMP R0, #...
        (fwval(fw,k+3) == 0x03A00005) &&
        isSTR_cond(fw, k+4) && (fwRn(fw,k+4) == fwRd(fw,k)) &&  // STRxx R0, [Rx,ofst]
        (LDR2val(fw,k) < fw->base))
    {
        uint32_t base = LDR2val(fw,k);
        uint32_t ofst = fwOp2(fw,k+4);
        print_stubs_min(fw,"movie_status",base+ofst,idx2adr(fw,k));
        return 1;
    }
    return 0;
}

int match_full_screen_refresh(firmware *fw, int k, uint32_t v1, uint32_t v2)
{
    if (((fw->buf[k] & 0xFF1FF000) == 0xE51F0000) &&    // LDR R0, =base
        (fw->buf[k+1] == 0xE5D01000) &&                 // LDRB R1, [R0]
        (fw->buf[k+2] == 0xE3811002) &&                 // ORR R1, R1, #2
        (fw->buf[k+3] == 0xE5C01000) &&                 // STRB R1, [R0]
        isBX_LR(fw,k+4))                                // BX LR
    {
        uint32_t base = LDR2val(fw,k);
        print_stubs_min(fw,"full_screen_refresh",base,idx2adr(fw,k));
    }
    return 0;
}

int match_canon_shoot_menu_active(firmware *fw, int k, uint32_t v1, uint32_t v2)
{
    if (((fw->buf[k]   & 0xFF1FF000) == 0xE51F1000) &&  // LDR R1, =base
        ((fw->buf[k+1] & 0xFFFFF000) == 0xE5D10000) &&  // LDRB R0, [R1, #n]
        (fw->buf[k+2] == 0xE2800001) &&                 // ADD R0, R0, #1
        ((fw->buf[k+3] & 0xFFFFF000) == 0xE5C10000) &&  // STRB R0, [R1, #n]
        (isB(fw,k+4)))                                  // B
    {
        uint32_t base = LDR2val(fw,k);
        uint32_t ofst = fw->buf[k+1] & 0x00000FFF;
        print_stubs_min(fw,"canon_shoot_menu_active",base+ofst,idx2adr(fw,k));
    }
    else
    if (((fw->buf[k]   & 0xFF1FF000) == 0xE51F0000) &&  // LDR R0, =base
        ((fw->buf[k+1] & 0xFFFFF000) == 0xE5D01000) &&  // LDRB R1, [R0, #n]
        (fw->buf[k+2] == 0xE2811001) &&                 // ADD R1, R1, #1
        ((fw->buf[k+3] & 0xFFFFF000) == 0xE5C01000) &&  // STRB R1, [R0, #n]
        (isB(fw,k+4)))                                  // B
    {
        uint32_t base = LDR2val(fw,k);
        uint32_t ofst = fw->buf[k+1] & 0x00000FFF;
        print_stubs_min(fw,"canon_shoot_menu_active",base+ofst,idx2adr(fw,k));
    }
    return 0;
}

int match_playrec_mode(firmware *fw, int k, uint32_t v1, uint32_t v2)
{
    if (((fw->buf[k]    & 0xFF1FF000) == 0xE51F1000) && // LDR R1, =base
        ((fw->buf[k+1]  & 0xFFFFF000) == 0xE5810000) && // STR R0, [R1, #n]
        ((fw->buf[k+3]  & 0xFF1FF000) == 0xE51F0000) && // LDR R0, =base
        ((fw->buf[k+4]  & 0xFFFFF000) == 0xE5900000) && // LDR R0, [R0, #n]
        ((fw->buf[k+6]  & 0xFF1FF000) == 0xE51F1000) && // LDR R1, =base
        ((fw->buf[k+9]  & 0xFF1FF000) == 0xE51F0000) && // LDR R0, =base
        ((fw->buf[k+12] & 0xFF1FF000) == 0xE51F1000) && // LDR R1, =base
        ((fw->buf[k+15] & 0xFF1FF000) == 0xE51F0000) && // LDR R0, =base
        ((fw->buf[k+18] & 0xFF1FF000) == 0xE51F1000) && // LDR R1, =base
        (LDR2val(fw,k) == LDR2val(fw,k+3)) &&
        (LDR2val(fw,k) == LDR2val(fw,k+6)) &&
        (LDR2val(fw,k) == LDR2val(fw,k+9)) &&
        (LDR2val(fw,k) == LDR2val(fw,k+12)) &&
        (LDR2val(fw,k) == LDR2val(fw,k+15)) &&
        (LDR2val(fw,k) == LDR2val(fw,k+18)))
    {
        uint32_t base = LDR2val(fw,k);
        uint32_t ofst = fw->buf[k+1] & 0x00000FFF;
        print_stubs_min(fw,"playrec_mode",base+ofst,idx2adr(fw,k));
    }
    return 0;
}

int match_some_flag_for_af_scan(firmware *fw, int k, uint32_t v1, uint32_t v2)
{
    if (isB(fw,k)   &&  // B loc
        isB(fw,k+1) &&  // B loc
        isB(fw,k+2) &&  // B loc
        isB(fw,k+3) &&  // B loc
        isB(fw,k+4) &&  // B loc
        isB(fw,k+5) &&  // B loc
        isB(fw,k+6) &&  // B loc
        isB(fw,k+7) &&  // B loc
        (followBranch(fw,idx2adr(fw,k),1) != followBranch(fw,idx2adr(fw,k+1),1)) &&
        (followBranch(fw,idx2adr(fw,k),1) == followBranch(fw,idx2adr(fw,k+2),1)) &&
        (followBranch(fw,idx2adr(fw,k),1) == followBranch(fw,idx2adr(fw,k+3),1)) &&
        (followBranch(fw,idx2adr(fw,k),1) != followBranch(fw,idx2adr(fw,k+4),1)) &&
        (followBranch(fw,idx2adr(fw,k),1) != followBranch(fw,idx2adr(fw,k+5),1)) &&
        (followBranch(fw,idx2adr(fw,k),1) == followBranch(fw,idx2adr(fw,k+6),1)) &&
        (followBranch(fw,idx2adr(fw,k),1) == followBranch(fw,idx2adr(fw,k+7),1)) &&
        (isLDR_PC(fw,adr2idx(fw,followBranch(fw,idx2adr(fw,k),1)))))    // LDR R0, =base
    {
        uint32_t base = LDR2val(fw,adr2idx(fw,followBranch(fw,idx2adr(fw,k),1)));
        if (base < fw->base)
            print_stubs_min(fw,"some_flag_for_af_scan",base,followBranch(fw,idx2adr(fw,k),1));
    }
    return 0;
}

int match_palette_data(firmware *fw, int k, uint32_t v1, uint32_t v2)
{
    if ((fw->buf[k] == 0) && (fw->buf[k+1] == 0x00FF0000) &&
        (fw->buf[k+577] == 1) && (fw->buf[k+578] == 0x00FF0000) &&
        (fw->buf[k+1154] == 2) && (fw->buf[k+1155] == 0x00FF0000))
    {
        return idx2adr(fw,k);
    }
    else if ((fw->buf[k] == 0) && (fw->buf[k+1] == 0x00FF0000) &&
             (fw->buf[k+513] == 1) && (fw->buf[k+514] == 0x00FF0000) &&
             (fw->buf[k+1026] == 2) && (fw->buf[k+1027] == 0x00FF0000))
    {
        return idx2adr(fw,k);
    }
    return 0;
}

int match_palette_buffer_offset(firmware *fw, int k)
{
    int idx2 = idxFollowBranch(fw, k, 0x01000001);
    if (isLDR(fw, idx2+2) && isBL(fw, idx2+3))
    {
        uint32_t palette_size = LDR2val(fw,idx2+2);
        if (palette_size >= 0x400)
        {
            bprintf("// Offset from start of palette_buffer to color data = %d (Found @0x%08x)\n",palette_size-0x400,idx2adr(fw,idx2+2));
            return 1;
        }
    }
    else if (isADR(fw, idx2+2) && isBL(fw, idx2+3))
    {
        uint32_t palette_size = ALUop2(fw,idx2+2);
        if (palette_size >= 0x400)
        {
            bprintf("// Offset from start of palette_buffer to color data = %d (Found @0x%08x)\n",palette_size-0x400,idx2adr(fw,idx2+2));
            return 1;
        }
    }
    return 0;
}

int match_palette_data3(firmware *fw, int k, uint32_t palette_data, uint32_t v2)
{
    if (isLDR_PC(fw, k) && (LDR2val(fw,k) == palette_data) && isLDR_PC(fw,k-1) && isLDR_PC(fw,k-6) && isLDR(fw,k-5))
    {
        int palette_control = LDR2val(fw,k-6);
        int ptr_offset = fwOp2(fw,k-5);
        uint32_t fadr = find_inst_rev(fw, isSTMFD_LR, k-7, 30);
        if (fadr > 0)
        {
            int k1 = search_fw(fw, find_B, fadr, 0, 1);
            if ((k1 > 0) && isLDR_PC(fw,k1-2) && isLDR(fw,k1-1) && (LDR2val(fw,k1-2) == palette_control))
            {
                int active_offset = fwOp2(fw,k1-1);
                print_stubs_min(fw,"active_palette_buffer",palette_control+active_offset,idx2adr(fw,k1-1));
                print_stubs_min(fw,"palette_buffer_ptr",palette_control+ptr_offset,idx2adr(fw,k-5));
                if (isBL(fw,k+8))
                {
                    fadr = followBranch(fw, idx2adr(fw,k+8), 0x01000001);
                    int idx = adr2idx(fw, fadr);
                    if (isLDR(fw, idx+2) && isBL(fw, idx+3))
                    {
                        uint32_t palette_size = LDR2val(fw,idx+2);
                        if (palette_size >= 0x400)
                        {
                            bprintf("// Offset from start of palette_buffer to color data = %d (Found @0x%08x)\n",palette_size-0x400,idx2adr(fw,idx+2));
                        }
                    }
                }
                return 1;
            }
        }
    }
    return 0;
}

int match_palette_data2(firmware *fw, int k, uint32_t v1, uint32_t v2)
{
    if (isLDR(fw,k) && (LDR2val(fw,k) == v1))
    {
        int k1;
        for (k1=k+1; k1<k+20; k1++)
        {
            if (isBL(fw,k1) && isLDMFD(fw,k1+2))
            {
                uint32_t fadr = followBranch(fw, idx2adr(fw,k1), 0x01000001);
                int idx = adr2idx(fw, fadr);
                int k2;
                for (k2=idx; k2<idx+40; k2++)
                {
                    if (isSTR(fw,k2) && isLDMFD(fw,k2+1))
                    {
                        int ptr_offset = fwval(fw,k2) & 0xFFF;
                        print_stubs_min(fw,"palette_buffer_ptr",v2+ptr_offset,idx2adr(fw,k2));
                        int found = 0;
                        for (k=idx; (k<idx+16) && !found; k++)
                        {
                            if (isBL(fw,k) && isCMP(fw,k+1) && isADR_PC(fw,k+2))
                            {
                                fadr = ADR2adr(fw,k+2);
                                idx = adr2idx(fw, fadr);
                                int k3;
                                for (k3=idx; (k3<idx+16) && !found; k3++)
                                {
                                    if (isBL(fw,k3))
                                    {
                                        if (match_palette_buffer_offset(fw,k3))
                                            return 1;
                                    }
                                }
                            }
                        }
                        return 1;
                    }
                }
            }
            else if (isLDR_cond(fw,k1) && isLDMFD(fw,k1+2) && isBL(fw,k1-2))
            {
                int ptr_offset = fwval(fw,k1) & 0xFFF;
                print_stubs_min(fw,"palette_buffer_ptr",v2+ptr_offset,idx2adr(fw,k1));
                match_palette_buffer_offset(fw, k1-2);
                return 1;
            }
        }
    }
    return 0;
}

int match_SavePaletteData(firmware *fw, int idx, int palette_data)
{
    if (isBL(fw,idx+13))
    {
        uint32_t fadr = followBranch(fw, idx2adr(fw,idx+13), 0x01000001);
        idx = adr2idx(fw, fadr);
        if (isLDR(fw,idx) && isLDR(fw,idx+1) && isB(fw,idx+2))
        {
            uint32_t palette_control = LDR2val(fw,idx);
            print_stubs_min(fw,"palette_control",palette_control,idx2adr(fw,idx));
            int active_offset = fwval(fw,idx+1) & 0xFFF;
            print_stubs_min(fw,"active_palette_buffer",palette_control+active_offset,idx2adr(fw,idx+1));
            fadr = followBranch(fw,idx2adr(fw,idx+2),1);
            idx = adr2idx(fw, fadr);
            if (isLDR(fw,idx+17) && isLDR(fw,idx+18) && isLDR(fw,idx+12) && (LDR2val(fw,idx+12) == palette_control))
            {
                if (isLDR(fw,idx+13))
                {
                    int ptr_offset = fwval(fw,idx+13) & 0xFFF;
                    print_stubs_min(fw,"palette_buffer_ptr",palette_control+ptr_offset,idx2adr(fw,idx+13));
                }
                int palette_buffer;
                if ((fwval(fw,idx+18) & 0x0000F000) == 0)
                {
                    palette_buffer = LDR2val(fw,idx+17);
                    print_stubs_min(fw,"palette_buffer",palette_buffer,idx2adr(fw,idx+17));
                }
                else
                {
                    palette_buffer = LDR2val(fw,idx+18);
                    print_stubs_min(fw,"palette_buffer",palette_buffer,idx2adr(fw,idx+18));
                }
                if (isBL(fw,idx+26))
                {
                    fadr = followBranch(fw, idx2adr(fw,idx+26), 0x01000001);
                    idx = adr2idx(fw, fadr);
                    if (isLDR(fw, idx+2) && isBL(fw, idx+3))
                    {
                        uint32_t palette_size = LDR2val(fw,idx+2);
                        if (palette_size >= 0x400)
                        {
                            bprintf("// Offset from start of palette_buffer to color data = %d (Found @0x%08x)\n",palette_size-0x400,idx2adr(fw,idx+2));
                        }
                    }
                }
            }
        }
        else if (isLDR(fw,idx) && isLDR(fw,idx+6) && isLDR(fw,idx+7) && isBX(fw,idx+8))
        {
            int active_offset = -1;
            if ((fwval(fw,idx+6) & 0x0000F000) == 0)
                active_offset = fwval(fw,idx+6) & 0xFFF;
            else if ((fwval(fw,idx+7) & 0x0000F000) == 0)
                active_offset = fwval(fw,idx+7) & 0xFFF;
            if (active_offset >= 0)
            {
                uint32_t palette_control = LDR2val(fw,idx);
                //print_stubs_min(fw,"palette_control",palette_control,idx2adr(fw,idx));
                print_stubs_min(fw,"active_palette_buffer",palette_control+active_offset,idx2adr(fw,idx+1));
                search_fw(fw, match_palette_data2, palette_data, palette_control, 1);
            }
        }
        return 1;
    }

    return 0;
}

int match_viewport_address3(firmware *fw, int k, uint32_t v1, uint32_t v2)
{
    if (isLDR_PC(fw,k) && (LDR2val(fw,k) == v1))
    {
        // Scan back to start of function
        int k1 = find_inst_rev(fw, isSTMFD_LR, k-1, 1000);
        if (k1 > 0)
        {
            // Check if function sig matches what we want
            if (isLDR_PC(fw,k1+1) && isLDR_PC(fw,k1+2) && isLDR(fw,k1+3) &&
                (fwRd(fw,k1+1) == fwRn(fw,k1+3)))
            {
                uint32_t a = LDR2val(fw,k1+1);
                print_stubs_min(fw,"viewport_buffers",v1,idx2adr(fw,k));
                print_stubs_min(fw,"active_viewport_buffer",a,idx2adr(fw,k1+1));
                return 1;
            }
            else
            if (isLDR_PC(fw,k1+1) && isLDR_PC(fw,k1+3) && isLDR(fw,k1+4) &&
                (fwRd(fw,k1+1) == fwRn(fw,k1+4)))
            {
                uint32_t a = LDR2val(fw,k1+1);
                print_stubs_min(fw,"viewport_buffers",v1,idx2adr(fw,k));
                print_stubs_min(fw,"active_viewport_buffer",a,idx2adr(fw,k1+1));
                return 1;
            }
            else
            if (isLDR_PC(fw,k1+1) && isLDR_PC(fw,k1+4) && isLDR(fw,k1+5) &&
                (fwRd(fw,k1+1) == fwRn(fw,k1+5)))
            {
                uint32_t a = LDR2val(fw,k1+1);
                print_stubs_min(fw,"viewport_buffers",v1,idx2adr(fw,k));
                print_stubs_min(fw,"active_viewport_buffer",a,idx2adr(fw,k1+1));
                return 1;
            }
        }
    }
    return 0;
}

int match_viewport_address2(firmware *fw, int k, uint32_t v1, uint32_t v2)
{
    if (fw->buf[k] == v1)
    {
        if (search_fw(fw, match_viewport_address3, v1, 0, 1))
            return 1;
    }
    return 0;
}

int match_viewport_address(firmware *fw, int k, uint32_t v1, uint32_t v2)
{
    if (fw->buf[k] == v1)
    {
        // Find location in firmware that points to viewport address
        if (search_fw(fw, match_viewport_address2, idx2adr(fw,k), 0, 1))
            return 1;
    }
    return 0;
}

int match_physw_status(firmware *fw, int k, int v)
{
    if (isLDR_PC(fw,k))
    {
        print_stubs_min(fw,"physw_status",LDR2val(fw,k),idx2adr(fw,k));
    }

    return 0;
}

int match_physw_run(firmware *fw, int k, int v)
{
    if (isLDR_PC(fw,k))
    {
        uint32_t base = LDR2val(fw,k);
        uint32_t fadr = followBranch(fw, idx2adr(fw,k+1), 1);
        uint32_t ofst = fw->buf[adr2idx(fw,fadr)] & 0x00000FFF;
        print_stubs_min(fw,"physw_run",base+ofst,idx2adr(fw,k));

        ofst = fw->buf[k+2] & 0x00000FFF;
        print_stubs_min(fw,"physw_sleep_delay",base+ofst,idx2adr(fw,k));
    }

    return 0;
}

int match_canon_menu_active(firmware *fw, int k, int v)
{
    if (isLDR_PC(fw,k))
    {
        uint32_t base = LDR2val(fw,k);
        int k1;
        for (k1=k+1; k1<k+5; k1++)
        {
            if (isLDR(fw,k1))
            {
                uint32_t ofst = fw->buf[k1] & 0x00000FFF;
                print_stubs_min(fw,"canon_menu_active",base+ofst,idx2adr(fw,k));
            }
        }
    }

    return 0;
}

int match_zoom_busy(firmware *fw, int k, int v)
{
    if (isBL(fw,k))
    {
        int idx1 = idxFollowBranch(fw,k,0x01000001);
        int k1;
        for (k1=idx1; k1<idx1+50; k1++)
        {
            if ((fw->buf[k1] & 0xFFFF0000) == 0xE8BD0000)   // LDMFD
            {
                uint32_t fadr = 0;
                if (isADR_PC(fw,k1+1))
                {
                    fadr = ADR2adr(fw,k1+1);
                }
                else if (isADR_PC(fw,k1+2))
                {
                    fadr = ADR2adr(fw,k1+2);
                }
                else if (isADR_PC(fw,k1-3))
                {
                    fadr = ADR2adr(fw,k1-3);
                }
                else if (isLDR_PC(fw,k1+1))
                {
                    fadr = LDR2val(fw,k1+1);
                }
                else if (isLDR_PC(fw,k1+2))
                {
                    fadr = LDR2val(fw,k1+2);
                }
                if (fadr != 0)
                {
                    int idx2 = adr2idx(fw,fadr);
                    if (isLDR_PC(fw,idx2+1) && isLDR(fw,idx2+2))
                    {
                        int base = LDR2val(fw,idx2+1);
                        int ofst = fw->buf[idx2+2] & 0xFFF;
                           print_stubs_min(fw,"zoom_busy",base+ofst-4,fadr);
                        break;
                    }
                }
            }
        }
    }

    return 0;
}

int match_focus_busy(firmware *fw, int k, int v)
{
    if ((fw->buf[k] & 0xFFFF0000) == 0xE8BD0000)   // LDMFD
    {
        int k1 = 0;
        if (isBL(fw,k-2))
        {
            k1 = idxFollowBranch(fw,k-2,0x01000001);
        }
        if (isBL(fw,k-1))
        {
            k1 = idxFollowBranch(fw,k-1,0x01000001);
        }
        if (k1 != 0)
        {
            if (isLDR_PC(fw,k1+1) && isLDR(fw,k1+3))
            {
                int base = LDR2val(fw,k1+1);
                int ofst = fw->buf[k1+3] & 0xFFF;
                   print_stubs_min(fw,"focus_busy",base+ofst-4,idx2adr(fw,k1));
                return 1;
            }
        }
    }

    return 0;
}

int match_bitmap_buffer2(firmware *fw, int k, int v)
{
    uint32_t screen_lock = idx2adr(fw,k);
    if (isBL(fw,v) && (followBranch(fw,idx2adr(fw,v),0x01000001) == screen_lock) && isBL(fw,v+2) && isBL(fw,v+3))
    {
        uint32_t fadr = followBranch2(fw,idx2adr(fw,v+3),0x01000001);
        int k1 = adr2idx(fw,fadr);
        if (isLDR_PC(fw,k1+1))
        {
            int reg = (fwval(fw,k1+1) & 0x0000F000) >> 12;
            uint32_t adr = LDR2val(fw,k1+1);
            int k2;
            for (k2=k1; k2<k1+32; k2++)
            {
                if (isLDR_PC(fw,k2) && isLDR(fw,k2+1) && (((fwval(fw,k2+1) & 0x000F0000) >> 16) == reg))
                {
                    uint32_t bitmap_buffer = LDR2val(fw,k2);
                    if (bitmap_buffer == (adr + 0x1C))
                    {
                        uint32_t active_bitmap_buffer = adr + (fwval(fw,k2+1) & 0xFFF);
                        print_stubs_min(fw,"bitmap_buffer",bitmap_buffer,idx2adr(fw,k2));
                        print_stubs_min(fw,"active_bitmap_buffer",active_bitmap_buffer,idx2adr(fw,k2+1));
                    }
                }
            }
        }
    }

    return 0;
}

int match_bitmap_buffer(firmware *fw, int k, int v)
{
    search_saved_sig(fw, "ScreenLock", match_bitmap_buffer2, k, 0, 1);
    return 0;
}

int match_raw_buffer(firmware *fw, int k, uint32_t rb1, uint32_t v2)
{
    if (((fwval(fw,k) == rb1) && (fwval(fw,k+4) == rb1) && (fwval(fw,k-2) != 1) && (fwval(fw,k+2) >= fw->uncached_adr)) ||
        ((fwval(fw,k) == rb1) && (fwval(fw,k+4) == rb1) && (fwval(fw,k+20) == rb1)))
    {
        uint32_t rb2 = fwval(fw,k+1);
        if ((rb1 != rb2) && (rb2 > 0))
        {
            bprintf("// Camera has 2 RAW buffers @ 0x%08x & 0x%08x\n", rb1, rb2, idx2adr(fw,k));
            bprintf("//  Note: active buffer --> raw_buffers[active_raw_buffer]\n");
            bprintf("//        other buffer  --> raw_buffers[active_raw_buffer^1]\n");
            print_stubs_min(fw,"raw_buffers",idx2adr(fw,k),idx2adr(fw,k));
        }
        return rb2;
    }
    else if ((fwval(fw,k) == rb1) && (fwval(fw,k-2) == 2) && (fwval(fw,k-7) == rb1))
    {
        uint32_t rb2 = fwval(fw,k+3);
        if ((rb1 != rb2) && (rb2 > 0))
        {
            bprintf("// Camera has 2 RAW buffers @ 0x%08x & 0x%08x\n", rb1, rb2, idx2adr(fw,k));
            bprintf("//  Note: active buffer --> raw_buffers[ active_raw_buffer   *3]\n");
            bprintf("//        other buffer  --> raw_buffers[(active_raw_buffer^1)*3]\n");
            print_stubs_min(fw,"raw_buffers",idx2adr(fw,k),idx2adr(fw,k));
        }
        return rb2;
    }
    return 0;
}

int match_fileiosem(firmware *fw, int k, uint32_t fadr, uint32_t nadr)
{
    if ((k > 5) && isADR_PC(fw, k) && isBL(fw,k+1) && (ADR2adr(fw,k) == nadr) && (followBranch2(fw, idx2adr(fw,k+1), 0x01000001) == fadr))
    {
        int j, rn = -1;
        for (j = k-1; j > k-5; j++)
        {
            if (isLDR(fw,j))
            {
                if (fwRd(fw,j) == 0)
                    rn = fwRn(fw, j);
                else if (fwRd(fw,j) == rn)
                {
                    int v = LDR2val(fw,j);
                    print_stubs_min(fw,"fileio_semaphore",v,idx2adr(fw,j));
                    return 1;
                }
            }
        }
    }
    return 0;
}

// Search for things that go in 'stubs_min.S'
void find_stubs_min(firmware *fw)
{
    int k,k1;

    out_hdr = 1;
    add_blankline();

    bprintf("// Values below can be overridden in 'stubs_min.S':\n");

    // Find 'physw_status'
    search_saved_sig(fw, "kbd_read_keys", match_physw_status, 0, 0, 5);

    // Find 'physw_run' & 'physw_sleep_delay'
    search_saved_sig(fw, "task_PhySw", match_physw_run, 0, 0, 5);

    // Find 'levent_table'
    search_fw(fw, match_levent_table, 0, 0, 1);

    // Find 'FlashParamsTable'
    if (FlashParamsTable_address != 0)
        print_stubs_min(fw,"FlashParamsTable",FlashParamsTable_address,FlashParamsTable_address);

    // Find 'movie_status'
    search_fw(fw, match_movie_status, 0, 0, 1);

    // Find 'video_compression_rate'
    uint32_t sadr = find_str(fw, "CompressionRateAdjuster.c");
    k = find_nxt_str_ref(fw, sadr, -1);
    int found = 0;
    while ((k >= 0) && !found)
    {
        int f = find_inst_rev(fw, isSTMFD_LR, k-1, 100);
        if (f != -1)
        {
            f = search_fw(fw, find_BL, f, 0, 1);
            if (f > 0)
            {
                f--;
                if ((fwval(fw,f) & 0xFFF00000) == 0xE2400000)     // SUB
                {
                    int src = fwRn(fw,f);
                    for (k1 = f-1; (k1 > f-10) && !found; k1--)
                    {
                        if (isLDR_PC(fw,k1) && (fwRd(fw,k1) == src))
                        {
                            uint32_t v = LDR2val(fw,k1) - ALUop2(fw,f);
                            print_stubs_min(fw,"video_compression_rate",v,idx2adr(fw,k1));
                            found = 1;
                        }
                    }
                }
            }
        }
        k = find_nxt_str_ref(fw, sadr, k);
    }

    // Find 'full_screen_refresh'
    search_fw(fw, match_full_screen_refresh, 0, 0, 1);

    // Find 'canon_menu_active'
    search_saved_sig(fw, "StartRecModeMenu", match_canon_menu_active, 0, 0, 5);

    // Find 'canon_shoot_menu_active'
    search_fw(fw, match_canon_shoot_menu_active, 0, 0, 1);

    // Find 'playrec_mode'
    int found_playrec_mode = 0;
    k = find_str_ref(fw, "AFFChg");
    if ((k >= 0) && isBL(fw,k+6))
    {
        k = idxFollowBranch(fw, k+6, 0x01000001);
        if (isLDR_PC(fw,k) && isLDR(fw,k+1))
        {
            uint32_t base = LDR2val(fw,k);
            uint32_t ofst = fw->buf[k+1] & 0x00000FFF;
            print_stubs_min(fw,"playrec_mode",base+ofst,idx2adr(fw,k));
            found_playrec_mode = 1;
        }
    }
    if (!found_playrec_mode)
    {
        search_fw(fw, match_playrec_mode, 0, 0, 1);
    }

    // Find 'zoom_status'
    int found_zoom_status = 0;

    k = find_str_ref(fw, "m_ZoomState            :%d\n");
    if (k >= 0)
    {
        if (isLDR(fw,k-1))
        {
            uint32_t ofst = fw->buf[k-1] & 0x00000FFF;
            uint32_t reg = (fw->buf[k-1] & 0x000F0000) >> 16;
            uint32_t ldr_inst = 0xE51F0000 | (reg << 12);
            for (k1=k-2; k1>k-20; k1--)
            {
                if ((fw->buf[k1] & 0xFF1FF000) == ldr_inst)
                {
                    uint32_t base = LDR2val(fw,k1);
                    print_stubs_min(fw,"zoom_status",base+ofst,idx2adr(fw,k));
                    found_zoom_status = 1;
                    break;
                }
            }
        }
    }

    if (!found_zoom_status)
    {
        for (k=0; k<fw->size; k++)
        {
            if (((fw->buf[k] & 0xFF1FF000) == 0xE51F0000) &&    // LDR R0, =base
                (fw->buf[k+1] == 0xE5D00000) &&                 // LDRB R0, [R0]
                (fw->buf[k+2] == 0xE1B00000) &&                 // MOVS R0, R0
                (fw->buf[k+3] == 0x13A00001) &&                 // MOVNE R0, #1
                isBX_LR(fw,k+4))                                // BX LR
            {
                uint32_t base = LDR2val(fw,k);
                print_stubs_min(fw,"zoom_status",base,idx2adr(fw,k));
                found_zoom_status = 1;
                //break;
            }
        }
    }

    if (!found_zoom_status)
    {
        k = find_str_ref(fw, "TerminateDeliverToZoomController");
        if (k >= 0)
        {
            for (k1=0; k1<5; k1++)
            {
                if (isLDR_PC(fw,k+k1))
                {
                    uint32_t base = LDR2val(fw,k+k1);
                    print_stubs_min(fw,"zoom_status",base+0x20,idx2adr(fw,k+k1));
                    found_zoom_status = 1;
                    break;
                }
            }
        }
    }

    // Find 'some_flag_for_af_scan'
    search_fw(fw, match_some_flag_for_af_scan, 0, 0, 1);

    // focus_len_table
    if (fw->sv->min_focus_len != 0)
    {
        int found = 0, pos = 0, len = 0, size = 0;
        for (k=0; k<fw->size; k++)
        {
            if (fw->buf[k] == fw->sv->min_focus_len)
            {
                int mul = 1;
                if ((fw->buf[k+1] == 100) && (fw->buf[k+2] == 0)) mul = 3;
                if ((fw->buf[k+1] == 100) && (fw->buf[k+2] != 0)) mul = 2;
                if ((fw->buf[k+1] ==   0) && (fw->buf[k+2] != 0)) mul = 2;
                for (k1 = k + mul; (k1 < fw->size) && (fw->buf[k1] > fw->buf[k1-mul]) && (fw->buf[k1] > fw->sv->min_focus_len) && (fw->buf[k1] < fw->sv->max_focus_len); k1 += mul) ;
                if (fw->buf[k1] == fw->sv->max_focus_len)
                {
                    if ((found == 0) || ((size < mul) && (len < ((k1 - k) / mul) + 1)))
                    {
                        found = 1;
                        pos = k;
                        len = ((k1 - k) / mul) + 1;
                        size = mul;
                    }
                }
            }
        }
        if (found == 1)
        {
            bprintf("// focus_len_table contains zoom focus lengths for use in 'get_focal_length' (main.c).\n");
            if (size == 1)
                bprintf("// each entry contains 1 int value, which is the the zoom focus length.\n",size);
            else
                bprintf("// each entry contains %d int value(s), the first is the zoom focus length.\n",size);
            bprintf("// there are %d entries in the table - set NUM_FL to %d\n",len,len);
            print_stubs_min(fw,"focus_len_table",idx2adr(fw,pos),idx2adr(fw,pos));
        }
    }

    // Find 'zoom_busy'
    search_saved_sig(fw, "ResetZoomLens", match_zoom_busy, 0, 0, 5);

    // Find 'focus_busy'
    search_saved_sig(fw, "ResetFocusLens", match_focus_busy, 0, 0, 25);

    // Find 'recreview_hold'
    k = find_str_ref(fw, "ShootCon_NotifyStartReviewHold");
    if (k >= 0)
    {
        for (k1=k; k1<k+20; k1++)
        {
            if (isLDR_PC(fw,k1) && ((fw->buf[k1+1] & 0xFFFF0FFF) == 0xE3A00001) && isSTR(fw,k1+2) &&
                ((fw->buf[k1+1] & 0x0000F000) == (fw->buf[k1+2] & 0x0000F000)) &&
                ((fw->buf[k1] & 0x0000F000) == ((fw->buf[k1+2] & 0x000F0000) >> 4)))
            {
                uint32_t base = LDR2val(fw,k1);
                int ofst = fw->buf[k1+2] & 0x00000FFF;
                print_stubs_min(fw,"recreview_hold",base+ofst,idx2adr(fw,k1));
                break;
            }
        }
    }

    // Find palette colour data
    uint32_t palette_data = search_fw(fw, match_palette_data, 0, 0, 1);

    // Find 'palette buffer' info
    if (palette_data)
    {
        bprintf("// Palette colour tables  found @ 0x%08x\n", palette_data);
        if (search_saved_sig(fw, "SavePaletteData", match_SavePaletteData, palette_data, 0, 1) == 0)
        {
            search_fw(fw, match_palette_data3, palette_data, 0, 1);
        }
    }

    // Find 'bitmap buffer' info
    search_saved_sig(fw, "GUISrv_StartGUISystem", match_bitmap_buffer, 0, 0, 50);

    // Get viewport address
    uint32_t v = find_viewport_address(fw,&k);
    if (k >= 0)
    {
        search_fw(fw, match_viewport_address, v, 0, 1);
    }

    // find 1st RAW buffer address
    k = find_str_ref(fw, "CRAW BUFF       %p");
    if (k >= 0)
    {
        int rb1_idx;
        uint32_t rb1 =0, rb2 = 0;
        if (isLDR(fw,k-1))
        {
            rb1 = LDR2val(fw,k-1);
            rb1_idx = k - 1;
        }
        else if (isMOV_immed(fw,k-1))
        {
            rb1 = ALUop2(fw,k-1);
            rb1_idx = k - 1;
        }
        else if (isMOV(fw,k-1) && (fwRd(fw,k-1) == 1))
        {
            int reg = fwval(fw,k-1) & 0xF;
            for (k1=k-2; k1>k-50; k1--)
            {
                if (isLDR(fw,k1) && (fwRd(fw,k1) == reg))
                {
                    rb1 = LDR2val(fw,k1);
                    rb1_idx = k1;
                    break;
                }
            }
        }
        if (rb1 > 0)
        {
            found = 0;
            rb2 = search_fw(fw, match_raw_buffer, rb1, 0, 5);
            if ((rb2 > 0) && (rb1 != rb2))
            {
                // Find 'active_raw_buffer'
                sadr = find_str(fw, "SsImgProcBuf.c");
                k = find_nxt_str_ref(fw, sadr, -1);
                while ((k >= 0) && !found)
                {
                    int f = find_inst_rev(fw, isSTMFD_LR, k-1, 100);
                    if (f != -1)
                    {
                        int e = find_inst(fw, isLDMFD_PC, f+1, 200);
                        for (k1 = f+1; k1 < e; k1++)
                        {
                            if (
                                (
                                    ((fwval(fw,k1)   & 0xFFF00FFF) == 0xE2400001) &&    // SUB Rx, Rn, #1
                                    isLDR(fw,k1+1) &&                                   // LDR Ry, [Rz,
                                    ((fwval(fw,k1+2) & 0xFFF00000) == 0xE1500000) &&    // CMP Rx, Ry
                                    (((fwRd(fw,k1) == fwRd(fw,k1+2)) && (fwRd(fw,k1+1) == fwRn(fw,k1+2))) ||
                                     ((fwRd(fw,k1) == fwRn(fw,k1+2)) && (fwRd(fw,k1+1) == fwRd(fw,k1+2)))) &&
                                    ((fwval(fw,k1+3) & 0xFFF00FFF) == 0x12800001) &&    // ADDNE Ry, Ry, #1
                                    ((fwRd(fw,k1+3) == fwRn(fw,k1+3)) && (fwRd(fw,k1+3) == fwRd(fw,k1+1))) &&
                                    ((fwval(fw,k1+4) & 0xFFF00FFF) == 0x03A00000) &&    // MOVEQ Ry, #0
                                    (fwRd(fw,k1+4) == fwRd(fw,k1+1)) &&
                                    isSTR(fw,k1+5) &&                                   // STR Ry, [Rz,
                                    ((fwRd(fw,k1+5) == fwRd(fw,k1+1)) && (fwRn(fw,k1+5) == fwRn(fw,k1+1)) && (fwOp2(fw,k1+5) == fwOp2(fw,k1+1)))
                                ) ||
                                (
                                    ((fwval(fw,k1)   & 0xFFF00FFF) == 0xE2400001) &&    // SUB Rx, Rn, #1
                                    isLDR(fw,k1+1) &&                                   // LDR Ry, [Rz,
                                    ((fwval(fw,k1+3) & 0xFFF00000) == 0xE1500000) &&    // CMP Rx, Ry
                                    (((fwRd(fw,k1) == fwRd(fw,k1+3)) && (fwRd(fw,k1+1) == fwRn(fw,k1+3))) ||
                                     ((fwRd(fw,k1) == fwRn(fw,k1+3)) && (fwRd(fw,k1+1) == fwRd(fw,k1+3)))) &&
                                    ((fwval(fw,k1+4) & 0xFFF00FFF) == 0x12800001) &&    // ADDNE Ry, Ry, #1
                                    ((fwRd(fw,k1+4) == fwRn(fw,k1+4)) && (fwRd(fw,k1+4) == fwRd(fw,k1+1))) &&
                                    ((fwval(fw,k1+5) & 0xFFF00FFF) == 0x03A00000) &&    // MOVEQ Ry, #0
                                    (fwRd(fw,k1+5) == fwRd(fw,k1+1)) &&
                                    isSTR(fw,k1+7) &&                                   // STR Ry, [Rz,
                                    ((fwRd(fw,k1+7) == fwRd(fw,k1+1)) && (fwRn(fw,k1+7) == fwRn(fw,k1+1)) && (fwOp2(fw,k1+7) == fwOp2(fw,k1+1)))
                                )
                               )
                            {
                                int ofst = fwOp2(fw,k1+1);
                                int reg = fwRn(fw,k1+1);
                                int k2;
                                for (k2 = f+1; (k2 < e) && !found; k2++)
                                {
                                    if (isLDR_PC(fw,k2) && (fwRd(fw,k2) == reg))
                                    {
                                        uint32_t base = LDR2val(fw,k2);
                                        print_stubs_min(fw,"active_raw_buffer",base+ofst,idx2adr(fw,k1));
                                        found = 1;
                                        break;
                                    }
                                }
                            }
                        }
                    }
                    k = find_nxt_str_ref(fw, sadr, k);
                }
            }
            if (!found)
            {
                bprintf("// Camera appears to have only 1 RAW buffer @ 0x%08x (Found @0x%08x)\n", rb1, idx2adr(fw,rb1_idx));
            }
        }
    }

    // Find 'fileio_semaphore'
    k = get_saved_sig(fw, "TakeSemaphoreStrictly");
    if (k >= 0)
    {
        uint32_t fadr = func_names[k].val;
        k = find_str(fw, "FileSem.c");
        if (k >= 0)
        {
            uint32_t nadr = idx2adr(fw, k);
            search_fw(fw, match_fileiosem, fadr, nadr, 3);
        }
    }

    // Find UI property count
    search_saved_sig(fw, "PTM_SetCurrentItem", match_uiprop_count, 0, 0, 30);
}

//------------------------------------------------------------------------------------------------------------

int find_ctypes(firmware *fw, int k)
{
    static unsigned char ctypes[] =
    {
        0x20, 0x20, 0x20, 0x20, 0x20, 0x20, 0x20, 0x20, 0x20, 0x60, 0x60, 0x60, 0x60, 0x60, 0x20, 0x20,
        0x20, 0x20, 0x20, 0x20, 0x20, 0x20, 0x20, 0x20, 0x20, 0x20, 0x20, 0x20, 0x20, 0x20, 0x20, 0x20,
        0x48, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10,
        0x84, 0x84, 0x84, 0x84, 0x84, 0x84, 0x84, 0x84, 0x84, 0x84, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10,
        0x10, 0x81, 0x81, 0x81, 0x81, 0x81, 0x81, 1, 1, 1, 1, 1, 1, 1, 1, 1,
        1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 0x10, 0x10, 0x10, 0x10, 0x10,
        0x10, 0x82, 0x82, 0x82, 0x82, 0x82, 0x82, 2, 2, 2, 2, 2, 2, 2, 2, 2,
        2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 0x10, 0x10, 0x10, 0x10, 0x20
    };

    if (k < (fw->size*4 - sizeof(ctypes)))
    {
        if (memcmp(((char*)fw->buf)+k,ctypes,sizeof(ctypes)) == 0)
        {
            bprintf("DEF(ctypes, 0x%08x)\n", fw->base + k);
            return 1;
        }
    }
    return 0;
}

int match_nrflag3(firmware *fw, int k, uint32_t v1, uint32_t v2)
{
    if (isBL(fw,k) && (idxFollowBranch(fw,k,0x01000001) == v1))
    {
        // Found call to function, work out R3 value passed in
        int ofst1 = 0;
        int k3, k4 = 0;
        for (k3=k; k3>k-30; k3--)
        {
            if ((fwval(fw,k3) & 0x0F0FF000) == 0x020D3000)       // Dest = R3, Src = SP = skip
                break;
            if ((fwval(fw,k3) & 0xFF0FF000) == 0xE2033000)       // ADD/SUB R3,R3,x
            {
                k4 = k3;
                if ((fwval(fw,k3) & 0x00F00000) == 0x00400000)   // SUB
                    ofst1 -= (fwval(fw,k3) & 0x00000FFF);
                else
                    ofst1 += (fwval(fw,k3) & 0x00000FFF);
            }
            if (isLDR_PC(fw,k3) && (fwRd(fw,k3) == 3))
            {
                int ofst2 = LDR2val(fw,k3);
                bprintf("\n// For capt_seq.c\n");
                if (ofst1 == 0)
                {
                    bprintf("DEF(_nrflag,0x%04x) // Found @ %08x (0x%04x)\n",ofst2,idx2adr(fw,k3),ofst2);
                    bprintf("//static long *nrflag = (long*)(0x%04x);       // Found @ %08x\n",ofst2,idx2adr(fw,k3));
                }
                else if (ofst1 < 0)
                {
                    bprintf("DEF(_nrflag,0x%04x) // Found @ %08x (0x%04x) & %08x (-0x%02x)\n",ofst2+ofst1,idx2adr(fw,k3),ofst2,idx2adr(fw,k),-ofst1);
                    bprintf("//static long *nrflag = (long*)(0x%04x-0x%02x);  // Found @ %08x & %08x\n",ofst2,-ofst1,idx2adr(fw,k3),idx2adr(fw,k4));
                }
                else
                {
                    bprintf("DEF(_nrflag,0x%04x) // Found @ %08x (0x%04x) & %08x (+0x%02x)\n",ofst2+ofst1,idx2adr(fw,k3),ofst2,idx2adr(fw,k),ofst1);
                    bprintf("//static long *nrflag = (long*)(0x%04x+0x%02x);  // Found @ %08x & %08x\n",ofst2,ofst1,idx2adr(fw,k3),idx2adr(fw,k4));
                }
                return 1;
            }
        }
    }
    return 0;
}

int match_nrflag(firmware *fw, int idx, int v)
{
    int k1, k2, k3;
    int found = 0;

    if (isLDR(fw, idx+1) && isLDR(fw, idx+2))
    {
        k3 = idx+2;
        int ofst2 = LDR2val(fw, k3);

        for (k1=k3+1; k1<k3+8; k1++)
        {
            if (isB(fw, k1))
            {
                k2 = idxFollowBranch(fw,k1,0x01000001);
                if (isSTR(fw, k2))
                {
                    found = 1;
                    break;
                }
                k2++;
                if (isSTR(fw, k2))
                {
                    found = 1;
                    break;
                }
            }
        }

        if (found)
        {
            int ofst1 = fw->buf[k2] & 0x00000FFF;
            bprintf("\n// For capt_seq.c\n");
            bprintf("DEF(_nrflag,0x%04x) // Found @ %08x (0x%04x) & %08x (+0x%02x)\n",ofst2+ofst1,idx2adr(fw,k3),ofst2,idx2adr(fw,k2),ofst1);
            bprintf("//static long *nrflag = (long*)(0x%04x+0x%02x);  // Found @ %08x & %08x\n",ofst2,ofst1,idx2adr(fw,k3),idx2adr(fw,k2));
            bprintf("//#define NR_AUTO (0)                          // have to explictly reset value back to 0 to enable auto\n");
        }
    }

    return found;
}

int match_nrflag2(firmware *fw, int k, int v)
{
    // Found NR_GetDarkSubType function, now follow first BL call.
    if (isBL(fw,k))
    {
        k = idxFollowBranch(fw,k,0x01000001);
        return search_fw(fw, match_nrflag3, k, 0, 1);
    }

    return 0;
}

// find LEDs

// helpers for find_leds()
int isADD(firmware *fw, int offset)
{
    if ((fwval(fw,offset) & 0xfff00000) == (0xe2800000)) // ADD
    {
        return 1;
    }
    return 0;
}

int isSUB(firmware *fw, int offset)
{
    if ((fwval(fw,offset) & 0xfff00000) == (0xe2400000)) // SUB
    {
        return 1;
    }
    return 0;
}

int isSTRw(firmware *fw, int offset)
{
    if ((fwval(fw,offset) & 0xfff00000) == (0xe5800000)) // STR Rx, [Ry, #offs]
    {
        return 1;
    }
    return 0;
}

int isSTRB(firmware *fw, int offset)
{
    if ((fwval(fw,offset) & 0xfff00000) == (0xe5c00000)) // STRB Rx, [Ry, #offs]
    {
        return 1;
    }
    return 0;
}

typedef struct {
    uint32_t addr;  // LED GPIO address
    int reg;        // register used to assemble the address
    int offs;       // offset in the LED table
    int done;       // already written
} LED_s;

// an array of 16 LED candidates should be enough
#define LEDMAX 16

int find_leds(firmware *fw)
{
    int j1, j2, j3;
    LED_s leds[LEDMAX];
    int k0 = find_str_ref(fw,"LEDCon");
    if (k0<0)
        return 0;
    int k1 = find_inst_rev(fw,isSTMFD_LR,k0,110);
    if (k1<0)
        return 0;
    j1 = find_inst(fw,isBL,k1,80);
    j2 = find_Nth_inst(fw,isBL,k1,80,3);
    if ((j1<0) || (j2<0))
        return 0;
    j3 = 0; // highest leds[] index
    int found;

    memset(leds, 0, sizeof(LED_s)*LEDMAX);

    // in older versions of the routine, 1st and 3rd BL is memory allocation
    if (followBranch(fw,idx2adr(fw,j1),0x01000001) == followBranch(fw,idx2adr(fw,j2),0x01000001))
    {
        k1 = find_Nth_inst(fw,isBL,k1,80,2);
        // LED table initializer func
        k1 = idxFollowBranch(fw,k1,0x01000001);
        if (k1<0)
            return 0;
        bprintf("\n// LED table init @ 0x%x\n",idx2adr(fw,k1));

        // scan for MMIO addresses loaded via LDR, should work on DIGIC 4 and III cams
        // some cameras may have specially handled LEDs, they will not be found
        // DIGIC 5 cams will get no hits here
        j2 = 0;
        j1 = 0;
        while (j2 < 32)
        {
            if (isLDR_PC(fw,k1+j2))
            {
                uint32_t l1 = LDR2val(fw,k1+j2);
                if (l1 >= 0xc0220000)
                {
                    leds[j3].addr = l1;
                    leds[j3].reg = fwRd(fw,k1+j2);
                    j3++;
                }
            }
            else if (isBX_LR(fw,k1+j2) || isB(fw,k1+j2))
            {
                break;
            }
            j2++;
            if (j3>=LEDMAX)
                break;
        }

        j3--;
        j1 = 0;
        if (j3 >= 0)
        {
            int repeatfrom = 0;
            int repeatreg = 0;
            int repeataddr = 0;
            while (j3 >= 0)
            {
                // main cycle to parse the LED table init function
                // tries to work out the LED MMIO addresses

                j2 = 0;
                int gotit = 0;
                if (repeatfrom)
                {
                    j2 = repeatfrom;
                    leds[j3].reg = repeatreg;
                    leds[j3].addr = repeataddr;
                    leds[j3].done = 0;
                    gotit = 1;
                    repeatfrom = 0;
                }
                while (j2 < 32)
                {
                    found = 0;
                    if (isLDR_PC(fw,k1+j2))
                    {
                        if (!gotit)
                        {
                            uint32_t l1 = LDR2val(fw,k1+j2);
                            if (l1 >= 0xc0220000)
                            {
                                if ((leds[j3].reg == fwRd(fw,k1+j2)) && (leds[j3].addr == LDR2val(fw,k1+j2)))
                                {
                                    leds[j3].done = 0;
                                    gotit = 1;
                                }
                            }
                        }
                        else
                        {
                            if (leds[j3].reg == fwRd(fw,k1+j2))
                            {
                                break;
                            }
                        }
                    }
                    else if (isBX_LR(fw,k1+j2) || isB(fw,k1+j2))
                    {
                        break;
                    }
                    if (!gotit)
                    {
                        // fast forward until the LDR in question is found
                        j2++;
                        continue;
                    }
                    if (isADD(fw,k1+j2))
                    {
                        if (leds[j3].reg == fwRd(fw,k1+j2))
                        {
                            leds[j3].addr += ALUop2a(fw,k1+j2);
                            leds[j3].done = 0;
                        }
                        else if (leds[j3].reg == fwRn(fw,k1+j2))
                        {
                            // MMIO address passed to another register, schedule re-entry if possible
                            if (!repeatfrom)
                            {
                                repeataddr = leds[j3].addr + ALUop2a(fw,k1+j2);
                                repeatreg = fwRd(fw,k1+j2);
                                repeatfrom = j2 + 1;
                            }
                        }
                    }
                    else if (isSUB(fw,k1+j2))
                    {
                        if (leds[j3].reg == fwRd(fw,k1+j2))
                        {
                            leds[j3].addr -= ALUop2a(fw,k1+j2);
                            leds[j3].done = 0;
                        }
                        else if (leds[j3].reg == fwRn(fw,k1+j2))
                        {
                            // MMIO address passed to another register, schedule re-entry if possible
                            if (!repeatfrom)
                            {
                                repeataddr = leds[j3].addr - ALUop2a(fw,k1+j2);
                                repeatreg = fwRd(fw,k1+j2);
                                repeatfrom = j2 + 1;
                            }
                        }
                    }
                    else if (isSTR(fw,k1+j2))
                    {
                        // LED references are always stored with STR, not STRB or STRH
                        // check for matching register
                        if (leds[j3].reg == fwRd(fw,k1+j2))
                        {
                            leds[j3].offs = fwval(fw,k1+j2) & 0xfff;
                            found = 1;
                        }
                    }
                    else if (isMOV_immed(fw,k1+j2) && (leds[j3].reg == fwRd(fw,k1+j2)))
                    {
                        // the register holding the MMIO address gets a new value, start again with the next MMIO, if any
                        break;
                    }
                    j2++;
                    // output data if valid
                    if (found && (!leds[j3].done))
                    {
                        j1++;
                        bprintf("// LED #%i: 0x%08x, offset 0x%x\n",j1, leds[j3].addr, leds[j3].offs);
                        leds[j3].done = 1;
                    }
                }
                if (!repeatfrom)
                {
                    j3--;
                }
            }
        }
        else
        {
            // DIGIC 5
            // LEDs are identified by their location in the GPIO table, not their address
            // some LEDs might be "special cased" and not appear in the GPIO table or the LED table init function
            // those special cases are not currently handled

            // locate GPIO table first
            int gpiotbladdr = 0;
            j2 = find_str_ref(fw,"\n\n Set LCD Driver: Address 0x%04x <-- Data 0x%04x\n");
            if (j2 > 0)
            {
                j2 = find_inst_rev(fw, isBL, j2, 8);
                if (j2 > 0)
                {
                    j3 = 2;
                    while (j3 > 0)
                    {
                        if ( !((fwval(fw,j2-1)&0xfffff000)==0xe3a01000) && !((fwval(fw,j2-2)&0xfffff000)==0xe3a01000) ) // MOV R1, #imm
                        {
                            j2 = find_inst_rev(fw, isBL, j2-1, 6);
                        }
                        else
                        {
                            break;
                        }
                        j3--;
                    }
                    if (j2 > 0)
                    {
                        // j2 points to a function that is used to poke GPIOs, based on a GPIO table
                        j2 = idxFollowBranch(fw, j2, 0x01000001);
                        if (isLDR_PC(fw,j2))
                        {
                            // 1st instruction references the table's address (could change in the future?)
                            gpiotbladdr = adr2idx(fw, LDR2val(fw, j2));
                        }
                    }
                }
            }
            // identify the LED(s)
            /*
             * r0 is a pointer to the LED struct array
             * below pattern inits an LED:
             * mov rx, #imm
             * ...
             * strb ry, [r0, #imm2]
             * str rx, [r0, #imm2+4]
             * the strb pair might be at a different location
             */
            j2 = 0;
            j3 = 0;
            while (j2 < 32)
            {
                // find str rx, [r0, #imm2+4]
                if (isSTRw(fw,k1+j2))
                {
                    uint32_t o1 = fwval(fw,k1+j2)&0xfff; // may not be the correct way of determining the offset
                    uint32_t o2 = 0;
                    int n = 1;
                    int r1 = -1;
                    while (n < 32)
                    {
                        if (isSTRB(fw,k1+n))
                        {
                            o2 = fwval(fw,k1+n)&0xfff;
                            if (o1-4 == o2)
                            {
                                r1 = fwRd(fw,k1+j2); // dest. register
                            }
                        }
                        else if (isBX_LR(fw,k1+n) || isB(fw,k1+n))
                        {
                            break;
                        }
                        n++;
                    }
                    if (r1 >= 0)
                    {
                        // search for that register's content
                        n = k1+j2-1;
                        while (n >= k1)
                        {
                            if (isMOV_immed(fw, n) && (fwRd(fw, n) == r1))
                            {
                                leds[j3].addr = ALUop2a(fw,n);
                                leds[j3].offs = o1;
                                j3++;
                                break;
                            }
                            n--;
                        }
                    }
                }
                else if (isBX_LR(fw,k1+j2) || isB(fw,k1+j2))
                {
                    break;
                }
                j2++;
                if (j3 >= LEDMAX)
                    break;
            }
            j3--;
            while (j3 >= 0)
            {
                j1++;
                if (leds[j3].addr > 0xfff)
                {
                    bprintf("// LED #%i: unknown (index: 0x%08x), offset 0x%x\n",j1, leds[j3].addr, leds[j3].offs);
                }
                else if (gpiotbladdr)
                {
                    bprintf("// LED #%i: 0x%08x (#%d in GPIO table), offset 0x%x\n",j1, fwval(fw, leds[j3].addr + gpiotbladdr), leds[j3].addr, leds[j3].offs);
                }
                else
                {
                    bprintf("// LED #%i:  #%d in GPIO table, offset 0x%x\n",j1, leds[j3].addr, leds[j3].offs);
                }
                j3--;
            }
            if (gpiotbladdr)
            {
                bprintf("// GPIO table @ 0x%x\n",idx2adr(fw, gpiotbladdr));
            }
        }
    }
    else
    {
        // DIGIC 4+
        // LEDs are identified by their location in the GPIO table, not their address
        // some LEDs might be "special cased" and not appear in the GPIO table or the LED table init function
        // those special cases are not currently handled

        int f1 = get_saved_sig(fw,"CreateMessageQueueStrictly");
        if (f1 < 0)
            return 0;
        f1 = adr2idx(fw, func_names[f1].val);

        int n;
        k1 = k0; // ref. to "LEDCon"
        for (n=0; n<5; n++)
        {
            k1 = find_inst_rev(fw,isBL,k1-1,80);
            if (k1 > 0)
            {
                if (idx2adr(fw,idxFollowBranch(fw,k1,0x01000001)) == idx2adr(fw,f1)) // BL CreateMessageQueueStrictly
                {
                    n = -1;
                    break;
                }
            }
        }
        if (n >= 0)
            return 0;
        // LED table initializer func, right before CreateMessageQueueStrictly
        k1 = find_inst_rev(fw,isBL,k1-1,80);
        if (k1<0)
            return 0;
        k1 = idxFollowBranch(fw,k1,0x01000001);
        if (k1<0)
            return 0;
        bprintf("\n// LED table init @ 0x%x\n",idx2adr(fw,k1));

        // locate GPIO table
        j1 = 0;
        int gpiotbladdr = 0;
        j2 = find_str_ref(fw,"\n\n Set LCD Driver: Address 0x%04x <-- Data 0x%04x\n");
        if (j2 > 0)
        {
            j2 = find_inst_rev(fw, isBL, j2, 8);
            if (j2 > 0)
            {
                j3 = 2;
                while (j3 > 0)
                {
                    if ( !((fwval(fw,j2-1)&0xfffff000)==0xe3a00000) && !((fwval(fw,j2-2)&0xfffff000)==0xe3a00000) ) // MOV R0, #imm
                    {
                        j2 = find_inst_rev(fw, isBL, j2-1, 6);
                    }
                    else
                    {
                        break;
                    }
                    j3--;
                }
                if (j2 > 0)
                {
                    // j2 points to a function that is used to poke GPIOs, based on a GPIO table
                    j2 = idxFollowBranch(fw, j2, 0x01000001);
                    if (isLDR_PC(fw,j2))
                    {
                        // 1st instruction references the table's address (could change in the future?)
                        gpiotbladdr = adr2idx(fw, LDR2val(fw, j2));
                    }
                }
            }
        }
        // identify the LED(s)
        /*
         * r0 is a pointer to the LED struct array
         * below pattern inits an LED:
         * mov rx, #imm
         * ...
         * strb ry, [r0, #imm2]
         * str rx, [r0, #imm2+4]
         * the strb pair might be at a different location
         */
        j2 = 0;
        j3 = 0;
        while (j2 < 32)
        {
            // find str rx, [r0, #imm2+4]
            if (isSTRw(fw,k1+j2))
            {
                uint32_t o1 = fwval(fw,k1+j2)&0xfff; // may not be the correct way of determining the offset
                uint32_t o2 = 0;
                int n = 1;
                int r1 = -1;
                while (n < 32)
                {
                    if (isSTRB(fw,k1+n))
                    {
                        o2 = fwval(fw,k1+n)&0xfff;
                        if (o1-4 == o2)
                        {
                            r1 = fwRd(fw,k1+j2); // dest. register
                        }
                    }
                    else if (isBX_LR(fw,k1+n) || isB(fw,k1+n))
                    {
                        break;
                    }
                    n++;
                }
                if (r1 >= 0)
                {
                    // search for that register's content
                    n = k1+j2-1;
                    while (n >= k1)
                    {
                        if (isMOV_immed(fw, n) && (fwRd(fw, n) == r1))
                        {
                            leds[j3].addr = ALUop2a(fw,n);
                            leds[j3].offs = o1;
                            j3++;
                            break;
                        }
                        n--;
                    }
                }
            }
            else if (isBX_LR(fw,k1+j2) || isB(fw,k1+j2))
            {
                break;
            }
            j2++;
            if (j3 >= LEDMAX)
                break;
        }
        j3--;
        while (j3 >= 0)
        {
            j1++;
            if (leds[j3].addr > 0xfff)
            {
                bprintf("// LED #%i: unknown (index: 0x%08x), offset 0x%x\n",j1, leds[j3].addr, leds[j3].offs);
            }
            else if (gpiotbladdr)
            {
                bprintf("// LED #%i: 0x%08x (#%d in GPIO table), offset 0x%x\n",j1, fwval(fw, leds[j3].addr + gpiotbladdr), leds[j3].addr, leds[j3].offs);
            }
            else
            {
                bprintf("// LED #%i:  #%d in GPIO table, offset 0x%x\n",j1, leds[j3].addr, leds[j3].offs);
            }
            j3--;
        }
        if (gpiotbladdr)
        {
            bprintf("// GPIO table @ 0x%x\n",idx2adr(fw, gpiotbladdr));
        }
    }
    return 0;
}

int find_task_related_info(firmware *fw)
{
    int i = get_saved_sig(fw,"get_self_task_id");
    uint32_t u, v;
    if (i < 0)
    {
        return 0;
    }
    i = adr2idx(fw, func_names[i].val);
    if ( (fwval(fw,i)&0xffff0000)==0xe59f0000 ) // ldr r0, [pc, #imm]
    {
        // "interrupt service routine" flag
        u = LDR2val(fw, i);
        if ( (fwval(fw,i+3)&0xffff0000)==0x059f0000 ) // ldreq r0, [pc, #imm]
        {
            // pointer to current task's control block
            v = LDR2val(fw, i+3);
            bprintf("// ISR flag: 0x%x, pointer to current task's control block: 0x%x\n",u, v);
        }
    }
    // part 2, find the TCB area
    int j, k, n, fnd;
    int m = 0;
    i = find_str(fw, "DRYOS version 2.3, release ");
    j = find_nxt_str_ref(fw, i, -1);
    if (j == -1)
    {
        // special case: some r50 cams have the string in RAM and all references point there
        u = idx2adr(fw,i);
        if ( (u > fw->base_copied) && ((u-fw->base_copied)/4 < fw->size2))
        {
            i = adr2idx(fw, fw->base2 + (u-fw->base_copied));
            j = find_nxt_str_ref(fw, i, -1);
        }
    }
    fnd = 0;
    while (!fnd) {
        if (j != -1)
        {
            k = find_nxt_str_ref(fw, i, j+1);
            if (k != -1)
            {
                if (k-j>5)
                {
                    // refs too far, try again
                    j = k;
                }
                else
                {
                    m = find_inst_rev(fw, isSTMFD_LR, j, 42);
                    if (j-m>24)
                    {
                        fnd = 1;
                    }
                }
            }
        }
        else
        {
            break;
        }
    }
    u = 0;
    if (fnd)
    {
        n = find_Nth_inst(fw, isBL, m, 6, 2);
        if (n != -1)
        {
            n = idxFollowBranch(fw,n,0x01000001);
            n = find_inst(fw, isSTR, n, 8);
            if (n != -1)
            {
                m = fwRn(fw, n);    // this register holds the base address pointer of TCB area
                n = find_inst_rev(fw, isLDR_PC, n-1, 4);
                if (n != -1)
                {
                    if (fwRd(fw, n) != m)
                    {
                        n = find_inst_rev(fw, isLDR_PC, n-1, 3);
                        if ((n != -1) && (fwRd(fw, n) == m))
                        {
                            u = LDR2val(fw, n);
                            v = idx2adr(fw, n);
                            bprintf("// pointer to TCB area: 0x%x, found @ 0x%x\n",u,v);
                        }
                    }
                }
            }
        }
    }


    return 0;
}

void find_AdditionAgent_RAM(firmware *fw)
{
    int i = get_saved_sig(fw,"AdditionAgentRAM_FW");
    uint32_t r, sizeloc = 0, startloc = 0;
    uint32_t ramsize = 0;
    uint32_t ramstart = 0;
    if (i >= 0)
    {
        int j1 = adr2idx(fw, func_names[i].val);
        int n;
        for (n=1; n<16; n++)
        {
            if (fwval(fw,j1+n) == 0xe3500a32) // cmp  r0, #0x32000 
            {
                ramsize = 0x32000;
                sizeloc = idx2adr(fw,j1+n);
                break;
            }
            else if (fwval(fw,j1+n) == 0xe3500a22) // cmp  r0, #0x22000 
            {
                ramsize = 0x22000;
                sizeloc = idx2adr(fw,j1+n);
                break;
            }
        }
        if (n >= 15)
            n = 0;
        j1 += n;
        for (n=0; n<=16; n++)
        {
            r = LDR2val(fw,j1+n);
            if ( isLDR_PC(fw,j1+n) && (r>fw->memisostart) && (r<fw->maxram) )
            {
                ramstart = r;
                startloc = idx2adr(fw,j1+n);
            }
        }
        if (ramstart>0)
        {
            bprintf("//   ARAM_HEAP_START = 0x%x  // Found @ 0x%08x\n",ramstart,startloc);
            bprintf("//   ARAM_HEAP_SIZE  = 0x%x   // Found @ 0x%08x\n",ramsize,sizeloc);
        }
    }
}

// Search for things
void find_other_vals(firmware *fw)
{
    out_hdr = 1;
    add_blankline();

    bprintf("// Misc stuff\n");

    if (!search_fw_bytes(fw, find_ctypes))
    {
        bprintf("//DEF(ctypes, *** Not Found ***)\n");
    }

    add_blankline();
    find_task_related_info(fw);
    find_leds(fw);

    // Look for nrflag (for capt_seq.c)
    int found = 0;
    if (fw->dryos_ver >= 45)
    {
        found = search_saved_sig(fw, "NR_SetDarkSubType", match_nrflag, 0, 0, 1);
    }
    if (!found)
    {
        search_saved_sig(fw, "NR_GetDarkSubType", match_nrflag2, 0, 0, 20);
    }
}

//------------------------------------------------------------------------------------------------------------

void print_kval(firmware *fw, uint32_t tadr, int tsiz, int tlen, uint32_t ev, const char *name, char *sfx)
{
    int tidx = adr2idx(fw,tadr);
    int k, kval = 0;
    for (k=0; k<tlen; k+=tsiz)
    {
        if (fw->buf[tidx+k+1] == ev)
        {
            kval = fw->buf[tidx+k];
            tadr = idx2adr(fw,tidx+k);
            break;
        }
    }
    if (kval > 0)
    {
        char fn[100], rn[100];
        strcpy(fn,name); strcat(fn,sfx);
        strcpy(rn,name); strcat(rn,"_IDX");

        int r = (kval >> 5) & 7;
        uint32_t b = (1 << (kval & 0x1F));
        int i = (kval >> 16) & 1;

        bprintf("//#define %-20s0x%08x // Found @0x%08x, levent 0x%x%s\n",fn,b,tadr,ev,i?" (non-inverted logic)":"");
        bprintf("//#define %-20s%d\n",rn,r);
    }
}

void print_physw_raw_vals(firmware *fw, uint32_t tadr, int tsiz, int tlen)
{
    int tidx = adr2idx(fw,tadr);
    int k, kval = 0;
    uint32_t ev;
    FILE *out_fp = fopen("physw_bits.txt", "w");
    if (out_fp == NULL) return;

    for (k=0; k<tlen; k+=tsiz)
    {
        ev = fw->buf[tidx+k+1];
        kval = fw->buf[tidx+k];
        tadr = idx2adr(fw,tidx+k);
        if (kval > 0)
        {

            int r = (kval >> 5) & 7;
            uint32_t b = (1 << (kval & 0x1F));
            int i = (kval >> 16) & 1;

            fprintf(out_fp, "levent 0x%08x, 0x%08x, index %d%s\n", ev, b, r, i?" (non-inverted logic)":"");
        }
    }
    fclose(out_fp);
}

typedef struct {
    int         reg;
    uint32_t    bits;
    char        nm[32];
    uint32_t    fadr;
    uint32_t    ev;
    int         inv;
} kinfo;

int     kmask[3];
kinfo   key_info[100];
int     kcount = 0;
uint32_t kshutter_min_bits = 0xFFFFFFFF;

void add_kinfo(int r, uint32_t b, const char *nm, uint32_t adr, uint32_t ev, int inv)
{
    key_info[kcount].reg = r;
    key_info[kcount].bits = b;
    strcpy(key_info[kcount].nm, nm);
    key_info[kcount].fadr = adr;
    key_info[kcount].ev = ev;
    key_info[kcount].inv = inv;
    kcount++;
    kmask[r] |= b;
    if ((ev <= 1) && (b < kshutter_min_bits)) kshutter_min_bits = b;
}

uint32_t add_kmval(firmware *fw, uint32_t tadr, int tsiz, int tlen, uint32_t ev, const char *name, uint32_t xtra)
{
    int tidx = adr2idx(fw,tadr);
    int r, k, kval = 0;
    uint32_t b = 0;
    int inv = 0;
    for (k=0; k<tlen; k+=tsiz)
    {
        if (fw->buf[tidx+k+1] == ev)
        {
            kval = fw->buf[tidx+k];
            tadr = idx2adr(fw,tidx+k);
            break;
        }
    }
    if (kval > 0)
    {
        r = (kval >> 5) & 7;
        b = (1 << (kval & 0x1F));
        inv = ((kval&0xff0000)==0x10000)?0:1;

        add_kinfo(r,b|xtra,name,tadr,ev,inv);
    }

    return b;
}

int kinfo_compare(const kinfo *p1, const kinfo *p2)
{
    if (p1->reg > p2->reg)
    {
        return 1;
    }
    else if (p1->reg < p2->reg)
    {
        return -1;
    }
    if ((p1->ev <= 1) && (p2->ev <= 1))    // output shutter entries in reverse order
    {
        if (p1->bits > p2->bits)
        {
            return -1;
        }
        else if (p1->bits < p2->bits)
        {
            return 1;
        }
    }
    // if one entry is shutter then compare to min shutter bits
    if (p1->ev <= 1)
    {
        if (kshutter_min_bits > p2->bits)
        {
            return 1;
        }
        else if (kshutter_min_bits < p2->bits)
        {
            return -1;
        }
    }
    if (p2->ev <= 1)
    {
        if (p1->bits > kshutter_min_bits)
        {
            return 1;
        }
        else if (p1->bits < kshutter_min_bits)
        {
            return -1;
        }
    }
    if (p1->bits > p2->bits)
    {
        return 1;
    }
    else if (p1->bits < p2->bits)
    {
        return -1;
    }

    return 0;
}

void print_kmvals()
{
    qsort(key_info, kcount, sizeof(kinfo), (void*)kinfo_compare);

    bprintf("//static KeyMap keymap[] = {\n");

    int k;
    for (k=0; k<kcount; k++)
    {
        bprintf("//    { %d, %-20s,0x%08x }, // Found @0x%08x, levent 0x%02x%s\n",key_info[k].reg,key_info[k].nm,key_info[k].bits,key_info[k].fadr,key_info[k].ev,(key_info[k].inv==0)?"":" (uses inverted logic in physw_status)");
    }

    bprintf("//    { 0, 0, 0 }\n//};\n");
}

int match_GetSDProtect(firmware *fw, int k, int v)
{
    if (isB(fw,k))    // B
    {
        k = idxFollowBranch(fw,k,1);
        if (isLDR_PC(fw,k))
        {
            return LDR2val(fw,k);
        }
    }

    return 0;
}

void find_key_vals(firmware *fw)
{
    int k,k1;

    out_hdr = 1;
    add_blankline();

    // find 'SD_READONLY_FLAG'
    uint32_t tadr = search_saved_sig(fw, "GetSDProtect", match_GetSDProtect, 0, 1, 1);
    if (tadr == 0)
    {
        k = find_str_ref(fw,"SD Not Exist\n");
        if (k >= 0)
        {
            for (k1=k-1; k1>k-5; k1--)
            {
                if (isBL(fw,k1))    // BL
                {
                    uint32_t fadr = followBranch(fw,idx2adr(fw,k1),0x01000001);
                    int k2 = adr2idx(fw,fadr);
                    if (isLDR_PC(fw,k2))
                    {
                        tadr = LDR2val(fw,k2);
                    }
                }
            }
        }
    }
    if (tadr != 0)
    {
        int tsiz = 2;
        if (fw->buf[adr2idx(fw,tadr)+2] == 0) tsiz = 3;

        uint32_t madr = fw->base + (fw->size*4-4);
        for (k=0; k<(tadr-fw->base)/4; k++)
        {
            if (isLDR_PC(fw,k))
            {
                uint32_t adr = LDR2val(fw,k);
                if ((adr > tadr) && (adr < madr))
                {
                    madr = adr;
                }
            }
        }
        int tlen = (madr - tadr) / 4;
        if (tsiz == 2)
        {
            k1 = adr2idx(fw,tadr);
            for (k=0; k<tlen/3; k+=3)
            {
                if ((fw->buf[k1+k+1] == 0xFFFFFFFF) && (fw->buf[k1+k+4] == 0xFFFFFFFF))
                {
                    tsiz = 3;
                    break;
                }
            }
        }
        if (tlen > 50*tsiz) tlen = 50*tsiz;

#ifdef LIST_PHYSW_TABLE
        // output all physw events from the table if enabled
        print_physw_raw_vals(fw, tadr, tsiz, tlen);
#endif
        bprintf("// Bitmap masks and physw_status index values for SD_READONLY and USB power flags (for kbd.c).\n");
        if (fw->dryos_ver >= 49)
        {
            // Event ID's have changed in DryOS R49 **********
            print_kval(fw,tadr,tsiz,tlen,0x20A,"SD_READONLY","_FLAG");
            print_kval(fw,tadr,tsiz,tlen,0x202,"USB","_MASK");
            print_kval(fw,tadr,tsiz,tlen,0x205,"BATTCOVER","_FLAG");
            print_kval(fw,tadr,tsiz,tlen,0x204,"HOTSHOE","_FLAG");
        }
        else
        {
            print_kval(fw,tadr,tsiz,tlen,0x90A,"SD_READONLY","_FLAG");
            print_kval(fw,tadr,tsiz,tlen,0x902,"USB","_MASK");
            print_kval(fw,tadr,tsiz,tlen,0x905,"BATTCOVER","_FLAG");
            print_kval(fw,tadr,tsiz,tlen,0x904,"HOTSHOE","_FLAG");
        }

        uint32_t key_half = add_kmval(fw,tadr,tsiz,tlen,0,"KEY_SHOOT_HALF",0);
        add_kmval(fw,tadr,tsiz,tlen,1,"KEY_SHOOT_FULL",key_half);
        add_kmval(fw,tadr,tsiz,tlen,1,"KEY_SHOOT_FULL_ONLY",0);
        
        if (fw->dryos_ver == 52)  // unclear if this applies any other ver
        {
            add_kmval(fw,tadr,tsiz,tlen,3,"KEY_ZOOM_IN",0);
            add_kmval(fw,tadr,tsiz,tlen,4,"KEY_ZOOM_OUT",0);
            add_kmval(fw,tadr,tsiz,tlen,6,"KEY_UP",0);
            add_kmval(fw,tadr,tsiz,tlen,7,"KEY_DOWN",0);
            add_kmval(fw,tadr,tsiz,tlen,8,"KEY_LEFT",0);
            add_kmval(fw,tadr,tsiz,tlen,9,"KEY_RIGHT",0);
            add_kmval(fw,tadr,tsiz,tlen,0xA,"KEY_SET",0);
            add_kmval(fw,tadr,tsiz,tlen,0xB,"KEY_MENU",0);
            add_kmval(fw,tadr,tsiz,tlen,0xC,"KEY_DISPLAY",0);
            add_kmval(fw,tadr,tsiz,tlen,0x12,"KEY_HELP",0);
            add_kmval(fw,tadr,tsiz,tlen,0x19,"KEY_ERASE",0);
            add_kmval(fw,tadr,tsiz,tlen,2,"KEY_VIDEO",0);
        }
        else if (fw->dryos_ver < 54)
        {
            add_kmval(fw,tadr,tsiz,tlen,2,"KEY_ZOOM_IN",0);
            add_kmval(fw,tadr,tsiz,tlen,3,"KEY_ZOOM_OUT",0);
            add_kmval(fw,tadr,tsiz,tlen,4,"KEY_UP",0);
            add_kmval(fw,tadr,tsiz,tlen,5,"KEY_DOWN",0);
            add_kmval(fw,tadr,tsiz,tlen,6,"KEY_LEFT",0);
            add_kmval(fw,tadr,tsiz,tlen,7,"KEY_RIGHT",0);
            add_kmval(fw,tadr,tsiz,tlen,8,"KEY_SET",0);
            add_kmval(fw,tadr,tsiz,tlen,9,"KEY_MENU",0);
            add_kmval(fw,tadr,tsiz,tlen,0xA,"KEY_DISPLAY",0);
        }
        else if (fw->dryos_ver < 55)
        {
            add_kmval(fw,tadr,tsiz,tlen,3,"KEY_ZOOM_IN",0);
            add_kmval(fw,tadr,tsiz,tlen,4,"KEY_ZOOM_OUT",0);
            add_kmval(fw,tadr,tsiz,tlen,6,"KEY_UP",0);
            add_kmval(fw,tadr,tsiz,tlen,7,"KEY_DOWN",0);
            add_kmval(fw,tadr,tsiz,tlen,8,"KEY_LEFT",0);
            add_kmval(fw,tadr,tsiz,tlen,9,"KEY_RIGHT",0);
            add_kmval(fw,tadr,tsiz,tlen,0xA,"KEY_SET",0);
            add_kmval(fw,tadr,tsiz,tlen,0xE,"KEY_MENU",0);
            add_kmval(fw,tadr,tsiz,tlen,2,"KEY_VIDEO",0);
            add_kmval(fw,tadr,tsiz,tlen,0xD,"KEY_HELP",0);
            //add_kmval(fw,tadr,tsiz,tlen,?,"KEY_DISPLAY",0);
        }
        else
        {
            add_kmval(fw,tadr,tsiz,tlen,3,"KEY_ZOOM_IN",0);
            add_kmval(fw,tadr,tsiz,tlen,4,"KEY_ZOOM_OUT",0);
            add_kmval(fw,tadr,tsiz,tlen,6,"KEY_UP",0);
            add_kmval(fw,tadr,tsiz,tlen,7,"KEY_DOWN",0);
            add_kmval(fw,tadr,tsiz,tlen,8,"KEY_LEFT",0);
            add_kmval(fw,tadr,tsiz,tlen,9,"KEY_RIGHT",0);
            add_kmval(fw,tadr,tsiz,tlen,0xA,"KEY_SET",0);
            add_kmval(fw,tadr,tsiz,tlen,0x14,"KEY_MENU",0);
            add_kmval(fw,tadr,tsiz,tlen,2,"KEY_VIDEO",0);
            add_kmval(fw,tadr,tsiz,tlen,0xD,"KEY_HELP",0);
            //add_kmval(fw,tadr,tsiz,tlen,?,"KEY_DISPLAY",0);
        }
        if (fw->dryos_ver <= 47)
        {
            add_kmval(fw,tadr,tsiz,tlen,0x601,"KEY_PLAYBACK",0);
            add_kmval(fw,tadr,tsiz,tlen,0x600,"KEY_POWER",0);
            add_kmval(fw,tadr,tsiz,tlen,0x12,"KEY_VIDEO",0);
        }
        else
        {
            add_kmval(fw,tadr,tsiz,tlen,0x101,"KEY_PLAYBACK",0);
            add_kmval(fw,tadr,tsiz,tlen,0x100,"KEY_POWER",0);
            if (fw->dryos_ver == 49)
            {
                add_kmval(fw,tadr,tsiz,tlen,0x19,"KEY_VIDEO",0);
            }
            else if(fw->dryos_ver == 50)
            {
                add_kmval(fw,tadr,tsiz,tlen,0x1A,"KEY_VIDEO",0);
                add_kmval(fw,tadr,tsiz,tlen,0x14,"KEY_HELP",0);
            }
        }

        bprintf("\n// Keymap values for kbd.c. Additional keys may be present, only common values included here.\n");
        print_kmvals();
    }
}

//------------------------------------------------------------------------------------------------------------

uint32_t nadr;
uint32_t eadr;

int get_eventproc_val(firmware *fw, int k)
{
    if (isADR_PC(fw,k) && (fwRd(fw,k) == 0))
        nadr = ADR2adr(fw,k);
    else if (isADR_PC(fw,k) && (fwRd(fw,k) == 1))
        eadr = ADR2adr(fw,k);
    else if (isLDR_PC(fw,k) && (fwRd(fw,k) == 0))
        nadr = LDR2val(fw,k);
    else if (isLDR_PC(fw,k) && (fwRd(fw,k) == 1))
        eadr = LDR2val(fw,k);
    else
        return 0;
    return 1;
}

void add_func_name(char *n, uint32_t eadr, char *suffix)
{
    int k;

    char *s = n;
    if (suffix != 0)
    {
        s = malloc(strlen(n) + strlen(suffix) + 1);
        sprintf(s, "%s%s", n, suffix);
    }

    for (k=0; func_names[k].name != 0; k++)
    {
        if (strcmp(func_names[k].name, s) == 0)
        {
            if (func_names[k].val == 0)             // same name, no address
            {
                func_names[k].val = eadr;
                func_names[k].flags |= EV_MATCH;
                if (s != n) free(s);
                return;
            }
            else if (func_names[k].val == eadr)     // same name, same address
            {
                if (s != n) free(s);
                return;
            }
        }
    }

    func_names[next_func_entry].name = s;
    func_names[next_func_entry].flags = OPTIONAL|UNUSED;
    func_names[next_func_entry].val = eadr;
    next_func_entry++;
    func_names[next_func_entry].name = 0;
}

void add_func_name2(firmware *fw, uint32_t nadr, uint32_t eadr, char *suffix)
{
    char *n = (char*)adr2ptr(fw,nadr);
    if (isB(fw,adr2idx(fw,eadr)))
    {
        char *s = malloc(strlen(n) + 3);
        sprintf(s,"j_%s",n);
        add_func_name(s, eadr, suffix);
        eadr = followBranch(fw,eadr,1);
    }
    add_func_name(n, eadr, suffix);
}

int match_eventproc(firmware *fw, int k, uint32_t fadr, uint32_t v2)
{
    if (isBorBL(fw,k))
    {
        uint32_t adr = followBranch(fw,idx2adr(fw,k),0x01000001);
        if (adr == fadr)
        {
            nadr = 0;
            eadr = 0;
            k--;
            if (get_eventproc_val(fw, k) == 0)
            {
                int k1 = find_inst_rev(fw, isB, k, 500);
                if (k1 >= 0)
                {
                    k = k1 - 1;
                    get_eventproc_val(fw, k);
                }
            }
            k--;
            if (get_eventproc_val(fw, k) == 0)
            {
                int k1 = find_inst_rev(fw, isB, k, 500);
                if (k1 >= 0)
                {
                    k = k1 - 1;
                    get_eventproc_val(fw, k);
                }
            }
            if ((nadr != 0) && (eadr != 0))
            {
                add_func_name2(fw, nadr, eadr, "_FW");
            }
        }
    }
    return 0;
}

int match_registerproc2(firmware *fw, int k, uint32_t fadr, uint32_t v2)
{
    if (isBorBL(fw,k))
    {
        uint32_t adr = followBranch(fw,idx2adr(fw,k),0x01000001);
        if (adr == fadr)
        {
            nadr = 0;
            eadr = 0;
            k--;
            if (get_eventproc_val(fw, k) == 0)
            {
                int k1 = find_inst_rev(fw, isB, k, 500);
                if (k1 >= 0)
                {
                    k = k1 - 1;
                    get_eventproc_val(fw, k);
                }
            }
            k--;
            if (get_eventproc_val(fw, k) == 0)
            {
                int k1 = find_inst_rev(fw, isB, k, 500);
                if (k1 >= 0)
                {
                    k = k1 - 1;
                    get_eventproc_val(fw, k);
                }
            }
            if ((nadr != 0) && (eadr != 0))
            {
                add_func_name2(fw, nadr, eadr, "_FW");
            }
        }
    }
    return 0;
}

int match_registerproc(firmware *fw, int k, uint32_t fadr, uint32_t v2)
{
    if (isB(fw,k+1) && isMOV_immed(fw,k) && (fwRd(fw,k) == 2))
    {
        uint32_t adr = followBranch(fw,idx2adr(fw,k+1),1);
        if (adr == fadr)
        {
            search_fw(fw, match_registerproc2, idx2adr(fw,k), 0, 2);
        }
    }
    return 0;
}

int match_registerlists(firmware *fw, int k, uint32_t fadr, uint32_t v2)
{
    if (isBorBL(fw,k+1) && isLDR_PC(fw,k) && (fwRd(fw,k) == 0))
    {
        uint32_t adr = followBranch2(fw,idx2adr(fw,k+1),0x01000001);
        if (adr == fadr)
        {
            int j = adr2idx(fw,LDR2val(fw,k));
            if (!idx_valid(fw,j))
            {
                j = adr2idx(fw,LDR2val(fw,k) - fw->data_start + fw->data_init_start);
            }
            if (idx_valid(fw,j))
            {
                while (fwval(fw,j) != 0)
                {
                    add_func_name2(fw, fwval(fw,j), fwval(fw,j+1), "_FW");
                    j += 2;
                }
            }
        }
    }
    else if (isBorBL(fw,k+1) && isLDMFD(fw,k) && isLDR_PC(fw,k-1) && (fwRd(fw,k-1) == 0))
    {
        uint32_t adr = followBranch2(fw,idx2adr(fw,k+1),0x01000001);
        if (adr == fadr)
        {
            int j = adr2idx(fw,LDR2val(fw,k-1));
            if (!idx_valid(fw,j))
            {
                j = adr2idx(fw,LDR2val(fw,k-1) - fw->data_start + fw->data_init_start);
            }
            if (idx_valid(fw,j))
            {
                while (fwval(fw,j) != 0)
                {
                    add_func_name2(fw, fwval(fw,j), fwval(fw,j+1), "_FW");
                    j += 2;
                }
            }
        }
    }
    return 0;
}

void find_eventprocs(firmware *fw)
{
    int j = get_saved_sig(fw,"ExportToEventProcedure_FW");
    if (j >= 0)
    {
        uint32_t fadr = func_names[j].val;
        search_fw(fw, match_eventproc, fadr, 0, 1);

        if (isB(fw,adr2idx(fw,fadr)+2))
        {
            fadr = followBranch(fw, fadr+8, 1);
            add_func_name("RegisterEventProcedure", fadr, 0);
            search_fw(fw, match_registerproc, fadr, 0, 2);
        }

        j = get_saved_sig(fw,"SS.Create_FW");
        if (j >= 0)
        {
            j = adr2idx(fw,func_names[j].val);
            int offsets[] = { 1, 3, 7, 8 };
            int i;
            for (i=0; i<4; i++)
            {
                if (isLDR_PC(fw,j+offsets[i]) && (fwRd(fw,j+offsets[i]) == 0) && isBL(fw,j+offsets[i]+1))
                {
                    fadr = followBranch(fw,idx2adr(fw,j+offsets[i]+1),0x01000001);
                    search_fw(fw, match_registerlists, fadr, 0, 2);
                    break;
                }
            }
        }
        else
        {
            // S5IS
            j = find_strptr_ref(fw,"ResetZoomLens");
            if (j >= 0)
            {
                if (isBorBL(fw,j+1))
                {
                    fadr = followBranch(fw,idx2adr(fw,j+1),0x01000001);
                    search_fw(fw, match_registerlists, fadr, 0, 2);
                }
            }
        }

        j = get_saved_sig(fw,"TerminateAdjustmentSystem_FW");
        if (j >= 0)
        {
            j = adr2idx(fw,func_names[j].val);
            int k;
            for (k=j; k<j+8; k++)
            {
                if (isBL(fw,k))
                {
                    int k1 = adr2idx(fw,followBranch(fw,idx2adr(fw,k),0x01000001));
                    int k2;
                    for (k2=k1; k2<k1+20; k2++)
                    {
                        if (isLDR_PC(fw,k2) && (fwRd(fw,k2) == 0) && isLDMFD(fw,k2+1))
                        {
                            int k3 = adr2idx(fw, LDR2val(fw,k2) - fw->data_start + fw->data_init_start);
                            if (idx_valid(fw,k3))
                            {
                                while (fwval(fw,k3) != 0)
                                {
                                    add_func_name2(fw, fwval(fw,k3), fwval(fw,k3+1), "_FW");
                                    k3 += 2;
                                }
                            }
                        }
                    }
                }
            }
        }
    }
}

uint32_t findTaskAddress(firmware *fw, int k, int reg)
{
    int o;

    for (o=-1; o>-7; o--)
    {
        if (isLDR_PC(fw,k+o) && (fwRd(fw,k+o) == reg))
        {
            uint32_t adr = LDR2val(fw,k+o);
            int i;
            for (i=o+1; i<0; i++)
            {
                if (fwval(fw,k+i) == (0xE5900000 | (reg << 12) | (reg << 16)))    // LDR Rx,[Rx]
                {
                    adr = fwval(fw,adr2idx(fw,adr));
                }
            }
            return adr;
        }
        else if (isADR_PC(fw,k+o) && (fwRd(fw,k+o) == reg))
        {
            return(ADR2adr(fw,k+o));
        }
    }

    return 0;
}

int match_createtask(firmware *fw, int k, uint32_t fadr, uint32_t v2)
{
    if (isBorBL(fw,k))
    {
        uint32_t adr = followBranch2(fw,idx2adr(fw,k),0x01000001);
        if (adr == fadr)
        {
            fadr = findTaskAddress(fw, k, 3);
            if (fadr != 0)
            {
                uint32_t sadr = findTaskAddress(fw, k, 0);
                if (sadr != 0)
                {
                    char *s = adr2ptr(fw,sadr);
                    char *nm = malloc(strlen(s)+6);
                    sprintf(nm,"task_%s",s);
                    add_func_name(nm, fadr, 0);
                }
            }
        }
    }

    return 0;
}

void find_tasks(firmware *fw)
{
    int k = get_saved_sig(fw,"CreateTask");
    if (k >= 0)
    {
        search_fw(fw, match_createtask, func_names[k].val, 0, 7);
    }
    k = get_saved_sig(fw,"CreateTaskStrictly");
    if (k >= 0)
    {
        search_fw(fw, match_createtask, func_names[k].val, 0, 7);
    }
}

void find_builddate(firmware *fw)
{
    int j = get_saved_sig(fw,"GetBuildDate_FW");
    if (j >= 0)
    {
        int idx = adr2idx(fw, func_names[j].val);
        uint32_t adr = ADR2adr(fw, idx);
        idx = adr2idx(fw, adr);
        fw->fw_build_date = (char*)&fw->buf[idx];
    }
    else
        fw->fw_build_date = 0;

    j = get_saved_sig(fw,"GetBuildTime_FW");
    if (j >= 0)
    {
        int idx = adr2idx(fw, func_names[j].val);
        uint32_t adr = ADR2adr(fw, idx);
        idx = adr2idx(fw, adr);
        fw->fw_build_time = (char*)&fw->buf[idx];
    }
    else
        fw->fw_build_time = 0;
}

void write_levent_table_dump(firmware *fw, uint32_t tadr)
{
    char *str;
    uint32_t lid = 0;
    uint32_t val;
    if (!tadr) {
        return;
    }
    FILE *f=fopen("levent_table.txt","w");
    if(!f) {
        return;
    }
    fprintf(f,"address    ID     (unknown)  name\n");

    for(;;tadr += 12) {
        val = *(uint32_t*)adr2ptr(fw, tadr);
        if ((val == 0xffffffff) || (val == 0) || (*(uint32_t*)adr2ptr(fw, tadr+4) < lid)) {
            break;
        }
        lid = *(uint32_t*)adr2ptr(fw, tadr+4);
        str = (char*)adr2ptr(fw,val);
        if (str) {
            fprintf(f,"0x%08x 0x%04x 0x%08x %s\n",tadr,lid,*(uint32_t*)adr2ptr(fw, tadr+8),str);
        }
    }
    fclose(f);
}

//------------------------------------------------------------------------------------------------------------

// Write out firmware info
void output_firmware_vals(firmware *fw)
{
    bprintf("// Camera info:\n");

    if (fw->dryos_ver == 0)
    {
        bprintf("//   Can't find DRYOS version !!!\n\n");
    }
    else
    {
        if (fw->dryos_ver > 58) // ***** UPDATE for new DryOS version *****
            bprintf("//   DRYOS R%d (%s) *** New DRYOS Version - please update finsig_dryos.c ***\n",fw->dryos_ver,fw->dryos_ver_str);
        else
            bprintf("//   DRYOS R%d (%s)\n",fw->dryos_ver,fw->dryos_ver_str);
    }

    if (fw->firmware_ver_str == 0)
    {
        bprintf("//   Can't find firmware version !!!\n\n");
    }
    else
    {
        uint32_t j = idx2adr(fw,fw->fwver_idx);
        char *c = strrchr(fw->firmware_ver_str,' ') + 1; // points after the last space char
        uint32_t k = j + c - fw->firmware_ver_str;
        if ( (k>=j) && (k<j+32) )
        {
            bprintf("//   %s   // Found @ 0x%08x, \"%s\" @ 0x%08x\n",fw->firmware_ver_str,j,c,k);
        }
        else
        {
            // no space found in string (shouldn't happen)
            bprintf("//   %s   // Found @ 0x%08x, \"%s\" @ 0x%08x\n",fw->firmware_ver_str,j,fw->firmware_ver_str,j);
        }
    }

    if (fw->fw_build_date != 0)
    {
        bprintf("//   Firmware build timestamp: %s %s\n",fw->fw_build_date, (fw->fw_build_time==0)?"":fw->fw_build_time);
    }

    if (fw->fsize > (fw->size + 256))
    {
        bprintf("//   Possible corrupt firmware dump - file size too small for start address 0x%08x\n",fw->base);
        bprintf("//     file size = %.2fMB, should be %.2fMB\n", ((double)fw->size*4.0)/(1024.0*1024.0),((double)fw->fsize*4.0)/(1024.0*1024.0));
    }

    if (fw->cam != 0)
    {
        bprintf("//   %s\n",fw->cam);
    }
    else
    {
        bprintf("//   Could not find Camera name - possible corrupt firmware dump\n");
    }

    bprintf("\n// Values for makefile.inc\n");
    bprintf("//   PLATFORMOSVER = %d\n",fw->dryos_ver);

    if (fw->pid != 0)
    {
        bprintf("//   PLATFORMID = %d (0x%04x) // Found @ 0x%08x\n",fw->pid,fw->pid,fw->pid_adr);
    }
    else
    {
        bprintf("//   PLATFORMID = ?           // Not found @ 0x%08x\n",fw->pid_adr);
    }

    if (fw->maxram != 0)
        bprintf("//   MAXRAMADDR = 0x%08x\n",fw->maxram);

    if (fw->memisostart != 0)
    {
        osig *o = find_match(fw->sv->makevals, "MEMISOSTART", fw->memisostart);
        if (o && (o->val == fw->memisostart))
            bprintf("//   MEMISOSTART = 0x%08x\n",fw->memisostart);
        else
            bprintf("//   MEMISOSTART = 0x%08x  (*** DOES NOT MATCH MAKEFILE VALUE 0x%08x***)\n",fw->memisostart,(o)?o->val:0);
    }

    if (fw->ksys != 0)
    {
        bprintf("//   KEYSYS = %s              // Found @ 0x%08x\n",fw->ksys,idx2adr(fw,fw->ksys_idx));

        if (fw->dancing_bits_idx != 0)
        {
            if (fw->dancing_bits)
            {
                bprintf("//   NEED_ENCODED_DISKBOOT = %d   // Found @ 0x%08x",fw->dancing_bits,idx2adr(fw,fw->dancing_bits_idx));
                osig *o = find_sig(fw->sv->makevals,"NEED_ENCODED_DISKBOOT");
                if (o == 0)
                    bprintf(" (*** NOT IN MAKEFILE.INC ***)");
                else if (o->val != fw->dancing_bits)
                    bprintf(" (*** DOES NOT MATCH MAKEFILE.INC VALUE %d ***)",o->val);
                bprintf("\n");
            }
            else
            {
                bprintf("//   NEED_ENCODED_DISKBOOT = ? Not found, possible new 'dancing bits' entry needed. // Found @ 0x%08x\n",idx2adr(fw,fw->dancing_bits_idx));
            }
        }
    }

    if ((fw->dryos_ver >= 50) && (fw->base2 != 0))
    {
        bprintf("\n// Note, ROM copied to RAM :- from 0x%08x, to 0x%08x, len %d words.\n",fw->base_copied,fw->base2,fw->size2);
    }

    find_AdditionAgent_RAM(fw);
    
    bprintf("\n");
}

int compare_func_names(const func_entry **p1, const func_entry **p2)
{
    int rv = strcasecmp((*p1)->name, (*p2)->name);     // Case insensitive
    if (rv != 0)
        return rv;
    rv = strcmp((*p1)->name, (*p2)->name);          // Case sensitive (if equal with insensitive test)
    if (rv != 0)
        return rv;
    if ((*p1)->val < (*p2)->val)
        return -1;
    else if ((*p1)->val > (*p2)->val)
        return 1;
    return 0;
}

int compare_func_addresses(const func_entry **p1, const func_entry **p2)
{
    if ((*p1)->val < (*p2)->val)
        return -1;
    else if ((*p1)->val > (*p2)->val)
        return 1;
    return compare_func_names(p1,p2);
}

void write_funcs(firmware *fw, char *filename, func_entry *fns[], int (*compare)(const func_entry **p1, const func_entry **p2))
{
    int k;

    qsort(fns, next_func_entry, sizeof(func_entry*), (void*)compare);

    FILE *out_fp = fopen(filename, "w");
    for (k=0; k<next_func_entry; k++)
    {
        if (strncmp(fns[k]->name,"hook_",5) != 0)
        {
            if (fns[k]->val != 0)
            {
                if (fns[k]->flags & BAD_MATCH)
                {
                    osig* ostub2 = find_sig(fw->sv->stubs,fns[k]->name);
                    if (ostub2 && ostub2->val)
                        fprintf(out_fp, "0x%08x,%s,(stubs_entry_2.s)\n", ostub2->val, fns[k]->name);
                }
                else
                    fprintf(out_fp, "0x%08x,%s\n", fns[k]->val, fns[k]->name);
            }
#ifdef LIST_IMPORTANT_FUNCTIONS
            else if (fns[k]->flags & LIST_ALWAYS)
            {
                // helps development by listing important functions even when not found
                fprintf(out_fp, "0,%s,(NOT FOUND)\n", fns[k]->name);
            }
#endif
        }
    }
    fclose(out_fp);
}

int main(int argc, char **argv)
{
    firmware fw;
    int k;
    int ret = 0;
    const char *curr_name;

    clock_t t1 = clock();

    if ((argc < 4) || (argc > 5))
        usage("args");

    out_fp = fopen(argv[3],"w");
    if (out_fp == NULL) usage("failed to open outputfile");

    for (next_func_entry = 0; func_names[next_func_entry].name != 0; next_func_entry++);
    int max_find_func = next_func_entry;

    fw.sv = new_stub_values();
    load_stubs(fw.sv, "stubs_entry_2.S", 1);
    load_stubs_min(fw.sv);
    load_modemap(fw.sv);
    load_platform(fw.sv);
    load_makefile(fw.sv);

    bprintf("// !!! THIS FILE IS GENERATED. DO NOT EDIT. !!!\n");
    bprintf("#include \"stubs_asm.h\"\n\n");

    load_firmware(&fw,argv[1],argv[2],(argc==5)?argv[4]:0, OS_DRYOS);
    find_eventprocs(&fw);
    find_builddate(&fw);
    output_firmware_vals(&fw);

    out_hdr = 1;
    bprintf("// Stubs below should be checked. Stub not matched 100%%, or difference found to current 'stubs_entry_2.S'\n");
    bprintf("//    Name                                     Address      Rule  %%  Comp to stubs_entry_2.S\n");
    out_hdr = 0;
    bprintf("// Stubs below matched 100%%.\n");
    bprintf("//    Name                                     Address                Comp to stubs_entry_2.S\n");

    find_tasks(&fw);

    for (k = 0; k < max_find_func; k++)
    {
        count = 0;
        curr_name = func_names[k].name;

        if ((fw.dryos_ver >= find_min_ver(curr_name)) && (fw.dryos_ver <= find_max_ver(curr_name)))
        {
            find_matches(&fw, curr_name);
            print_results(&fw,curr_name,k);

            if (count == 0)
            {
                ret = 1;
            }
        }
    }

    fw.uncached_adr = 0;
    fw.uncached_adr_idx = 0;
    search_saved_sig(&fw, "FreeUncacheableMemory", match_CAM_UNCACHED_BIT, 0, 0, 8);
    
    find_modemap(&fw);
    find_stubs_min(&fw);
    find_lib_vals(&fw);
    find_key_vals(&fw);
    find_platform_vals(&fw);
    find_other_vals(&fw);

    write_output();

    fclose(out_fp);

    func_entry *fns[MAX_FUNC_ENTRY];
    for (k=0; k<next_func_entry; k++)
        fns[k] = &func_names[k];

    write_funcs(&fw, "funcs_by_name.csv", fns, compare_func_names);
    write_funcs(&fw, "funcs_by_address.csv", fns, compare_func_addresses);

    clock_t t2 = clock();

    printf("Time to generate stubs %.2f seconds\n",(double)(t2-t1)/(double)CLOCKS_PER_SEC);

    return ret;
}

//------------------------------------------------------------------------------------------------------------
