#include <stdlib.h>
#include <stdio.h>
#include <stdint.h>
#include <string.h>
#include <time.h>
#include <stdarg.h>
#include <ctype.h>

#include "stubs_load.h"
#include "firmware_load.h"

//------------------------------------------------------------------------------------------------------------
// #define DEBUG_PRINT_ALL_FUNC_NAMES 1 // enable debug output on stderr for development
// #define LIST_IMPORTANT_FUNCTIONS   1 // always list functions with 'LIST_ALWAYS' flag, even when not found
// #define PRINT_LEVENT_TABLE 1         // print all levents to file
//------------------------------------------------------------------------------------------------------------

// Buffer output into header and body sections

FILE    *out_fp;
char    out_buf[32*1024] = "";
int     out_len = 0;
char    hdr_buf[32*1024] = "";
int     hdr_len = 0;
int     out_hdr = 1;

void bprintf(char *fmt, ...)
{
    va_list argp;
    va_start(argp, fmt);

    if (out_hdr)
        hdr_len += vsprintf(hdr_buf+hdr_len,fmt,argp);
    else
        out_len += vsprintf(out_buf+out_len,fmt,argp);

    va_end(argp);
}

void add_blankline()
{
    if (strcmp(hdr_buf+hdr_len-2,"\n\n") != 0)
    {
        hdr_buf[hdr_len++] = '\n';
        hdr_buf[hdr_len] = 0;
    }
}

void write_output()
{
    add_blankline();
    if (out_fp)
    {
        fprintf(out_fp,"%s",hdr_buf);
        fprintf(out_fp,"%s",out_buf);
    }
}

//------------------------------------------------------------------------------------------------------------

void usage(char *err)
{
    bprintf("finsig <primary> <base> <outputfilename> [alt base] - Error = %s\n",err);
    write_output();
    fprintf(stderr,"finsig <primary> <base> <outputfilename> [alt base] - Error = %s\n",err);
    exit(1);
}

void error(char *fmt, int n)
{
    bprintf(fmt, n);
    write_output();
    exit(1);
}

//------------------------------------------------------------------------------------------------------------

// Signature match handling

// Structure to store match info
typedef struct {
    uint32_t ptr;
    uint32_t fail;
    uint32_t success;
    int sig;
} Match;

// qsort compare function the Match struture
int match_compare(const Match *p1, const Match *p2)
{
    /* NOTE: If a function has *more* matches, it will be prefered, even if it has a lower percent matches */
    if (p1->success > p2->success)
    {
        if ((p2->fail == 0) && (p1->fail > 0))
        {
            return 1;
        }
        else
        {
            return -1;
        }
    }
    else if (p1->success < p2->success)
    {
        if ((p1->fail == 0) && (p2->fail > 0))
        {
            return -1;
        }
        else
        {
            return 1;
        }
    }
    else
    {
        if (p1->fail < p2->fail)
        {
            return -1;
        }
        else if (p1->fail > p2->fail)
        {
            return 1;
        }
    }

    if (p1->sig < p2->sig)
    {
        return -1;
    }
    else if (p1->sig > p2->sig)
    {
        return 1;
    }

    /* scores are equal. prefer lower address */

    if (p1->ptr < p2->ptr)
    {
        return -1;
    }
    else if (p1->ptr > p2->ptr)
    {
        return 1;
    }

    return 0;
}

#define MAX_MATCHES (8192)

Match matches[MAX_MATCHES];
int count;

// Add a new Match
void addMatch(uint32_t fadr, int s, int f, int sig)
{
    matches[count].ptr = fadr;
    matches[count].success = s;
    matches[count].fail = f;
    matches[count].sig = sig;
    count++;
}

// Add a Match, adjust address for stuff in the block copied to RAM
void fwAddMatch(firmware *fw, uint32_t fadr, int s, int f, int sig)
{
    if ((fadr >= fw->base_copied) && (fadr < (fw->base_copied + fw->size2*4)))
    {
        addMatch(fadr - fw->base_copied + fw->base2,s,f,sig);
    }
    else
    {
        addMatch(fadr,s,f,sig);
    }
}

//------------------------------------------------------------------------------------------------------------

// Signature structures generated by gensig2.exe

typedef struct {
    uint32_t offs;
    uint32_t value;
    uint32_t mask;
} FuncSig;

typedef struct {
    const char *name;
    FuncSig *sig;
    int ver;
} FuncsList;

//------------------------------------------------------------------------------------------------------------

// Master list of functions / addresses to find

#define DONT_EXPORT     1
#define OPTIONAL        2
#define UNUSED          4
#define BAD_MATCH       8
#define EV_MATCH        16
#define LIST_ALWAYS     32

typedef struct {
    char        *name;
    int         flags;
    uint32_t    val;
} func_entry;

int next_func_entry = 0;

#define MAX_FUNC_ENTRY  5000

func_entry  func_names[MAX_FUNC_ENTRY] =
{
    // Do these first as they are needed to find others
    { "ExportToEventProcedure_FW", UNUSED|DONT_EXPORT },
    { "RegisterEventProcedure_FW", UNUSED|DONT_EXPORT },
    //{ "CreateJumptable", UNUSED|DONT_EXPORT },
    { "_uartr_req", UNUSED|DONT_EXPORT },
    { "StartRecModeMenu", UNUSED|DONT_EXPORT },
    { "LogCameraEvent", UNUSED|DONT_EXPORT },
    { "IsControlEventActive", UNUSED|DONT_EXPORT }, // helper, vx
    { "GetLogicalEventName", UNUSED|DONT_EXPORT }, // helper, vx, name made up

    { "AllocateMemory", UNUSED },
    { "AllocateUncacheableMemory" },
    { "Close" },
    { "CreateBinarySemaphore", UNUSED|DONT_EXPORT },
    { "CreateCountingSemaphore", UNUSED|DONT_EXPORT },
    { "CreateTask" },
    { "DebugAssert", OPTIONAL },
    { "DeleteDirectory_Fut", UNUSED|DONT_EXPORT },
    { "DeleteFile_Fut" },
    { "DeleteSemaphore", UNUSED|DONT_EXPORT },
    { "DoAELock" },
    { "DoAFLock" },
    { "EnterToCompensationEVF" },
    { "ExecuteEventProcedure" },
    { "ExitFromCompensationEVF" },
    { "ExitTask" },
    { "ExpCtrlTool_StartContiAE", OPTIONAL },
    { "ExpCtrlTool_StopContiAE", UNUSED|DONT_EXPORT },
    { "Fclose_Fut" },
    { "Feof_Fut" },
    { "Fflush_Fut" },
    { "Fgets_Fut" },
    { "Fopen_Fut" },
    { "Fread_Fut" },
    { "FreeMemory", UNUSED },
    { "FreeUncacheableMemory" },
    { "Fseek_Fut" },
    { "Fwrite_Fut" },
    { "GetBatteryTemperature" },
    { "GetCCDTemperature" },
    { "GetCurrentAvValue" },
    { "GetDrive_ClusterSize" },
    { "GetDrive_FreeClusters" },
    { "GetDrive_TotalClusters" },
    { "GetFocusLensSubjectDistance" },
    { "GetFocusLensSubjectDistanceFromLens" },
    { "GetImageFolder", OPTIONAL },
    { "GetKbdState", UNUSED|DONT_EXPORT },
    { "GetMemInfo", UNUSED|DONT_EXPORT },
    { "GetOpticalTemperature" },
    { "GetParameterData" },
    { "GetPropertyCase" },
    { "GetSystemTime" },
    { "GetVRAMHPixelsSize", UNUSED|DONT_EXPORT },
    { "GetVRAMVPixelsSize", UNUSED|DONT_EXPORT },
    { "GetZoomLensCurrentPoint" },
    { "GetZoomLensCurrentPosition" },
    { "GiveSemaphore", OPTIONAL },
    { "IsStrobeChargeCompleted" },
    { "LEDDrive", OPTIONAL },
    { "LocalTime" },
    { "LockMainPower" },
    { "Lseek", UNUSED|DONT_EXPORT },
    { "MakeDirectory_Fut" },
    { "MakeSDCardBootable", OPTIONAL },
    { "MoveFocusLensToDistance" },
    { "MoveIrisWithAv", OPTIONAL },
    { "MoveZoomLensWithPoint" },
    { "NewTaskShell", UNUSED|DONT_EXPORT },
    { "Open" },
    { "PB2Rec", UNUSED|DONT_EXPORT },
    { "PT_MoveDigitalZoomToWide", OPTIONAL },
    { "PT_MoveOpticalZoomAt", OPTIONAL },
    { "PT_PlaySound" },
    { "PostLogicalEventForNotPowerType" },
    { "PostLogicalEventToUI" },
    { "PutInNdFilter", OPTIONAL },
    { "PutOutNdFilter", OPTIONAL },
    { "Read" },
//    { "ReadFastDir", UNUSED|DONT_EXPORT },
    { "Rec2PB", UNUSED|DONT_EXPORT },
    { "RefreshPhysicalScreen" },
    { "Remove", OPTIONAL|UNUSED },
    { "RenameFile_Fut" },
    { "Restart" },
    { "ScreenLock", OPTIONAL },
    { "ScreenUnlock", OPTIONAL },
    { "SetAE_ShutterSpeed" },
    { "SetAutoShutdownTime" },
    { "SetCurrentCaptureModeType" },
    { "SetFileAttributes" },
    { "SetFileTimeStamp", UNUSED|DONT_EXPORT },
    { "SetLogicalEventActive" },
    { "SetParameterData" },
    { "SetPropertyCase" },
    { "SetScriptMode" },
    { "SleepTask" },
    { "TakeSemaphore" },
    { "TurnOffBackLight" },
    { "TurnOnBackLight" },
    { "TurnOnDisplay" },
    { "TurnOffDisplay" },
    { "UIFS_WriteFirmInfoToFile", UNUSED|DONT_EXPORT },
    { "UnlockAE" },
    { "UnlockAF" },
    { "UnlockMainPower" },
    { "UnsetZoomForMovie", OPTIONAL },
    { "UpdateMBROnFlash" },
    { "VbattGet" },
    { "Write" },
    { "WriteSDCard" },

    { "_log" },
    { "_log10" },
    { "_pow" },
    { "_sqrt" },
    { "add_ptp_handler" },
    { "apex2us" },
    { "close" },
    { "err_init_task", OPTIONAL },
    { "exmem_alloc" },
    { "exmem_free", OPTIONAL },
    { "free" },

    { "kbd_p1_f" },
    { "kbd_p1_f_cont" },
    { "kbd_p2_f" },
    { "kbd_read_keys", UNUSED|DONT_EXPORT },
    { "kbd_read_keys_r2" },

    { "kbd_pwr_off" },
    { "kbd_pwr_on" },
    { "lseek" },
    { "malloc" },
    { "memcmp" },
    { "memcpy" },
    { "memset" },
    { "mkdir" },
    { "mktime_ext", UNUSED|DONT_EXPORT },
    { "open" },
//    { "OpenFastDir", UNUSED|DONT_EXPORT },
    { "closedir" },
    { "qsort" },
    { "rand" },
    { "read", UNUSED|DONT_EXPORT },
    { "realloc", OPTIONAL },
    { "reboot_fw_update", UNUSED|DONT_EXPORT },
    { "set_control_event", UNUSED|DONT_EXPORT },
    { "srand" },
    { "stat" },
    { "strcat" },
    { "strchr" },
    { "strcmp" },
    { "strcpy" },
    { "strftime" },
    { "strlen" },
    { "strncmp" },
    { "strncpy" },
    { "strrchr" },
    { "strtol" },
//    { "strtolx", UNUSED|DONT_EXPORT },

    { "isdigit" },
    { "isspace" },
    { "isalpha" },
    { "isupper" },
    { "islower" },
    { "ispunct" },
    { "isxdigit" },
    { "toupper" },
    { "tolower" },
    { "taskLock" },
    { "taskUnlock" },
    { "taskName" },
    { "taskIdListGet" },
    { "errnoOfTaskGet" },
    { "readdir" },
    { "strtoul" },
    { "strpbrk" },
    { "utime" },
    { "localtime" },
    { "mktime" },
    { "opendir" },
    { "memchr" },
    { "memPartInfoGet" },
    { "iosDrvInstall" },
    { "iosDevAdd" },
    { "iosDevDelete" },
    { "GetCurrentTargetDistance" },
    { "taskCreateHookAdd" },
    { "taskDeleteHookAdd" },
    { "Mount_FileSystem" },
    { "Unmount_FileSystem" },
    { "MakeAFScan" },
    { "taskSuspend" },
    { "taskResume" },

    { "task_CaptSeq", UNUSED|DONT_EXPORT },
    { "task_ExpDrv", UNUSED|DONT_EXPORT },
    { "task_FileWrite", UNUSED|DONT_EXPORT },
    { "task_InitFileModules", UNUSED|DONT_EXPORT },
    { "task_MovieRecord", UNUSED|DONT_EXPORT },
    { "task_PhySw", UNUSED|DONT_EXPORT },
    { "task_SwitchCheck", UNUSED|DONT_EXPORT }, // old name for task_PhySw
    { "task_RotaryEncoder", UNUSED|DONT_EXPORT },
//    { "task_TouchPanel", OPTIONAL },

    //{ "hook_CreateTask" },

    { "time" },
    { "vsprintf" },
    { "write", UNUSED|DONT_EXPORT },

    { "EngDrvIn", OPTIONAL|UNUSED },
    { "EngDrvOut", OPTIONAL|UNUSED },
    { "EngDrvRead" },
    { "EngDrvBits", OPTIONAL|UNUSED },

    { "PTM_GetCurrentItem" },
    { "PTM_SetCurrentItem", UNUSED },
    { "PTM_NextItem", OPTIONAL|UNUSED },
    { "PTM_PrevItem", OPTIONAL|UNUSED },
    { "PTM_SetPropertyEnable", OPTIONAL|UNUSED },

    // OS functions, mostly to aid firmware analysis. Order is important!
    { "PT_GetSystemTime", OPTIONAL|UNUSED }, // only for locating timer functions
    { "_GetSystemTime", OPTIONAL|UNUSED }, // only for locating timer functions
    { "SetTimerAfter", OPTIONAL|UNUSED },
    { "SetTimerWhen", OPTIONAL|UNUSED },
    { "CancelTimer", OPTIONAL|UNUSED },
    { "CancelHPTimer" },
    { "SetHPTimerAfterTimeout", OPTIONAL|UNUSED },
    { "SetHPTimerAfterNow" },
    { "CreateTaskStrictly", OPTIONAL|UNUSED },
    { "CreateMessageQueue", OPTIONAL|UNUSED },
    { "CreateRecursiveLock", OPTIONAL|UNUSED },
    { "GetSemaphoreValue", OPTIONAL|UNUSED },
    { "TryTakeSemaphore", OPTIONAL|UNUSED },
    { "CreateMessageQueueStrictly", OPTIONAL|UNUSED },
    { "CreateEventFlagStrictly", OPTIONAL|UNUSED },
    { "CreateBinarySemaphoreStrictly", OPTIONAL|UNUSED },
    { "CreateCountingSemaphoreStrictly", OPTIONAL|UNUSED },
    { "CreateRecursiveLockStrictly", OPTIONAL|UNUSED },
    { "DeleteMessageQueue", OPTIONAL|UNUSED },
    { "PostMessageQueue", OPTIONAL|UNUSED },
    { "ReceiveMessageQueue", OPTIONAL|UNUSED },
    { "TryReceiveMessageQueue", OPTIONAL|UNUSED },
    { "TryPostMessageQueue", OPTIONAL|UNUSED },
    { "GetNumberOfPostedMessages", OPTIONAL|UNUSED },
    { "DeleteRecursiveLock", OPTIONAL|UNUSED },
    { "AcquireRecursiveLock", OPTIONAL|UNUSED },
    { "ReleaseRecursiveLock", OPTIONAL|UNUSED },
    { "WaitForAnyEventFlag", OPTIONAL|UNUSED },
    { "WaitForAllEventFlag", OPTIONAL|UNUSED },
    { "ClearEventFlag", OPTIONAL|UNUSED },
    { "SetEventFlag", OPTIONAL|UNUSED },
    { "GetEventFlagValue", OPTIONAL|UNUSED },
    { "CreateEventFlag", OPTIONAL|UNUSED },
    { "DeleteEventFlag", OPTIONAL|UNUSED },
    { "CheckAnyEventFlag", OPTIONAL|UNUSED },
    { "CheckAllEventFlag", OPTIONAL|UNUSED },
    { "RegisterInterruptHandler", OPTIONAL|UNUSED },
    { "UnregisterInterruptHandler", OPTIONAL|UNUSED },
    { "GetSRAndDisableInterrupt", OPTIONAL|UNUSED }, // disables IRQ, returns a value
    { "SetSR", OPTIONAL|UNUSED }, // enables IRQ, puts back value returned by GetSR
    { "EnableInterrupt", OPTIONAL|UNUSED }, // enables IRQ
    { "GetCurrentMachineTime", OPTIONAL|UNUSED }, // reads usec counter, name from ixus30
    { "HwOcReadICAPCounter", OPTIONAL|UNUSED }, // reads usec counter, name from ixus30

    // Other stuff needed for finding misc variables - don't export to stubs_entry.S
    { "GetSDProtect", UNUSED|DONT_EXPORT },
    { "DispCon_ShowBitmapColorBar", UNUSED },
    { "ResetZoomLens", OPTIONAL|UNUSED },
    { "ResetFocusLens", OPTIONAL|UNUSED },
    { "NR_GetDarkSubType", OPTIONAL|UNUSED },
    { "NR_SetDarkSubType", OPTIONAL|UNUSED },
    { "SavePaletteData", OPTIONAL|UNUSED },
    { "GUISrv_StartGUISystem", OPTIONAL|UNUSED },
    { "get_resource_pointer", OPTIONAL|UNUSED }, // name made up, gets a pointer to a certain resource (font, dialog, icon)

    { "wrapped_malloc", UNUSED|DONT_EXPORT},    // helper to find malloc, finds the other allocator on ixus30/40
    { "malloc_alt", UNUSED|DONT_EXPORT },       // should be the same as malloc when found
    { "WaitForEventFlag", UNUSED|DONT_EXPORT }, // helper to find other eventflag functions

    { "filesem_init", OPTIONAL|UNUSED }, // file semaphore init function, needed for verification

    { "MFOn", OPTIONAL },
    { "MFOff", OPTIONAL },

    { "GetAdChValue", OPTIONAL },

    { 0, 0, 0 }
};

// Return the array index of a named function in the array above
uint32_t find_saved_sig(const char *name)
{
    int i;
    for (i=0; func_names[i].name != 0; i++)
    {
        if (strcmp(name,func_names[i].name) == 0)
        {
            return i;
        }
    }
    return -1;
}

// Save the address value found for a function in the above array
void save_sig(const char *name, uint32_t val)
{
    int i = find_saved_sig(name);
    if (i >= 0)
    {
        func_names[i].val = val;
    }
}

// Get the saved address value for a named function
// If the address value is 0 then assume the function search has not occurred yet, so go search for it.
uint32_t get_saved_sig(firmware *fw, const char *name)
{
    int i = find_saved_sig(name);
    if (i >= 0)
    {
        if (func_names[i].val == 0)
        {
            // See if the function is in the 'func_list' array below
            int find_func(const char* name);
            int k1 = find_func(name);
            if (k1 >= 0)
            {
                // if found do full search
                void find_matches(firmware*,const char*);
                find_matches(fw, name);
                count = 0;
            }
            else
            {
                // not found, only do string matching search
                void find_str_sig_matches(firmware*,const char*);
                find_str_sig_matches(fw, name);
                count = 0;
            }
        }
        if (func_names[i].val == 0)
        {
            // If not found return invalid index
            i = -1;
        }
    }
    return i;
}

// Search for something relative to the location stored for a previously matched function
// Matching is done via the 'func' function, searching continues until 'func' returns non-zero
// Starts searching at 'ofst' from the saved address for a max of 'len' instructions
// Returns the 'func' value or 0 if no match found
int search_saved_sig(firmware *fw, char *sig, int (*func)(firmware*, int, int), int v, int ofst, int len)
{
    int k = get_saved_sig(fw, sig);
    if (k >= 0)
    {
        int idx = adr2idx(fw, func_names[k].val);
        for (k=idx+ofst; k<idx+ofst+len; k++)
        {
            int rv = func(fw, k, v);
            if (rv)
                return rv;
        }
    }
    return 0;
}

//------------------------------------------------------------------------------------------------------------

// New string / signature matching structure

// Structure for matching

typedef struct {
    int     type;               // 1 = func*, string, 2 = string, ... string*, func*, 3 = ADR Rx, func, ADR Ry, string, BL, ... string, etc
    char    *name;              // function name
    char    *ev_name;           // event / other name to match in the firmware
    int     offset;             // offset for following branches, or other tests
    // VxWorks version specific offsets?
    int     vxworks_offset;
} string_sig;

// Load old signature matching data generated by gensig_vxworks
#include "signatures_vxworks.h"

//------------------------------------------------------------------------------------------------------------

// Data for matching 'apex2us' function
uint32_t apex2us_test[] = { 0x3D09000, 0x3BBA304, 0x3A728D2, 0x3931EF4, 0x37F8303, 0x36C52A2, 0x3598B85, 0x3472B6A, 0 };

// Special case for apex2us
int match_apex2us(firmware *fw, int k, uint32_t v1, uint32_t v2)
{
    if (isLDR_PC(fw,k) && (LDR2val(fw,k) == v1) && ((fwRd(fw,k) == 1) || (fwRd(fw,k) == 2)))
    {
        k = find_inst_rev(fw, isSTMFD_LR, k, 200);
        if (k != 0)
        {
            if (fwval(fw,k-2) == 0xE3700D09)    // CMN R0, #0x240
                k -= 2;
            uint32_t fadr = idx2adr(fw,k);
            fwAddMatch(fw,fadr,32,0,121);
            return 1;
        }
    }
    return 0;
}
int find_apex2us(firmware *fw, string_sig *sig, int j)
{
    int i;
    for (i=0; apex2us_test[i] != 0; i++)
        if (fwval(fw,j+i) != apex2us_test[i])
            return 0;

    return search_fw(fw, match_apex2us, idx2adr(fw,j), 0, 1);
}

// Special case for mkdir
int find_mkdir(firmware *fw, string_sig *sig, int k)
{
    if (fwval(fw,k) == 0x12CEA600)
    {
        k = find_inst_rev(fw, isSTMFD_LR, k-20, 200);
        if (k != 0)
        {
            if ((((fwval(fw,k+12) & 0xFFF0FFFF) == 0xE350002F) && ((fwval(fw,k+15) & 0xFFF0FFFF) == 0xE3500021) && ((fwval(fw,k+19) & 0xFFF0FFFF) == 0xE3500020)) ||
                (((fwval(fw,k+11) & 0xFFF0FFFF) == 0xE350002F) && ((fwval(fw,k+14) & 0xFFF0FFFF) == 0xE3500021) && ((fwval(fw,k+18) & 0xFFF0FFFF) == 0xE3500020)))
            {
                uint32_t fadr = 0;
                if (isBL(fw,k+47))
                {
                    fadr = followBranch(fw, idx2adr(fw,k+47), 0x01000001);
                }
                else if (isBL(fw,k+48))
                {
                    fadr = followBranch(fw, idx2adr(fw,k+48), 0x01000001);
                }
                if (fadr != 0)
                {
                    fwAddMatch(fw,fadr,32,0,121);
                    return 1;
                }
            }
        }
    }
    return 0;
}

// Special case for _pow
int find_pow(firmware *fw, string_sig *sig, int j)
{
    if (!idx_valid(fw,j) || !idx_valid(fw,j+3)) return 0;
    // Find values passed to _pow
    if ((fwval(fw,j) == 0x00000000) && (fwval(fw,j+1) == 0x40000000) && (fwval(fw,j+2) == 0x00000000) && (fwval(fw,j+3) == 0x408F4000))
    {
        uint32_t adr1 = idx2adr(fw,j);      // address of 1st value
        uint32_t adr2 = idx2adr(fw,j+2);    // address of 2nd value
        int j1;

        for (j1 = j-5; j1>0; j1--)
        {
            if (isADR_PC_cond(fw,j1) &&                 // ADR ?
                (fwval(fw,j1+1) == 0xE8900003) &&       // LDMIA R0,{R0,R1}
                isBL(fw,j1+2) &&                        // BL
                isADR_PC_cond(fw,j1+4))                 // ADR ?
            {
                if ((ADR2adr(fw,j1) == adr1) && (ADR2adr(fw,j1+4) == adr2))
                {
                    uint32_t fadr = followBranch(fw,idx2adr(fw,j1+2),0x01000001);
                    fwAddMatch(fw,fadr,32,0,121);
                    return 1;
                }
            }
            else
            if (isADR_PC_cond(fw,j1) &&                 // ADR ?
                (fwval(fw,j1+1) == 0xE8900003) &&       // LDMIA R0,{R0,R1}
                isBL(fw,j1+2) &&                        // BL
                isADR_PC_cond(fw,j1+3))                 // ADR ?
            {
                if ((ADR2adr(fw,j1) == adr1) && (ADR2adr(fw,j1+3) == adr2))
                {
                    uint32_t fadr = followBranch(fw,idx2adr(fw,j1+2),0x01000001);
                    fwAddMatch(fw,fadr,32,0,121);
                    return 1;
                }
            }
            else
            if (isADR_PC_cond(fw,j1) &&                 // ADR ?
                (fwval(fw,j1+2) == 0xE8900003) &&       // LDMIA R0,{R0,R1}
                isBL(fw,j1+3) &&                        // BL
                isADR_PC_cond(fw,j1+4))                 // ADR ?
            {
                if ((ADR2adr(fw,j1) == adr1) && (ADR2adr(fw,j1+4) == adr2))
                {
                    uint32_t fadr = followBranch(fw,idx2adr(fw,j1+3),0x01000001);
                    fwAddMatch(fw,fadr,32,0,121);
                    return 1;
                }
            }
        }
    }

    return 0;
}

// Special case for _log & _log10
int find_log(firmware *fw, string_sig *sig, int j)
{
    // Find values passed to _log
    if (isBL(fw,j) && isLDR_PC(fw,j+1) && (LDR2val(fw,j+1) == 0x3FDBCB7B) && isLDR_PC(fw,j+2) && (LDR2val(fw,j+2) == 0x1526E50E))
    {
        uint32_t fadr = followBranch(fw,idx2adr(fw,j),0x01000001);
        fwAddMatch(fw,fadr,32,0,121);
        return 1;
    }

    return 0;
}
int find_log10(firmware *fw, string_sig *sig, int j)
{
    // Find values passed to _log
    if (isBL(fw,j) && isLDR_PC(fw,j+1) && (LDR2val(fw,j+1) == 0x3FDBCB7B) && isLDR_PC(fw,j+2) && (LDR2val(fw,j+2) == 0x1526E50E))
    {
        int k = find_inst_rev(fw, isSTMFD_LR, j-1, 100);
        uint32_t fadr = idx2adr(fw,k);
        fwAddMatch(fw,fadr,32,0,121);
        return 1;
    }

    return 0;
}

// Special case for 'closedir' (DryOS)
int find_closedir(firmware *fw)
{
    int j = get_saved_sig(fw,"OpenFastDir");
    if (j >= 0)
    {
        int k = find_inst(fw, isSTMFD_LR, adr2idx(fw,func_names[j].val)+1, 100);
        if (isB(fw,k-1) && isBL(fw,k-2))
        {
            uint32_t fadr = followBranch(fw, idx2adr(fw, k-2), 0x01000001);
            fwAddMatch(fw,fadr,32,0,121);
            return 1;
        }
    }

    return 0;
}

// Special case for 'add_ptp_handler'
int find_add_ptp_handler(firmware *fw, string_sig *sig, int k)
{
    uint32_t vals[] = { 0x9801, 0x9802, 0x9803, 0x9804, 0x9805, 0 };
    uint32_t fadr = 0;

    int i = 0;
    while ((vals[i] != 0) && isLDR_PC(fw,k) && (fwRd(fw,k) == 0) && (LDR2val(fw,k) == vals[i]))
    {
        k = find_inst(fw, isBL, k+1, 5);
        if (k == 0) return 0;
        if (fadr == 0)
            fadr = followBranch(fw, idx2adr(fw,k), 0x01000001);
        k = find_inst(fw, isLDR_PC, k+1, 5);
        if (k == 0) return 0;
        i++;
    }

    if (fadr != 0)
    {
        fwAddMatch(fw,fadr,32,0,121);
        return 1;
    }

    return 0;
}

// Special case for 'PT_PlaySound'
int find_PT_PlaySound(firmware *fw)
{
    int j, k;
    int k1 = get_saved_sig(fw,"LogCameraEvent");

    if (k1 >= 0)
    {
        j = find_str_ref(fw,"BufAccBeep");
        if (j >= 0)
        {
            k = find_inst(fw, isBL, j+1, 4);
            if (k >= 0)
            {
                uint32_t fadr = followBranch(fw, idx2adr(fw,k), 0x01000001);
                if (func_names[k1].val == fadr)
                {
                    k = find_inst(fw, isB, k+1, 10);
                    fadr = followBranch(fw, idx2adr(fw, k), 1);
                    fwAddMatch(fw,fadr,32,0,122);
                    return 1;
                }
            }
        }
    }

    return 0;
}

// Special case for 'ExportToEventProcedure'
int find_ExportToEventProcedure(firmware *fw)
{
    int k = find_str_ref(fw,"ExportToEventProcedure");

    if (k >= 0)
    {
        if (isLDR_PC(fw,k+1) && (isBorBL(fw,k+2) || isBorBL(fw,k+3)) && (fwRd(fw,k+1) == 1))
        {
            uint32_t fadr = LDR2val(fw,k+1);
            fwAddMatch(fw,fadr,32,0,122);
            return 1;
        }
    }

    return 0;
}

// Special case for 'RegisterEventProcedure'
int find_RegisterEventProcedure(firmware *fw)
{
    int k = find_str_ref(fw,"CreateProxyOfEventProcedure");

    if (k >= 0)
    {
        if (isLDR_PC(fw,k+1) && (isBorBL(fw,k+2) || isBorBL(fw,k+3)) && (fwRd(fw,k) == 0))
        {
            uint32_t fadr = 0;
            if (isBorBL(fw,k+2))
                fadr = followBranch(fw,idx2adr(fw,k+2),0x01000001);
            else
                fadr = followBranch(fw,idx2adr(fw,k+3),0x01000001);
            fwAddMatch(fw,fadr,32,0,122);
            return 1;
        }
    }

    return 0;
}

// Special case for set_control_event (Vx)
int find_set_control_event(firmware *fw)
{
    int j1, j2, k, found, n;
    int k1 = get_saved_sig(fw,"IsControlEventActive");
    int k2 = get_saved_sig(fw,"GetLogicalEventName");

    if ((k1 >= 0) && (k2 >= 0))
    {
        j1 = adr2idx(fw, func_names[k1].val);
        j2 = adr2idx(fw, func_names[k2].val);
        found = 0;
        // search from IsControlEventActive backwards
        for (k=j1-1; k>j1-256; k--)
        {
            if (isBL(fw,k))
            {
                if (j2 == idxFollowBranch(fw,k,0x01000001))
                {
                    // reference to GetLogicalEventName found
                    found = 1;
                    break;
                }
            }
        }
        if (found)
        {
            found = 0;
            // locate function start
            k = find_inst_rev(fw,isSTMFD_LR,k,60);
            // locate function end
            j2 = find_inst(fw,isLDMFD_PC,k+1,128);
            if ( (k<0) || (j2<0) )
                return 0;
            int prevbldest = 0;
            n = 0;
            k++;
            // look for not more than 7 BLs
            while (n<7)
            {
                k = find_inst(fw,isBL,k,32);
                // too late hit or no hit
                if ( (k>j2) || (k<0) )
                    return 0;
                j1 = idxFollowBranch(fw,k,0x01000001);
                if (prevbldest==j1)
                {
                    // two consecutive calls to set_control_event
                    found = 1;
                    break;
                }
                prevbldest = j1;
                n++;
                k++;
            }
            if (found)
            {
                fwAddMatch(fw,idx2adr(fw,j1),32,0,122);
                return 1;
            }
        }
    }
    return 0;
}

// see also find_FileAccessSem()
int find_filesem_init(firmware *fw)
{
    int s1 = find_str(fw, "FileSem.c");
    if (s1 < 0)
        return 0;
    s1 = find_inst(fw, isLDR_PC, s1+2, 16);
    if (s1 < 0)
        return 0;
    s1 = find_inst_rev(fw, isSTMFD_LR, s1-1, 16);
    if (s1 < 0)
        return 0;
    fwAddMatch(fw,idx2adr(fw,s1),32,0,122);
    return 1;
}

int find_getcurrentmachinetime(firmware *fw)
{
    int f1 = get_saved_sig(fw,"SetHPTimerAfterNow");
    if (f1 < 0)
        return 0;
    f1 = adr2idx(fw, func_names[f1].val);
    f1 = find_inst(fw, isBL, f1, 16);
    if (f1>0)
    {
        f1 = idxFollowBranch(fw,f1,0x01000001);
        fwAddMatch(fw,idx2adr(fw,f1),32,0,122);
        return 1;
    }
    return 0;
}

//------------------------------------------------------------------------------------------------------------

// Data for matching the '_log' function
uint32_t log_test[] = {
    0x1526E50E, 0x3FDBCB7B, 0
};

// Data for matching 'Fut' functions
uint32_t DeleteDirectory_Fut_test[] = { 0x09400017 };
uint32_t MakeDirectory_Fut_test[]   = { 0x09400015 };
uint32_t RenameFile_Fut_test[]      = { 0x09400013 };

//------------------------------------------------------------------------------------------------------------

// Signature matching data
string_sig string_sigs[] =
{
    // Same as previously found eventproc - do these first
    {20, "AllocateMemory", "AllocateMemory_FW", 1 },
    {20, "Close", "Close_FW", 1 },
    {20, "CreateCountingSemaphore", "CreateCountingSemaphore_FW", 1 },
    {20, "CreateTask", "CreateTask_FW", 1 },
    {20, "DeleteSemaphore", "DeleteSemaphore_FW", 1 },
    {20, "DispCon_ShowBitmapColorBar", "DispCon_ShowBitmapColorBar_FW", 1 },
    {20, "ExitTask", "ExitTask_FW", 1 },
    {20, "Fclose_Fut", "Fclose_Fut_FW", 1 },
    {20, "Fopen_Fut", "Fopen_Fut_FW", 1 },
    {20, "Fread_Fut", "Fread_Fut_FW", 1 },
    {20, "FreeMemory", "FreeMemory_FW", 1 },
    {20, "Fseek_Fut", "Fseek_Fut_FW", 1 },
    {20, "Fwrite_Fut", "Fwrite_Fut_FW", 1 },
    {20, "GetSDProtect", "GetSDProtect_FW", 1 },
    //{20, "GetSystemTime", "GetSystemTime_FW", 1 },
    {20, "GetCurrentAvValue", "GetCurrentAvValue_FW", 1 },
    {20, "GetOpticalTemperature", "GetOpticalTemperature_FW", 1 },
    {20, "GetVRAMHPixelsSize", "GetVRAMHPixelsSize_FW", 1 },
    {20, "GetVRAMVPixelsSize", "GetVRAMVPixelsSize_FW", 1 },
    {20, "GetZoomLensCurrentPoint", "GetZoomLensCurrentPoint_FW", 1 },
    {20, "GiveSemaphore", "GiveSemaphore_FW", 1 },
    {20, "GUISrv_StartGUISystem", "GUISrv_StartGUISystem_FW", 1 },
    {20, "LEDDrive", "LEDDrive_FW", 1 },
    {20, "LockMainPower", "LockMainPower_FW", 1 },
    //{20, "lseek", "Lseek_FW", 1 },
    {20, "Lseek", "Lseek_FW", 1 },
    {20, "MoveIrisWithAv", "MoveIrisWithAv_FW", 1 },
    {20, "MoveZoomLensWithPoint", "MoveZoomLensWithPoint_FW", 1 },
    {20, "memcmp", "memcmp_FW", 1 },
    {20, "memcpy", "memcpy_FW", 1 },
    {20, "memset", "memset_FW", 1 },
    {20, "NewTaskShell", "NewTaskShell_FW", 1 },
    {20, "NR_GetDarkSubType", "NR_GetDarkSubType_FW", 1 },
    {20, "NR_SetDarkSubType", "NR_SetDarkSubType_FW", 1 },
    {20, "Open", "Open_FW", 1 },
    {20, "PostLogicalEventForNotPowerType", "PostLogicalEventForNotPowerType_FW", 1 },
    {20, "PostLogicalEventToUI", "PostLogicalEventToUI_FW", 1 },
    {20, "PT_MoveDigitalZoomToWide", "PT_MoveDigitalZoomToWide_FW", 1 },
    {20, "PT_MoveOpticalZoomAt", "PT_MoveOpticalZoomAt_FW", 1 },
    {20, "Read", "Read_FW", 1 },
    {20, "RefreshPhysicalScreen", "RefreshPhysicalScreen_FW", 1 },
    {20, "ResetFocusLens", "ResetFocusLens_FW", 1 },
    {20, "ResetZoomLens", "ResetZoomLens_FW", 1 },
    {20, "SavePaletteData", "SavePaletteData_FW", 1 },
    {20, "SetAutoShutdownTime", "SetAutoShutdownTime_FW", 1 },
    {20, "SetCurrentCaptureModeType", "SetCurrentCaptureModeType_FW", 1 },
    {20, "SetScriptMode", "SetScriptMode_FW", 1 },
    {20, "SleepTask", "SleepTask_FW", 1 },
    {20, "strcmp", "j_strcmp_FW", 0 },
    {20, "strcmp", "strcmp_FW", 0 },
    {20, "strcpy", "strcpy_FW", 1 },
    {20, "strlen", "strlen_FW", 1 },
    {20, "StartRecModeMenu", "StartRecModeMenu_FW", 1 },
    //{20, "TakeSemaphore", "TakeSemaphore_FW", 1 },
    {20, "UIFS_WriteFirmInfoToFile", "UIFS_WriteFirmInfoToFile_FW", 1 },
    {20, "UnlockMainPower", "UnlockMainPower_FW", 1 },
    {20, "VbattGet", "VbattGet_FW", 1 },
    {20, "write", "Write_FW", 1 },
    {20, "Write", "Write_FW", 1 },
    {20, "task_CaptSeq", "task_CaptSeqTask", 1 },
    {20, "task_ExpDrv", "task_ExpDrvTask", 1 },
    {20, "task_FileWrite", "task_FileWriteTask", 1 },
    {20, "task_RotaryEncoder", "task_JogDial", 1 },
    {20, "task_RotaryEncoder", "task_RotarySw", 1 },
    {20, "task_SwitchCheck", "task_SwitchCheckTask", 1 },
    {20, "IsControlEventActive", "IsControlEventActive_FW", 0 },
    {20, "GetLogicalEventName", "ShowLogicalEventName_FW", 0x01000002 },
    {20, "TurnOnDisplay", "DispCon_TurnOnDisplay_FW", 0 },
    {20, "TurnOffDisplay", "DispCon_TurnOffDisplay_FW", 0 },
    {20, "UnlockAF", "PT_UnlockAF_FW", 0x01000002 }, // newer Vx
    {20, "DoAFLock", "PT_DoAFLock_FW", 0x01000002 }, // newer Vx
    {20, "DoAFLock", "DoAFLock_FW", 1 },
    {20, "UnlockAF", "UnlockAF_FW", 1 },
    {20, "UnlockAE", "PT_UnlockAE_FW", 0x01000002 }, // newer Vx
    {20, "DoAELock", "PT_DoAELock_FW", 0x01000002 }, // newer Vx
    {20, "DoAELock", "DoAELock_FW", 1 },
    {20, "UnlockAE", "UnlockAE_FW", 1 },
    {20, "MFOn", "MFOn_FW", 1 },
    {20, "MFOff", "MFOff_FW", 1 },
    {20, "GetAdChValue", "GetAdChValue_FW", 0 },
    {20, "HwOcReadICAPCounter", "GetCurrentMachineTime", 3 },

    { 1, "ExportToEventProcedure_FW", "ExportToEventProcedure", 1 },
    { 1, "AllocateMemory", "AllocateMemory", 1 },
    { 1, "Close", "Close", 1 },
    { 1, "CreateTask", "CreateTask", 1 },
    { 1, "DoAFLock", "PT_DoAFLock", 0x01000002 },
    { 1, "ExitTask", "ExitTask", 1 },
    { 1, "exmem_alloc", "ExMem.AllocCacheable", 4 },
    { 1, "exmem_free", "ExMem.FreeCacheable", 0x01000003 },
    { 1, "Fclose_Fut", "Fclose_Fut", 1 },
    { 1, "Feof_Fut", "Feof_Fut", 1 },
    { 1, "Fflush_Fut", "Fflush_Fut", 1 },
    { 1, "Fgets_Fut", "Fgets_Fut", 1 },
    { 1, "Fopen_Fut", "Fopen_Fut", 1 },
    { 1, "Fread_Fut", "Fread_Fut", 1 },
    { 1, "FreeMemory", "FreeMemory", 1 },
    { 1, "Fseek_Fut", "Fseek_Fut", 1 },
    { 1, "Fwrite_Fut", "Fwrite_Fut", 1 },
    { 1, "GetParameterData", "PTM_RestoreUIProperty", 0xF0000004 },
    { 1, "GetPropertyCase", "PT_GetPropertyCaseString", 1 },
    { 1, "GetPropertyCase", "PT_GetPropertyCaseInt", 0x0100000F },
    { 1, "GetPropertyCase", "GetPropertyCase", 0x0100000F },
    { 1, "GetSDProtect", "GetSDProtect", 1 },
    { 1, "GetSystemTime", "GetSystemTime", 1 },
    { 1, "LEDDrive", "LEDDrive", 1 },
    { 1, "LockMainPower", "LockMainPower", 1 },
    { 1, "Lseek", "Lseek", 1 },
    { 1, "lseek", "Lseek", 1 },
    { 1, "memcpy", "memcpy", 1 },
    { 1, "memcmp", "memcmp", 1 },
    { 1, "memset", "memset", 1 },
    { 1, "NewTaskShell", "NewTaskShell", 1 },
    { 1, "Open", "Open", 1 },
    { 1, "PostLogicalEventToUI", "PostLogicalEventToUI", 1 },
    { 1, "PostLogicalEventForNotPowerType", "PostLogicalEventForNotPowerType", 1 },
    { 1, "Read", "Read", 1 },
    { 1, "read", "Read", 1 },
    { 1, "RefreshPhysicalScreen", "RefreshPhysicalScreen", 1 },
    { 1, "SetAutoShutdownTime", "SetAutoShutdownTime", 1 },
    { 1, "SetCurrentCaptureModeType", "SetCurrentCaptureModeType", 1 },
    { 1, "SetLogicalEventActive", "UiEvnt_SetLogicalEventActive", 1 },
    { 1, "SetParameterData", "PTM_BackupUIProperty", 1 },
    { 1, "SetPropertyCase", "PT_SetPropertyCaseInt", 0x01000003 },
    { 1, "SetPropertyCase", "SetPropertyCase", 0x01000004 },
    { 1, "SetScriptMode", "SetScriptMode", 1 },
    { 1, "SleepTask", "SleepTask", 1 },
    { 1, "strcmp", "strcmp", 0 },
    { 1, "strcpy", "strcpy", 1 },
    { 1, "strlen", "strlen", 1 },
    { 1, "strtol", "atol", 3 },
    { 1, "TakeSemaphore", "TakeSemaphore", 1 },
    { 1, "UIFS_WriteFirmInfoToFile", "UIFS_WriteFirmInfoToFile", 1 },
    //{ 1, "UnlockAF", "PT_UnlockAF", 0x01000002 },
    { 1, "UnlockMainPower", "UnlockMainPower", 1 },
    { 1, "VbattGet", "VbattGet", 1 },
    { 1, "Write", "Write", 1 },
    { 1, "write", "Write", 1 },
    { 1, "GUISrv_StartGUISystem", "GUISrv_StartGUISystem", 1 },

    { 2, "GetBatteryTemperature", "GetBatteryTemperature", 1 },
    { 2, "GetCCDTemperature", "GetCCDTemperature", 1 },
    { 2, "GetOpticalTemperature", "GetOpticalTemperature", 1 },
    //{ 2, "GetFocusLensSubjectDistance", "GetCurrentTargetDistance", 1 },
    { 2, "GetZoomLensCurrentPoint", "GetZoomLensCurrentPoint", 1 },
    { 2, "GetZoomLensCurrentPosition", "GetZoomLensCurrentPosition", 1 },
    { 2, "MoveFocusLensToDistance", "MoveFocusLensToDistance", 1 },
    { 2, "MoveZoomLensWithPoint", "MoveZoomLensWithPoint", 1 },
    //{ 2, "GetCurrentAvValue", "GetCurrentAvValue", 1 },
    { 2, "PT_MoveOpticalZoomAt", "PT_MoveOpticalZoomAt", 1 },
    //{ 2, "PT_MoveOpticalZoomAt", "SS.MoveOpticalZoomAt", 1 },
    { 2, "PT_MoveDigitalZoomToWide", "PT_MoveDigitalZoomToWide", 1 },
    //{ 2, "PT_MoveDigitalZoomToWide", "SS.MoveDigitalZoomToWide", 1 },
    { 2, "MoveIrisWithAv", "MoveIrisWithAv", 1},
    //{ 2, "PutInNdFilter", "TurnOnNdFilter", 1 },
    //{ 2, "PutOutNdFilter", "TurnOffNdFilter", 1 },
    //{ 2, "PutInNdFilter", "PutInNdFilter", 1 },
    //{ 2, "PutOutNdFilter", "PutOutNdFilter", 1 },
    //{ 2, "IsStrobeChargeCompleted", "EF.IsChargeFull", 1 },
    { 2, "GetPropertyCase", "PT_GetPropertyCaseInt", 0x01000012 },
    { 2, "SetPropertyCase", "PT_SetPropertyCaseInt", 0x01000008 },
    { 2, "SetPropertyCase", "PT_SetPropertyCaseInt", 0x01000009 },
    //{ 2, "UnlockAF", "PT_UnlockAF", 0x01000002 },
    //{ 2, "UnlockAF", "SS.UnlockAF", 0x01000002 },
    //{ 2, "DoAFLock", "PT_DoAFLock", 0x01000002 },
    //{ 2, "DoAFLock", "SS.DoAFLock", 0x01000002 },
    { 2, "GetSystemTime", "PT_GetSystemTime", 0x01000003 },
    { 2, "PT_PlaySound", "PT_PlaySound", 0x01000005 },
    { 2, "StartRecModeMenu", "StartRecModeMenu", 1 },
    { 2, "GetSDProtect", "GetSDProtect", 1 },
    { 2, "DispCon_ShowBitmapColorBar", "DispCon_ShowBitmapColorBar", 1 },
    { 2, "SetAE_ShutterSpeed", "SetAE_ShutterSpeed", 1 },
    { 2, "ResetZoomLens", "ResetZoomLens", 1 },
    { 2, "ResetFocusLens", "ResetFocusLens", 1 },
    { 2, "NR_GetDarkSubType", "NR_GetDarkSubType", 1 },
    //{ 2, "NR_GetDarkSubType", "NRTBL.GetDarkSubType", 1 },
    { 2, "NR_SetDarkSubType", "NR_SetDarkSubType", 1 },
    //{ 2, "NR_SetDarkSubType", "NRTBL.SetDarkSubType", 1 },
    { 2, "SavePaletteData", "SavePaletteData", 1 },
    { 2, "GetVRAMHPixelsSize", "GetVRAMHPixelsSize", 1 },
    { 2, "GetVRAMVPixelsSize", "GetVRAMVPixelsSize", 1 },
    { 2, "EngDrvIn", "EngDrvIn", 4 },
    { 2, "EngDrvOut", "EngDrvOut", 0x01000005 },
    { 2, "EngDrvRead", "EngDrvRead", 4 },
    { 2, "EngDrvBits", "EngDrvBits", 0x01000006 },
    { 2, "exmem_alloc", "ExMem.AllocCacheable", 4 },
    { 2, "exmem_free", "ExMem.FreeCacheable", 0x01000003 },

    { 2, "PTM_GetCurrentItem", "PTM_GetCurrentItem", 0x01000003 },
    { 2, "PTM_SetCurrentItem", "PTM_SetCurrentItem", 8 },
    { 2, "PTM_NextItem", "PTM_NextItem", 0x01000003 },
    { 2, "PTM_PrevItem", "PTM_PrevItem", 0x01000003 },
    { 2, "PTM_SetPropertyEnable", "PTM_SetProprietyEnable", 8 },

    { 3, "AllocateMemory", "AllocateMemory", 1 },
    { 3, "FreeMemory", "FreeMemory", 1 },
    { 3, "PostLogicalEventToUI", "PostLogicalEventToUI", 1 },
    { 3, "PostLogicalEventForNotPowerType", "PostLogicalEventForNotPowerType", 1 },
    { 3, "LockMainPower", "LockMainPower", 1 },
    { 3, "UnlockMainPower", "UnlockMainPower", 1 },
    { 3, "SetAutoShutdownTime", "SetAutoShutdownTime", 1 },
    { 3, "NewTaskShell", "NewTaskShell", 1 },
    { 3, "VbattGet", "VbattGet", 1 },
    { 3, "LEDDrive", "LEDDrive", 1 },
    { 3, "SetPropertyCase", "PT_SetPropertyCaseInt", 0x01000003 },
    //{ 3, "UnlockAF", "PT_UnlockAF", 0x01000002 },
    //{ 3, "DoAFLock", "PT_DoAFLock", 0x01000002 },
    { 3, "UIFS_WriteFirmInfoToFile", "UIFS_WriteFirmInfoToFile", 1 },
    { 3, "PT_MoveOpticalZoomAt", "PT_MoveOpticalZoomAt", 1 },
    { 3, "PT_MoveDigitalZoomToWide", "PT_MoveDigitalZoomToWide", 1 },
    { 3, "PT_PlaySound", "PT_PlaySound", 1 },
    { 3, "exmem_alloc", "ExMem.AllocCacheable", 4 },
    { 3, "exmem_free", "ExMem.FreeCacheable", 0x01000003 },
    { 3, "GetSDProtect", "GetSDProtect", 1 },

    { 4, "TurnOnBackLight", "TurnOnBackLight", 1 },
    { 4, "TurnOffBackLight", "TurnOffBackLight", 1 },
    { 4, "EnterToCompensationEVF", "SSAPI::EnterToCompensationEVF", 1 },
    { 4, "EnterToCompensationEVF", "ExpComp On", 1 },
    { 4, "EnterToCompensationEVF", "ExpOn", 1 },
    { 4, "ExitFromCompensationEVF", "SSAPI::ExitFromCompensationEVF", 1 },
    { 4, "ExitFromCompensationEVF", "ExpComp Off", 1 },
    { 4, "ExitFromCompensationEVF", "ExpOff", 1 },
    //{ 4, "PB2Rec", "AC:PB2Rec", 1 },
    //{ 4, "PB2Rec", "AC:PB2Rec", 6 },
    //{ 4, "PB2Rec", "AC:PB2Rec", 11 },
    //{ 4, "Rec2PB", "AC:Rec2PB", 1 },
    { 4, "RefreshPhysicalScreen", "ScreenUnLock", 1 },
    { 4, "RefreshPhysicalScreen", "ScreenUnLock", 7 },
    { 4, "RefreshPhysicalScreen", "ScreenUnLock", 15 },
    { 4, "RefreshPhysicalScreen", "Reduce ScreenUnLock", 5 },
    { 4, "RefreshPhysicalScreen", "Window:IneffectiveLockPhysicalScreen", 8 },
    { 4, "UnsetZoomForMovie", "ZoomCon_UnsetZoomForMovie", 1 },
    { 4, "ExpCtrlTool_StopContiAE", "StopContiAE", 9 },
    { 4, "ExpCtrlTool_StopContiAE", "StopContiAE", 10 },
    { 4, "ExpCtrlTool_StartContiAE", "StartContiAE", 9 },
    { 4, "ExpCtrlTool_StartContiAE", "StartContiAE", 10 },
    { 4, "ExecuteEventProcedure", "Can not Execute ", 14 },

    { 5, "UIFS_WriteFirmInfoToFile", "UIFS_WriteFirmInfoToFile", 1 },
    { 5, "CreateTask", "CreateTask", 1 },
    //{ 5, "hook_CreateTask", "CreateTask", 1 },
    { 5, "ExitTask", "ExitTask", 1 },
    { 5, "SleepTask", "SleepTask", 1 },
    { 5, "DeleteSemaphore", "DeleteSemaphore", 1 },
    { 5, "CreateCountingSemaphore", "CreateCountingSemaphore", 1 },
    //                                                                   ???
    { 5, "UpdateMBROnFlash", "MakeBootDisk", 0x01000003,                  11 },
    { 5, "MakeSDCardBootable", "MakeBootDisk", 0x01000003,                 1 },

    //{ 6, "Restart", "Bye", 0 },
    { 6, "GetImageFolder", "GetCameraObjectTmpPath ERROR[ID:%lx] [TRY:%lx]\n", 0 },
    { 6, "reboot_fw_update", "FirmUpgrade.c", 0 },

    { 7, "CreateTaskStrictly", "FileWriteTask", 0x01000001 },
    { 7, "RegisterInterruptHandler", "SdDmaInt", 0x01000001 },
    { 7, "LogCameraEvent", "BufAccBeep", 0x01000001 },
    { 7, "LogCameraEvent", "MyCamFunc_PlaySound_MYCAM_COVER_OPEN", 0x01000001 },

    { 8, "WriteSDCard", "Mounter.c", 0 },

    // Ensure ordering in func_names is correct for dependencies here
    //                                                                   ???
    //{ 9, "kbd_p1_f", "task_PhySw", 0,                                      5 },
    //{ 9, "kbd_p2_f", "task_PhySw", 0,                                      7 },
    { 9, "kbd_read_keys", "kbd_p1_f", 0,                                   2 },
    { 9, "kbd_p1_f_cont", "kbd_p1_f", -1,                                  3 },
    //{ 9, "kbd_read_keys_r2", "kbd_read_keys", 0,                          11 },
    { 9, "GetKbdState", "kbd_read_keys", 0,                                8 },
    { 9, "GetKbdState", "kbd_read_keys", 0,                                9 },
    //{ 9, "strtolx", "strtol", 0,                                           1 },
    //{ 9, "mkdir", "MakeDirectory_Fut", 0x01000001,                        17 },
    //{ 9, "mkdir", "MakeDirectory_Fut", 0x01000002,                        17 },
    //{ 9, "time", "MakeDirectory_Fut", 0,                                  12 },
    { 9, "stat", "_uartr_req", 0,                                          0 },
    //{ 9, "PostMessageQueue", "PostMessageQueueStrictly", 0,                3 },
    //{ 9, "WaitForAnyEventFlag", "WaitForAnyEventFlagStrictly", 0,          3 },
    //{ 9, "WaitForAllEventFlag", "WaitForAllEventFlagStrictly", 0,          3 },
    //{ 9, "CreateMessageQueue", "CreateMessageQueueStrictly", 0,            1 },
    { 9, "CreateRecursiveLock", "CreateRecursiveLockStrictly", 0,       1 },
    { 9, "CreateRecursiveLock", "CreateRecursiveLockStrictly", 0,       6 },    // old vx
    { 9, "CreateEventFlag", "CreateEventFlagStrictly", 0,               1 },
    { 9, "_GetSystemTime", "PT_GetSystemTime", 0,                       3 },
    { 9, "close", "Close", 0,                                              2 },
    { 9, "open", "Open", 0,                                                3 },
    { 9, "open", "Open", 0,                                                3 },
    { 9, "malloc_alt", "wrapped_malloc", 0,                                2 },

    //                                                                   ???
    { 11, "err_init_task", "\n-- %s() error in init_task() --", 0,         2 },
    { 11, "set_control_event", "Button:0x%08X:%s", 0x01000001,            14 },
    { 11, "set_control_event", "Button:0x%08X:%s", 0xf1000001,            15 },
    { 11, "set_control_event", "Button:0x%08X:%s", 0x01000001,            19 },
    { 11, "set_control_event", "Button:0x%08X:%s", 0x01000001,            20 },
    { 11, "_log", (char*)log_test, 0x01000001,                             1 },
    { 11, "_uartr_req", "A/uartr.req", 0,                                  3 },

    //                                                                   ???
    //{ 12, "DeleteFile_Fut", "DeleteFile_Fut", 1,                        0x38 },
    //{ 12, "AllocateUncacheableMemory", "AllocateUncacheableMemory", 1,  0x2C },
    //{ 12, "FreeUncacheableMemory", "FreeUncacheableMemory", 1,          0x30 },
    //{ 12, "free", "free", 1,                                            0x28 },
    //{ 12, "malloc", "malloc", 0x01000003,                               0x24 },
    //{ 12, "TakeSemaphore", "TakeSemaphore", 1,                          0x14 },
    //{ 12, "GiveSemaphore", "GiveSemaphore", 1,                          0x18 },
    //{ 12, "ClearEventFlag", "ClearEventFlag", 1,                        0x04 },
    //{ 12, "SetEventFlag", "SetEventFlag", 1,                            0x08 },
    //{ 12, "WaitForAnyEventFlag", "WaitForAnyEventFlag", 1,              0x0c },
    //{ 12, "WaitForAllEventFlag", "WaitForAllEventFlag", 1,              0x10 },

    { 13, "strftime", "Sunday", 1 },

    { 15, "LocalTime", "%04d:%02d:%02d %02d:%02d:%02d", 0x01000001 },
    { 15, "GetMemInfo", "Malloc Information\n", 0x01000001 },
    { 15, "GetMemInfo", "Malloc Information (%s type)\n", 0x01000001 },
    { 15, "vsprintf", "\nCPrintf Size Over!!", 0x01000001 },
    //{ 15, "ReadFastDir", "ReadFast_ERROR\n", 0x01000001 },
    //{ 15, "OpenFastDir", "OpenFastDir_ERROR\n", 0x01000001 },
    { 15, "realloc", "fatal error - scanner input buffer overflow", 0x01000001 },
    { 15, "CreateBinarySemaphore", "SdPower.c", 0x01000001 },
    { 15, "CreateBinarySemaphore", "DoWBForEVFOnce CreateBinarySemaphore Error\r\n", 0x01000001 }, // old Vx
    { 15, "CreateBinarySemaphore", "DoWBForEVFOnce CreateBinarySemaphore Error\n", 0x01000001 }, // old Vx
    //                                                                           Vx
    { 15, "SetHPTimerAfterTimeout", "FrameRateGenerator.c", 0x01000001,          0x0008 },
    { 15, "wrapped_malloc", "\n malloc error \n", 0x01000001,                    0x0010 },
    { 15, "IsStrobeChargeCompleted", "\r\nCaptSeq::ChargeNotCompleted!!", 0x01000001 }, // ixus30, 40
    { 15, "get_resource_pointer", "Not found icon resource.\r\n", 0x01000001,    0x0008 },

    { 16, "DeleteDirectory_Fut", (char*)DeleteDirectory_Fut_test, 0x01000001 },
    { 16, "MakeDirectory_Fut", (char*)MakeDirectory_Fut_test, 0x01000001 },
    { 16, "RenameFile_Fut", (char*)RenameFile_Fut_test, 0x01000001 },

    { 17, "ScreenLock", "StartRecModeMenu", 0 },
    { 17, "ScreenUnlock", "StartRecModeMenu", 0 },

    // Ensure ordering in func_names is correct for dependencies here
    //                                                                           Vx
    { 19, "CreateMessageQueueStrictly", "CreateTaskStrictly", 0,                 0x0210 },
    { 19, "CreateMessageQueueStrictly", "CreateTaskStrictly", 0,                 0x0720 }, // old vx
    { 19, "CreateEventFlagStrictly", "CreateMessageQueueStrictly", 0,            0x010b },
    { 19, "CreateEventFlagStrictly", "CreateMessageQueueStrictly", 0,            0x0619 }, // old vx
    { 19, "CreateBinarySemaphoreStrictly", "CreateEventFlagStrictly", 0,         0x010b },
    { 19, "CreateBinarySemaphoreStrictly", "CreateEventFlagStrictly", 0,         0x081b }, // old vx
    { 19, "CreateCountingSemaphoreStrictly", "CreateBinarySemaphoreStrictly", 0, 0x010b },
    { 19, "CreateCountingSemaphoreStrictly", "CreateBinarySemaphoreStrictly", 0, 0x081b }, // old vx
    { 19, "CreateRecursiveLockStrictly", "CreateCountingSemaphoreStrictly", 0,   0x010b },
    { 19, "CreateRecursiveLockStrictly", "CreateCountingSemaphoreStrictly", 0,   0x071a }, // old vx
    //{ 19, "PostMessageQueue", "TryReceiveMessageQueue", 0,                       0x091f },
    //{ 19, "DeleteMessageQueue", "CreateMessageQueue", 0,                         0x1021 },
    //{ 19, "ReceiveMessageQueue", "DeleteMessageQueue", 0,                        0x1024 },
    //{ 19, "TryReceiveMessageQueue", "ReceiveMessageQueue", 0,                    0x002b },
    { 19, "TryPostMessageQueue", "PostMessageQueue", 0,                          0x058a },
    { 19, "GetNumberOfPostedMessages", "TryPostMessageQueue", 0,                 0x004b },
    { 19, "DeleteRecursiveLock", "CreateRecursiveLock", 0,                       0x0247 },
    { 19, "AcquireRecursiveLock", "DeleteRecursiveLock", 0,                      0x0136 }, // old vx
    { 19, "AcquireRecursiveLock", "DeleteRecursiveLock", 0,                      0x0026 }, // new vx
    { 19, "ReleaseRecursiveLock", "AcquireRecursiveLock", 0,                     0x154a }, // old vx
    { 19, "ReleaseRecursiveLock", "AcquireRecursiveLock", 0,                     0x002a }, // new vx
    { 19, "GetEventFlagValue", "ClearEventFlag", 0,                              0x0014 },
    //{ 19, "DeleteEventFlag", "CreateEventFlag", 0,                               0x0016 },
    { 19, "CheckAnyEventFlag", "CheckAllEventFlag", 0,                          -0x0004 },
    { 19, "CheckAllEventFlag", "WaitForEventFlag", 0,                           -0x1028 },
    { 19, "WaitForAnyEventFlag", "WaitForEventFlag", 0,                          0x0569 },
    { 19, "WaitForAllEventFlag", "WaitForAnyEventFlag", 0,                       0x1004 },
    { 19, "ClearEventFlag", "SetEventFlag", 0,                                   0x0167 }, // new vx
    { 19, "ClearEventFlag", "SetEventFlag", 0,                                   0x0165 }, // old vx
    { 19, "SetEventFlag", "WaitForAllEventFlag", 0,                              0x1004 },
    { 19, "TryTakeSemaphore", "GiveSemaphore", 0,                               -0x0619 }, // old vx
    { 19, "TryTakeSemaphore", "GiveSemaphore", 0,                               -0x071c }, // new vx
    { 19, "GetSemaphoreValue", "GiveSemaphore", 0,                               0x0035 }, // old vx
    { 19, "GetSemaphoreValue", "GiveSemaphore", 0,                               0x013b }, // new vx
    { 19, "SetHPTimerAfterNow", "SetHPTimerAfterTimeout", 0,                    -0x0223 },
    { 19, "CancelHPTimer", "SetHPTimerAfterTimeout", 0,                          0x09fa },
    { 19, "SetTimerAfter", "_GetSystemTime", 0,                                  0x0355 },
    { 19, "SetTimerWhen", "SetTimerAfter", 0,                                    0x0327 },
    { 19, "CancelTimer", "SetTimerWhen", 0,                                      0x026d }, // older vx
    { 19, "CancelTimer", "SetTimerWhen", 0,                                      0x021e }, // newer vx (no unused functions inbetween)

    { 19, "UnregisterInterruptHandler", "RegisterInterruptHandler", 0,           0x0909 },
    { 19, "GetSRAndDisableInterrupt", "UnregisterInterruptHandler", 0,           0x0f06 },
    { 19, "SetSR", "GetSRAndDisableInterrupt", 0,                                0x1003 },
    { 19, "EnableInterrupt", "SetSR", 0,                                         0x1508 },
    //{ 19, "UnregisterInterruptHandler", "RegisterInterruptHandler", 0,           0x0009 },
    //{ 19, "GetSRAndDisableInterrupt", "UnregisterInterruptHandler", 0,           0x0006 },
    //{ 19, "SetSR", "UnregisterInterruptHandler", 0,                              0x1007 },
    //{ 19, "EnableInterrupt", "UnregisterInterruptHandler", 0,                    0x170f },
    { 19, "GetDrive_TotalClusters", "GetDrive_FreeClusters", 0,                 -0x120f },
    { 19, "GetDrive_TotalClusters", "GetDrive_FreeClusters", 0,                 -0x130f }, // s80
    { 19, "GetDrive_TotalClusters", "GetDrive_FreeClusters", 0,                 -0x120e }, // ixus950
    { 19, "GetDrive_TotalClusters", "GetDrive_FreeClusters", 0,                 -0x1310 }, // old vx
    { 19, "GetDrive_ClusterSize", "GetDrive_TotalClusters", 0,                  -0x020f },
    { 19, "GetDrive_ClusterSize", "GetDrive_TotalClusters", 0,                  -0x0217 }, // s80
    { 19, "GetDrive_ClusterSize", "GetDrive_TotalClusters", 0,                  -0x0310 }, // old vx

    { 21, "add_ptp_handler", (char*)find_add_ptp_handler, 0 },
    //{ 21, "apex2us", (char*)find_apex2us, 0 },
    { 21, "mkdir", (char*)find_mkdir, 0 },
    { 21, "_pow", (char*)find_pow, 0 },
    { 21, "_log", (char*)find_log, 0 },
    { 21, "_log10", (char*)find_log10, 0 },

    { 22, "closedir", (char*)find_closedir, 0 },
    { 22, "PT_PlaySound", (char*)find_PT_PlaySound, 0 },
    { 22, "ExportToEventProcedure_FW", (char*)find_ExportToEventProcedure, 0 },
    { 22, "RegisterEventProcedure_FW", (char*)find_RegisterEventProcedure, 0 },

    { 22, "set_control_event", (char*)find_set_control_event, 0 }, // vx
    { 22, "filesem_init", (char*)find_filesem_init, 0 }, // vx
    { 22, "GetCurrentMachineTime", (char*)find_getcurrentmachinetime, 0},

    //                                                                                          Vx
    { 100, "DebugAssert", "\nAssert: File %s Line %d\n", 0,                                     10 },
    { 100, "DebugAssert", "\aAssert: File %s,  Expression %s,  Line %d\n", 0,                   14 }, // ixus30,40
    { 100, "CreateMessageQueue", "CreateMessageQueue : call from interrupt handler", 0,         8 },
    { 100, "DeleteMessageQueue", "DeleteMessageQueue : call from interrupt handler", 0,         8 },
    { 100, "PostMessageQueue", "PostMessageQueue : call from interrupt handler", 0,             12 },
    { 100, "ReceiveMessageQueue", "ReceiveMessageQueue : NULL buffer address", 0,               11 },
    { 100, "TryReceiveMessageQueue", "TryReceiveMessageQueue : NULL buffer address", 0,         10 },
    { 100, "CreateEventFlag", "CreateEventFlag : call from interrupt handler", 0,               5 },
    { 100, "DeleteEventFlag", "DeleteEventFlag : call from interrupt handler", 0,               8 },
    { 100, "WaitForEventFlag", "WaitForEventFlag : call from interrupt handler", 0,             12 },
    { 100, "CancelHPTimer", "EventProcedure", 0,                                                101 }, // newest vx
    
    { 101, "DeleteSemaphore", "DeleteSemaphore", 0 },
    { 101, "CreateCountingSemaphore", "CreateCountingSemaphore", 0 },
    { 101, "TakeSemaphore", "TakeSemaphore", 0 },
    { 101, "GiveSemaphore", "GiveSemaphore", 0 },
    { 101, "CreateTask", "CreateTask", 0 },
    { 101, "ExitTask", "ExitTask", 0 },
    { 101, "PT_GetSystemTime", "PT_GetSystemTime", 0 },

    { 102, "PB2Rec", "AC:PB2Rec", 0,                                                            0x20 },
    { 102, "Rec2PB", "AC:Rec2PB", 0,                                                            0x20 },

    { 103, "GetDrive_FreeClusters", "AvailClusters.c", 0,                                       0x04 },

    { 104, "SetLogicalEventActive", "EventReciever.c", 16,                                      0x0000 },

    { 0, 0, 0, 0 }
};

// Find the named function in the 'func_list' array above, return index of entry
int find_func(const char* name)
{
    int i;
    for (i=0; func_list[i].name != 0; i++)
    {
        if (strcmp(name, func_list[i].name) == 0)
        {
            return i;
        }
    }
    return -1;  // Not found
}

// Get VxWorks version specific offset?
int vxworks_offset(firmware *fw, string_sig *sig)
{
    return sig->vxworks_offset;
}

//------------------------------------------------------------------------------------------------------------

// Loop through firmware looking for instances of a sig string
// For each one found call the check_match function to see if it matches the sig
// Return 1 if match found, else return 0
int fw_string_process(firmware *fw, string_sig *sig, int (*check_match)(firmware *fw, string_sig *sig, int j), int inc_eos)
{
    int nlen = strlen(sig->ev_name);
    uint32_t nm0 = *((uint32_t*)sig->ev_name);
    uint32_t *p;
    int j;
    BufRange *br;

    for (br = fw->br; br != 0; br = br->next)
    {
        for (p = br->p, j = br->off; j < br->off+br->len-nlen/4; p++, j++)
        {
            if ((nm0 == *p) && (memcmp(p+1,sig->ev_name+4,nlen-4+inc_eos) == 0))
            {
                if (check_match(fw,sig,j))
                    return 1;
            }
        }
    }

    return 0;
}

// As above; but scan the firmware byte-by-byte rather than by words
// Slower; but required when strings are not stores on 32 bit boundaries
int fw_string_process_unaligned(firmware *fw, string_sig *sig, int (*check_match)(firmware *fw, string_sig *sig, int j))
{
    int nlen = strlen(sig->ev_name);
    char *p;
    int j;
    BufRange *br;

    for (br = fw->br; br != 0; br = br->next)
    {
        for (p = (char*)br->p, j = 0; j < br->len*4-nlen; p++, j++)
        {
            if (strcmp(p,sig->ev_name) == 0)
            {
                if (check_match(fw,sig,j+br->off*4))
                    return 1;
            }
        }
    }

    return 0;
}

// Loop through the firmware from the start, calling 'check_match' for each location
// If 'check_match' returns non-zero, exit returning 1.
// Otherwise returns 0 - no match found.
int fw_process(firmware *fw, string_sig *sig, int (*check_match)(firmware *fw, string_sig *sig, int j))
{
    uint32_t *p;
    int j;
    BufRange *br;

    for (br = fw->br; br != 0; br = br->next)
    {
        for (p = br->p, j = br->off; j < br->off+br->len; p++, j++)
        {
            if (check_match(fw,sig,j))
                return 1;
        }
    }

    return 0;
}

//------------------------------------------------------------------------------------------------------------

// New string / signature matching functions

// Sig pattern:
//      Function pointer    -   DCD func
//      String              -   DCB "func"
// Note: 'func' may not be the target address, the sig->offset value allows selection of another function
//       called at a fixed number of instructions from the 'func' address found
int match_strsig1(firmware *fw, string_sig *sig, int j)
{
    uint32_t fadr = fwval(fw,j-1);      // function address
    if (idx_valid(fw,adr2idx(fw,fadr))) // is function address valid
    {
        // If function address is a B, and we are following branches, then follow the first B
        if (sig->offset > 1) fadr = followBranch(fw, fadr, 1);
        // Follow any subsequent branch at the given offset
        fadr = followBranch2(fw, fadr, sig->offset);
        fwAddMatch(fw,fadr,32,0,101);
        return 1;
    }
    return 0;
}

// Sig pattern:
//      String pointer      -       DCD str
//      Function pointer    -       DCD func
//                ...
//      String              -   str DCB "func"
// Note: 'func' may not be the target address, the offset value allows selection of another function
//       called at a fixed number of instructions from the 'func' address found
int match_strsig2a(firmware *fw, int k, uint32_t sadr, uint32_t offset)
{
    if (fwval(fw,k) == sadr)                // pointer to string?
    {
        uint32_t fadr = fwval(fw,k+1);      // function address
        if (idx_valid(fw,adr2idx(fw,fadr))) // is function address valid
        {
            uint32_t bfadr = followBranch2(fw, fadr, offset);
            if ((offset <= 1) || (bfadr != fadr))
            {
                fwAddMatch(fw,bfadr,32,0,102);
                return 1;
            }
        }
    }
    return 0;
}
int match_strsig2(firmware *fw, string_sig *sig, int j)
{
    // Note - 'j' is byte offset in firmware not instruction index (called from fw_string_process_unaligned)
    return search_fw(fw, match_strsig2a, fw->base + j, sig->offset, 2);
}

// Sig pattern:
//      Load Func Address   -   ADR Rx, func
//      Load String Address -   ADR Rx, "func"
//      Branch              -   BL
//              ...
//      String              -   DCB "func"
// or
//      Load Func Address   -   ADR Rx, func
//                              B   loc
//              ...
//                          loc:
//      Load String Address -   ADR    Rx, "func"
//      Branch              -   BL
//              ...
//      String              -   DCB "func"
// 'sadr' = address of "func" string
// Note: 'func' may not be the target address, the offset value allows selection of another function
//       called at a fixed number of instructions from the 'func' address found
int match_strsig3a(firmware *fw, int k, uint32_t sadr, uint32_t offset)
{
    if (isADR_PC(fw,k+1) &&    // ADR ?
        isBorBL(fw,k+2))       // B or BL ?
    {
        uint32_t padr = ADR2adr(fw,k+1);    // get address pointed to by 2nd ADR instructioin
        if (padr == sadr)                   // does it match target string
        {
            int j2 = k;
            int found = 0;
            if (isADR_PC(fw,k))             // ADR ?
                found = 1;
            else
            {
                // May be DCD block between 1st and 2nd ADR
                for (j2 = k-2; j2 >= 0 && j2 >= k-4096; j2--)
                {
                    if (isADR_PC(fw,j2) &&  // ADR ?
                        isB(fw,j2+1))       // B
                    {
                        uint32_t fa = idx2adr(fw,j2+1);
                        fa = followBranch(fw,fa,1);
                        if (adr2idx(fw,fa) == k+1)
                        {
                            found = 1;
                            break;
                        }
                    }
                }
            }
            if (found)
            {
                uint32_t fadr = ADR2adr(fw,j2);
                if (offset > 1) fadr = followBranch(fw, fadr, 1);
                fadr = followBranch2(fw, fadr, offset);
                fwAddMatch(fw,fadr,32,0,103);
                return 1;
            }
        }
    }
    return 0;
}
int match_strsig3(firmware *fw, string_sig *sig, int j)
{
    return search_fw(fw, match_strsig3a, idx2adr(fw,j), sig->offset, 3);
}

// Sig pattern:
//      Save Regs           -   STMFD
//                ... (offset)
//      Load String Address -   ADR Rx, "func"
//                ...
//      String              -   DCB "func"
int match_strsig4a(firmware *fw, int k, uint32_t sadr, uint32_t offset)
{
    if (isSTMFD(fw,k) &&        // STMFD
        isADR_PC(fw,k+offset))  // ADR ?
    {
        uint32_t padr = ADR2adr(fw,k+offset);
        if (padr == sadr)
        {
            uint32_t fadr = idx2adr(fw,k);
            fwAddMatch(fw,fadr,32,0,104);
            return 1;
        }
    }
    return 0;
}
int match_strsig4(firmware *fw, string_sig *sig, int j)
{
    return search_fw(fw, match_strsig4a, idx2adr(fw,j), sig->offset, sig->offset+1);
}

// Sig pattern:
//      Load Func Address   -   LDR Rx, =func
//      Load String Address -   xDR Rx, "func"  (LDR or ADR)
//      Branch              -   BL
//              ...
//      String              -   DCB "func"
// or
//      Load Func Address   -   LDR Rx, =func
//                              B   loc
//              ...
//                          loc:
//      Load String Address -   xDR Rx, "func"  (LDR or ADR)
//      Branch              -   BL
//              ...
//      String              -   DCB "func"
static int dryos_ofst;
int match_strsig5a(firmware *fw, int k, uint32_t sadr, uint32_t offset)
{
    if ((isADR_PC(fw,k+1) || isLDR_PC(fw,k+1)) &&   // LDR or ADR ?
        isBorBL(fw,k+2))                            // B or BL ?
    {
        uint32_t padr;
        if (isLDR_PC(fw,k+1))                       // LDR ?
            padr = LDR2val(fw,k+1);
        else
            padr = ADR2adr(fw,k+1);
        if (padr == sadr)
        {
            int j2 = k;
            int found = 0;
            if (isLDR_PC(fw,k))                     // LDR ?
                found = 1;
            else
            {
                for (j2 = k-2; j2 >= 0 && j2 >= k-4096; j2--)
                {
                    if (isLDR_PC(fw,j2) &&  // LDR ?
                        isB(fw,j2+1))       // B
                    {
                        if (idxFollowBranch(fw,j2+1,1) == k+1)
                        {
                            found = 1;
                            break;
                        }
                    }
                }
            }
            if (found)
            {
                uint32_t fadr = LDR2val(fw,j2);
                if (offset > 1) fadr = followBranch(fw, fadr, 1);
                fadr = followBranch2(fw, fadr, offset);
                if (dryos_ofst != 0)
                {
                    uint32_t fadr2 = followBranch(fw, fadr, dryos_ofst);
                    if (fadr == fadr2) return 0;
                    fadr = fadr2;
                }
                fwAddMatch(fw,fadr,32,0,105);
                return 1;
            }
        }
    }
    return 0;
}
int match_strsig5(firmware *fw, string_sig *sig, int j)
{
    dryos_ofst = vxworks_offset(fw,sig);
    return search_fw(fw, match_strsig5a, idx2adr(fw,j), sig->offset, 3);
}

// Sig pattern:
//    Function immediately preceeding string
int match_strsig6(firmware *fw, string_sig *sig, int j)
{
    int j1 = find_inst_rev(fw, isSTMFD_LR, j-1, j-1);
    if (j1 > 0)
    {
        uint32_t fadr = idx2adr(fw,j1);
        fwAddMatch(fw,fadr,32,0,106);
        return 1;
    }

    return 0;
}

// Sig pattern:
//      Str ref -   xDRnn Rx, =str_ptr
//            ...
//                  BL  func
//            ...
//      String      DCB "str"
int match_strsig7a(firmware *fw, int k, uint32_t sadr, uint32_t offset)
{
    if (isADR_PC_cond(fw,k) || isLDR_PC_cond(fw,k)) // LDR or ADR ?
    {
        uint32_t padr;
        if (isLDR_PC_cond(fw,k)) // LDR ?
            padr = LDR2val(fw,k);
        else
            padr = ADR2adr(fw,k);
        if (padr == sadr)
        {
            int j2 = find_inst(fw, isBL, k+1, 10);
            if (j2 > 0)
            {
                uint32_t fa = idx2adr(fw,j2);
                fa = followBranch2(fw,fa,offset);
                fwAddMatch(fw,fa,32,0,107);
                return 1;
            }
        }
    }
    return 0;
}
int match_strsig7(firmware *fw, string_sig *sig, int j)
{
    return search_fw(fw, match_strsig7a, idx2adr(fw,j), sig->offset, 2);
}

// Sig pattern:
//      Special case for WriteSDCard
int ofst;
int match_strsig8(firmware *fw, string_sig *sig, int j)
{
    int j1;
    for (j1=j-2; j1<j+8; j1++)
    {
        uint32_t fadr = idx2adr(fw,j1);
        if (fwval(fw,j1) >= fw->base)   // pointer ??
        {
            int j2;
            for (j2=j1-1; j2>=j1-1000 && j2>=0; j2--)
            {
                if (isLDR_PC(fw,j2) && (LDR2adr(fw,j2) == fadr))    // LDR ?
                {
                    if ((isSTR(fw,j2+1) && (fwOp2(fw,j2+1) == ofst)) ||    // STR ?
                        (isSTR(fw,j2+2) && (fwOp2(fw,j2+2) == ofst)))      // STR ?
                    {
                        fadr = fwval(fw,j1);
                        if (idx_valid(fw,adr2idx(fw,fadr)))
                        {
                            fwAddMatch(fw,fadr,32,0,108);
                            return 1;
                        }
                    }
                }
            }
        }
    }

    return 0;
}
int find_strsig8(firmware *fw, string_sig *sig)
{
    uint32_t fadr = 0;
    int srch = 20;

    // Find "UpdateMBROnFlash" code
    int j = get_saved_sig(fw,"UpdateMBROnFlash");
    if (j >= 0)
    {
        fadr = func_names[j].val;
    }
    else
    {
        j = get_saved_sig(fw,"MakeSDCardBootable");
        if (j >= 0)
        {
            fadr = func_names[j].val;
            srch = 32;
        }
    }

    if (fadr == 0) return 0;

    int idx = adr2idx(fw, fadr);
    ofst = -1;

    for (j=idx+srch; j<idx+srch+12; j++)
    {
        if (isLDR(fw,j) && isLDR(fw,j+1) && isLDR(fw,j+2))
        {
            ofst = fwOp2(fw,j) + fwOp2(fw,j+1) + fwOp2(fw,j+2);
            break;
        }
    }

    if (ofst == -1) return 0;

    return fw_string_process(fw, sig, match_strsig8, 1);
}

// Sig pattern:
//      Func is B/BL @ offset from previous found func
//          prev_func
//              ... (offset)
//              BL    func
int find_strsig9(firmware *fw, string_sig *sig)
{
    int j = get_saved_sig(fw,sig->ev_name);
    if (j >= 0)
    {
        if (func_names[j].val != 0)
        {
            int ofst = vxworks_offset(fw, sig);
            uint32_t fadr = followBranch(fw, func_names[j].val+ofst*4, 0xF1000001);
            if ((sig->offset == -1) || (fadr != func_names[j].val+ofst*4))
            {
                uint32_t fadr2 = fadr;
                if (sig->offset != -1) fadr2 = followBranch2(fw, fadr2, sig->offset);
                if ((sig->offset <= 0) || (fadr2 != fadr))
                {
                    fwAddMatch(fw,fadr2,32,0,109);
                    return 1;
                }
            }
        }
    }

    return 0;
}

// Sig pattern:
//      Func            -   func
//            .... (offset)
//      Ref to string   -       ADR Rx, str
//            ....
//      String          -       DCB "str"
int match_strsig11(firmware *fw, string_sig *sig, int j)
{
    int ofst = vxworks_offset(fw, sig);

    uint32_t sadr = idx2adr(fw,j);        // string address
    int j1;
    for (j1 = j-1; j1 >= 0; j1--)
    {
        if (isADR_PC_cond(fw,j1))   // ADR ?
        {
            uint32_t padr = ADR2adr(fw,j1);
            if (padr == sadr)
            {
                uint32_t fadr = idx2adr(fw,j1-ofst);
                uint32_t bfadr = followBranch(fw,fadr,sig->offset);
                // special case for 'set_control_event'
                int found = 0;
                if (strcmp(sig->name,"set_control_event") == 0)
                {
                    int j2 = j1 - ofst;
                    if (isBL_cond(fw,j2) &&                                             // BLxx
                        isLDR_SP(fw,j2+1) && (fwRd(fw,j2+1) == 0) &&                    // LDR R0,[SP,x]
                        isBL(fw,j2+2) &&                                                // BL
                        isMOV(fw,j2+3) && (fwRd(fw,j2+3) == 4) && (fwRn(fw,j2+3) == 0)) // LDR R4, R0
                    {
                        found = 1;
                    }
                }
                else
                    found = 1;
                if (found && ((sig->offset == 0) || (bfadr != fadr)))
                {
                    fwAddMatch(fw,bfadr,32,0,111);
                    return 1;
                }
            }
        }
    }

    return 0;
}

//// Sig pattern:
////      Func is referenced in 'CreateJumptable'
////          LDR R1, =func
////          STR R1, [R0,nnn]
////      nnn - dryos version dependant offset
//int find_strsig12(firmware *fw, string_sig *sig)
//{
//    int j = get_saved_sig(fw,"CreateJumptable");
//
//    int ofst = vxworks_offset(fw, sig);
//
//    if (ofst == 0) return 0;
//
//    if (j >= 0)
//    {
//        if (func_names[j].val != 0)
//        {
//            int idx = adr2idx(fw, func_names[j].val);
//            for(; !isBX_LR(fw,idx); idx++)  // BX LR
//            {
//                if (((fwval(fw,idx+1) & 0xFFFFF000) == 0xE5801000) && // STR R1,[R0,nnn]
//                    (fwOp2(fw,idx+1) == ofst))
//                {
//                    uint32_t fadr = LDR2val(fw,idx);
//                    uint32_t bfadr = followBranch2(fw,fadr,sig->offset);
//                    if ((sig->offset <= 1) || ((bfadr != fadr) && ((fw->buf[adr2idx(fw,fadr)] & 0xFFFF0000) == 0xE92D0000)))
//                    {
//                        fwAddMatch(fw,bfadr,32,0,112);
//                        return 1;
//                    }
//                }
//                else if (isB(fw,idx))    // B
//                {
//                    idx = adr2idx(fw,followBranch(fw,idx2adr(fw,idx),1)) - 1;
//                }
//            }
//        }
//    }
//
//    return 0;
//}

// Sig pattern:
//      Func    -   func
//          ... (offset)
//      Str ref -       LDR Rx, =str_ptr
//          ...
//      Str ptr -       DCD str_ptr
//          ...
//      Str ptr -   str_ptr
//                      DCD str
//          ...
//      String          DCB "str"
int match_strsig13a(firmware *fw, int k, uint32_t sadr, uint32_t offset)
{
    if (fwval(fw,k) == sadr)    // string ptr
    {
        uint32_t padr = idx2adr(fw,k);      // string ptr address
        int j2;
        for (j2 = k-1; j2 >= 0; j2--)
        {
            if (fwval(fw,j2) == padr)       // string ptr address
            {
                uint32_t ppadr = idx2adr(fw,j2);        // string ptr ptr address
                int j3;
                for (j3 = j2-1; j3 >= 0; j3--)
                {
                    if (isLDR_PC(fw,j3) && (LDR2adr(fw,j3) == ppadr))
                    {
                        uint32_t fadr = idx2adr(fw,j3-offset);
                        fwAddMatch(fw,fadr,32,0,113);
                        return 1;
                    }
                }
            }
        }
    }
    return 0;
}
int match_strsig13(firmware *fw, string_sig *sig, int j)
{
    // Note - 'j' is offset in firmware not instruction index (called from fw_string_process_unaligned)
    return search_fw(fw, match_strsig13a, fw->base + j, sig->offset, 1);
}

// Sig pattern:
//                  BL  func
//            ...
//      Str ref -   xDR Rx, =str_ptr
//            ...
//      String      DCB "str"
// dryos_ofst: search range limit for the previous bl instruction
int match_strsig15a(firmware *fw, int k, uint32_t sadr, uint32_t offset)
{
    if (isADR_PC_cond(fw,k) || isLDR_PC_cond(fw,k))   // LDR or ADR ?
    {
        uint32_t padr;
        if (isLDR_PC_cond(fw,k)) // LDR ?
            padr = LDR2val(fw,k);
        else
            padr = ADR2adr(fw,k);
        if (padr == sadr)
        {
            int j2 = find_inst_rev(fw, isBL, k-1, dryos_ofst);
            if (j2 > 0)
            {
                uint32_t fa = idx2adr(fw,j2);
                fa = followBranch2(fw,fa,offset);
                fwAddMatch(fw,fa,32,0,115);
                return 1;
            }
        }
    }
    return 0;
}
int match_strsig15(firmware *fw, string_sig *sig, int j)
{
    dryos_ofst = vxworks_offset(fw,sig);
    if (dryos_ofst == 0) dryos_ofst = 50;
    return search_fw(fw, match_strsig15a, idx2adr(fw,j), sig->offset, 1);
}

// Sig pattern:
//      Function immediately preceeding usage of hex value
int match_strsig16(firmware *fw, string_sig *sig, int j)
{
    uint32_t nm0 = *((uint32_t*)sig->ev_name);

    if (isADR_PC_cond(fw,j) || isLDR_PC_cond(fw,j))   // LDR or ADR ?
    {
        uint32_t padr = 0;
        if (isLDR_PC_cond(fw,j)) // LDR ?
        {
            int k = adr2idx(fw,LDR2adr(fw,j));
            if (idx_valid(fw,k))
                padr = fwval(fw,k);
        }
        else
            padr = ADR2adr(fw,j);
        if (padr == nm0)
        {
            int j2 = find_inst_rev(fw, isSTMFD_LR, j-1, 50);
            if (j2 > 0)
            {
                uint32_t fa = idx2adr(fw,j2);
                fwAddMatch(fw,fa,32,0,116);
                return 1;
            }
        }
    }

    return 0;
}

// Sig pattern:
//      Special case for ScreenLock & ScreenUnlock
int find_strsig17(firmware *fw, string_sig *sig)
{
    int j = get_saved_sig(fw,"StartRecModeMenu");

    if (j >= 0)
    {
        if (func_names[j].val != 0)
        {
            int idx = adr2idx(fw, func_names[j].val);
            int k = 0;
            if (isLDR_PC(fw,idx-3) && isMOV_immed(fw,idx-2) && isB(fw,idx-1))
            {
                k = adr2idx(fw,LDR2val(fw,idx-3));
            }
            else if (isMOV_immed(fw,idx-3) && isADR_PC(fw,idx-2) && isB(fw,idx-1))
            {
                k = adr2idx(fw,ADR2adr(fw,idx-2));
            }
            if (k != 0)
            {
                uint32_t fadr;
                if (strcmp(sig->name,"ScreenLock") == 0)
                    fadr = followBranch(fw,idx2adr(fw,k+1),0x01000001);
                else
                {
                    k = find_inst(fw, isLDMFD, k+1, 60);
                    fadr = followBranch(fw,idx2adr(fw,k-1),0x01000001);
                }
                fwAddMatch(fw,fadr,32,0,117);
                return 1;
            }
        }
    }

    return 0;
}

// Sig pattern:
//      Func is offset from previously found func
//          prev_func
//              ... (offset, less than 0xff)
//          func
// vxworks_offset to be encoded as: 0xQRPP
// PP) offset, 0x00...0xff, in words
// Q) 'previous instruction flag': 0 for LDMFD, 1 for B
// R) 'additional gap size': 0x0 ... 0xf, to skip an additional gap, in words
// offset is negative if vxworks_offset is negative
int find_strsig19(firmware *fw, string_sig *sig)
{
    int j = get_saved_sig(fw,sig->ev_name);
    if (j >= 0)
    {
        if (func_names[j].val != 0) // address, not idx
        {
            int ofst = vxworks_offset(fw, sig);
            int neg = 1;
            if (ofst<0)
            {
                neg=-1;
                ofst=-ofst;
            }
            int addoffs = (ofst & 0xf00)>>8;
            uint32_t fadr = func_names[j].val+neg*(ofst&0xff)*4;
            int k = 0;
            switch ((ofst&0xf000)>>12)
            {
                case 0: k = isLDMFD_PC(fw, adr2idx(fw, fadr)-1-addoffs); break;
                case 1: k = isB(fw, adr2idx(fw, fadr)-1-addoffs); break;
            }
            if (k)
            {
                fwAddMatch(fw,fadr,32,0,119);
                return 1;
            }
        }
    }

    return 0;
}

// Sig pattern (Vx specific?):
//      Func            -   func
//            .... (offset)
//      Ref to string   -       LDR Rx, pstr
//            ....
//      Ptr to string   -       DCD pstr
// based on method 11
int match_strsig100(firmware *fw, string_sig *sig, int j)
{
    int ofst = vxworks_offset(fw, sig);

    uint32_t sadr = idx2adr(fw,j);        // string address
    int j1;
    for (j1 = j+256; j1 >= 0; j1--)
    {
        if (isLDR(fw,j1))   // LDR ?
        {
            uint32_t pval = LDR2val(fw,j1);
            if (pval == sadr)
            {
                uint32_t fadr = idx2adr(fw,j1-ofst);
                uint32_t bfadr = followBranch(fw,fadr,sig->offset);
                if ((sig->offset == 0) || (bfadr != fadr))
                {
                    fwAddMatch(fw,bfadr,32,0,1100);
                    return 1;
                }
            }
        }
    }

    return 0;
}

// Sig pattern (VxWorks specific):
// still needed due to ambiguous parts of code (indirect ExportToEventProcedure, RegisterEventProcedure calls)
//      String               -   DCB "func"
//              ...
//      Load String Address  -   LDR R0, ="func"
//      Load func Address    -   LDR R1, =func
// these are parameters for ExportToEventProcedure, RegisterEventProcedure
int match_strsig101(firmware *fw, string_sig *sig, int j)
{
    uint32_t sadr = idx2adr(fw,j);        // string address
    int j1;
    for (j1 = j+1024; j1 >= 0; j1--)
    {
        if (isLDR(fw,j1) && (fwRd(fw,j1)==0))   // LDR R0,
        {
            uint32_t pval = LDR2val(fw,j1);
            if (pval == sadr) // our string's address
            {
                int j2;
                for (j2 = j1+2; j2 > j1; j2--)
                {
                    if (isLDR(fw,j2) && (fwRd(fw,j2)==1)) // LDR R1,
                    {
                        uint32_t pval2 = LDR2val(fw,j2);
                        fwAddMatch(fw,pval2,32,0,1101);
                        return 1;
                    }
                }
            }
        }
    }

    return 0;
}

// Sig pattern to identify AC: functions (Vx)
// Function starts directly after its debug string
int match_strsig102(firmware *fw, string_sig *sig, int j)
{
    uint32_t sadr = idx2adr(fw,j);        // string address
    char *n = (char*)adr2ptr(fw,sadr);
    int nl = 0;
    if (n) {
        nl = strlen(n);
        nl = (((nl+3)>>2)<<2);
        if (nl > 0) {
            fwAddMatch(fw,sadr+nl,32,0,1102);
            return 1;
        }
    }
    return 0;
}

// Func is offset from a string, special case for GetDrive_FreeClusters
//      Func            -   func
//            ....
//      string          -   DCB "str"
int match_strsig103(firmware *fw, string_sig *sig, int j)
{
    int ofst = vxworks_offset(fw, sig);

    int j1 = j+ofst;
    if (isSTMFD_LR(fw,j1))
    {
        // get third BL
        j1 = find_Nth_inst(fw,isBL,j1+1,20,3);
        if (j1<0)
            return 0;
        j1 = idxFollowBranch(fw,j1,0x01000001);
        // get first B
        j1 = find_Nth_inst(fw,isB,j1+1,32,1);
        if (j1<0)
            return 0;
        j1 = idxFollowBranch(fw,j1,0x01000001);
        fwAddMatch(fw,idx2adr(fw,j1),32,0,1103);
        return 1;
    }
    return 0;
}

// Sig pattern:
//      Func            -   func
//            .... [offset]
//      Nth instruction backwards from string ref
//      Ref to string   -       LDR Rx, pstr
//            ....
//      Ptr to string   -       DCD pstr
// based on method 11
// vxworks_offset to be encoded as: 0xQNPP
// P) additional offset
// Q) instruction to locate: 0 for STMFD_LR
// N) search the instruction this many times - 1 (i.e. 0 = first instruction backwards)
// prev instruction search range is limited by sig->offset
// the reference to the string has to occur within 256 words (fixed range)
int match_strsig104(firmware *fw, string_sig *sig, int j)
{
    int ofst = vxworks_offset(fw, sig);
    int prinst = (ofst&0xf000)>>12;
    int ninst = ((ofst&0xf00)>>8)+1;
    ofst &= 0xff;
    void *instid;
    switch (prinst) {
        case 0:
        default:
            instid = (void*)isSTMFD_LR;
    }

    uint32_t sadr = idx2adr(fw,j);        // string address
    int j1;
    for (j1 = j+256; j1 >= 0; j1--)
    {
        if (isLDR(fw,j1))   // LDR ?
        {
            uint32_t pval = LDR2val(fw,j1);
            if (pval == sadr)
            {
                int j2 = find_Nth_inst_rev(fw,instid,j1-1,sig->offset,ninst);
                if (j2>0)
                {
                    fwAddMatch(fw,idx2adr(fw,j2-ofst),32,0,1104);
                    return 1;
                }
            }
        }
    }

    return 0;
}

// Call processing function based on type
int find_strsig(firmware *fw, string_sig *sig)
{
    switch (sig->type)
    {
    case 1:     return fw_string_process(fw, sig, match_strsig1, 1);
    case 2:     return fw_string_process_unaligned(fw, sig, match_strsig2);
    case 3:     return fw_string_process(fw, sig, match_strsig3, 1);
    case 4:     return fw_string_process(fw, sig, match_strsig4, 1);
    case 5:     return fw_string_process(fw, sig, match_strsig5, 1);
    case 6:     return fw_string_process(fw, sig, match_strsig6, 1);
    case 7:     return fw_string_process(fw, sig, match_strsig7, 1);
    case 8:     return find_strsig8(fw, sig);
    case 9:     return find_strsig9(fw, sig);
    case 11:    return fw_string_process(fw, sig, match_strsig11, 0);
    //case 12:    return find_strsig12(fw, sig);
    case 13:    return fw_string_process_unaligned(fw, sig, match_strsig13);
    case 15:    return fw_string_process(fw, sig, match_strsig15, 1);
    case 16:    return fw_process(fw, sig, match_strsig16);
    case 17:    return find_strsig17(fw, sig);
    case 19:    return find_strsig19(fw, sig);
    case 20:
        {
            int j = find_saved_sig(sig->ev_name);
            if (j >= 0)
            {
                uint32_t fadr = followBranch2(fw,func_names[j].val,sig->offset);
                fwAddMatch(fw,fadr,32,0,120);
                return 1;
            }
            return 0;
        }
    case 21:    return fw_process(fw, sig, (int (*)(firmware*, string_sig*, int))(sig->ev_name));
    case 22:    return ((int (*)(firmware*))(sig->ev_name))(fw);
    case 100:   return fw_string_process(fw, sig, match_strsig100, 0);
    case 101:   return fw_string_process(fw, sig, match_strsig101, 0);
    case 102:   return fw_string_process(fw, sig, match_strsig102, 0);
    case 103:   return fw_string_process(fw, sig, match_strsig103, 0);
    case 104:   return fw_string_process(fw, sig, match_strsig104, 0);
    }

    return 0;
}

//------------------------------------------------------------------------------------------------------------

// Matching functions

// Search using new style matching only
void find_str_sig_matches(firmware *fw, const char *curr_name)
{
    int i;

    int found_ev = 0;

    count = 0;

    for (i = 0; string_sigs[i].ev_name != 0 && !found_ev; i++)
    {
        if (strcmp(curr_name, string_sigs[i].name) == 0)
        {
            if (find_strsig(fw, &string_sigs[i]))
            {
                found_ev = 1;
                break;
            }
        }
    }

    if (count > 1)
    {
        qsort(matches, count, sizeof(Match), (void*)match_compare);
    }

    if (count > 0)
    {
        save_sig(curr_name, matches->ptr);
    }
}

// Search using new style matching first
// If not found used old style matching
void find_matches(firmware *fw, const char *curr_name)
{
    FuncSig *sig, *s;
    BufRange *n;
    uint32_t *p;
    int i, j;
    int fail, success;

    int found_ev = 0;

    count = 0;

    // Already found (eventproc)?
    i = find_saved_sig(curr_name);
    if (i >= 0)
    {
        if ((func_names[i].val != 0) && (func_names[i].flags & EV_MATCH) != 0)
        {
            fwAddMatch(fw,func_names[i].val,32,0,120);
            found_ev = 1;
        }
    }


    // Try and match using 'string' based signature matching first
    for (i = 0; string_sigs[i].ev_name != 0 && !found_ev; i++)
    {
        if (strcmp(curr_name, string_sigs[i].name) == 0)
        {
            if (find_strsig(fw, &string_sigs[i]))
            {
                found_ev = 1;
                break;
            }
        }
    }

    // If not found see if the name is in the old style instruction compare match table
    // Set start value for j in next section if found
    if (!found_ev)
    {
        found_ev = 1;
        for (j=0; func_list[j].name; j++)
        {
            if (strcmp(curr_name,func_list[j].name) == 0)
            {
                found_ev = 0;
                break;
            }
        }
    }

    // Not found so far, try instruction comparison matching
    while (!found_ev)
    {
        sig = func_list[j].sig;

        for (n = fw->br; n != 0; n = n->next)
        {
            for (p = n->p, i = 0; i < n->len; p++, i++)
            {
                fail = 0;
                success = 0;
                for (s = sig; s->offs != -1; s++)
                {
                    if ((p[s->offs] & s->mask) != s->value)
                        fail++;
                    else
                        success++;
                }
                // If sig starts with STMFD and first instruction does not match ignore it
                if (((p[sig->offs] & sig->mask) != sig->value) && (sig->offs == 0) && (sig->value == 0xe92d0000)) success = 0;
                if (success > fail)
                {
                    if (s->mask == -2)
                    {
                        int end_branch = 0;
                        int idx = 0;
                        uint32_t *p1 = 0;
                        if ((fw->buf[n->off+i+s->value] & 0x0F000000) == 0x0A000000)   // B
                        {
                            idx = adr2idx(fw, followBranch2(fw, idx2adr(fw,n->off+i+s->value), 0xF0000001));
                            if ((idx >= 0) && (idx < fw->size))
                            {
                                end_branch = 1;
                                p1 = &fw->buf[idx];
                            }
                        }
                        int fail2 = 0;
                        int success2 = 0;
                        //fprintf(stderr,"\t%s %d %08x %08x %d %d\n",curr_name,idx,idx2adr(fw,idx),idx2adr(fw,i+n->off),success,fail);
                        s++;
                        for (; s->offs != -1; s++)
                        {
                            if (!end_branch || (p1[s->offs] & s->mask) != s->value){
                                fail2++;
                            } else {
                                success2++;
                            }
                        }
                        if (fail2 == 0)
                        {
                            success = success + fail + success2;
                            fail = 0;
                        }
                        else
                        {
                            success = success + success2;
                            fail = fail + fail2;
                        }
                        //fprintf(stderr,"\t%s %d %08x %08x %d %d\n",curr_name,idx,idx2adr(fw,idx),idx2adr(fw,i+n->off),success,fail);
                    }
                }
                if (success > fail)
                {
                    // Special case for drive space functions, see if there is a refernce to "Mounter.c" in the function
                    // Increase match % if so, increase fail count if not
                    if ((strcmp(curr_name, "GetDrive_ClusterSize") == 0) ||
                        (strcmp(curr_name, "GetDrive_FreeClusters") == 0) ||
                        (strcmp(curr_name, "GetDrive_TotalClusters") == 0))
                    {
                        int fnd = 0;
                        for (s = sig; s->offs != -1; s++)
                        {
                            if (isLDR_PC_cond(fw,n->off+i+s->offs))
                            {
                                int m = adr2idx(fw,LDR2val(fw,n->off+i+s->offs));
                                if ((m >= 0) && (m < fw->size) && (strcmp((char*)(&fw->buf[m]),"Mounter.c") == 0))
                                {
                                    fnd = 1;
                                }
                            }
                            else if (isADR_PC_cond(fw,n->off+i+s->offs))
                            {
                                int m = adr2idx(fw,ADR2adr(fw,n->off+i+s->offs));
                                if ((m >= 0) && (m < fw->size) && (strcmp((char*)(&fw->buf[m]),"Mounter.c") == 0))
                                {
                                    fnd = 1;
                                }
                            }
                        }
                        if (fnd)
                            success++;
                        else
                            fail++;
                    }
                    fwAddMatch(fw,idx2adr(fw,i+n->off),success,fail,func_list[j].ver);
                    if (count >= MAX_MATCHES)
                    {
                        bprintf("// WARNING: too many matches for %s!\n", func_list[j].name);
                        break;
                    }
                }
            }
        }

        // same name, so we have another version of the same function
        if ((func_list[j+1].name == NULL) || (strcmp(curr_name, func_list[j+1].name) != 0))
        {
            found_ev = 1;
            break;
        }
        j++;
    }

    if (count > 1)
    {
        qsort(matches, count, sizeof(Match), (void*)match_compare);
    }

    if (count > 0)
    {
        save_sig(curr_name, matches->ptr);
    }
}

// Output match results for function
void print_results(firmware *fw, const char *curr_name, int k)
{
    int i;
    int err = 0;
    char line[500] = "";

    if (func_names[k].flags & DONT_EXPORT) return;

    // find best match and report results
    osig* ostub2 = find_sig(fw->sv->stubs,curr_name);

    if ((count == 0)
        || (matches->fail > 0)
        || (ostub2 && (matches->ptr != ostub2->val))
       )
    {
        if (!ostub2 || (ostub2->type != TYPE_IGNORE))
            err = 1;
        func_names[k].flags |= BAD_MATCH;
    }
    else
    {
        if (func_names[k].flags & UNUSED) return;
    }

    // write to header (if error) or body buffer (no error)
    out_hdr = err;

    char *macro = "NHSTUB";
    if (strncmp(curr_name,"task_",5) == 0 ||
        strncmp(curr_name,"hook_",5) == 0) macro = "   DEF";

    if (count == 0)
    {
        if (func_names[k].flags & OPTIONAL) return;
        char fmt[50] = "";
        sprintf(fmt, "// ERROR: %%s is not found. %%%ds//--- --- ", (int)(34-strlen(curr_name)));
        sprintf(line+strlen(line), fmt, curr_name, "");
    }
    else
    {
        if (ostub2 || (func_names[k].flags & UNUSED))
            sprintf(line+strlen(line),"//%s(%-37s,0x%08x) //%3d ", macro, curr_name, matches->ptr, matches->sig);
        else
            sprintf(line+strlen(line),"%s(%-39s,0x%08x) //%3d ", macro, curr_name, matches->ptr, matches->sig);

        if (matches->fail > 0)
            sprintf(line+strlen(line),"%2d%% ", matches->success*100/(matches->success+matches->fail));
        else
            sprintf(line+strlen(line),"    ");
    }

    if (ostub2)
    {
        if (ostub2->type == TYPE_IGNORE)
            sprintf(line+strlen(line),"       Overridden");
        else if ((count > 0) && (matches->ptr == ostub2->val))
            sprintf(line+strlen(line),"       == 0x%08x    ",ostub2->val);
        else
            sprintf(line+strlen(line),"   *** != 0x%08x    ",ostub2->val);
    }
    else
        sprintf(line+strlen(line),"                        ");

    for (i=strlen(line)-1; i>=0 && line[i]==' '; i--) line[i] = 0;
    bprintf("%s\n",line);

    for (i=1;i<count && matches[i].fail==matches[0].fail;i++)
    {
        if (matches[i].ptr != matches->ptr)
        {
            bprintf("// ALT: %s(%s, 0x%x) // %d %d/%d\n", macro, curr_name, matches[i].ptr, matches[i].sig, matches[i].success, matches[i].fail);
        }
    }
}

//------------------------------------------------------------------------------------------------------------

typedef struct
{
    uint16_t mode;
    char *nm;
} ModeMapName;

ModeMapName mmnames[] = {
    { 32768,"MODE_AUTO" },
    { 32769,"MODE_M" },
    { 32770,"MODE_AV" },
    { 32771,"MODE_TV" },
    { 32772,"MODE_P" },

    { 65535,"" }
};

char* mode_name(uint16_t v)
{
    int i;
    for (i=0; mmnames[i].mode != 65535; i++)
    {
        if (mmnames[i].mode == v)
            return mmnames[i].nm;
    }

    return "";
}

void output_modemap(firmware *fw, int k, int l)
{
    int cnt = 0;

    bprintf("%08x\n",idx2adr(fw,k));
    // detect table member size
    if (fw->buf[k] & 0xffff0000)
    {
        // 16bit entries
        uint16_t *p = (uint16_t*)(&fw->buf[k]);
        k = 0;
        while ((*p != 0xFFFF) && (k < l*2))
        {
            osig *m = find_sig_val(fw->sv->modemap, *p);
            if (!m)
            {
                char *s = mode_name(*p);
                bprintf("// Mode %5d in firmware but not in current modemap",*p);
                if (strcmp(s,"") != 0)
                    bprintf(" (%s)",s);
                bprintf("\n");
                cnt++;
            }
            else
            {
                m->pct = 100;
            }
            p++;
            k++;
        }
    }
    else
    {
        // 32bit entries (first one is 0)
        uint32_t *p = (uint32_t*)(&fw->buf[k]);
        k = 0;
        while (k < l)
        {
            osig *m = find_sig_val(fw->sv->modemap, *p);
            if (!m)
            {
                osig *m = find_sig_val(fw->sv->modemap, (*p)&0xffff);
                if (!m)
                {
                    char *s = mode_name(*p);
                    if ((*p)&0xffff0000)
                        bprintf("// Strange mode 0x%08x in firmware but not in current modemap",*p);
                    else
                        bprintf("// Mode %d in firmware but not in current modemap",*p);
                    if (strcmp(s,"") != 0)
                        bprintf(" (%s)",s);
                    bprintf("\n");
                }
                else
                {
                    char *s = mode_name((*p)&0xffff);
                    if ((*p)&0xffff0000)
                        bprintf("// Strange mode 0x%08x in firmware, found in modemap as %d (%s)",*p,m->val,m->nm);
                    else
                        bprintf("// Mode %d in firmware, found in modemap as %d (%s)",*p,m->val,m->nm); // impossible?
                    if (strcmp(s,"") != 0)
                        bprintf(" (%s)",s);
                    bprintf("\n");
                    m->pct = 100;
                }
                cnt++;
            }
            else
            {
                m->pct = 100;
            }
            p++;
            k++;
        }
    }
    osig *m = fw->sv->modemap;
    while (m)
    {
        if (m->pct != 100)    // not matched above?
        {
            bprintf("// Current modemap entry not found in firmware - %-24s %5d\n",m->nm,m->val);
            cnt++;
        }
        m = m->nxt;
    }
    if (cnt == 0)
    {
        bprintf("// No problems found with modemap table.\n");
    }
}

int find_modelist(firmware *fw, uint32_t fadr)
{
    uint32_t j1;
    int k1, k2;
    // flashparamstable 1st entry
    j1 = fwval(fw,adr2idx(fw,fadr));
    k1 = adr2idx(fw,j1);
    if (k1<0)
        return 0;
    k2 = find_inst_rev(fw,isLDMFD_PC,k1-1,50);
    if (k2<0)
    {
        // newest vx models
        k2 = find_Nth_inst_rev(fw,isMOV,k1-1,16,2);
    }
    if (k2<0)
        return 0;
    k1 = find_inst_rev(fw,isLDR,k2,5);
    if (k1<0)
        return 0;
    j1 = LDR2val(fw,k1);
    k1 = adr2idx(fw,j1);
    if (k1<0)
        return 0;
    j1 = fwval(fw,k1);
    k1 = adr2idx(fw,j1);
    if (k1<0)
        return 0;
    k1 = find_inst_rev(fw,isMOV,k1-1,120);
    if (k1<0)
        return 0;
    if (fwval(fw,k1) != 0xe1a0f00e) // mov pc, lr
        return 0;
    k1--;
    if (!(isLDR(fw,k1) && (fwRd(fw,k1)==0))) // ldr r0,
        return 0;
    k2 = k1;
    j1 = LDR2val(fw,k1);
    k1 = adr2idx(fw,j1);
    if (k1<0)
        return 0;
    
    bprintf("// Firmware modemap table found @%08x -> ",idx2adr(fw,k1));
    output_modemap(fw,k1,k2-k1);
    return 1;

}

static uint32_t FlashParamsTable_address = 0;

int match_FlashParamsTable2(firmware *fw, int k, uint32_t v1, uint32_t v2)
{
    if (fw->buf[k] == v1)
    {
        FlashParamsTable_address = idx2adr(fw,k);
        return 1;
    }
    return 0;
}

int match_FlashParamsTable(firmware *fw, int k, uint32_t v1, uint32_t v2)
{
    if ((fw->buf[k] > fw->base) && (fw->buf[k+1] == 0x00010000) && ((fw->buf[k+2] == 0xFFFF0000)||(fw->buf[k+2] == 0xFFFF0002)))
    {
        if (search_fw(fw, match_FlashParamsTable2, idx2adr(fw,k), 0, 1))
            return 1;
    }
    // ixus30, 40
    if ((fw->buf[k+1] > fw->base) && (fw->buf[k] == 0x00000000) && (fw->buf[k+2] == 0x00000001) &&
        (fw->buf[k+3] == 0x00000000) && (fw->buf[k+4] == 0xffffffff))
    {
        if (search_fw(fw, match_FlashParamsTable2, idx2adr(fw,k), 0, 1))
            return 1;
    }
    return 0;
}

void find_modemap(firmware *fw)
{
    out_hdr = 1;

    // Find 'FlashParamsTable'
    search_fw(fw, match_FlashParamsTable, 0, 0, 1);
    // Find 'modelist'
    // get flashparamstable 1st entry -> jump to that address -> search backwards 1st ldmfd_pc (mov pc,lr in newest vx) ->
    // search backwards ldr address -> jump to that address (this is the propertytable, not researched) ->
    // get 1st entry -> jump to that address -> search backwards 1st mov pc,lr ->
    // get modemaptable address from the prev ldr
    // old vx mode entries are 32 bit, all newer cameras have 16 bit entries
    if (FlashParamsTable_address != 0)
    {
        add_blankline();
        bprintf("// Check of modemap from 'platform/CAMERA/shooting.c':\n");
        find_modelist(fw,FlashParamsTable_address);
    }
}

//------------------------------------------------------------------------------------------------------------

int find_FileAccessSem(firmware *fw)
{
    void print_stubs_min(firmware *fw, const char *name, uint32_t fadr, uint32_t atadr);
    int s1 = find_str(fw, "FileSem.c");
    if (s1 < 0)
        return 0;
    s1 = find_inst(fw, isLDR_PC, s1+2, 16);
    if (s1 < 0)
        return 0;
    uint32_t u1 = LDR2val(fw, s1);
    if (u1 > fw->memisostart)
        return 0;
    print_stubs_min(fw,"fileio_semaphore",u1,idx2adr(fw,s1));
    return 1;
}

//------------------------------------------------------------------------------------------------------------

/*
int match_CAM_UNCACHED_BIT(firmware *fw, int k, int v)
{
    if ((fw->buf[k] & 0x0FFFF000) == 0x03C00000)    // BIC
    {
        uint32_t fadr = ALUop2(fw,k);
        bprintf("//#undef  CAM_UNCACHED_BIT\n");
        bprintf("//#define CAM_UNCACHED_BIT  0x%08x // Found @0x%08x\n",fadr,idx2adr(fw,k));
    }

    return 0;
}
*/

// Search for things that go in 'platform_camera.h'
void find_platform_vals(firmware *fw)
{
    int k,k1;

    out_hdr = 1;
    add_blankline();

    bprintf("// Values below go in 'platform_camera.h':\n");

/*
    bprintf("//#define CAM_DRYOS         1\n");
    if (fw->dryos_ver >= 39)
        bprintf("//#define CAM_DRYOS_2_3_R39 1 // Defined for cameras with DryOS version R39 or higher\n");
    if (fw->dryos_ver >= 47)
        bprintf("//#define CAM_DRYOS_2_3_R47 1 // Defined for cameras with DryOS version R47 or higher\n");

    // Find 'RAW' image size
    uint32_t raw_width = 0;
    uint32_t raw_height = 0;
    uint32_t kw=0, kh=0;

    k = find_str_ref(fw, "\r[%ld] AdjDrvType[%02ld] -> DrvType[%02");
    if (k >= 0)
    {
        // Width
        for (k1 = k-1; k1 >= k-20; k1--)
        {
            if ((fw->buf[k1] & 0x0FFF0FFF) == 0x058D0034)           // STRxx Rn, [SP,#0x34]
            {
                if ((fw->buf[k1-1] & 0x0FFF0000) == 0x03A00000)     // MOVxx Rn, #YYY
                {
                    raw_width = ALUop2(fw, k1-1);
                    kw = k1-1;
                }
                else if ((fw->buf[k1-2] & 0x0FFF0000) == 0x03A00000)// MOVxx Rn, #YYY
                {
                    raw_width = ALUop2(fw, k1-2);
                    kw = k1-2;
                }
                else if (isLDR_PC_cond(fw,k1-1))
                {
                    raw_width = LDR2val(fw,k1-1);
                    kw = k1-1;
                }
                else if (isLDR_PC_cond(fw,k1-2))
                {
                    raw_width = LDR2val(fw,k1-2);
                    kw = k1-2;
                }
            }
        }
        // Height
        for (k1 = k-1; k1 >= k-20; k1--)
        {
            if ((fw->buf[k1] & 0x0FFF0FFF) == 0x058D0030)           // STRxx Rn, [SP,#0x30]
            {
                if ((fw->buf[k1-1] & 0x0FFF0000) == 0x03A00000)     // MOVxx Rn, #YYY
                {
                    raw_height = ALUop2(fw, k1-1);
                    kh = k1-1;
                }
                else if ((fw->buf[k1-2] & 0x0FFF0000) == 0x03A00000)// MOVxx Rn, #YYY
                {
                    raw_height = ALUop2(fw, k1-2);
                    kh = k1-2;
                }
                else if (isLDR_PC_cond(fw,k1-1))
                {
                    raw_height = LDR2val(fw,k1-1);
                    kh = k1-1;
                }
                else if (isLDR_PC_cond(fw,k1-2))
                {
                    raw_height = LDR2val(fw,k1-2);
                    kh = k1-2;
                }
                if ((fw->buf[k1-1] & 0x0FFF0000) == 0x02400000)     // SUBxx Rn, #YYY
                {
                    raw_height = raw_width - ALUop2(fw, k1-1);
                    kh = k1-1;
                }
            }
        }
    }

    if ((raw_width == 0) && (raw_height == 0))
    {
        k = find_str_ref(fw, " CrwAddress %lx, CrwSize H %ld V %ld\r");
        if (k >= 0)
        {
            // Width
            for (k1=k-1; k1>=k-5; k1--)
            {
                if ((fw->buf[k1] & 0xFFFFF000) == 0xE3A02000)       // MOV R2, #nnn
                {
                    raw_width = ALUop2(fw,k1);
                    kw = k1;
                }
                else
                if (isLDR_PC(fw,k1) && ((fw->buf[k1]& 0x0000F000) == 0x00002000))   // LDR R2, =nnn
                {
                    raw_width = LDR2val(fw,k1);
                    kw = k1;
                }
            }
            // Height
            for (k1=k-1; k1>=k-5; k1--)
            {
                if ((fw->buf[k1] & 0xFFFFF000) == 0xE3A03000)       // MOV R3, #nnn
                {
                    raw_height = ALUop2(fw,k1);
                    kh = k1;
                }
                else
                if (isLDR_PC(fw,k1) && ((fw->buf[k1]& 0x0000F000) == 0x00003000))   // LDR R3, =nnn
                {
                    raw_height = LDR2val(fw,k1);
                    kh = k1;
                }
                else
                if ((fw->buf[k1] & 0xFFFFF000) == 0xE2423000)       // SUB R3, R2, #nnn
                {
                    raw_height = raw_width - ALUop2(fw,k1);
                    kh = k1;
                }
            }
        }
    }

    if (raw_width != 0)
    {
        bprintf("//#define CAM_RAW_ROWPIX    %d // Found @0x%08x\n",raw_width,idx2adr(fw,kw));
    }
    else
    {
        bprintf("//#define CAM_RAW_ROWPIX    *** Not Found ***\n");
    }
    if (raw_height != 0)
    {
        bprintf("//#define CAM_RAW_ROWS      %d // Found @0x%08x\n",raw_height,idx2adr(fw,kh));
    }
    else
    {
        bprintf("//#define CAM_RAW_ROWS      *** Not Found ***\n");
    }

    // Find 'CAM_UNCACHED_BIT'
    search_saved_sig(fw, "FreeUncacheableMemory", match_CAM_UNCACHED_BIT, 0, 0, 8);
*/

    // Find 'PARAM_CAMERA_NAME'
    if (FlashParamsTable_address != 0)
    {
        k1 = adr2idx(fw,FlashParamsTable_address);
        for (k=k1; k<k1+20; k++)
        {
            uint32_t fadr = fwval(fw,k);
            int k2 = adr2idx(fw,fadr);
            if (idx_valid(fw,k2))
            {
                uint32_t sadr = fwval(fw,k2);
                k2 = adr2idx(fw,sadr);
                if (idx_valid(fw,k2))
                {
                    char *s = adr2ptr(fw,sadr);
                    if (((fw->cam != 0) && (strcmp(s,fw->cam) == 0)) || (strcmp(s,"Unknown") == 0))
                    {
                        bprintf("//#define PARAM_CAMERA_NAME %d // Found @0x%08x\n",k-k1,fadr);
                        break;
                    }
                }
            }
        }
    }
}

//------------------------------------------------------------------------------------------------------------

uint32_t find_viewport_address(firmware *fw, int *kout)
{
    int k, k1;

    // find viewwport address for 'vid_get_viewport_fb'
    k = find_str_ref(fw, "VRAM Address  : %p\r");
    if (k >= 0)
    {
        for (k1=k-1; k1>k-8; k1--)
        {
            if (isLDR(fw,k1) && isLDR(fw,k1+1))
            {
                uint32_t v1 = LDR2val(fw,k1);
                uint32_t v2 = LDR2val(fw,k1+1);
                if (v2 > v1) v1 = v2;
                *kout = k1;
                return v1;
            }
        }
    }

    *kout = -1;
    return 0;
}

int match_vid_get_bitmap_fb(firmware *fw, int k, int v)
{
    if (isBL(fw,k-1) && // BL
        isLDR_PC(fw,k))
    {
        uint32_t v1 = LDR2val(fw,k);
        bprintf("//void *vid_get_bitmap_fb()        { return (void*)0x%08x; }             // Found @0x%08x\n",v1,idx2adr(fw,k));
        return 1;
    }
    else
    if (isBL(fw,k-1) && // BL
        (isLDR_PC(fw,k+1)))
    {
        uint32_t v1 = LDR2val(fw,k+1);
        bprintf("//void *vid_get_bitmap_fb()        { return (void*)0x%08x; }             // Found @0x%08x\n",v1,idx2adr(fw,k));
        return 1;
    }

    return 0;
}

int match_get_flash_params_count(firmware *fw, int k, int v)
{
    if ((fw->buf[k] & 0xFFF00FFF) == 0xE3C00901)    // BIC Rn, Rn, #0x4000
    {
        uint32_t r = (fw->buf[k] & 0x0000F000)<<4;       // Register
        int n;
        for (n=k+1; n<k+20; n++)
        {
            if (((fw->buf[n] & 0xFFF00000) == 0xE3500000) && ((fw->buf[n] & 0x000F0000) == r))  // CMP, Rn #val
            {
                // the found value needs to be increased by one (BLS instruction follows)
                bprintf("//int get_flash_params_count(void) { return 0x%02x; }                          // Found @0x%08x\n",(fw->buf[n]&0xFFF)+1,idx2adr(fw,n));
                return 1;
            }
        }
    }

    return 0;
}

// based on match_get_flash_params_count
int match_uiprop_count(firmware *fw, int k, int v)
{
    uint32_t uic = 0;
    int j = -1;
    if (isB(fw, k+6))
    {
        j = idxFollowBranch(fw,k+6,0x01000001);
    }
    if (j > 0)
    {
        int m;
        for (m=j; m<j+32; m++)
        {
            if (isLDMFD_PC(fw,m)) break;
            if (isCMP(fw,m))
            {
                uint32_t v1 = ALUop2(fw,m);
                if ((v1 > 0x10)&&(v1 < 0xa0))
                {
                    uic = v1+1; // BLS follows this CMP instruction
                    j = m;
                    break;
                }
            }
        }
    }

    if (uic == 0) return 0;

    char *name = "uiprop_count";
    osig *o = find_sig(fw->sv->stubs_min,name);
    if (o)
    {
        bprintf("//DEF_CONST(%-34s,0x%08x) // Found @0x%08x",name,uic,idx2adr(fw,j));
        if (uic != o->val)
        {
            bprintf(", ** != ** stubs_min = 0x%08x (%s)",o->val,o->sval);
        }
        else
        {
            bprintf(",          stubs_min = 0x%08x (%s)",o->val,o->sval);
        }
    }
    else
    {
        bprintf("DEF_CONST(%-34s,0x%08x) // Found @0x%08x",name,uic,idx2adr(fw,j));
    }
    bprintf("\n");
    return 1;
}

// Search for things that go in 'lib.c'
void find_lib_vals(firmware *fw)
{
    //int k,k1;

    out_hdr = 1;
    add_blankline();

    bprintf("// Values below go in 'lib.c':\n");
/*
    // Find 'vid_get_bitmap_fb'
    search_saved_sig(fw, "DispCon_ShowBitmapColorBar", match_vid_get_bitmap_fb, 0, 1, 30);

    // find 'vid_get_viewport_fb'
    uint32_t v = find_viewport_address(fw,&k);
    if (k >= 0)
    {
        bprintf("//void *vid_get_viewport_fb()      { return (void*)0x%08x; }             // Found @0x%08x\n",v,idx2adr(fw,k));
    }

    // find 'vid_get_viewport_fb_d'
    static int fbd[3][3] =
    {
        { -2, -3,  1 },
        {  1,  3,  4 },
        { -1, -2,  1 },
    };
    int sadr = find_str(fw, "ImagePlayer.c");
    k = find_nxt_str_ref(fw, sadr, -1);
    int found = 0;
    while ((k >= 0) && !found)
    {
        int f;
        for (f=0; f<3 && !found; f++)
        {
            if (isLDR(fw,k+fbd[f][0]) && isLDR(fw,k+fbd[f][1]) && isLDR(fw,k+fbd[f][2]))
            {
                int reg = fw->buf[k+fbd[f][2]] & 0x000F0000;    // Index register used
                int ka = 0;
                if (((fw->buf[k+fbd[f][0]] & 0x0000F000) << 4) == reg)      { ka = k+fbd[f][0]; }
                else if (((fw->buf[k+fbd[f][1]] & 0x0000F000) << 4) == reg) { ka = k+fbd[f][1]; }
                if (ka > 0)
                {
                    uint32_t adr = LDR2val(fw,ka);
                    for (k1=k+2; k1<k+20; k1++)
                    {
                        if (isSTR(fw,k1) && ((fw->buf[k1] & 0x000F0000) == reg))
                        {
                            uint32_t ofst = fw->buf[k1] & 0x00000FFF;
                            bprintf("DEF(%-40s,0x%08x) // Found 0x%04x (@0x%08x) + 0x%02x (@0x%08x)\n","viewport_fb_d",adr+ofst,adr,idx2adr(fw,ka),ofst,idx2adr(fw,k1));
                            bprintf("//void *vid_get_viewport_fb_d()    { return (void*)(*(int*)(0x%04x+0x%02x)); } // Found @0x%08x & 0x%08x\n",adr,ofst,idx2adr(fw,ka),idx2adr(fw,k1));
                            found = 1;
                            break;
                        }
                    }
                }
            }
        }
        k = find_nxt_str_ref(fw, sadr, k);
    }

    // find 'camera_jpeg_count_str'
    k = find_str_ref(fw, "9999");
    if (k >= 0)
    {
        if (isLDR(fw,k-1) && isBL(fw,k+1))
        {
            uint32_t v1 = LDR2val(fw,k-1);
            bprintf("DEF(%-40s,0x%08x) // Found @0x%08x\n","jpeg_count_str",v1,idx2adr(fw,k-1));
            bprintf("//char *camera_jpeg_count_str()    { return (char*)0x%08x; }             // Found @0x%08x\n",v1,idx2adr(fw,k-1));
        }
    }

    // find 'hook_raw_size'
    k = find_str_ref(fw, "CRAW BUFF SIZE  %p");
    if (k >= 0)
    {
        if (isLDR(fw,k-1))
        {
            uint32_t craw_bufsize = LDR2val(fw,k-1);
            bprintf("//long hook_raw_size()             { return 0x%08x; }                    // Found @0x%08x\n",craw_bufsize,idx2adr(fw,k-1));
        }
    }
*/
    // Find value for 'get_flash_params_count'
    search_saved_sig(fw, "GetParameterData", match_get_flash_params_count, 0, 0, 30);
}

//------------------------------------------------------------------------------------------------------------

void print_stubs_min(firmware *fw, const char *name, uint32_t fadr, uint32_t atadr)
{
    osig *o = find_sig(fw->sv->stubs_min,name);
    if (o)
    {
        bprintf("//DEF(%-40s,0x%08x) // Found @0x%08x",name,fadr,atadr);
        if (fadr != o->val)
        {
            bprintf(", ** != ** stubs_min = 0x%08x (%s)",o->val,o->sval);
        }
        else
        {
            bprintf(",          stubs_min = 0x%08x (%s)",o->val,o->sval);
        }
    }
    else
    {
        bprintf("DEF(%-40s,0x%08x) // Found @0x%08x",name,fadr,atadr);
    }
    bprintf("\n");
}

int match_levent_table(firmware *fw, int k, uint32_t v1, uint32_t v2)
{
    if ((fw->buf[k] > fw->base) && (fw->buf[k+1] == 0x00000800) && (fw->buf[k+2] == 0x00000002))
    {
        print_stubs_min(fw,"levent_table",idx2adr(fw,k),idx2adr(fw,k));
#ifdef PRINT_LEVENT_TABLE
        uint32_t levent_tbl = idx2adr(fw,k);
        void write_levent_table_dump(firmware*, uint32_t);
        write_levent_table_dump(fw, levent_tbl);
#endif
    }
    return 0;
}

int match_movie_status(firmware *fw, int k, uint32_t v1, uint32_t v2)
{
    if (isLDR_PC(fw, k) &&                              // LDR R0, =base
        ((fw->buf[k+1] & 0xFE0F0000) == 0xE20F0000) &&  // ADR R1, =sub
        isSTR(fw, k+2) &&                               // STR R1, [R0,N]
        (fw->buf[k+3] == 0xE3A01003) &&                 // MOV R1, 3
        isSTR(fw, k+4) &&                               // STR R1, [R0,ofst]
        (LDR2val(fw,k) < fw->base))
    {
        uint32_t base = LDR2val(fw,k);
        uint32_t ofst = fw->buf[k+4] & 0x00000FFF;
        print_stubs_min(fw,"movie_status",base+ofst,idx2adr(fw,k));
        return 1;
    }
    else
    if (isLDR_PC(fw, k) &&                              // LDR R1, =sub
        isLDR_PC(fw, k+1) &&                            // LDR R0, =base
        isSTR(fw, k+2) &&                               // STR R1, [R0,N]
        (fw->buf[k+3] == 0xE3A01003) &&                 // MOV R1, 3
        isSTR(fw, k+4) &&                               // STR R1, [R0,ofst]
        (LDR2val(fw,k+1) < fw->base))
    {
        uint32_t base = LDR2val(fw,k+1);
        uint32_t ofst = fw->buf[k+4] & 0x00000FFF;
        print_stubs_min(fw,"movie_status",base+ofst,idx2adr(fw,k));
        return 1;
    }
    else
    if (isLDR_PC(fw, k) &&                                      // LDR Rx, =base
        isLDR(fw, k+1) && (fwRd(fw,k) == fwRn(fw,k+1)) &&       // LDR R0, [Rx, ...]
        isCMP(fw, k+2) && (fwRd(fw,k+2) == fwRd(fw,k+1)) &&     // CMP R0, #...
        (fwval(fw,k+3) == 0x03A00005) &&
        isSTR_cond(fw, k+4) && (fwRn(fw,k+4) == fwRd(fw,k)) &&  // STRxx R0, [Rx,ofst]
        (LDR2val(fw,k) < fw->base))
    {
        uint32_t base = LDR2val(fw,k);
        uint32_t ofst = fwOp2(fw,k+4);
        print_stubs_min(fw,"movie_status",base+ofst,idx2adr(fw,k));
        return 1;
    }
    return 0;
}

int match_full_screen_refresh(firmware *fw, int k, uint32_t v1, uint32_t v2)
{
    if (((fw->buf[k] & 0xFF1FF000) == 0xE51F0000) &&    // LDR R0, =base
        (fw->buf[k+1] == 0xE5D01000) &&                 // LDRB R1, [R0]
        (fw->buf[k+2] == 0xE3811002) &&                 // ORR R1, R1, #2
        (fw->buf[k+3] == 0xE5C01000) &&                 // STRB R1, [R0]
        isBX_LR(fw,k+4))                                // BX LR
    {
        uint32_t base = LDR2val(fw,k);
        print_stubs_min(fw,"full_screen_refresh",base,idx2adr(fw,k));
    }
    return 0;
}

int match_canon_shoot_menu_active(firmware *fw, int k, uint32_t v1, uint32_t v2)
{
    if (((fw->buf[k]   & 0xFF1FF000) == 0xE51F1000) &&  // LDR R1, =base
        ((fw->buf[k+1] & 0xFFFFF000) == 0xE5D10000) &&  // LDRB R0, [R1, #n]
        (fw->buf[k+2] == 0xE2800001) &&                 // ADD R0, R0, #1
        ((fw->buf[k+3] & 0xFFFFF000) == 0xE5C10000) &&  // STRB R0, [R1, #n]
        (isB(fw,k+4)))                                  // B
    {
        uint32_t base = LDR2val(fw,k);
        uint32_t ofst = fw->buf[k+1] & 0x00000FFF;
        print_stubs_min(fw,"canon_shoot_menu_active",base+ofst,idx2adr(fw,k));
    }
    else
    if (((fw->buf[k]   & 0xFF1FF000) == 0xE51F0000) &&  // LDR R0, =base
        ((fw->buf[k+1] & 0xFFFFF000) == 0xE5D01000) &&  // LDRB R1, [R0, #n]
        (fw->buf[k+2] == 0xE2811001) &&                 // ADD R1, R1, #1
        ((fw->buf[k+3] & 0xFFFFF000) == 0xE5C01000) &&  // STRB R1, [R0, #n]
        (isB(fw,k+4)))                                  // B
    {
        uint32_t base = LDR2val(fw,k);
        uint32_t ofst = fw->buf[k+1] & 0x00000FFF;
        print_stubs_min(fw,"canon_shoot_menu_active",base+ofst,idx2adr(fw,k));
    }
    return 0;
}

int match_playrec_mode(firmware *fw, int k, uint32_t v1, uint32_t v2)
{
    if (((fw->buf[k]    & 0xFF1FF000) == 0xE51F1000) && // LDR R1, =base
        ((fw->buf[k+1]  & 0xFFFFF000) == 0xE5810000) && // STR R0, [R1, #n]
        ((fw->buf[k+3]  & 0xFF1FF000) == 0xE51F0000) && // LDR R0, =base
        ((fw->buf[k+4]  & 0xFFFFF000) == 0xE5900000) && // LDR R0, [R0, #n]
        ((fw->buf[k+6]  & 0xFF1FF000) == 0xE51F1000) && // LDR R1, =base
        ((fw->buf[k+9]  & 0xFF1FF000) == 0xE51F0000) && // LDR R0, =base
        ((fw->buf[k+12] & 0xFF1FF000) == 0xE51F1000) && // LDR R1, =base
        ((fw->buf[k+15] & 0xFF1FF000) == 0xE51F0000) && // LDR R0, =base
        ((fw->buf[k+18] & 0xFF1FF000) == 0xE51F1000) && // LDR R1, =base
        (LDR2val(fw,k) == LDR2val(fw,k+3)) &&
        (LDR2val(fw,k) == LDR2val(fw,k+6)) &&
        (LDR2val(fw,k) == LDR2val(fw,k+9)) &&
        (LDR2val(fw,k) == LDR2val(fw,k+12)) &&
        (LDR2val(fw,k) == LDR2val(fw,k+15)) &&
        (LDR2val(fw,k) == LDR2val(fw,k+18)))
    {
        uint32_t base = LDR2val(fw,k);
        uint32_t ofst = fw->buf[k+1] & 0x00000FFF;
        print_stubs_min(fw,"playrec_mode",base+ofst,idx2adr(fw,k));
    }
    return 0;
}

int match_some_flag_for_af_scan(firmware *fw, int k, uint32_t v1, uint32_t v2)
{
    if (isB(fw,k)   &&  // B loc
        isB(fw,k+1) &&  // B loc
        isB(fw,k+2) &&  // B loc
        isB(fw,k+3) &&  // B loc
        isB(fw,k+4) &&  // B loc
        isB(fw,k+5) &&  // B loc
        isB(fw,k+6) &&  // B loc
        isB(fw,k+7) &&  // B loc
        (followBranch(fw,idx2adr(fw,k),1) != followBranch(fw,idx2adr(fw,k+1),1)) &&
        (followBranch(fw,idx2adr(fw,k),1) == followBranch(fw,idx2adr(fw,k+2),1)) &&
        (followBranch(fw,idx2adr(fw,k),1) == followBranch(fw,idx2adr(fw,k+3),1)) &&
        (followBranch(fw,idx2adr(fw,k),1) != followBranch(fw,idx2adr(fw,k+4),1)) &&
        (followBranch(fw,idx2adr(fw,k),1) != followBranch(fw,idx2adr(fw,k+5),1)) &&
        (followBranch(fw,idx2adr(fw,k),1) == followBranch(fw,idx2adr(fw,k+6),1)) &&
        (followBranch(fw,idx2adr(fw,k),1) == followBranch(fw,idx2adr(fw,k+7),1)) &&
        (isLDR_PC(fw,adr2idx(fw,followBranch(fw,idx2adr(fw,k),1)))))    // LDR R0, =base
    {
        uint32_t base = LDR2val(fw,adr2idx(fw,followBranch(fw,idx2adr(fw,k),1)));
        if (base < fw->base)
            print_stubs_min(fw,"some_flag_for_af_scan",base,followBranch(fw,idx2adr(fw,k),1));
    }
    return 0;
}

int match_palette_data(firmware *fw, int k, uint32_t v1, uint32_t v2)
{
    if ((fw->buf[k] == 0) && (fw->buf[k+1] == 0x00FF0000) &&
        (fw->buf[k+577] == 1) && (fw->buf[k+578] == 0x00FF0000) &&
        (fw->buf[k+1154] == 2) && (fw->buf[k+1155] == 0x00FF0000))
    {
        return idx2adr(fw,k);
    }
    else if ((fw->buf[k] == 0) && (fw->buf[k+1] == 0x00FF0000) &&
             (fw->buf[k+513] == 1) && (fw->buf[k+514] == 0x00FF0000) &&
             (fw->buf[k+1026] == 2) && (fw->buf[k+1027] == 0x00FF0000))
    {
        return idx2adr(fw,k);
    }
    return 0;
}

int match_palette_buffer_offset(firmware *fw, int k)
{
    int idx2 = idxFollowBranch(fw, k, 0x01000001);
    if (isLDR(fw, idx2+2) && isBL(fw, idx2+3))
    {
        uint32_t palette_size = LDR2val(fw,idx2+2);
        if (palette_size >= 0x400)
        {
            bprintf("// Offset from start of palette_buffer to color data = %d (Found @0x%08x)\n",palette_size-0x400,idx2adr(fw,idx2+2));
            return 1;
        }
    }
    else if (isADR(fw, idx2+2) && isBL(fw, idx2+3))
    {
        uint32_t palette_size = ALUop2(fw,idx2+2);
        if (palette_size >= 0x400)
        {
            bprintf("// Offset from start of palette_buffer to color data = %d (Found @0x%08x)\n",palette_size-0x400,idx2adr(fw,idx2+2));
            return 1;
        }
    }
    return 0;
}

int match_palette_data3(firmware *fw, int k, uint32_t palette_data, uint32_t v2)
{
    if (isLDR_PC(fw, k) && (LDR2val(fw,k) == palette_data) && isLDR_PC(fw,k-1) && isLDR_PC(fw,k-6) && isLDR(fw,k-5))
    {
        int palette_control = LDR2val(fw,k-6);
        int ptr_offset = fwOp2(fw,k-5);
        uint32_t fadr = find_inst_rev(fw, isSTMFD_LR, k-7, 30);
        if (fadr > 0)
        {
            int k1 = search_fw(fw, find_B, fadr, 0, 1);
            if ((k1 > 0) && isLDR_PC(fw,k1-2) && isLDR(fw,k1-1) && (LDR2val(fw,k1-2) == palette_control))
            {
                int active_offset = fwOp2(fw,k1-1);
                print_stubs_min(fw,"active_palette_buffer",palette_control+active_offset,idx2adr(fw,k1-1));
                print_stubs_min(fw,"palette_buffer_ptr",palette_control+ptr_offset,idx2adr(fw,k-5));
                if (isBL(fw,k+8))
                {
                    fadr = followBranch(fw, idx2adr(fw,k+8), 0x01000001);
                    int idx = adr2idx(fw, fadr);
                    if (isLDR(fw, idx+2) && isBL(fw, idx+3))
                    {
                        uint32_t palette_size = LDR2val(fw,idx+2);
                        if (palette_size >= 0x400)
                        {
                            bprintf("// Offset from start of palette_buffer to color data = %d (Found @0x%08x)\n",palette_size-0x400,idx2adr(fw,idx+2));
                        }
                    }
                }
                return 1;
            }
        }
    }
    return 0;
}

int match_palette_data2(firmware *fw, int k, uint32_t v1, uint32_t v2)
{
    if (isLDR(fw,k) && (LDR2val(fw,k) == v1))
    {
        int k1;
        for (k1=k+1; k1<k+20; k1++)
        {
            if (isBL(fw,k1) && isLDMFD(fw,k1+2))
            {
                uint32_t fadr = followBranch(fw, idx2adr(fw,k1), 0x01000001);
                int idx = adr2idx(fw, fadr);
                int k2;
                for (k2=idx; k2<idx+40; k2++)
                {
                    if (isSTR(fw,k2) && isLDMFD(fw,k2+1))
                    {
                        int ptr_offset = fwval(fw,k2) & 0xFFF;
                        print_stubs_min(fw,"palette_buffer_ptr",v2+ptr_offset,idx2adr(fw,k2));
                        int found = 0;
                        for (k=idx; (k<idx+16) && !found; k++)
                        {
                            if (isBL(fw,k) && isCMP(fw,k+1) && isADR_PC(fw,k+2))
                            {
                                fadr = ADR2adr(fw,k+2);
                                idx = adr2idx(fw, fadr);
                                int k3;
                                for (k3=idx; (k3<idx+16) && !found; k3++)
                                {
                                    if (isBL(fw,k3))
                                    {
                                        if (match_palette_buffer_offset(fw,k3))
                                            return 1;
                                    }
                                }
                            }
                        }
                        return 1;
                    }
                }
            }
            else if (isLDR_cond(fw,k1) && isLDMFD(fw,k1+2) && isBL(fw,k1-2))
            {
                int ptr_offset = fwval(fw,k1) & 0xFFF;
                print_stubs_min(fw,"palette_buffer_ptr",v2+ptr_offset,idx2adr(fw,k1));
                match_palette_buffer_offset(fw, k1-2);
                return 1;
            }
        }
    }
    return 0;
}

int match_SavePaletteData(firmware *fw, int idx, int palette_data)
{
    if (isBL(fw,idx+13))
    {
        uint32_t fadr = followBranch(fw, idx2adr(fw,idx+13), 0x01000001);
        idx = adr2idx(fw, fadr);
        if (isLDR(fw,idx) && isLDR(fw,idx+1) && isB(fw,idx+2))
        {
            uint32_t palette_control = LDR2val(fw,idx);
            print_stubs_min(fw,"palette_control",palette_control,idx2adr(fw,idx));
            int active_offset = fwval(fw,idx+1) & 0xFFF;
            print_stubs_min(fw,"active_palette_buffer",palette_control+active_offset,idx2adr(fw,idx+1));
            fadr = followBranch(fw,idx2adr(fw,idx+2),1);
            idx = adr2idx(fw, fadr);
            if (isLDR(fw,idx+17) && isLDR(fw,idx+18) && isLDR(fw,idx+12) && (LDR2val(fw,idx+12) == palette_control))
            {
                if (isLDR(fw,idx+13))
                {
                    int ptr_offset = fwval(fw,idx+13) & 0xFFF;
                    print_stubs_min(fw,"palette_buffer_ptr",palette_control+ptr_offset,idx2adr(fw,idx+13));
                }
                int palette_buffer;
                if ((fwval(fw,idx+18) & 0x0000F000) == 0)
                {
                    palette_buffer = LDR2val(fw,idx+17);
                    print_stubs_min(fw,"palette_buffer",palette_buffer,idx2adr(fw,idx+17));
                }
                else
                {
                    palette_buffer = LDR2val(fw,idx+18);
                    print_stubs_min(fw,"palette_buffer",palette_buffer,idx2adr(fw,idx+18));
                }
                if (isBL(fw,idx+26))
                {
                    fadr = followBranch(fw, idx2adr(fw,idx+26), 0x01000001);
                    idx = adr2idx(fw, fadr);
                    if (isLDR(fw, idx+2) && isBL(fw, idx+3))
                    {
                        uint32_t palette_size = LDR2val(fw,idx+2);
                        if (palette_size >= 0x400)
                        {
                            bprintf("// Offset from start of palette_buffer to color data = %d (Found @0x%08x)\n",palette_size-0x400,idx2adr(fw,idx+2));
                        }
                    }
                }
            }
        }
        else if (isLDR(fw,idx) && isLDR(fw,idx+6) && isLDR(fw,idx+7) && isBX(fw,idx+8))
        {
            int active_offset = -1;
            if ((fwval(fw,idx+6) & 0x0000F000) == 0)
                active_offset = fwval(fw,idx+6) & 0xFFF;
            else if ((fwval(fw,idx+7) & 0x0000F000) == 0)
                active_offset = fwval(fw,idx+7) & 0xFFF;
            if (active_offset >= 0)
            {
                uint32_t palette_control = LDR2val(fw,idx);
                //print_stubs_min(fw,"palette_control",palette_control,idx2adr(fw,idx));
                print_stubs_min(fw,"active_palette_buffer",palette_control+active_offset,idx2adr(fw,idx+1));
                search_fw(fw, match_palette_data2, palette_data, palette_control, 1);
            }
        }
        return 1;
    }

    return 0;
}

int match_viewport_address3(firmware *fw, int k, uint32_t v1, uint32_t v2)
{
    if (isLDR_PC(fw,k) && (LDR2val(fw,k) == v1))
    {
        // Scan back to start of function
        int k1 = find_inst_rev(fw, isSTMFD_LR, k-1, 1000);
        if (k1 > 0)
        {
            // Check if function sig matches what we want
            if (isLDR_PC(fw,k1+1) && isLDR_PC(fw,k1+2) && isLDR(fw,k1+3) &&
                (fwRd(fw,k1+1) == fwRn(fw,k1+3)))
            {
                uint32_t a = LDR2val(fw,k1+1);
                print_stubs_min(fw,"viewport_buffers",v1,idx2adr(fw,k));
                print_stubs_min(fw,"active_viewport_buffer",a,idx2adr(fw,k1+1));
                return 1;
            }
            else
            if (isLDR_PC(fw,k1+1) && isLDR_PC(fw,k1+3) && isLDR(fw,k1+4) &&
                (fwRd(fw,k1+1) == fwRn(fw,k1+4)))
            {
                uint32_t a = LDR2val(fw,k1+1);
                print_stubs_min(fw,"viewport_buffers",v1,idx2adr(fw,k));
                print_stubs_min(fw,"active_viewport_buffer",a,idx2adr(fw,k1+1));
                return 1;
            }
            else
            if (isLDR_PC(fw,k1+1) && isLDR_PC(fw,k1+4) && isLDR(fw,k1+5) &&
                (fwRd(fw,k1+1) == fwRn(fw,k1+5)))
            {
                uint32_t a = LDR2val(fw,k1+1);
                print_stubs_min(fw,"viewport_buffers",v1,idx2adr(fw,k));
                print_stubs_min(fw,"active_viewport_buffer",a,idx2adr(fw,k1+1));
                return 1;
            }
        }
    }
    return 0;
}

int match_viewport_address2(firmware *fw, int k, uint32_t v1, uint32_t v2)
{
    if (fw->buf[k] == v1)
    {
        if (search_fw(fw, match_viewport_address3, v1, 0, 1))
            return 1;
    }
    return 0;
}

int match_viewport_address(firmware *fw, int k, uint32_t v1, uint32_t v2)
{
    if (fw->buf[k] == v1)
    {
        // Find location in firmware that points to viewport address
        if (search_fw(fw, match_viewport_address2, idx2adr(fw,k), 0, 1))
            return 1;
    }
    return 0;
}

int match_physw_status(firmware *fw, int k, int v)
{
    if (isLDR_PC(fw,k))
    {
        print_stubs_min(fw,"physw_status",LDR2val(fw,k),idx2adr(fw,k));
    }

    return 0;
}

int match_physw_run(firmware *fw, int k, int v)
{
    if (isLDR_PC(fw,k))
    {
        uint32_t base = LDR2val(fw,k);
        uint32_t fadr = followBranch(fw, idx2adr(fw,k+1), 1);
        uint32_t ofst = fw->buf[adr2idx(fw,fadr)] & 0x00000FFF;
        print_stubs_min(fw,"physw_run",base+ofst,idx2adr(fw,k));

        ofst = fw->buf[k+2] & 0x00000FFF;
        print_stubs_min(fw,"physw_sleep_delay",base+ofst,idx2adr(fw,k));
    }

    return 0;
}

int match_canon_menu_active(firmware *fw, int k, int v)
{
    if (isLDR_PC(fw,k))
    {
        uint32_t base = LDR2val(fw,k);
        int k1;
        for (k1=k+1; k1<k+5; k1++)
        {
            if (isLDR(fw,k1))
            {
                uint32_t ofst = fw->buf[k1] & 0x00000FFF;
                print_stubs_min(fw,"canon_menu_active",base+ofst,idx2adr(fw,k));
            }
        }
    }

    return 0;
}

int match_zoom_busy(firmware *fw, int k, int v)
{
    if (isBL(fw,k))
    {
        int idx1 = idxFollowBranch(fw,k,0x01000001);
        int k1;
        for (k1=idx1; k1<idx1+50; k1++)
        {
            if ((fw->buf[k1] & 0xFFFF0000) == 0xE8BD0000)   // LDMFD
            {
                uint32_t fadr = 0;
                if (isADR_PC(fw,k1+1))
                {
                    fadr = ADR2adr(fw,k1+1);
                }
                else if (isADR_PC(fw,k1+2))
                {
                    fadr = ADR2adr(fw,k1+2);
                }
                else if (isADR_PC(fw,k1-3))
                {
                    fadr = ADR2adr(fw,k1-3);
                }
                else if (isLDR_PC(fw,k1+1))
                {
                    fadr = LDR2val(fw,k1+1);
                }
                else if (isLDR_PC(fw,k1+2))
                {
                    fadr = LDR2val(fw,k1+2);
                }
                if (fadr != 0)
                {
                    int idx2 = adr2idx(fw,fadr);
                    if (isLDR_PC(fw,idx2+1) && isLDR(fw,idx2+2))
                    {
                        int base = LDR2val(fw,idx2+1);
                        int ofst = fw->buf[idx2+2] & 0xFFF;
                           print_stubs_min(fw,"zoom_busy",base+ofst-4,fadr);
                        break;
                    }
                }
            }
        }
    }

    return 0;
}

int match_focus_busy(firmware *fw, int k, int v)
{
    if ((fw->buf[k] & 0xFFFF0000) == 0xE8BD0000)   // LDMFD
    {
        int k1 = 0;
        if (isBL(fw,k-2))
        {
            k1 = idxFollowBranch(fw,k-2,0x01000001);
        }
        if (isBL(fw,k-1))
        {
            k1 = idxFollowBranch(fw,k-1,0x01000001);
        }
        if (k1 != 0)
        {
            if (isLDR_PC(fw,k1+1) && isLDR(fw,k1+3))
            {
                int base = LDR2val(fw,k1+1);
                int ofst = fw->buf[k1+3] & 0xFFF;
                   print_stubs_min(fw,"focus_busy",base+ofst-4,idx2adr(fw,k1));
                return 1;
            }
        }
    }

    return 0;
}

int match_bitmap_buffer2(firmware *fw, int k, int v)
{
    uint32_t screen_lock = idx2adr(fw,k);
    if (isBL(fw,v) && (followBranch(fw,idx2adr(fw,v),0x01000001) == screen_lock) && isBL(fw,v+2) && isBL(fw,v+3))
    {
        uint32_t fadr = followBranch2(fw,idx2adr(fw,v+3),0x01000001);
        int k1 = adr2idx(fw,fadr);
        if (isLDR_PC(fw,k1+1))
        {
            int reg = (fwval(fw,k1+1) & 0x0000F000) >> 12;
            uint32_t adr = LDR2val(fw,k1+1);
            int k2;
            for (k2=k1; k2<k1+32; k2++)
            {
                if (isLDR_PC(fw,k2) && isLDR(fw,k2+1) && (((fwval(fw,k2+1) & 0x000F0000) >> 16) == reg))
                {
                    uint32_t bitmap_buffer = LDR2val(fw,k2);
                    if (bitmap_buffer == (adr + 0x1C))
                    {
                        uint32_t active_bitmap_buffer = adr + (fwval(fw,k2+1) & 0xFFF);
                        print_stubs_min(fw,"bitmap_buffer",bitmap_buffer,idx2adr(fw,k2));
                        print_stubs_min(fw,"active_bitmap_buffer",active_bitmap_buffer,idx2adr(fw,k2+1));
                    }
                }
            }
        }
    }

    return 0;
}

int match_bitmap_buffer(firmware *fw, int k, int v)
{
    search_saved_sig(fw, "ScreenLock", match_bitmap_buffer2, k, 0, 1);
    return 0;
}

int match_raw_buffer(firmware *fw, int k, uint32_t rb1, uint32_t v2)
{
    if (((fwval(fw,k) == rb1) && (fwval(fw,k+4) == rb1) && (fwval(fw,k-2) != 1)) ||
        ((fwval(fw,k) == rb1) && (fwval(fw,k+4) == rb1) && (fwval(fw,k+20) == rb1)))
    {
        uint32_t rb2 = fwval(fw,k+1);
        if ((rb1 != rb2) && (rb2 > 0))
        {
            bprintf("// Camera has 2 RAW buffers @ 0x%08x & 0x%08x\n", rb1, rb2, idx2adr(fw,k));
            bprintf("//  Note: active buffer --> raw_buffers[active_raw_buffer]\n");
            bprintf("//        other buffer  --> raw_buffers[active_raw_buffer^1]\n");
            print_stubs_min(fw,"raw_buffers",idx2adr(fw,k),idx2adr(fw,k));
        }
        return rb2;
    }
    else if ((fwval(fw,k) == rb1) && (fwval(fw,k-2) == 2) && (fwval(fw,k-7) == rb1))
    {
        uint32_t rb2 = fwval(fw,k+3);
        if ((rb1 != rb2) && (rb2 > 0))
        {
            bprintf("// Camera has 2 RAW buffers @ 0x%08x & 0x%08x\n", rb1, rb2, idx2adr(fw,k));
            bprintf("//  Note: active buffer --> raw_buffers[ active_raw_buffer   *3]\n");
            bprintf("//        other buffer  --> raw_buffers[(active_raw_buffer^1)*3]\n");
            print_stubs_min(fw,"raw_buffers",idx2adr(fw,k),idx2adr(fw,k));
        }
        return rb2;
    }
    return 0;
}

// Search for things that go in 'stubs_min.S'
void find_stubs_min(firmware *fw)
{
    //int k,k1;

    out_hdr = 1;
    add_blankline();

    bprintf("// Values below can be overridden in 'stubs_min.S':\n");

    // Find 'levent_table'
    search_fw(fw, match_levent_table, 0, 0, 1);

    // Find 'FlashParamsTable'
    if (FlashParamsTable_address != 0)
        print_stubs_min(fw,"FlashParamsTable",FlashParamsTable_address,FlashParamsTable_address);

    find_FileAccessSem(fw);
/*
    // Find 'physw_status'
    search_saved_sig(fw, "kbd_read_keys", match_physw_status, 0, 0, 5);

    // Find 'physw_run' & 'physw_sleep_delay'
    search_saved_sig(fw, "task_PhySw", match_physw_run, 0, 0, 5);

    // Find 'movie_status'
    search_fw(fw, match_movie_status, 0, 0, 1);

    // Find 'video_compression_rate'
    uint32_t sadr = find_str(fw, "CompressionRateAdjuster.c");
    k = find_nxt_str_ref(fw, sadr, -1);
    int found = 0;
    while ((k >= 0) && !found)
    {
        int f = find_inst_rev(fw, isSTMFD_LR, k-1, 100);
        if (f != -1)
        {
            f = search_fw(fw, find_BL, f, 0, 1);
            if (f > 0)
            {
                f--;
                if ((fwval(fw,f) & 0xFFF00000) == 0xE2400000)     // SUB
                {
                    int src = fwRn(fw,f);
                    for (k1 = f-1; (k1 > f-10) && !found; k1--)
                    {
                        if (isLDR_PC(fw,k1) && (fwRd(fw,k1) == src))
                        {
                            uint32_t v = LDR2val(fw,k1) - ALUop2(fw,f);
                            print_stubs_min(fw,"video_compression_rate",v,idx2adr(fw,k1));
                            found = 1;
                        }
                    }
                }
            }
        }
        k = find_nxt_str_ref(fw, sadr, k);
    }

    // Find 'full_screen_refresh'
    search_fw(fw, match_full_screen_refresh, 0, 0, 1);

    // Find 'canon_menu_active'
    search_saved_sig(fw, "StartRecModeMenu", match_canon_menu_active, 0, 0, 5);

    // Find 'canon_shoot_menu_active'
    search_fw(fw, match_canon_shoot_menu_active, 0, 0, 1);

    // Find 'playrec_mode'
    int found_playrec_mode = 0;
    k = find_str_ref(fw, "AFFChg");
    if ((k >= 0) && isBL(fw,k+6))
    {
        k = idxFollowBranch(fw, k+6, 0x01000001);
        if (isLDR_PC(fw,k) && isLDR(fw,k+1))
        {
            uint32_t base = LDR2val(fw,k);
            uint32_t ofst = fw->buf[k+1] & 0x00000FFF;
            print_stubs_min(fw,"playrec_mode",base+ofst,idx2adr(fw,k));
            found_playrec_mode = 1;
        }
    }
    if (!found_playrec_mode)
    {
        search_fw(fw, match_playrec_mode, 0, 0, 1);
    }

    // Find 'zoom_status'
    int found_zoom_status = 0;

    k = find_str_ref(fw, "m_ZoomState            :%d\n");
    if (k >= 0)
    {
        if (isLDR(fw,k-1))
        {
            uint32_t ofst = fw->buf[k-1] & 0x00000FFF;
            uint32_t reg = (fw->buf[k-1] & 0x000F0000) >> 16;
            uint32_t ldr_inst = 0xE51F0000 | (reg << 12);
            for (k1=k-2; k1>k-20; k1--)
            {
                if ((fw->buf[k1] & 0xFF1FF000) == ldr_inst)
                {
                    uint32_t base = LDR2val(fw,k1);
                    print_stubs_min(fw,"zoom_status",base+ofst,idx2adr(fw,k));
                    found_zoom_status = 1;
                    break;
                }
            }
        }
    }

    if (!found_zoom_status)
    {
        for (k=0; k<fw->size; k++)
        {
            if (((fw->buf[k] & 0xFF1FF000) == 0xE51F0000) &&    // LDR R0, =base
                (fw->buf[k+1] == 0xE5D00000) &&                 // LDRB R0, [R0]
                (fw->buf[k+2] == 0xE1B00000) &&                 // MOVS R0, R0
                (fw->buf[k+3] == 0x13A00001) &&                 // MOVNE R0, #1
                isBX_LR(fw,k+4))                                // BX LR
            {
                uint32_t base = LDR2val(fw,k);
                print_stubs_min(fw,"zoom_status",base,idx2adr(fw,k));
                found_zoom_status = 1;
                //break;
            }
        }
    }

    if (!found_zoom_status)
    {
        k = find_str_ref(fw, "TerminateDeliverToZoomController");
        if (k >= 0)
        {
            for (k1=0; k1<5; k1++)
            {
                if (isLDR_PC(fw,k+k1))
                {
                    uint32_t base = LDR2val(fw,k+k1);
                    print_stubs_min(fw,"zoom_status",base+0x20,idx2adr(fw,k+k1));
                    found_zoom_status = 1;
                    break;
                }
            }
        }
    }

    // Find 'some_flag_for_af_scan'
    search_fw(fw, match_some_flag_for_af_scan, 0, 0, 1);

    // focus_len_table
    if (fw->sv->min_focus_len != 0)
    {
        int found = 0, pos = 0, len = 0, size = 0;
        for (k=0; k<fw->size; k++)
        {
            if (fw->buf[k] == fw->sv->min_focus_len)
            {
                int mul = 1;
                if ((fw->buf[k+1] == 100) && (fw->buf[k+2] == 0)) mul = 3;
                if ((fw->buf[k+1] == 100) && (fw->buf[k+2] != 0)) mul = 2;
                if ((fw->buf[k+1] ==   0) && (fw->buf[k+2] != 0)) mul = 2;
                for (k1 = k + mul; (k1 < fw->size) && (fw->buf[k1] > fw->buf[k1-mul]) && (fw->buf[k1] > fw->sv->min_focus_len) && (fw->buf[k1] < fw->sv->max_focus_len); k1 += mul) ;
                if (fw->buf[k1] == fw->sv->max_focus_len)
                {
                    if ((found == 0) || ((size < mul) && (len < ((k1 - k) / mul) + 1)))
                    {
                        found = 1;
                        pos = k;
                        len = ((k1 - k) / mul) + 1;
                        size = mul;
                    }
                }
            }
        }
        if (found == 1)
        {
            bprintf("// focus_len_table contains zoom focus lengths for use in 'get_focal_length' (main.c).\n");
            if (size == 1)
                bprintf("// each entry contains 1 int value, which is the the zoom focus length.\n",size);
            else
                bprintf("// each entry contains %d int value(s), the first is the zoom focus length.\n",size);
            bprintf("// there are %d entries in the table - set NUM_FL to %d\n",len,len);
            print_stubs_min(fw,"focus_len_table",idx2adr(fw,pos),idx2adr(fw,pos));
        }
    }

    // Find 'zoom_busy'
    search_saved_sig(fw, "ResetZoomLens", match_zoom_busy, 0, 0, 5);

    // Find 'focus_busy'
    search_saved_sig(fw, "ResetFocusLens", match_focus_busy, 0, 0, 25);

    // Find 'recreview_hold'
    k = find_str_ref(fw, "ShootCon_NotifyStartReviewHold");
    if (k >= 0)
    {
        for (k1=k; k1<k+20; k1++)
        {
            if (isLDR_PC(fw,k1) && ((fw->buf[k1+1] & 0xFFFF0FFF) == 0xE3A00001) && isSTR(fw,k1+2) &&
                ((fw->buf[k1+1] & 0x0000F000) == (fw->buf[k1+2] & 0x0000F000)) &&
                ((fw->buf[k1] & 0x0000F000) == ((fw->buf[k1+2] & 0x000F0000) >> 4)))
            {
                uint32_t base = LDR2val(fw,k1);
                int ofst = fw->buf[k1+2] & 0x00000FFF;
                print_stubs_min(fw,"recreview_hold",base+ofst,idx2adr(fw,k1));
                break;
            }
        }
    }

    // Find palette colour data
    uint32_t palette_data = search_fw(fw, match_palette_data, 0, 0, 1);

    // Find 'palette buffer' info
    if (palette_data)
    {
        bprintf("// Palette colour tables  found @ 0x%08x\n", palette_data);
        if (search_saved_sig(fw, "SavePaletteData", match_SavePaletteData, palette_data, 0, 1) == 0)
        {
            search_fw(fw, match_palette_data3, palette_data, 0, 1);
        }
    }

    // Find 'bitmap buffer' info
    search_saved_sig(fw, "GUISrv_StartGUISystem", match_bitmap_buffer, 0, 0, 50);

    // Get viewport address
    uint32_t v = find_viewport_address(fw,&k);
    if (k >= 0)
    {
        search_fw(fw, match_viewport_address, v, 0, 1);
    }

    // find 1st RAW buffer address
    k = find_str_ref(fw, "CRAW BUFF       %p");
    if (k >= 0)
    {
        int rb1_idx;
        uint32_t rb1 =0, rb2 = 0;
        if (isLDR(fw,k-1))
        {
            rb1 = LDR2val(fw,k-1);
            rb1_idx = k - 1;
        }
        else if (isMOV_immed(fw,k-1))
        {
            rb1 = ALUop2(fw,k-1);
            rb1_idx = k - 1;
        }
        else if (isMOV(fw,k-1) && (fwRd(fw,k-1) == 1))
        {
            int reg = fwval(fw,k-1) & 0xF;
            for (k1=k-2; k1>k-50; k1--)
            {
                if (isLDR(fw,k1) && (fwRd(fw,k1) == reg))
                {
                    rb1 = LDR2val(fw,k1);
                    rb1_idx = k1;
                    break;
                }
            }
        }
        if (rb1 > 0)
        {
            found = 0;
            rb2 = search_fw(fw, match_raw_buffer, rb1, 0, 5);
            if ((rb2 > 0) && (rb1 != rb2))
            {
                // Find 'active_raw_buffer'
                sadr = find_str(fw, "SsImgProcBuf.c");
                k = find_nxt_str_ref(fw, sadr, -1);
                while ((k >= 0) && !found)
                {
                    int f = find_inst_rev(fw, isSTMFD_LR, k-1, 100);
                    if (f != -1)
                    {
                        int e = find_inst(fw, isLDMFD_PC, f+1, 200);
                        for (k1 = f+1; k1 < e; k1++)
                        {
                            if (
                                (
                                    ((fwval(fw,k1)   & 0xFFF00FFF) == 0xE2400001) &&    // SUB Rx, Rn, #1
                                    isLDR(fw,k1+1) &&                                   // LDR Ry, [Rz,
                                    ((fwval(fw,k1+2) & 0xFFF00000) == 0xE1500000) &&    // CMP Rx, Ry
                                    (((fwRd(fw,k1) == fwRd(fw,k1+2)) && (fwRd(fw,k1+1) == fwRn(fw,k1+2))) ||
                                     ((fwRd(fw,k1) == fwRn(fw,k1+2)) && (fwRd(fw,k1+1) == fwRd(fw,k1+2)))) &&
                                    ((fwval(fw,k1+3) & 0xFFF00FFF) == 0x12800001) &&    // ADDNE Ry, Ry, #1
                                    ((fwRd(fw,k1+3) == fwRn(fw,k1+3)) && (fwRd(fw,k1+3) == fwRd(fw,k1+1))) &&
                                    ((fwval(fw,k1+4) & 0xFFF00FFF) == 0x03A00000) &&    // MOVEQ Ry, #0
                                    (fwRd(fw,k1+4) == fwRd(fw,k1+1)) &&
                                    isSTR(fw,k1+5) &&                                   // STR Ry, [Rz,
                                    ((fwRd(fw,k1+5) == fwRd(fw,k1+1)) && (fwRn(fw,k1+5) == fwRn(fw,k1+1)) && (fwOp2(fw,k1+5) == fwOp2(fw,k1+1)))
                                ) ||
                                (
                                    ((fwval(fw,k1)   & 0xFFF00FFF) == 0xE2400001) &&    // SUB Rx, Rn, #1
                                    isLDR(fw,k1+1) &&                                   // LDR Ry, [Rz,
                                    ((fwval(fw,k1+3) & 0xFFF00000) == 0xE1500000) &&    // CMP Rx, Ry
                                    (((fwRd(fw,k1) == fwRd(fw,k1+3)) && (fwRd(fw,k1+1) == fwRn(fw,k1+3))) ||
                                     ((fwRd(fw,k1) == fwRn(fw,k1+3)) && (fwRd(fw,k1+1) == fwRd(fw,k1+3)))) &&
                                    ((fwval(fw,k1+4) & 0xFFF00FFF) == 0x12800001) &&    // ADDNE Ry, Ry, #1
                                    ((fwRd(fw,k1+4) == fwRn(fw,k1+4)) && (fwRd(fw,k1+4) == fwRd(fw,k1+1))) &&
                                    ((fwval(fw,k1+5) & 0xFFF00FFF) == 0x03A00000) &&    // MOVEQ Ry, #0
                                    (fwRd(fw,k1+5) == fwRd(fw,k1+1)) &&
                                    isSTR(fw,k1+7) &&                                   // STR Ry, [Rz,
                                    ((fwRd(fw,k1+7) == fwRd(fw,k1+1)) && (fwRn(fw,k1+7) == fwRn(fw,k1+1)) && (fwOp2(fw,k1+7) == fwOp2(fw,k1+1)))
                                )
                               )
                            {
                                int ofst = fwOp2(fw,k1+1);
                                int reg = fwRn(fw,k1+1);
                                int k2;
                                for (k2 = f+1; (k2 < e) && !found; k2++)
                                {
                                    if (isLDR_PC(fw,k2) && (fwRd(fw,k2) == reg))
                                    {
                                        uint32_t base = LDR2val(fw,k2);
                                        print_stubs_min(fw,"active_raw_buffer",base+ofst,idx2adr(fw,k1));
                                        found = 1;
                                        break;
                                    }
                                }
                            }
                        }
                    }
                    k = find_nxt_str_ref(fw, sadr, k);
                }
            }
            if (!found)
            {
                bprintf("// Camera appears to have only 1 RAW buffer @ 0x%08x (Found @0x%08x)\n", rb1, idx2adr(fw,rb1_idx));
            }
        }
    }
*/
    // Find UI property count
    search_saved_sig(fw, "PTM_SetCurrentItem", match_uiprop_count, 0, 0, 30);

}

//------------------------------------------------------------------------------------------------------------

int find_ctypes(firmware *fw, int k)
{
    static unsigned char ctypes[] =
    {
        0x20, 0x20, 0x20, 0x20, 0x20, 0x20, 0x20, 0x20, 0x20, 0x60, 0x60, 0x60, 0x60, 0x60, 0x20, 0x20,
        0x20, 0x20, 0x20, 0x20, 0x20, 0x20, 0x20, 0x20, 0x20, 0x20, 0x20, 0x20, 0x20, 0x20, 0x20, 0x20,
        0x48, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10,
        0x84, 0x84, 0x84, 0x84, 0x84, 0x84, 0x84, 0x84, 0x84, 0x84, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10,
        0x10, 0x81, 0x81, 0x81, 0x81, 0x81, 0x81, 1, 1, 1, 1, 1, 1, 1, 1, 1,
        1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 0x10, 0x10, 0x10, 0x10, 0x10,
        0x10, 0x82, 0x82, 0x82, 0x82, 0x82, 0x82, 2, 2, 2, 2, 2, 2, 2, 2, 2,
        2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 0x10, 0x10, 0x10, 0x10, 0x20
    };

    if (k < (fw->size*4 - sizeof(ctypes)))
    {
        if (memcmp(((char*)fw->buf)+k,ctypes,sizeof(ctypes)) == 0)
        {
            bprintf("DEF(ctypes, 0x%08x)\n", fw->base + k);
            return 1;
        }
    }
    return 0;
}

int match_nrflag3(firmware *fw, int k, uint32_t v1, uint32_t v2)
{
    if (isBL(fw,k) && (idxFollowBranch(fw,k,0x01000001) == v1))
    {
        // Found call to function, work out R3 value passed in
        int ofst1 = 0;
        int k3, k4 = 0;
        for (k3=k; k3>k-30; k3--)
        {
            if ((fwval(fw,k3) & 0x0F0FF000) == 0x020D3000)       // Dest = R3, Src = SP = skip
                break;
            if ((fwval(fw,k3) & 0xFF0FF000) == 0xE2033000)       // ADD/SUB R3,R3,x
            {
                k4 = k3;
                if ((fwval(fw,k3) & 0x00F00000) == 0x00400000)   // SUB
                    ofst1 -= (fwval(fw,k3) & 0x00000FFF);
                else
                    ofst1 += (fwval(fw,k3) & 0x00000FFF);
            }
            if (isLDR_PC(fw,k3) && (fwRd(fw,k3) == 3))
            {
                int ofst2 = LDR2val(fw,k3);
                if (ofst2 > (fw->data_len*4 + fw->data_start)) // has to be in the preinited data section
                    return 0;
                bprintf("\n// For capt_seq.c\n");
                if (ofst1 == 0)
                {
                    bprintf("DEF(_nrflag,0x%04x) // Found @ %08x (0x%04x)\n",ofst2,idx2adr(fw,k3),ofst2);
                    bprintf("//static long *nrflag = (long*)(0x%04x);       // Found @ %08x\n",ofst2,idx2adr(fw,k3));
                }
                else if (ofst1 < 0)
                {
                    bprintf("DEF(_nrflag,0x%04x) // Found @ %08x (0x%04x) & %08x (-0x%02x)\n",ofst2+ofst1,idx2adr(fw,k3),ofst2,idx2adr(fw,k),-ofst1);
                    bprintf("//static long *nrflag = (long*)(0x%04x-0x%02x);  // Found @ %08x & %08x\n",ofst2,-ofst1,idx2adr(fw,k3),idx2adr(fw,k4));
                }
                else
                {
                    bprintf("DEF(_nrflag,0x%04x) // Found @ %08x (0x%04x) & %08x (+0x%02x)\n",ofst2+ofst1,idx2adr(fw,k3),ofst2,idx2adr(fw,k),ofst1);
                    bprintf("//static long *nrflag = (long*)(0x%04x+0x%02x);  // Found @ %08x & %08x\n",ofst2,ofst1,idx2adr(fw,k3),idx2adr(fw,k4));
                }
                return 1;
            }
        }
    }
    return 0;
}

int match_nrflag(firmware *fw, int idx, int v)
{
    int k1, k2, k3;
    int found = 0;

    if (isLDR(fw, idx+1) && isLDR(fw, idx+2))
    {
        k3 = idx+2;
        int ofst2 = LDR2val(fw, k3);

        for (k1=k3+1; k1<k3+8; k1++)
        {
            if (isB(fw, k1))
            {
                k2 = idxFollowBranch(fw,k1,0x01000001);
                if (isSTR(fw, k2))
                {
                    found = 1;
                    break;
                }
                k2++;
                if (isSTR(fw, k2))
                {
                    found = 1;
                    break;
                }
            }
        }

        if (found)
        {
            int ofst1 = fw->buf[k2] & 0x00000FFF;
            bprintf("\n// For capt_seq.c\n");
            bprintf("DEF(_nrflag,0x%04x) // Found @ %08x (0x%04x) & %08x (+0x%02x)\n",ofst2+ofst1,idx2adr(fw,k3),ofst2,idx2adr(fw,k2),ofst1);
            bprintf("//static long *nrflag = (long*)(0x%04x+0x%02x);  // Found @ %08x & %08x\n",ofst2,ofst1,idx2adr(fw,k3),idx2adr(fw,k2));
            bprintf("//#define NR_AUTO (0)                          // have to explictly reset value back to 0 to enable auto\n");
        }
    }

    return found;
}

int match_nrflag2(firmware *fw, int k, int v)
{
    // Found NR_GetDarkSubType function, now follow first BL call.
    if (isBL(fw,k))
    {
        k = idxFollowBranch(fw,k,0x01000001);
        return search_fw(fw, match_nrflag3, k, 0, 1);
    }

    return 0;
}

// find LEDs, Vx specific

// ADD Rx, Rx, #0x220000
int isADD_0x220000(firmware *fw, int offset)
{
    return ((fwval(fw,offset) & 0xfff00fff) == (0xe2800822));
}

typedef struct {
    uint32_t addr;  // LED GPIO address
    int reg;        // register used to assemble the address
    int offs;       // offset in the LED table
} LED_s;

int find_leds(firmware *fw)
{
    int j1, j2, m, n;
    LED_s led;
    int k1 = find_str_ref(fw,"LEDCon");
    if (k1<0)
        return 0;
    k1 = find_inst_rev(fw,isSTMFD_LR,k1,96);
    if (k1<0)
        return 0;
    j1 = find_inst(fw,isBL,k1,80);
    j2 = find_Nth_inst(fw,isBL,k1,80,3);
    if ((j1<0) || (j2<0))
        return 0;
    // 1st and 3rd BL is memory allocation
    if (followBranch(fw,idx2adr(fw,j1),0x01000001) != followBranch(fw,idx2adr(fw,j2),0x01000001))
        return 0;
    k1 = find_Nth_inst(fw,isBL,k1,80,2);
    // LED table initializer func
    k1 = idxFollowBranch(fw,k1,0x01000001);
    if (k1<0)
        return 0;
    bprintf("// LED table init @ 0x%x\n",idx2adr(fw,k1));
    j2 = 1;
    while (1)
    {
        j1 = find_Nth_inst(fw,isADD_0x220000,k1,40,j2);
        if (j1>0)
        {
            led.reg = fwRd(fw,j1);
            led.addr = 0x220000;
            led.offs = 0;
            n = j1-1;
            while (!isSTMFD_LR(fw,n))
            {
                if ((fwval(fw,n)&0xfffff000) == (0xe2800000+(led.reg<<12)+(led.reg<<16))) // ADD Rx, Rx, #0xc00000yz
                {
                    if ( ALUop2a(fw,n) >= 0xc0000000 )
                    {
                        led.addr += ALUop2a(fw,n);
                    }
                }
                else if ((fwval(fw,n)&0xfffff000) == (0xe3a00000+(led.reg<<12))) // MOV Rx, #imm
                {
                    led.addr += ALUop2a(fw,n);
                    m = n+1;
                    while (!isLDMFD_PC(fw,m))
                    {
                        if ((fwval(fw,m)&0xfff0f000) == (0xe5800000+(led.reg<<12))) // STR Rx, [Ry, imm]
                        {
                            led.offs = fwval(fw,m) & 0xfff;
                            break;
                        }
                        m++;
                    }
                    if (led.offs != 0)
                        break;
                }
                n--;
            }
            // output data if valid
            if (led.offs != 0)
            {
                bprintf("// LED #%i: 0x%08x, offset 0x%x\n",j2, led.addr, led.offs);
            }
            j2++;
        }
        else
        {
            break;
        }
    }
    bprintf("\n");
    return 0;
}

// Search for things
void find_other_vals(firmware *fw)
{
    out_hdr = 1;
    add_blankline();

    bprintf("// Misc stuff\n");
    find_leds(fw);
/*
    if (!search_fw_bytes(fw, find_ctypes))
    {
        bprintf("//DEF(ctypes, *** Not Found ***)\n");
    }
*/   
    bprintf("// canon_data_src: 0x%x, canon_data_len: 0x%x\n",fw->data_init_start,fw->data_len*4);


    // Look for nrflag (for capt_seq.c)
    search_saved_sig(fw, "NR_GetDarkSubType", match_nrflag2, 0, 0, 20);
}

//------------------------------------------------------------------------------------------------------------

void print_kval(firmware *fw, uint32_t tadr, int tsiz, int tlen, uint32_t ev, const char *name, char *sfx)
{
    int tidx = adr2idx(fw,tadr);
    int k, kval = 0;
    for (k=0; k<tlen; k+=tsiz)
    {
        if (fw->buf[tidx+k+1] == ev)
        {
            kval = fw->buf[tidx+k];
            tadr = idx2adr(fw,tidx+k);
            break;
        }
    }
    if (kval > 0)
    {
        char fn[100], rn[100];
        strcpy(fn,name); strcat(fn,sfx);
        strcpy(rn,name); strcat(rn,"_IDX");

        int r = (kval >> 5) & 7;
        uint32_t b = (1 << (kval & 0x1F));

        bprintf("//#define %-20s0x%08x // Found @0x%08x, levent 0x%x\n",fn,b,tadr,ev);
        bprintf("//#define %-20s%d\n",rn,r);
    }
}

typedef struct {
    int         reg;
    uint32_t    bits;
    char        nm[32];
    uint32_t    fadr;
    uint32_t    ev;
    int         inv;
} kinfo;

int     kmask[3];
kinfo   key_info[100];
int     kcount = 0;
uint32_t kshutter_min_bits = 0xFFFFFFFF;

void add_kinfo(int r, uint32_t b, const char *nm, uint32_t adr, uint32_t ev, int inv)
{
    key_info[kcount].reg = r;
    key_info[kcount].bits = b;
    strcpy(key_info[kcount].nm, nm);
    key_info[kcount].fadr = adr;
    key_info[kcount].ev = ev;
    key_info[kcount].inv = inv;
    kcount++;
    kmask[r] |= b;
    if ((ev <= 1) && (b < kshutter_min_bits)) kshutter_min_bits = b;
}

uint32_t add_kmval(firmware *fw, uint32_t tadr, int tsiz, int tlen, uint32_t ev, const char *name, uint32_t xtra)
{
    int tidx = adr2idx(fw,tadr);
    int r, k, kval = 0;
    uint32_t b = 0;
    int inv = 0;
    for (k=0; k<tlen; k+=tsiz)
    {
        if (fw->buf[tidx+k+1] == ev)
        {
            kval = fw->buf[tidx+k];
            tadr = idx2adr(fw,tidx+k);
            break;
        }
    }
    if (kval > 0)
    {
        r = (kval >> 5) & 7;
        b = (1 << (kval & 0x1F));
        inv = ((kval&0xff0000)==0x10000)?0:1;

        add_kinfo(r,b|xtra,name,tadr,ev,inv);
    }

    return b;
}

int kinfo_compare(const kinfo *p1, const kinfo *p2)
{
    if (p1->reg > p2->reg)
    {
        return 1;
    }
    else if (p1->reg < p2->reg)
    {
        return -1;
    }
    if ((p1->ev <= 1) && (p2->ev <= 1))    // output shutter entries in reverse order
    {
        if (p1->bits > p2->bits)
        {
            return -1;
        }
        else if (p1->bits < p2->bits)
        {
            return 1;
        }
    }
    // if one entry is shutter then compare to min shutter bits
    if (p1->ev <= 1)
    {
        if (kshutter_min_bits > p2->bits)
        {
            return 1;
        }
        else if (kshutter_min_bits < p2->bits)
        {
            return -1;
        }
    }
    if (p2->ev <= 1)
    {
        if (p1->bits > kshutter_min_bits)
        {
            return 1;
        }
        else if (p1->bits < kshutter_min_bits)
        {
            return -1;
        }
    }
    if (p1->bits > p2->bits)
    {
        return 1;
    }
    else if (p1->bits < p2->bits)
    {
        return -1;
    }

    return 0;
}

void print_kmvals()
{
    qsort(key_info, kcount, sizeof(kinfo), (void*)kinfo_compare);

    bprintf("//static KeyMap keymap[] = {\n");

    int k;
    for (k=0; k<kcount; k++)
    {
        bprintf("//    { %d, %-20s,0x%08x }, // Found @0x%08x, levent 0x%02x%s\n",key_info[k].reg,key_info[k].nm,key_info[k].bits,key_info[k].fadr,key_info[k].ev,(key_info[k].inv==0)?"":" (uses inverted logic in physw_status)");
    }

    bprintf("//    { 0, 0, 0 }\n//};\n");
}

int match_GetSDProtect(firmware *fw, int k, int v)
{
    if (isB(fw,k))    // B
    {
        k = idxFollowBranch(fw,k,1);
        if (isLDR_PC(fw,k))
        {
            return LDR2val(fw,k);
        }
    }

    return 0;
}

void find_key_vals(firmware *fw)
{
    int k,k1;

    out_hdr = 1;
    add_blankline();

    // find 'SD_READONLY_FLAG'
    uint32_t tadr = search_saved_sig(fw, "GetSDProtect", match_GetSDProtect, 0, 1, 1);
    if (tadr == 0)
    {
        k = find_str_ref(fw,"SD Not Exist\n");
        if (k >= 0)
        {
            for (k1=k-1; k1>k-5; k1--)
            {
                if (isBL(fw,k1))    // BL
                {
                    uint32_t fadr = followBranch(fw,idx2adr(fw,k1),0x01000001);
                    int k2 = adr2idx(fw,fadr);
                    if (isLDR_PC(fw,k2))
                    {
                        tadr = LDR2val(fw,k2);
                    }
                }
            }
        }
    }
    if (tadr != 0)
    {
        int tsiz = 2;
        if (fw->buf[adr2idx(fw,tadr)+2] == 0) tsiz = 3;

        uint32_t madr = fw->base + (fw->size*4-4);
        for (k=0; k<(tadr-fw->base)/4; k++)
        {
            if (isLDR_PC(fw,k))
            {
                uint32_t adr = LDR2val(fw,k);
                if ((adr > tadr) && (adr < madr))
                {
                    madr = adr;
                }
            }
        }
        int tlen = (madr - tadr) / 4;
        if (tsiz == 2)
        {
            k1 = adr2idx(fw,tadr);
            for (k=0; k<tlen/3; k+=3)
            {
                if ((fw->buf[k1+k+1] == 0xFFFFFFFF) && (fw->buf[k1+k+4] == 0xFFFFFFFF))
                {
                    tsiz = 3;
                    break;
                }
            }
        }
        if (tlen > 50*tsiz) tlen = 50*tsiz;

        bprintf("// Bitmap masks and physw_status index values for SD_READONLY and USB power flags (for kbd.c).\n");
        if (fw->dryos_ver >= 49)
        {
            // Event ID's have changed in DryOS R49 **********
            print_kval(fw,tadr,tsiz,tlen,0x20A,"SD_READONLY","_FLAG");
            print_kval(fw,tadr,tsiz,tlen,0x202,"USB","_MASK");
        }
        else
        {
            print_kval(fw,tadr,tsiz,tlen,0x90A,"SD_READONLY","_FLAG");
            print_kval(fw,tadr,tsiz,tlen,0x902,"USB","_MASK");
        }

        uint32_t key_half = add_kmval(fw,tadr,tsiz,tlen,0,"KEY_SHOOT_HALF",0);
        add_kmval(fw,tadr,tsiz,tlen,1,"KEY_SHOOT_FULL",key_half);
        add_kmval(fw,tadr,tsiz,tlen,1,"KEY_SHOOT_FULL_ONLY",0);
        add_kmval(fw,tadr,tsiz,tlen,2,"KEY_ZOOM_IN",0);
        add_kmval(fw,tadr,tsiz,tlen,3,"KEY_ZOOM_OUT",0);
        add_kmval(fw,tadr,tsiz,tlen,4,"KEY_UP",0);
        add_kmval(fw,tadr,tsiz,tlen,5,"KEY_DOWN",0);
        add_kmval(fw,tadr,tsiz,tlen,6,"KEY_LEFT",0);
        add_kmval(fw,tadr,tsiz,tlen,7,"KEY_RIGHT",0);
        add_kmval(fw,tadr,tsiz,tlen,8,"KEY_SET",0);
        add_kmval(fw,tadr,tsiz,tlen,9,"KEY_MENU",0);
        add_kmval(fw,tadr,tsiz,tlen,0xA,"KEY_DISPLAY",0);
        if (fw->dryos_ver <= 47)
        {
            add_kmval(fw,tadr,tsiz,tlen,0x601,"KEY_PLAYBACK",0);
            add_kmval(fw,tadr,tsiz,tlen,0x600,"KEY_POWER",0);
            add_kmval(fw,tadr,tsiz,tlen,0x12,"KEY_VIDEO",0);
        }
        else
        {
            add_kmval(fw,tadr,tsiz,tlen,0x101,"KEY_PLAYBACK",0);
            add_kmval(fw,tadr,tsiz,tlen,0x100,"KEY_POWER",0);
            if (fw->dryos_ver == 49)
            {
                add_kmval(fw,tadr,tsiz,tlen,0x19,"KEY_VIDEO",0);
            }
            else if(fw->dryos_ver == 50)
            {
                add_kmval(fw,tadr,tsiz,tlen,0x1A,"KEY_VIDEO",0);
                add_kmval(fw,tadr,tsiz,tlen,0x14,"KEY_HELP",0);
            }
        }

        bprintf("\n// Keymap values for kbd.c. Additional keys may be present, only common values included here.\n");
        print_kmvals();
    }
}

//------------------------------------------------------------------------------------------------------------

uint32_t nadr;
uint32_t eadr;

int get_eventproc_val(firmware *fw, int k)
{
    if (isADR_PC(fw,k) && (fwRd(fw,k) == 0))
        nadr = ADR2adr(fw,k);
    else if (isADR_PC(fw,k) && (fwRd(fw,k) == 1))
        eadr = ADR2adr(fw,k);
    else if (isLDR_PC(fw,k) && (fwRd(fw,k) == 0))
        nadr = LDR2val(fw,k);
    else if (isLDR_PC(fw,k) && (fwRd(fw,k) == 1))
        eadr = LDR2val(fw,k);
    else
        return 0;
    return 1;
}

void add_func_name(char *n, uint32_t eadr, char *suffix)
{
    int k;

    char *s = n;
    if (suffix != 0)
    {
        s = malloc(strlen(n) + strlen(suffix) + 1);
        sprintf(s, "%s%s", n, suffix);
    }

    for (k=0; func_names[k].name != 0; k++)
        if (strcmp(func_names[k].name, s) == 0)
        {
            if (func_names[k].val == 0)
            {
                func_names[k].val = eadr;
                func_names[k].flags |= EV_MATCH;
            }
            if (s != n) free(s);
            return;
        }

    func_names[next_func_entry].name = s;
    func_names[next_func_entry].flags = OPTIONAL|UNUSED;
    func_names[next_func_entry].val = eadr;
    next_func_entry++;
    func_names[next_func_entry].name = 0;
}

void add_func_name2(firmware *fw, uint32_t nadr, uint32_t eadr, char *suffix)
{
    char *n = (char*)adr2ptr(fw,nadr);
    if (*n)
    {
        if (isB(fw,adr2idx(fw,eadr)))
        {
            char *s = malloc(strlen(n) + 3);
            sprintf(s,"j_%s",n);
            add_func_name(s, eadr, suffix);
            eadr = followBranch(fw,eadr,1);
        }
        add_func_name(n, eadr, suffix);
    }
}

int match_eventproc(firmware *fw, int k, uint32_t fadr, uint32_t v2)
{
    if (isBorBL(fw,k))
    {
        uint32_t adr = followBranch(fw,idx2adr(fw,k),0x01000001);
        if (adr == fadr)
        {
            nadr = 0;
            eadr = 0;
            k--;
            get_eventproc_val(fw, k);
            k--;
            get_eventproc_val(fw, k);
            if ((nadr == 0) || (eadr == 0))
            {
                k--;
                get_eventproc_val(fw, k);
            }
            if ((nadr == 0) || (eadr == 0))
            {
                k--;
                get_eventproc_val(fw, k);
            }
            if ((nadr != 0) && (eadr != 0))
            {
                add_func_name2(fw, nadr, eadr, "_FW");
            }
        }
    }
    else if (isLDR_PC(fw,k) && (fwRd(fw,k) == 0) && isLDR_PC(fw,k+1) && (fwRd(fw,k+1) == 1) &&
             (fwval(fw,k+2) == 0xE1A0E00F) && (fwval(fw,k+3) == 0xE1A0F004))
    {
        nadr = LDR2val(fw,k);
        eadr = LDR2val(fw,k+1);
        add_func_name2(fw, nadr, eadr, "_FW");
    }
    return 0;
}

int match_registerlists(firmware *fw, int k, uint32_t fadr, uint32_t v2)
{
    if (isBorBL(fw,k))
    {
        uint32_t adr = followBranch2(fw,idx2adr(fw,k),0x01000001);
        if (adr == fadr)
        {
            int k1;
            for (k1=k-1; k1>k-6; k1--)
            {
                if (isLDR_PC(fw,k1) && (fwRd(fw,k1) == 0))
                {
                    int j = adr2idx(fw,LDR2val(fw,k1));
                    if (!idx_valid(fw,j))
                    {
                        j = adr2idx(fw,LDR2val(fw,k1) - fw->data_start + fw->data_init_start);
                    }
                    if (idx_valid(fw,j))
                    {
                        while (fwval(fw,j) != 0)
                        {
                            add_func_name2(fw, fwval(fw,j), fwval(fw,j+1), "_FW");
                            j += 2;
                        }
                    }
                    break;
                }
            }
        }
    }
    return 0;
}

int match_registerlistproc(firmware *fw, int k, uint32_t fadr, uint32_t v2)
{
    if (isSTMFD_LR(fw,k) && isBL(fw,k+6) && isLDMFD_PC(fw,k+11))
    {
        uint32_t adr = followBranch2(fw,idx2adr(fw,k+6),0x01000001);
        if (adr == fadr)
        {
            fadr = idx2adr(fw,k);
            search_fw(fw, match_registerlists, fadr, 0, 6);
            return 1;
        }
    }
    return 0;
}

void find_eventprocs(firmware *fw)
{
    int j = get_saved_sig(fw,"ExportToEventProcedure_FW");
    if (j >= 0)
    {
        uint32_t fadr = func_names[j].val;
        search_fw(fw, match_eventproc, fadr, 0, 1);
    }
    j = get_saved_sig(fw,"RegisterEventProcedure_FW");
    if (j >= 0)
    {
        uint32_t fadr = func_names[j].val;
        search_fw(fw, match_eventproc, fadr, 0, 1);
        search_fw(fw, match_registerlistproc, fadr, 0, 12);
    }
}

// LDR R0, =...
int isLDR_PC_r0(firmware *fw, int offset)
{
    return ((fwval(fw,offset) & 0xFE1FF000) == (0xE41F0000));
}

// LDR R3, =...
int isLDR_PC_r3(firmware *fw, int offset)
{
    return ((fwval(fw,offset) & 0xFE1FF000) == (0xE41F3000));
}

int match_createtask(firmware *fw, int k, uint32_t fadr, uint32_t v2)
{
    uint32_t adr = followBranch(fw,idx2adr(fw,k),0x01000001);
    int j1, j2;
    // CreateTask* ?
    if (adr == fadr)
    {
        j1 = find_inst_rev(fw,isLDR_PC_r3,k,16);
        j2 = find_inst_rev(fw,isLDR_PC_r0,k,16);
        // param0 and param3 are needed
        if ((j1 > 0) && (j2 > 0))
        {
            // check for unwanted function calls
            if ((find_inst(fw,isBL,j1,16)!=k) || (find_inst(fw,isBL,j2,16)!=k))
                return 0;
            // both parameters seem ok
            uint32_t sadr = LDR2val(fw,j2);
            uint32_t tadr = LDR2val(fw,j1);
            if (sadr != 0)
            {
                char *s = adr2ptr(fw,sadr);
                // check first 2 characters in task name, check valid task address
                if ((isalnum(s[0])&&isalnum(s[1])) && (idx_valid(fw,adr2idx(fw,tadr))))
                {
                    char *nm = malloc(strlen(s)+6);
                    sprintf(nm,"task_%s",s);
                    add_func_name(nm, tadr, 0);
                }
            }
        }
    }
    return 0;
}

void find_tasks(firmware *fw)
{
    int k = get_saved_sig(fw,"CreateTask");
    if (k >= 0)
    {
        search_fw(fw, match_createtask, func_names[k].val, 0, 5);
    }
    k = get_saved_sig(fw,"CreateTaskStrictly");
    if (k >= 0)
    {
        search_fw(fw, match_createtask, func_names[k].val, 0, 5);
    }
}

void find_builddate(firmware *fw)
{
    int j = get_saved_sig(fw,"GetBuildDate_FW");
    if (j >= 0)
    {
        int idx = adr2idx(fw, func_names[j].val);
        idx = LDR2idx(fw, idx);
        uint32_t adr = fwval(fw, idx);
        idx = adr2idx(fw, adr);
        fw->fw_build_date = (char*)&fw->buf[idx];
    }
    else
        fw->fw_build_date = 0;

    j = get_saved_sig(fw,"GetBuildTime_FW");
    if (j >= 0)
    {
        int idx = adr2idx(fw, func_names[j].val);
        idx = LDR2idx(fw, idx);
        uint32_t adr = fwval(fw, idx);
        idx = adr2idx(fw, adr);
        fw->fw_build_time = (char*)&fw->buf[idx];
    }
    else
        fw->fw_build_time = 0;
}

void write_levent_table_dump(firmware *fw, uint32_t tadr)
{
    char *str;
    uint32_t lid = 0;
    uint32_t val;
    if (!tadr) {
        return;
    }
    FILE *f=fopen("levent_table.txt","w");
    if(!f) {
        return;
    }
    fprintf(f,"address    ID     (unknown)  name\n");

    for(;;tadr += 12) {
        val = *(uint32_t*)adr2ptr(fw, tadr);
        if ((val == 0xffffffff) || (val == 0) || (*(uint32_t*)adr2ptr(fw, tadr+4) < lid)) {
            break;
        }
        lid = *(uint32_t*)adr2ptr(fw, tadr+4);
        str = (char*)adr2ptr(fw,val);
        if (str) {
            fprintf(f,"0x%08x 0x%04x 0x%08x %s\n",tadr,lid,*(uint32_t*)adr2ptr(fw, tadr+8),str);
        }
    }
    fclose(f);
}

//------------------------------------------------------------------------------------------------------------

// Write out firmware info
void output_firmware_vals(firmware *fw)
{
    bprintf("// Camera info:\n");

    if (fw->firmware_ver_str == 0)
    {
        bprintf("//   Can't find firmware version !!!\n\n");
    }
    else
    {
        uint32_t j = idx2adr(fw,fw->fwver_idx);
        char *c = strrchr(fw->firmware_ver_str,' ') + 1; // points after the last space char
        uint32_t k = j + c - fw->firmware_ver_str;
        if ( (k>=j) && (k<j+32) )
        {
            bprintf("//   %s   // Found @ 0x%08x, \"%s\" @ 0x%08x\n",fw->firmware_ver_str,j,c,k);
        }
        else
        {
            // no space found in string (shouldn't happen)
            bprintf("//   %s   // Found @ 0x%08x, \"%s\" @ 0x%08x\n",fw->firmware_ver_str,j,fw->firmware_ver_str,j);
        }
    }

    if (fw->fw_build_date != 0)
    {
        bprintf("//   Firmware build timestamp: %s %s\n",fw->fw_build_date, (fw->fw_build_time==0)?"":fw->fw_build_time);
    }

    if (fw->fsize > (fw->size + 256))
    {
        bprintf("//   Possible corrupt firmware dump - file size too small for start address 0x%08x\n",fw->base);
        bprintf("//     file size = %.2fMB, should be %.2fMB\n", ((double)fw->size*4.0)/(1024.0*1024.0),((double)fw->fsize*4.0)/(1024.0*1024.0));
    }

    if (fw->cam != 0)
    {
        bprintf("//   %s\n",fw->cam);
    }
    else
    {
        bprintf("//   Could not find Camera name - possible corrupt firmware dump\n");
    }

    bprintf("\n// Values for makefile.inc\n");

    if (fw->pid != 0)
    {
        bprintf("//   PLATFORMID = %d (0x%04x) // Found @ 0x%08x\n",fw->pid,fw->pid,fw->pid_adr);
    }

    if (fw->maxram != 0)
        bprintf("//   MAXRAMADDR = 0x%08x\n",fw->maxram);

    if (fw->memisostart != 0)
    {
        osig *o = find_match(fw->sv->makevals, "MEMISOSTART", fw->memisostart);
        if (o && (o->val == fw->memisostart))
            bprintf("//   MEMISOSTART = 0x%08x\n",fw->memisostart);
        else
            bprintf("//   MEMISOSTART = 0x%08x  (*** DOES NOT MATCH MAKEFILE VALUE 0x%08x***)\n",fw->memisostart,(o)?o->val:0);
    }

    bprintf("\n");
}

int compare_func_names(const func_entry **p1, const func_entry **p2)
{
    int rv = strcasecmp((*p1)->name, (*p2)->name);     // Case insensitive
    if (rv != 0)
        return rv;
    return strcmp((*p1)->name, (*p2)->name);        // Case sensitive (if equal with insensitive test)
}

int compare_func_addresses(const func_entry **p1, const func_entry **p2)
{
    if ((*p1)->val < (*p2)->val)
        return -1;
    else if ((*p1)->val > (*p2)->val)
        return 1;
    return compare_func_names(p1,p2);
}

void write_funcs(firmware *fw, char *filename, func_entry *fns[], int (*compare)(const func_entry **p1, const func_entry **p2))
{
    int k;

    qsort(fns, next_func_entry, sizeof(func_entry*), (void*)compare);

    FILE *out_fp = fopen(filename, "w");
    for (k=0; k<next_func_entry; k++)
    {
        if (strncmp(fns[k]->name,"hook_",5) != 0)
        {
            if (fns[k]->val != 0)
            {
                if (fns[k]->flags & BAD_MATCH)
                {
                    osig* ostub2 = find_sig(fw->sv->stubs,fns[k]->name);
                    if (ostub2 && ostub2->val)
                        fprintf(out_fp, "0x%08x,%s,(stubs_entry_2.s)\n", ostub2->val, fns[k]->name);
                }
                else
                    fprintf(out_fp, "0x%08x,%s\n", fns[k]->val, fns[k]->name);
            }
#ifdef LIST_IMPORTANT_FUNCTIONS
            else if (fns[k]->flags & LIST_ALWAYS)
            {
                // helps development by listing important functions even when not found
                fprintf(out_fp, "0,%s,(NOT FOUND)\n", fns[k]->name);
            }
#endif
        }
    }
    fclose(out_fp);
}

int main(int argc, char **argv)
{
    firmware fw;
    int k;
    int ret = 0;
    const char *curr_name;

    clock_t t1 = clock();

    if ((argc < 4) || (argc > 5))
        usage("args");

    out_fp = fopen(argv[3],"w");
    if (out_fp == NULL) usage("failed to open outputfile");

    for (next_func_entry = 0; func_names[next_func_entry].name != 0; next_func_entry++);
    int max_find_func = next_func_entry;

    fw.sv = new_stub_values();
    load_stubs(fw.sv, "stubs_entry_2.S", 1);
    load_stubs_min(fw.sv);
    load_modemap(fw.sv);
    load_platform(fw.sv);
    load_makefile(fw.sv);

    bprintf("// !!! THIS FILE IS GENERATED. DO NOT EDIT. !!!\n");
    bprintf("#include \"stubs_asm.h\"\n\n");

    load_firmware(&fw,argv[1],argv[2],(argc==5)?argv[4]:0, OS_VXWORKS);
    find_eventprocs(&fw);
    find_builddate(&fw);
    output_firmware_vals(&fw);

    out_hdr = 1;
    bprintf("// Stubs below should be checked. Stub not matched 100%%, or difference found to current 'stubs_entry_2.S'\n");
    bprintf("//    Name                                     Address      Rule  %%  Comp to stubs_entry_2.S\n");
    out_hdr = 0;
    bprintf("// Stubs below matched 100%%.\n");
    bprintf("//    Name                                     Address                Comp to stubs_entry_2.S\n");

    find_tasks(&fw);

    for (k = 0; k < max_find_func; k++)
    {
        count = 0;
        curr_name = func_names[k].name;
#ifdef DEBUG_PRINT_ALL_FUNC_NAMES
fprintf(stderr,"%s:",curr_name);
#endif
        find_matches(&fw, curr_name);
        print_results(&fw,curr_name,k);

        if (count == 0)
        {
            ret = 1;
        }
    }

    find_modemap(&fw);
    find_stubs_min(&fw);
    find_lib_vals(&fw);
    //find_key_vals(&fw);
    find_platform_vals(&fw);
    find_other_vals(&fw);

    write_output();

    fclose(out_fp);

    func_entry *fns[MAX_FUNC_ENTRY];
    for (k=0; k<next_func_entry; k++)
        fns[k] = &func_names[k];

    write_funcs(&fw, "funcs_by_name.csv", fns, compare_func_names);
    write_funcs(&fw, "funcs_by_address.csv", fns, compare_func_addresses);

    clock_t t2 = clock();

    printf("Time to generate stubs %.2f seconds\n",(double)(t2-t1)/(double)CLOCKS_PER_SEC);

    return ret;
}

//------------------------------------------------------------------------------------------------------------
